--// $Id: FlexHub.lua 1457 2013-07-05 13:32:36Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011/2012 Daywalker, FlipFlop and en_dator / dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

--[[
	* Commandline options are ONLY for Linux
	* Windows users that don't want to use the Console-GUI have to edit the tArgs below for startup options
	* Linux users can also edit the tArgs section below instead of using the command line (but options on the command line will override whats
	  saved in this file.
	* There is no difference in how the settings work on Windows and Linux

	* NOTE! any setting entered on commandline or saved into this file will override the corresponding hub setting!
		Order of override is :
		1 - commandline
		2 - settings stored here in tArgs
		3 - settings set from inside the hub

	--// These settings will not be stored in hub settings, they must be used (in commandline or in tArgs below) everytime or they will reset back to default values
	--nogui				To disable the hub Console-GUI
	--noupdate				To disable the !update command in the hub
	--nochangeports			To disable the commands to add and remove ports inside a running hub
	--norestart				To disable the restart function from inside the hub
	--noshutdown				To disable the shutdown function from inside the hub
	--notouchrestartscript		To tell FlexHub to not touch the RestartScript.sh or RestartScript.bat used to restart the hub (it will still create one if its missing)

	--logdir=./logs/				The directory to store logfiles.
	--configdir=./settings/		The directory to store and load configurationfiles.
	--storagedir=./				The main directory to store and load settings, logs, import, this directory will contain the dirs: settings, logs, import and backup

	--// The following settings will be saved on first use and do not need to be used on the command line again (unless you want to change them)
	--port=<port1>[,<port2>]				Regular port(s), to use multiple ports, just add them one after another comma separated, no spaces
	--sport=<secure port1>[,<secure port2>]	Secure port(s): EXPERIMENTAL (Can cause the hub to freeze when a secure socket is 'ghosting'.)
	--bind=<ip>								Will bind the port to a specific interface (IP), default = * and will bind to all local interfaces
	--userlimit=<number>					Host userlimit, there's no way to override this value from within the hub, but a lower value can be used in the hub to control max users with !set maxusers <number>

	--// Use these when you want to add a new Hubowner account from the commandline (needed on first run if you use --nogui)
	--nick=<nickname>						Nickname of master account
	--pass=<password>						Password of master account	(always use a unique and safe pass)
	--nickip=<ip>							Only allow connections from this ip for the master account (same as the bindip command in the hub)
]]--
package.cpath = "./lib/?.so;./lib/?.dll;./lib/?.dylib;"
package.path = "./lib/?.lua"

--// Note: these settings can be overridden by using commandline parameters in linux, for windows they must be edited here
tArgs = {
	["nick"] = "",
	["pass"] = "",
	["profile"] = "Hubowner",
	["nickip"] = "",

	["port"] = "",							--// Regular port(s). If multiple ports are specified, they must be enclosed in quotes without spaces:  ["port"] = "8000,8001,80002",
	["bind"] = "*",							--// Will bind the port to a specific interface (IP), default is "*" and will bind to all local interfaces
	["sport"] = "",							--// Secure port(s): EXPERIMENTAL (Known issue: can cause the hub to freeze when a secure socket is 'ghosting'.) If multiple ports are specified, they must be enclosed in quotes without spaces:  ["sport"] = "8000,8001,80002",
	["userlimit"] = "",						--// Host userlimit, there's no way to override this value from within the hub

	-- // the following settings are NOT avaliable through the Console-GUI and can only be set here
	["nogui"] = false,						--// To disable the hub Console-GUI use: true, To enable the hub Console-GUI use: false
	["noupdate"] = false,				--// To disable updates started from the hub use: true, To enable updates started from the hub use: false
	["norestart"] = false,				--// To disable the restart function from inside the hub use: true, To enable use: false
	["noshutdown"] = false,			--// To disable the shutdown function from inside the hub use: true, To enable use: false
	["nochangeports"] = false,		--// To disable the commands to add and remove ports inside a running hub
	["notouchrestartscript"] = false,	--// To tell FlexHub to not touch the RestartScript.sh or RestartScript.bat used to restart the hub (it will still create one if its missing)

	["logdir"] = "./logs/",					--// Can override this by using commandline parameter: --logdir=./MyHub/logs/
	["configdir"] = "./settings/",		--// Can override this by using commandline parameter: --configdir=./MyHub/settings/
	["storagedir"] = "./",					--// This contains the path to the folders: logs,settings,import and backup. This setting overrides the previous 2 path settings. End with /
}

--// Do not change lines below

dofile("./core/Hub-Start.lua")

--// Daywalker and FlipFlop
