--// $Id: Hub-AccountManager.lua 1411 2012-12-26 19:02:10Z en_dator $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tHubFiles["Hub-AccountManager.lua"] = {
	["sRev"] = "$Rev: 1411 $",
	["bRestart"] = false,
	["sPath"] = "./core/Hub-AccountManager.lua",
	}

AccountManager = {
	["Protected"] = { ["User"] = 2, ["Gui"] = 1, ["Bot"] = 2},		--// 2 = Can't register users to profile, 1 = Can only be registered with Console

	["AddComment"] = function(self, ssid, nick, comment)
		if not nick or nick == "" then return false, "No nick specified." end
		local vnick, reg = AccountManager:GetName(nick)

		if not vnick then return false, "Account with name <"..nick..">  doesn't exist." end
		if tProfiles[reg.sProfile].iLevel < tUser:GetProfileLevel(ssid) or tUser:GetProfileLevel(ssid) == 10 then
			if not comment or comment == "" or comment:lower() == "clear" then 
				reg.sComment = "" 
				return true, "Comments for account <"..vnick.."> cleared."
			else
				reg.sComment = (reg.sComment or "")..comment.."\n"
				return true, "Comment for account <"..vnick.."> added, currently containing: \n"..reg.sComment
			end
		else
			return false, Lang(ssid, 0, "This user is too powerful for you")
		end
	end,

	["GetName"] = function(self, nick)
		if not nick or nick == "" then return false, "No nick specified." end

		--// Quick return
		if tAccounts[nick] then return nick, tAccounts[nick] end

		--// Loop over accountnames to find lowercase match
		for vnick, reg in pairs(tAccounts) do
			if nick:lower() == vnick:lower() then
				return vnick, reg
			end
		end
		return false, "Account with name <"..nick..">  doesn't exist."
	end,

	["Check"] = function(self, nick)
		local result = {}
		if nick then
			local reg = tAccounts[nick:lower()]
			if not reg then return false, "Account not found" end
			if not tProfiles[reg.sProfile] then 
				tAccounts[lnick].sProfile = "Reg"
				tFuncs:SaveSettings("accounts")
				return false, "Account: "..nick.." has a profile '"..reg.sProfile.."' that doesn't exist (anymore), changed to Reg profile."
			end
			return true, "Account checked."
		else
			--// Check if sNick exists and accounts are stored lowercase
			local delaccounts = {}
			local addaccounts = {}
			local changed = false
			for nick, reg in pairs(tAccounts) do
				if nick == nick:lower() then
					if not reg.sNick then 
						reg.sNick = nick 
						changed = true
					elseif nick ~= string.lower(reg.sNick) then	-- nick and sNick should also match
						reg.sNick = nick 
						changed = true
					end
				else
					reg.sNick = nick
					table.insert(addaccounts, reg)
					table.insert(delaccounts, nick)
					changed = true
				end

				--// CHECK NICKLENGTH AND CHARS
				local res, err = tCheck:NickChars(reg.sNick)
				if not res then
					table.insert(result, "Account: "..reg.sNick.." contains invalid characters, error: "..(err or "no msg") ..". Please change the nick manually with !changereg "..reg.sNick.." <newnick>")
				end

				--// test conversion of sNick
				local ok, convnick, errmsg = tFuncs:ToUTF8(reg.sNick, true)
				local sok, orgnick, errmsgg = tFuncs:ToLocale(convnick, true)

				if not ok or not sok or reg.sNick ~= orgnick then
					local new = orgnick:lower()
					if not tAccounts[new] then
						reg.sNick = new
						changed = true
						table.insert(result, "Account: "..reg.sNick.." is causing nick conversion error: "..(errmsgg or errmsg or " no msg") ..". Nick changed to "..new)
					else
						table.insert(result, "Account: "..reg.sNick.." is causing nick conversion error: "..(errmsgg or errmsg or " no msg") ..". Please change the nick manually with !changereg "..reg.sNick.." <newnick>")
					end
				end

				--// if conversion fails, also change reg, todo / fredo

				--// Check if profile exists
				if not tProfiles[reg.sProfile] then 
					table.insert(result, "Account: "..reg.sNick.." has a profile '"..reg.sProfile.."' that doesn't exist (anymore), changed to Reg profile.")
					tAccounts[nick].sProfile = "Reg"
					reg.sProfile = "Reg"
					changed = true
				end

				--// Check levelnumber
				local level = tProfiles[reg.sProfile].iLevel
				if not tAccounts[nick].iLevel or tAccounts[nick].iLevel ~= level then
					table.insert(result, "Account: "..reg.sNick.." with profile "..reg.sProfile.." has a missing or mismatching profilelevel, fixed and set to level: "..level..".")
					tAccounts[nick].iLevel = level
					changed = true
				end
				--// Check stats info
				if not tAccounts[nick].iLastSpoke then tAccounts[nick].iLastSpoke = 0 changed = true end
				if not tAccounts[nick].iLastLogin then tAccounts[nick].iLastLogin = 0 changed = true end
				
			end
			--// Delete unlowercased regs
			for num,nick in pairs(delaccounts) do
				tAccounts[nick] = nil
			end
			--// Add lowercased regs
			for num,reg in pairs(addaccounts) do
				tAccounts[reg.sNick:lower()] = reg
			end

			if changed then tFuncs:SaveSettings("accounts") end

			if #result > 0 then
				return false, "Accounts checked, following accounts had problems: \n\t"..table.concat(result, "\n\t")
			else
				return true, "All accounts OK."
			end
		end
	end,

	["Add"] = function(self, ssid, vnick, pass, prof, ip)
		local prof = prof or "Reg"

		if pass == "" and not tSecurity.General.AllowBlankPass then
			return false, Lang(ssid, 0, "Using blank password is disabled in settings, you must specify a password.")
		elseif pass ~= "" then
			local res, err = tCheck:PasswordChars(pass)
			if not res then
				return false, err
			end
		end

		local profdata, prof = ProfileManager:GetData(ssid, prof)
		if prof and not profdata then
			return false, Lang(ssid, 0, "You can't add an account to this profile. %s", prof)
		elseif not profdata then
			return false, prof
		end

		if ip and not tUser:IsAllowed(ssid, "bindip") then
			return false, Lang(ssid, 0, "You're not allowed to specify a binding IP.")
		end

		if pass == "" and not ip then
			return false, Lang(ssid, 0, "To create an account with no password you must specify an IP to bind to.")			-- comment out this line if you really want to allow creating accounts without passwords and no bind IP
		elseif pass == "" and profdata.iLevel == 10 then
			return false, Lang(ssid, 0, "You are not allowed to create a level 10 account without setting a password.")
		end

		if AccountManager.Protected[prof] == 2 then
			return false, Lang(ssid, 0, "You can't add an account to this profile.")
		elseif AccountManager.Protected[prof] == 1 and ssid ~= "AAAA" then
			return false, Lang(ssid, 0, "You can only add accounts with this profile through the console.")
		else
			--local level = profdata.iLevel or -1
--~ 			tFuncs:Debug("Checking nick: "..vnick.." with level: "..level)
			local res, _, reg, err, inuse = tCheck:Nick(nil, nil, vnick, (profdata.iLevel or -1))
			if not res then return false, "Registration failed: "..err end

			local vsid, snick = tNick:GetSid(vnick)
			if snick then vnick = snick end	--// If user is online, copy the casing from the user
			if ssid == "AAAA" or profdata.iLevel < tClient[ssid].iProfile or tClient[ssid].iProfile == 10 then
				if not vsid or ssid == "AAAA" or tUser:IsHigher(ssid,vsid) then		-- check if allowed to reg this profile
					if vnick and pass and prof then
						local lnick = vnick:lower()
						if not tNick:GetReg(vnick) then -- check if user is regged allready
							--local res,err = tCheck:PasswordChars(pass)
							--if not res then return false, err end

							-- start the reg process
							tAccounts[lnick] = {
									["sNick"] = vnick,
									["sProfile"] = prof,
									["sPassword"] = pass,
									["iLastLogin"] = 0,
									["iLastSpoke"] = 0,
									["iTotalOnlineTime"] = 0,
									["sReggedBy"] = (ssid~="AAAA" and tUser:GetNick(ssid) or "FlexHub Console"),
									["iRegDate"] = os.time(),
									["sComment"] = "",
									["sGreeting"] = "",
									["iLevel"] = (profdata.iLevel or 0),
							}

							local iptxt = ""
							if ip then 
									tAccounts[lnick]["sBindIP"] = ip
									iptxt = Lang(ssid, 0, "\n\tIP bound to:\t%s", ip)
							end

							tFuncs:Debug("[AccountManager] [Add] An account was created for: "..vnick.." with profile: "..prof.." by: "..(ssid~="AAAA" and tUser:GetNick(ssid) or "FlexHub Console"), "FI", -1)
							tFuncs:SaveSettings("accounts")

							local address = tFuncs:GetAddress()
							local pwd = pass
							local bNotifySetOwnPass = false
							if pwd == "" then
								pwd = "No password stored, use !setpass/!setmypass to set a password"
								bNotifySetOwnPass = true
							end

							local details = ""
							if tHubSettings.Main.sRegMessage and tHubSettings.Main.sRegMessage ~= "" then
								details = string.gsub(tHubSettings.Main.sRegMessage, "%%%[password%]", tFuncs:EscapePercent(pwd))
								details = string.gsub(details, "%%%[user_name%]", vnick)
								details = string.gsub(details, "%%%[user_profile%]", prof..iptxt)
								details = tFuncs:FillVars(details)
							else
								details = "\t\tNickname:\t"..vnick.."\n\t\tPassword:\t"..pwd.."\n\t\tProfile:\t\t"..prof..iptxt.."\n\t\tHubname:\t"..(SetManager:Get("hub_name") or "").."\n\t\tHubaddress:\t"..address.."\n"
							end

							vsid = tNick:GetSid(vnick)	-- check if victim is online
							if vsid then	-- send details in pm to victim

								if bNotifySetOwnPass then
									tClient[vsid].bNotifySetOwnPass = true
									prof = "Reg"
								end

								tClient[vsid].sProfile = prof
								tClient[vsid].iProfile = tProfiles[prof].iLevel
								tClient[vsid].bOperator = tProfiles[prof].bOperator or (tProfiles[prof].iLevel == 10 and true)
								tClient[vsid].bHasKey = tProfiles[prof].bHasKey
								tClient[vsid].bHasFakeKey = tProfiles[prof].bHasFakeKey

								local ct = tInfo:SetCT(tClient[vsid], vsid)

								tInfo:CreateADC(tClient[vsid], vsid)

								if tProfiles[prof].bHasKey then
									if tClient[vsid].sProtocol == "ADC" then
										if tClient[vsid].tInfo.tFiltered.CT > 3 then
											tSend:Raw("P_ADC", "BINF "..vsid.." CT"..ct.." OP1\n")
										elseif tClient[vsid].tInfo.tFiltered.CT > 0 then
											tSend:Raw(vsid, "BINF "..vsid.." CT"..ct.."\n")
										end
									else
										tSend:Raw("P_NMDC", "$OpList "..vnick.."|")
									end
								end

								if tProfiles[prof].bHasFakeKey then
									for sid, data in pairs(tClient) do
										if data.sIP == tClient[vsid].sIP then
											if tClient[vsid].sProtocol == "ADC" then
												tSend:Raw(sid, "BINF "..vsid.." CT16 OP1\n")
											elseif tClient[vsid].sProtocol == "NMDC" then
												tSend:Raw(sid, "$OpList "..vnick.."|")
											end
										end
									end
								end

								local coulddownload = tUser:CanDownload(vsid)
								if not coulddownload then
									tSend:ADCUserIps(sid)
								end

								tSend:UpdateRooms(vsid)

								tSend:PmToUser(Lang(ssid, 0, "\n\n\tYou are registered with the following details:\n\n%s\n\tPlease put this in your Favorite hub options (after you typed: /fav) and reconnect to this hub.", details), nil, vsid)
								tSend:ChatToUser(Lang(ssid, 0, "You are now registered, check a PM from "..SetManager:Get("bot_name").." for details."), nil, vsid)
								return true, Lang(ssid, 0, "\n\n\tYou have successfully registered this user with the following details:\n\n%s\n\tThese details were sent to the user in PM.", details)
							else	-- send it to yourself to paste it to victim
								return true, Lang(ssid, 0, "\n\n\tYou have successfully registered this user with the following details:\n\n%s\n\tPlease paste this info to the user you just registered.\n", details)
							end
						else
							return false, Lang(ssid, 0, "This user is already registered.")
						end
					else
						return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <password> <profile> <ip(optional)>")
					end
				else
					return false, Lang(ssid, 0, "You are not allowed to register a user with a profile of the same level or higher than your own.")
				end
			else
				return false, Lang(ssid, 0, "You are not allowed to register a user with a profile of the same level or higher than your own.")
			end
		end
	end,

	["Delete"] = function(self, ssid, vnick)
		local reg, vnick, lnick = tNick:GetReg(vnick)

		if reg then
			if tProfiles[reg.sProfile].iLevel < tUser:GetProfileLevel(ssid) or tUser:GetProfileLevel(ssid) == 10 then
				local prof = tAccounts[lnick].sProfile
				tAccounts[lnick] = nil

				tFuncs:Debug("[AccountManager] [Delete] Account was deleted for: "..vnick.." with profile: "..prof.." by: "..(ssid~="AAAA" and tUser:GetNick(ssid) or "FlexHub Console"), "FI", -1)
				tFuncs:SaveSettings("accounts")

				--// If user is online
				local vsid = tNick:GetSid(vnick)

				if vsid then
					local coulddownload = tUser:CanDownload(vsid)

					local wasop = tClient[vsid].bOperator
					local hadfakekey = tClient[vsid].bHasFakeKey
					local hadkey = tClient[vsid].bHasKey

					tClient[vsid].sProfile = "User"
					tClient[vsid].iProfile = -1
					tClient[vsid].bHasKey = false
					tClient[vsid].bOperator = false
					tClient[vsid].bHidden = nil

					tInfo:SetCT(tClient[vsid], vsid)
					tInfo:CreateADC(tClient[vsid], vsid)

					--// Remove key from user, only works for ADC, for NMDC it needs a $Quit
					if tClient[vsid].sProtocol == "ADC" then
						tSend:Raw(vsid, "BINF "..vsid.." CT\n")
					end

					tUser:Quit(vsid, true)
					tUser:Announce(vsid)

					tSend:OwnKey(vsid, true)
					tSend:UpdateRooms(vsid)

					if coulddownload or wasop then
						tSend:ADCUserIps(sid, true)	
					end
				end
				return true, Lang(ssid, 0, "Account: %s has been deleted.", vnick)
			else
				return false, Lang(ssid, 0, "This user is too powerful for you")
			end
		else
			return false, Lang(ssid, 0, "This user is not registered")
		end
	end,

	["Purge"] = function(self, ssid, profile, days, confirm)		--// use confirm = true to bypass confirmation check
		--// Validation
		if profile ~= "*" then profile = ProfileManager:GetName(ssid, profile) end
		if profile and profile ~= "*" and (profile == "User" or profile == "Bot" or tProfiles[profile].iLevel == 10) then 
			return false, "Error: you can't delete accounts with profile: "..profile 
		end
		if not profile then return false, "Error: specify a valid profilename to delete unused accounts." end

		local days = tonumber(days)		--// If days = 0 then only accounts with an iExpires will be checked for purging
		if not days then return false, "Error: specify the number of inactive days to delete unused accounts." end
		if confirm and (not tFuncs:ConfirmCode(confirm) or not tLogs.DeleteAccounts) then return false, "Wrong confirmation code. Use !purgeaccounts <profilename or *> <days> to get a new confirmation code." end
		if confirm then
			if not tLogs.DeleteAccountsDays or tLogs.DeleteAccountsDays ~= days then return false, "The amount of days doesn't match the amount of days of the previous command. Use !purgeaccounts "..profile.." "..days.." to get a new confirmation code." end
			if not tLogs.DeleteAccountsProfile or tLogs.DeleteAccountsProfile ~= profile then return false, "The profile doesn't match the profile of the previous command. Use !purgeaccounts "..profile.." "..days.." to get a new confirmation code." end
		end

		--// Init
		local ctime = socket.gettime()
		local purgetable = {}
		local result = ""

		--// Generate accounts for purging
		local PurgeAccounts = function(profname)
			for nick,data in pairs(tAccounts) do
				if data.sProfile == profname then
					if not tNick:GetSid(nick) then 	--// Skip online users
						if not tUser:IsProtected(nil, nick, data.sLastIP) then	--// Skip protected users and IP's
							if tProfiles[data.sProfile].iLevel < tUser:GetProfileLevel(ssid) then		--// Excludes level 10. will never be purged.
								local last = data.iLastLogout or data.iLastLogin or 0
								if data.iExpires and ctime > data.iExpires then
									table.insert(purgetable, { ["Last"] = last, ["sNick"] = data.sNick, ["sPassword"] = data.sPassword or "" , ["sProfile"] = data.sProfile or "Reg", ["sLastIP"] = data.sLastIP, ["iExpires"] = data.iExpires, ["sReggedBy"] = data.sReggedBy or "", ["iRegDate"] = data.iRegDate or 0, ["iTotalOnlineTime"] = data.iTotalOnlineTime or 0})
								elseif days > 0 and ((ctime - last) > (days*86400) or (last == 0 and data.iRegDate and ctime > (data.iRegDate + (days*86400)))) then
									table.insert(purgetable, { ["Last"] = last, ["sNick"] = data.sNick, ["sPassword"] = data.sPassword or "" , ["sProfile"] = data.sProfile or "Reg", ["sLastIP"] = data.sLastIP, ["sReggedBy"] = data.sReggedBy or "", ["iRegDate"] = data.iRegDate or 0, ["iTotalOnlineTime"] = data.iTotalOnlineTime or 0})
								end
							end
						end
					end
				end
			end
		end

		--// Sorting purgetable based on last online time
		local SortAccounts = function(a,b)
			if not a or not b or a.Last < b.Last then return true else return false end
		end
 
		local PumpMessage = function(str)	-- ok, its not pretty, but it works
			local pump = ""
			for w in str:gmatch("(.-\n)") do
				pump = pump .. w
				if #pump > 28000 then	-- (clients cant handle larger strings in one go than 32,767)
					tSend:PmToUser("\n"..pump, "AAAA", ssid)
					pump = ""
				end
			end
			tSend:PmToUser("\n"..pump, "AAAA", ssid)
		end

		--// Generate string to show accounts
		local ShowAccounts = function(purgetable)
			local res = ""
			table.sort(purgetable, SortAccounts)
			local tRes = {}

			if not confirm then
--				table.insert(tRes,"\n\tNick\t\t\tProfile\t\tLast online\tRegistered\t\tBy\t\tAccount expired\n\t"..string.rep("-",230))
				table.insert(tRes,"\n\tAccount expired\tLast Online\t\tRegistered Date\t\tProfile\tBy\t\tNick\n\t"..string.rep("-",230))
			else
				table.insert(tRes,"\n\tNick\t\t\tNick Password Profile\t\tLast online\tRegistered\tBy\t\tAccount expired\n\t"..string.rep("-",230))
			end

			for num, data in ipairs(purgetable) do
				if not confirm then
--					table.insert(tRes,"\n\t"..(PaddingT(data.sNick, 26)).." \t"..data.sProfile.." \t\t"..PaddingT((data.Last>0 and tFuncs:Duration(ctime - data.Last) or "Never"), 28).." \t"..ToDate(data.iRegDate).." \t"..(data.sReggedBy or "N/A")..(data.iExpires and " \t"..ToDate(data.iExpires) or ""))
					table.insert(tRes, "\n\t"..(data.iExpires and ToDate(data.iExpires) or "No expire date") .."\t"..PaddingT((data.Last>0 and tFuncs:Duration(ctime - data.Last) or "Never logged in"), 21).."\t"..ToDate(data.iRegDate).."\t"..data.sProfile.."\t"..PaddingT((data.sReggedBy or "N/A\t"), 18).."\t"..data.sNick)
				else
					table.insert(tRes,"\n\t"..(PaddingT(data.sNick, 26)).." \t"..PaddingT(data.sNick.." "..(data.sPassword or "\"\"").." "..data.sProfile, 28).." \t\t"..PaddingT((data.Last>0 and tFuncs:Duration(ctime - data.Last) or "Never"), 28).." \t"..ToDate(data.iRegDate).." \t "..data.sReggedBy or "N/A"..(data.iExpires and " \t"..ToDate(data.iExpires) or ""))
				end
			end

			res = table.concat(tRes)
			tFuncs:Debug("Purgeaccounts will return "..#res.." chars for tRes.")
			return res
		end

		--// Get previously generated purgetable or generate a new one
		if confirm and tLogs.DeleteAccounts then	--// Purge previously gathered accounts
				purgetable = tLogs.DeleteAccounts
		else
			if profile == "*" then
				for profname,profdata in pairs(tProfiles) do
					if profname ~= "User" then
						PurgeAccounts(profname)
					end
				end
			else
				PurgeAccounts(profile)
			end
		end

		--// Display accounts or delete accounts
		if #purgetable > 0 and confirm then		--// Delete accounts
			local amount = #purgetable
			result = ShowAccounts(purgetable)
			for num, data in ipairs(purgetable) do
				tFuncs:Debug("[AccountManager] [Purge] Expired/unused account deleted: "..data.sNick.." with profile: "..data.sProfile.." by: "..(ssid~="AAAA" and tUser:GetNick(ssid) or "FlexHub"), "FI", 0)
				tAccounts[data.sNick:lower()] = nil
			end
			tFuncs:SaveSettings("accounts")
			tLogs.DeleteAccounts = nil
			tLogs.DeleteAccountsDays = nil
			tLogs.DeleteAccountsProfile = nil

			local ret = ""
			if #result > 30000 then
				PumpMessage(result)
				ret = "\n\nDue to the large amount of accounts being deleted, the list has been pumped into a PM, you can use the PM log to review all accounts if the scollback is not enough\n\nNumber of deleted unused/expired accounts: "..amount.."\n"
			else 
				ret = "\n\nDeleted the following accounts:\n"..result.."\n\nNumber of deleted unused/expired accounts: "..amount.."\n"
			end

			return true, ret

		elseif #purgetable > 0 then			--// Preview unused/expired accounts
			local amount = #purgetable
			tLogs.DeleteAccounts = purgetable  	--// Have to store as a global table to prevent accounts being deleted that didn't show during generation, before confirmcode was given, because of the timedifference between commands
			tLogs.DeleteAccountsDays = days
			tLogs.DeleteAccountsProfile = profile
			result = ShowAccounts(purgetable)

			local ret = ""
			if #result > 30000 then
				PumpMessage(result)
				ret = "\n\nDue to the large amount of accounts being selected, the list has been pumped into a PM, you can use the PM log to view all accounts if the scollback is not enough\n\nTo confirm deletion of "..amount.." unused/expired accounts, use: !purgeaccounts "..profile.." "..days.." "..tHubSettings.Main.bConfirmCode.."\n"
			else 
				ret = "\n\nThe following accounts can be deleted:\n"..result.."\n\nTo confirm deletion of "..amount.." unused/expired accounts, use: !purgeaccounts "..profile.." "..days.." "..tHubSettings.Main.bConfirmCode.."\n"
			end

			return true, ret
		else
			tLogs.DeleteAccounts = nil
			tLogs.DeleteAccountsDays = nil
			tLogs.DeleteAccountsProfile = nil
			return false, "There were no unused/expired accounts to delete."
		end
	end,

	["SetProfile"] = function(self, ssid, vnick, prof)
		local reg, vnick, lnick = tNick:GetReg(vnick)
		if reg and vnick then
			if vnick:lower() == tClient[ssid].sNick:lower() and tClient[ssid].iProfile < 10 then
				return false, "You can't change your own profile."
			end

			local oldprof = reg.sProfile
			local ownlevel = tClient[ssid].iProfile
			if tProfiles[reg.sProfile] then
				if tProfiles[reg.sProfile].iLevel >= tClient[ssid].iProfile and tUser:GetProfileLevel(ssid) < 10 then
					return false, Lang(ssid, 0, "This user is too powerful for you.")
				end
			else
				--// Profile doesn't exist anymore, no levelcheck needed
				return false, Lang(ssid, 0, "Profile for this user doesn't exist anymore: "..oldprof..", command cancelled.")
					-- // what should we do here? we cant have it like this? fredo
			end

			local prof, err = ProfileManager:GetName(ssid, prof)
			if err and not prof then
				return false, Lang(ssid, 0, "Error: Profile doesn't exist, use !showprofiles for available profiles.")
			elseif not prof then
				return false, err
			end

			if AccountManager.Protected[prof] == 1 then
				return false, Lang(ssid, 0, "You can only add accounts with this profile through the console.")
			elseif AccountManager.Protected[prof] == 2 then
				return false, Lang(ssid, 0, "You can't change an account to this profile.")
			elseif tProfiles[prof] then
				if tProfiles[prof].iLevel <= ownlevel or ownlevel == 10 then
					-- check if profile is allready used
					if tAccounts[lnick].sProfile == prof then
						return false, Lang(ssid, 0, "User is already registered with that profile")
					else
						local oldprof = tAccounts[lnick].sProfile
						tAccounts[lnick].sProfile = prof
						tAccounts[lnick].iLevel = tProfiles[prof].iLevel
						tAccounts[lnick].iLastChange = os.time()
						tAccounts[lnick].iLastChangeBy = (tClient[ssid] and tClient[ssid].sNick or "Hub")

						--// Check to see if user is online
						local vsid = tNick:GetSid(vnick)
						if vsid then 
							local oldlevel = tClient[vsid].iProfile or -1
							tClient[vsid].sProfile = prof
							tClient[vsid].iProfile = tProfiles[prof].iLevel
							tClient[vsid].bOperator = tProfiles[prof].bOperator or (tProfiles[prof].iLevel == 10 and true)
							tClient[vsid].bHasKey = tProfiles[prof].bHasKey
							tClient[vsid].bHasFakeKey = tProfiles[prof].bHasFakeKey
							local ct = tInfo:SetCT(tClient[vsid], vsid)
							tInfo:CreateADC(tClient[vsid], vsid)
							tInfo:CreateNMDC(tClient[vsid], vsid)

							--// Announce the change to all users
							tUser:Quit(vsid, true)
							tUser:Announce(vsid)

							tSend:OwnKey(vsid, true)
							tSend:UpdateRooms(vsid)
							tSend:ADCUserIps(vsid, true)

							if (tClient[vsid].iProfile > oldlevel and tHubSettings.Main.bAnnounceHigherProfile) or (tClient[vsid].iProfile <= oldlevel and tHubSettings.Main.bAnnounceLowerProfile) then
								tSend:PmToUser("Your profile has been changed:\n\tNickname:\t"..vnick.."\n\tProfile:\t\t"..prof.."\n\tHubname:\t"..SetManager:Get("hub_name").."\n\tHub:\t\t"..tFuncs:GetAddress(), nil, vsid)
							end
						end

						tFuncs:Debug("[AccountManager] [SetProfile] Account for: "..vnick.." set from profile: "..oldprof.." to: "..prof.." by: "..(ssid~="AAAA" and tUser:GetNick(ssid) or "FlexHub Console"), "FI", -1)
						tFuncs:SaveSettings("accounts")

						return true, Lang(ssid, 0, "Profile of <%s> changed from %s to %s", vnick, oldprof, prof)
					end
				else
					return false, Lang(ssid, 0, "You can only change to profiles lower than your own profile.")
				end
			else
				return false, Lang(ssid, 0, "Error: Profile doesn't exist, use !showprofiles for available profiles.")
			end
		elseif vnick then
			return false, Lang(ssid, 0, "User isn't registered yet, use !addreg to register")
		else
			return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <profilename>")
		end	
	end,

	["GetProfileAccounts"] = function(self, ssid, prof)
		local prof, err = ProfileManager:GetName(ssid, prof)
		if not prof then
			return false, err
		end

		local profacc = {}
		for nick, data in pairs(tAccounts) do
			if data.sProfile:lower() == prof:lower() then
				profacc[nick] = data
			end
		end
		if profacc ~= {} then
			return profacc
		else
			return false, "No accounts found with profile: "..prof
		end
	end,

	["FindNoPassAccounts"] = function(self, ssid)
		local tbl = {}
		local count = 0

		for vnick, reg in pairs(tAccounts) do
			if reg.sPassword and reg.sPassword == "" then
				count = count + 1
				table.insert(tbl, "\t"..PaddingT(reg.sNick,24).."\t"..(ToDate(reg.iRegDate) or "<not specified>").."\t"..(ToDate(reg.iLastLogin) or "<never logged in>").."\t"..(reg.sBindIP or "<Not bound to IP>").."\t"..(reg.sChangedBy or reg.sReggedBy or "FlexHub"))
			if tSecurity.General.AllowBlankPass and tNick[reg.sNick] then
				local sid = tNick:GetSid(reg.sNick)
				if not tClient[sid].bNotifySetOwnPass then
					tClient[sid].bNotifySetOwnPass = true
				end
			end

			end
		end

		if count > 0 then
			return true, "Found the following "..count.." reg(s) without a password set:\n\n\tNick\t\tReg Date\t\t\tLast Login\tBound IP\t\tAdded/Changed by\n\t"..string.rep("-",180).."\n"..table.concat(tbl,"\n"), count
		else
			return true, "No accounts with empty passwords found."
		end
	end,
}
