--// $Id: Hub-Escape.lua 1314 2012-05-08 19:16:53Z en_dator $
--[[
	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.
]]--


tHubFiles["Hub-Escape.lua"] = {
	["sRev"] = "$Rev: 1314 $",
	["bRestart"] = false,
	["sPath"] = "./core/Hub-Escape.lua",
	}

tEsc = {

["Escape"] = function(self,sText, invert)
	-- Windows escape. Nevermind if it appears crappy on linux, it doesn't need it.
	tWinChars = {
	["€"] = "", ["‑"] = "", ["„"] = "", ["…"] = "", [" "] = "", ["‡"] = "", ["‰"] = "", [" "] = "",
	["‹"] = "", ["ő"] = "", ["Ť"] = "", ["Ž"] = "", ["Ź"] = "", ["‘"] = "", ["’"] = "", ["“"] = "",
	["”"] = "", ["•"] = "", ["–"] = "", ["—"] = "", ["™"] = "", ["š"] = "", ["›"] = "", ["ś"] = "",
	["ť"] = "", ["ž"] = "", ["ź"] = "", ["ˇ"] = "", ["˘"] = "", ["Ł"] = "", ["Ą"] = "", ["Ş"] = "", ["Ż"] = "",
	["˛"] = "", ["ł"] = "", ["ą"] = "", ["ş"] = "", ["Ľ"] = "", ["˝"] = "", ["ľ"] = "", ["ż"] = "", ["Ŕ"] = "",
	["Á"] = "", ["Â"] = "", ["Ă"] = "", ["Ä"] = "", ["Ĺ"] = "", ["Ć"] = "", ["Ç"] = "", ["Č"] = "", ["É"] = "",
	["Ę"] = "", ["Ë"] = "", ["đ"] = "", ["Í"] = "", ["Î"] = "", ["Ď"] = "", ["Đ"] = "", ["Ń"] = "", ["Ň"] = "",
	["Ó"] = "", ["Ô"] = "", ["Ő"] = "", ["Ö"] = "", ["×"] = "", ["Ř"] = "", ["Ů"] = "", ["Ñ"] = "", ["Ű"] = "",
	["Ü"] = "", ["Ý"] = "", ["Ţ"] = "", ["ß"] = "", ["ŕ"] = "", ["á"] = "", ["â"] = "", ["ă"] = "", ["ä"] = "",
	["ĺ"] = "", ["ć"] = "", ["ç"] = "", ["č"] = "", ["é"] = "", ["ę"] = "", ["ë"] = "", ["ě"] = "", ["í"] = "",
	["î"] = "", ["ď"] = "", ["đ"] = "", ["ń"] = "", ["ň"] = "", ["ó"] = "", ["ô"] = "", ["ő"] = "", ["ö"] = "",
	["÷"] = "", ["ř"] = "", ["ů"] = "", ["ú"] = "", ["ű"] = "", ["ü"] = "", ["ý"] = "", ["ţ"] = "", ["˙"] = ""
	}
	-- Linux escape. It should appear correctly (each value has to appear one character). If not, convert to UTF8 w/o BOM
	tNixChars = {
	[128] = "",[129] = "",[130] = "",[131] = "?",[132] = "",[133] = "",[134] = "",[135] = "",[136] = "?",
	[137] = "",[138] = "",[139] = "",[140] = "S",[141] = "T",[142] = "",[143] = "Z",[144] = "",[145] = "",
	[146] = "",[147] = "",[148] = "",[149] = "",[150] = "",[151] = "",[152] = "?",[153] = "",[154] = "",
	[155] = "",[156] = "s",[157] = "t",[158] = "",[159] = "z",[160] = "",[161] = "?",[162] = "?",[163] = "L",
	[164] = "",[165] = "A",[166] = "",[167] = "",[168] = "",[169] = "",[170] = "S",[171] = "",[172] = "",
	[173] = "",[174] = "",[175] = "Z",[176] = "",[177] = "",[178] = "?",[179] = "l",[180] = "",[181] = "",
	[182] = "",[183] = "",[184] = "",[185] = "a",[186] = "s",[187] = "",[188] = "L",[189] = "?",[190] = "l",
	[191] = "z",[192] = "R",[193] = "",[194] = "",[195] = "A",[196] = "",[197] = "L",[198] = "C",[199] = "",
	[200] = "C",[201] = "",[202] = "E",[203] = "",[204] = "E",[205] = "",[206] = "",[207] = "D",[208] = "",
	[209] = "N",[210] = "N",[211] = "",[212] = "",[213] = "O",[214] = "",[215] = "",[216] = "R",[217] = "U",
	[218] = "",[219] = "U",[220] = "",[221] = "",[222] = "T",[223] = "",[224] = "r",[225] = "",[226] = "",
	[227] = "a",[228] = "",[229] = "l",[230] = "c",[231] = "",[232] = "c",[233] = "",[234] = "e",[235] = "",
	[236] = "e",[237] = "",[238] = "",[239] = "d",[240] = "d",[241] = "n",[242] = "n",[243] = "",[244] = "",
	[245] = "o",[246] = "",[247] = "",[248] = "r",[249] = "u",[250] = "",[251] = "u",[252] = "",[253] = "",
	[254] = "t",[255] = "?",
	}
	if not invert then
		if sText and _OS == "Windows" then
			for k,v in pairs(tWinChars) do
				sText = sText:gsub(k,v)
			end
			sText = sText:gsub("(.)","%1")
		elseif sText then
			sText = string.gsub(sText,".",function(c)
				if string.byte(c) >= 128 then
					return tNixChars[string.byte(c)] or c
				else
					return c
				end
			end)
		else
			WriteDebug("Escape error: sText = nil")
			sText = ""
		end
	else
		if sText and _OS == "Windows" then
			for k,v in pairs(tNixChars) do
				sText = sText:gsub(v,k)
			end
--            sText = sText:gsub("Â(.)","%1")      -- dunno if this will work when commented out... don't know how to revert it
		elseif sText then
			tFuncs:Debug("Linux reverse escaping not implemented yet")
		else
			WriteDebug("Escape error: sText = nil")
			sText = ""
		end
	end
--~ 	collectgarbage("collect")
	return sText
end,
}

--[[  old
["Escape"] = function(self,sText, invert)
	-- Windows escape. Nevermind if it appears crappy on linux, it doesn't need it.
	tWinChars = {
	["€"] = "", ["‚"] = "", ["„"] = "", ["…"] = "", ["†"] = "", ["?"] = "", ["‰"] = "", ["L"] = "",
	["a"] = "", ["L"] = "S", ["L"] = "T", ["L?"] = "", ["La"] = "Z", ["?"] = "", ["’"] = "", ["s"] = "",
	["t"] = "", ["?"] = "", ["–"] = "", ["—"] = "", ["?"] = "", ["L?"] = "", ["s"] = "", ["L"] = "s",
	["LA"] = "t", ["Ll"] = "", ["Ls"] = "z", ["ˇ"] = "?", ["?"] = "?", ["L"] = "L", ["Ą"] = "A", ["L"] = "S", ["L"] = "Z",
	["˛"] = "?", ["L"] = "l", ["ą"] = "a", ["Lz"] = "s", ["?"] = "L", ["t"] = "?", ["l"] = "l", ["LL"] = "z", ["L"] = "R",
	["A"] = "", ["A"] = "", ["Ă"] = "A", ["A"] = "", ["a"] = "L", ["Ć"] = "C", ["A"] = "", ["S"] = "C", ["A"] = "",
	["?"] = "E", ["A"] = "", ["Ě"] = "E", ["AT"] = "", ["A"] = "", ["Ď"] = "D", ["Đ"] = "", ["L?"] = "N", ["L"] = "N",
	["A"] = "", ["A"] = "", ["L"] = "O", ["A"] = "", ["A"] = "", ["L?"] = "R", ["L"] = "U", ["A"] = "", ["L"] = "U",
	["As"] = "", ["At"] = "", ["L?"] = "T", ["Az"] = "", ["L"] = "r", ["A?"] = "", ["A?"] = "", ["?"] = "a", ["A"] = "",
	["s"] = "l", ["ć"] = "c", ["A"] = "", ["T"] = "c", ["A"] = "", ["ę"] = "e", ["A"] = "", ["ě"] = "e", ["A"] = "",
	["A"] = "", ["Z"] = "d", ["đ"] = "d", ["L"] = "n", ["L?"] = "n", ["Al"] = "", ["A"] = "", ["L"] = "o", ["A"] = "",
	["A"] = "", ["L"] = "r", ["LZ"] = "u", ["As"] = "", ["L"] = "u", ["AL"] = "", ["A?"] = "", ["LL"] = "t", ["˙"] = "?"
	}
	-- Linux escape. It should appear correctly (each value has to appear one character). If not, convert to UTF8 w/o BOM
	tNixChars = {
	[128] = "",[129] = "",[130] = "",[131] = "?",[132] = "",[133] = "",[134] = "",[135] = "",[136] = "?",
	[137] = "",[138] = "",[139] = "",[140] = "S",[141] = "T",[142] = "",[143] = "Z",[144] = "",[145] = "",
	[146] = "",[147] = "",[148] = "",[149] = "",[150] = "",[151] = "",[152] = "?",[153] = "",[154] = "",
	[155] = "",[156] = "s",[157] = "t",[158] = "",[159] = "z",[160] = "",[161] = "?",[162] = "?",[163] = "L",
	[164] = "",[165] = "A",[166] = "",[167] = "",[168] = "",[169] = "",[170] = "S",[171] = "",[172] = "",
	[173] = "",[174] = "",[175] = "Z",[176] = "",[177] = "",[178] = "?",[179] = "l",[180] = "",[181] = "",
	[182] = "",[183] = "",[184] = "",[185] = "a",[186] = "s",[187] = "",[188] = "L",[189] = "?",[190] = "l",
	[191] = "z",[192] = "R",[193] = "",[194] = "",[195] = "A",[196] = "",[197] = "L",[198] = "C",[199] = "",
	[200] = "C",[201] = "",[202] = "E",[203] = "",[204] = "E",[205] = "",[206] = "",[207] = "D",[208] = "",
	[209] = "N",[210] = "N",[211] = "",[212] = "",[213] = "O",[214] = "",[215] = "",[216] = "R",[217] = "U",
	[218] = "",[219] = "U",[220] = "",[221] = "",[222] = "T",[223] = "",[224] = "r",[225] = "",[226] = "",
	[227] = "a",[228] = "",[229] = "l",[230] = "c",[231] = "",[232] = "c",[233] = "",[234] = "e",[235] = "",
	[236] = "e",[237] = "",[238] = "",[239] = "d",[240] = "d",[241] = "n",[242] = "n",[243] = "",[244] = "",
	[245] = "o",[246] = "",[247] = "",[248] = "r",[249] = "u",[250] = "",[251] = "u",[252] = "",[253] = "",
	[254] = "t",[255] = "?",
	}
	if not invert then
		if sText and _OS == "Windows" then
			for k,v in pairs(tWinChars) do
				sText = sText:gsub(k,v)
			end
			sText = sText:gsub("Â(.)","%1")
		elseif sText then
			sText = string.gsub(sText,".",function(c)
				if string.byte(c) >= 128 then
					return tNixChars[string.byte(c)] or c
				else
					return c
				end
			end)
		else
			WriteDebug("Escape error: sText = nil")
			sText = ""
		end
	else
		if sText and _OS == "Windows" then
			for k,v in pairs(tNixChars) do
				sText = sText:gsub(v,k)
			end
--            sText = sText:gsub("Â(.)","%1")      -- dunno if this will work when commented out... don't know how to revert it
		elseif sText then
			tFuncs:Debug("Linux reverse escaping not implemented yet")
		else
			WriteDebug("Escape error: sText = nil")
			sText = ""
		end
	end
--~ 	collectgarbage("collect")
	return sText
end,
--]]