--// $Id: Hub-Funcs.lua 1468 2013-10-19 18:44:31Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011-2012 Daywalker, FlipFlop and en_dator

	dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.

]]--

tHubFiles["Hub-Funcs.lua"] = {
	["sRev"] = "$Rev: 1468 $",
	["bRestart"] = false,
	["sPath"] = "./core/Hub-Funcs.lua",
	}

--~ function Duration(stime)
--~ 	local t = os.date( "!*t", stime)
--~ 	return string.format( "%i day(s) %02i:%02i:%02i", stime / 86400, t.hour, t.min, t.sec )
--~ end

function LoadChunk(chunk)
	assert(loadstring(chunk))()
end

function SortNumber(a,b)
	a = tonumber(a) or 0
	b = tonumber(b) or 0
	if a < b then return true else return false end
end

function SortLowercase(a,b)
	a = tostring(a) or ""
	b = tostring(b) or ""
	if a:lower() < b:lower() then return true else return false end
end

function ToDate(num)
	num = tonumber(num) or os.time()
	if num and num > 0 then
		local dateformat = (tHubSettings and tHubSettings.Main and tHubSettings.Main.sDateTimeFormat) or "%x %X"
		return os.date(dateformat, num)
	else
		return false
	end
end

function Padding(txt, length)
	length = tonumber(length) or 0
	txt = tostring(txt)
	if length < #txt then
		return txt
	else
		length = length - #txt
		return txt..string.rep(" ", length)
	end
end

function PaddingT(txt, length, tabsize)
	local tabsize = tabsize or 8
	txt = tostring(txt)
	length = tonumber(length) or 0

	if length > #txt then
		length = length - #txt
		local t = math.floor(length/tabsize)
		length = length - (t*tabsize)
		txt = txt..string.rep(" ", length)..string.rep("\t", t)
	end
	return txt
end

function Clock()
	local timeformat = (tHubSettings and tHubSettings.Main and tHubSettings.Main.sTimeFormat) or "%X"
	return os.date("["..timeformat.."] ")
end

--~ function WriteDebug(msg)
--~ 	if tSettings.DebugToFile == 1 and tDebugFile then
--~ 		tDebugFile:write(Clock()..string.format("%.6f ", socket.gettime()).. ">> "..tostring(msg).."\n")
--~ 	end
--~ end

--~ function StartWriteDebug()
--~ 	if tSettings.DebugToFile == 1 then
--~ 		tDebugFile = io.open(_APPPATH.."debuginfo.txt","a+")
--~ 		if tDebugFile then
--~ 			print("*** Debugging to file started.")
--~ 			WriteDebug("*** APPLICATION STARTED ***")
--~ 		else
--~ 			print("*** Can't write to ".._APPPATH.."debuginfo.txt, debug to file disabled.")
--~ 		end
--~ 	end
--~ end

--~ if tSettings then StartWriteDebug() end

tFuncs = {

["iff"] = function(self, cond, a, b)
	return cond and a or b
end,

["DoMainTimer"] = function(self)
	if tHubSettings.tThreads.Checker then
		tFuncs:CheckListenThread()
	end
	if not _NOGUI then
		tGui:ExecuteConsoleEvent()
	end
end,

["UnlockMutexes"] = function(self, name)
	tFuncs:Debug("Unlocking mutexes...", "X", 0)
	if name then
		local res, err = pcall(ThreadUnlockMutex, _G[name])
		if not res then tFuncs:Debug("Error unlocking mutex "..tostring(name)..": "..err) return false, "Error unlocking mutex "..tostring(name)..": "..err end
		tFuncs:Debug("Unlocked mutex: "..name, "X", 0)
		return true, "Unlocked mutex: "..name
	else
		local ok = true
		for _,name in ipairs(tMutex) do
			local res, err = pcall(ThreadUnlockMutex, _G[name])
			if not res then ok = false tFuncs:Debug("Error unlocking mutex "..tostring(name)..": "..err) end
		end
		if ok then
			tFuncs:Debug("Unlocked all mutexes", "X", 0)
			return true, "Unlocked all mutexes"
		else
			tFuncs:Debug("An error occurred during unlocking of mutexes, not all mutexes could be unlocked.")
			return false, "Not all mutexes could be unlocked."
		end
	end
end,

["CheckShutdown"] = function(self)
	if _SHUTDOWN then
		if _NOW then
			_SHUTDOWN = nil
			tFuncs:StopHub(true)
			tFuncs:SaveSettings("*")
			OnStop()
				if _NOW == "r" and _OS == "Linux" then
					_NOW = nil
					local kill = io.popen (_CONFIGDIR.."RestartScript.sh")
				elseif _NOW == "r" and _OS == "Windows" and _NOGUI then --// When Flex is running as service
					_NOW = nil
					local kill = os.execute("cmd /c START \"FlexHub\" /B /D\"".._CONFIGDIR:gsub("/","\\").."\" RestartScript.bat\n")
				elseif _NOW == "r" then
					_NOW = nil
					local kill = io.popen ("START \"FlexHub\" /B /D\"".._CONFIGDIR:gsub("/","\\").."\" RestartScript.bat\n")
--					local kill = io.popen ("start /B \"".._APPPATH.."RestartScript.bat\"")
				end
			return false
		else
			_NOW = _SHUTDOWN
		end
	end
	return true
end,

["FreezeKiller"] = function(self, ctime, port)
	--// Kill listenthread
	ThreadDestroy(tThreads.Listen.Main)
	tThreads.Listen.Main = nil
	Test("*** Listenthread froze on status: "..tStats.Thread.Listen.Status.." Time: "..(ctime - (tStats.Thread.Listen.SocketTime or 0)))
	tFuncs:Debug("*** Listenthread froze on status: "..tStats.Thread.Listen.Status.." Time: "..(ctime - (tStats.Thread.Listen.SocketTime or 0)), "X", 2)

	Test("Listenthread killed.")

	tStats.Thread.Listen.SocketTime = false
	local sid = tStats.Thread.Listen.User
	if not port and sid and tClient[sid] then
		--// Gather data
		local xsocket = tClient[sid].xSocket
		local data = tUser:GetDataLine(sid)
		local stype = tClient[sid].bSecure and "Secure" or "Regular"

		Test(stype.." socket froze: "..data)

		--// Transfer socket to tKill table for handling by killer
		if not mutexkiller then mutexkiller = ThreadNewMutex() end
		ThreadLockMutex(mutexkiller)
		table.insert(tKill,{["xSocket"] = xsocket, ["data"] = data })
		ThreadUnlockMutex(mutexkiller)

		Test("Added socket to socketkiller table for user: "..data)
		tClient[sid].xSocket = nil	--// This will ensure the closing of socket will not be handled by ClientQuit

		--// Kill user
		Test("Killing user: "..data)
		tFuncs:ClientQuit(tClient[sid], sid)
		tClient[sid] = nil
		Test("User killed: "..data)
		tFuncs:Debug(data.."was disconnected because: "..stype.." socket not responding, user dropped.", nil, 2)

		--// Restart listenthread
		Test("Restarting listenthread...")
		tThreads.Listen.Main = ThreadCreate(tThread.TCPListen,{1})
		Test("Listenthread restarted.")
		tFuncs:Debug("Listenthread restarted.", "X", 0)
		tStats.Thread.Killer.DropCount = tStats.Thread.Killer.DropCount + 1

		--// Kill socket
		if not tThreads.Killer.Main then
			Test("Starting socketkiller...")
			tThreads.Killer.Main = ThreadCreate(tFuncs.FreezeKillerThread,{1})
		else
			--// Check if killer is running too long
			if ctime - (tStats.Thread.Killer.Time or 0) > 5 then	--// Check if killer is running too long
				Test("Socketkiller running too long, destroying socketkiller thread and creating a new one...")

				tStats.Thread.Killer.Failed = (tStats.Thread.Killer.Failed or 0) + (tStats.Thread.Killer.Pending or 0)
				ThreadLockMutex(mutexkiller)
				ThreadDestroy(tThreads.Killer.Main)
				ThreadUnlockMutex(mutexkiller)
				tStats.Thread.Killer.Status = "Killed"

				Test("Killed socketkiller, restarting...")
				tStats.Thread.Killer.KillCount = tStats.Thread.Killer.KillCount + 1
				tThreads.Killer.Main = ThreadCreate(tFuncs.FreezeKillerThread,{1})
			else
				Test("Socketkiller already running.")
			end

		end
	else
		Test("*** No user SID found, or frozen port thread triggered FreezeKiller.")
--~ 		tFuncs:UnlockMutexes()

		--// Kill socket
		if not tThreads.Killer.Main then
			Test("Starting socketkiller...")
			tThreads.Killer.Main = ThreadCreate(tFuncs.FreezeKillerThread,{1})
		else
			--// Check if killer is running too long
			if ctime - (tStats.Thread.Killer.Time or 0) > 5 then	--// Check if killer is running too long
				Test("Socketkiller running too long, destroying socketkiller thread and creating a new one...")

				tStats.Thread.Killer.Failed = (tStats.Thread.Killer.Failed or 0) + (tStats.Thread.Killer.Pending or 0)
				ThreadLockMutex(mutexkiller)
				ThreadDestroy(tThreads.Killer.Main)
				ThreadUnlockMutex(mutexkiller)
				tStats.Thread.Killer.Status = "Killed"

				Test("Killed socketkiller, restarting...")
				tThreads.Killer.Main = ThreadCreate(tFuncs.FreezeKillerThread,{1})
			else
				Test("Socketkiller already running.")
			end
		end

		--// Restart listenthread
		Test("Restarting listenthread...")
		tFuncs:UnlockMutexes()
		tThreads.Listen.Main = ThreadCreate(tThread.TCPListen,{1})
		Test("Listenthread restarted.")
		tFuncs:Debug("Listenthread restarted.", "X", 0)
	end
	Test("FreezeKiller in GUI thread ended.")
end,

["FreezeKillerThread"] = function(self)
	tStats.Thread.Killer.Status = "Running"
	Test("Socketkiller started.")

	ThreadLockMutex(mutexkiller)
	tStats.Thread.Killer.Time= socket:gettime()
	local kill = table.remove(tKill)
	ThreadUnlockMutex(mutexkiller)

	while kill do
		--// Kill socket
		tStats.Thread.Killer.Pending = (tStats.Thread.Killer.Pending or 0) +1
		tStats.Thread.Killer.Counter = (tStats.Thread.Killer.Counter or 0) +1

		Test("Attempting to kill socket: "..kill.data)
		if kill.xSocket then
			kill.xSocket:close()
			kill.xSocket = nil
		end

		Test("Socket killed: "..kill.data)
		tStats.Thread.Killer.CounterKilled = (tStats.Thread.Killer.CounterKilled or 0) +1
		tStats.Thread.Killer.Pending = (tStats.Thread.Killer.Pending or 1) -1

		tStats.Thread.Killer.Time= socket:gettime()
		ThreadLockMutex(mutexkiller)
		kill = table.remove(tKill)
		ThreadUnlockMutex(mutexkiller)
	end

	Test("Socketkiller thread finished, total killed: "..(tStats.Thread.Killer.CounterKilled or 0).." / Failed: "..(tStats.Thread.Killer.Failed or 0))

	--// Terminate this thread
	tThreads.Killer.Main = nil
	tStats.Thread.Killer.Status = "Finished"
	ThreadDestroy()
end,

["CheckListenThread"] = function(self)
	if tThreads.Listen.Main then
		local ctime = socket.gettime()

		--// Check if socket froze on read or write
		local sockettime = tStats.Thread.Listen.SocketTime
		local listenstatus = tStats.Thread.Listen.Status
		if sockettime and (ctime - sockettime) >= (tHubSettings.tThreads.TLSTimeout/1000) and (listenstatus == "Client-cycle: read from socket" or listenstatus == "Client-cycle: sending pending messages") then
			local res, err = pcall(tFuncs.FreezeKiller, 1, ctime)
			if not res then OnError(err) end
			ctime = socket.gettime()
		end

		--// Check if listenthread froze on deleting sockets
		local listentime = tStats.Thread.Listen.Time
		local listenstatus = tStats.Thread.Listen.Status
		if listentime and listentime > 0 and (ctime - listentime) >=  (tHubSettings.tThreads.TLSTimeout/1000) and listenstatus == "Deleting disconnected users" then
			Test("Listenthread froze on deleting of sockets. Status: "..(listenstatus or "unknown"))
			tFuncs:SafeDebug("Listenthread froze on deleting of sockets. Status: "..(listenstatus or "unknown"))
			--// Remove freezing sid from clientdelete table
			for num,data in pairs(tClientDelete) do
				if data.sid == tStats.Thread.Listen.User then
					tClientDelete[num] = nil
				end
			end
			--// Kill userthread and try to kill the usersocket
			local res, err = pcall(tFuncs.FreezeKiller, 1, ctime)
			if not res then OnError(err) end
			ThreadUnlockMutex(mutexdelete) --// IMPORTANT!  NOTE: killing the listenthread means unlocking all mutexes
			ctime = socket.gettime()
		end

		--// Check listenthread for other freezes
		local listentime = tStats.Thread.Listen.Time
		local listenstatus = tStats.Thread.Listen.Status
		if listentime and (ctime - listentime) >= tHubSettings.tThreads.ListenTimeout and (not tStats.Thread.Listen.LogsDumped or ctime > tStats.Thread.Listen.LogsDumped+600) then
			--// Listen thread has frozen, write stuff to an errorlog and restart the hub
			Test("Listenthread froze during: "..(listenstatus or "unknown"))
			tFuncs:SafeDebug("Listenthread froze during: "..(listenstatus or "unknown"))
			tStats.Thread.Listen.Time = nil
			local res, err = pcall(tFuncs.CoreDump, "CheckListenThread")
			if not res then OnError(err) end
			local res, err = pcall(tFuncs.FreezeKiller, 1, ctime)
			if not res then OnError(err) end
			ctime = socket.gettime()
		end

		--// Check portthreads
		if not tHubSettings.tThreads.OnePortThread then
			for port,data in pairs(tServer.Port) do
				if data.server then
					local porttime = tStats.Thread.Port[port].Time
					local portstatus = tStats.Thread.Port[port].Status
					if porttime and (ctime - porttime) >= (tHubSettings.tThreads.PortTimeout/1000) then
						Test("Portthread for port "..port.." froze during: "..(portstatus or "unknown").." / Data: "..(tostring(tStats.Thread.Port[port].Data) or "none"))
						tFuncs:SafeDebug("Portthread for port "..port.." running too long: "..(ctime - (porttime or 0)).." / Status: "..(portstatus or "unknown").." / Data: "..(tostring(tStats.Thread.Port[port].Data) or "none"))

						--// Kill port thread
						tFuncs:StopPort(port)

						--// Kill problem socket
						if tStats.Thread.Port[port].xSocket then
							--// Transfer socket to tKill table for handling by killer
							ThreadLockMutex(mutexkiller)
							table.insert(tKill,{["xSocket"] = tStats.Thread.Port[port].xSocket, ["data"] = (tostring(tStats.Thread.Port[port].Data) or "unknown") })
							ThreadUnlockMutex(mutexkiller)
							local res, err = pcall(tFuncs.FreezeKiller, 1, ctime, port)
							if not res then OnError(err) end
						end

						--// Restart port thread
						tFuncs:UnlockMutexes()
						tFuncs:StartPort(port, data.stype)

						ctime = socket.gettime()
					end
				end
			end
		else
			--// Check single portthread controlling all ports
		end

		--// Check SSL thread
		if tThreads.SSL.Handshake then
			local shaketime = tStats.Thread.SSL.Time
			if shaketime then shaketime = tonumber(shaketime) or ctime end
			if shaketime and (ctime - shaketime) >= (tHubSettings.tThreads.HandshakeTimeout/1000) then
				Test("SSL Handshake thread running too long: "..(ctime - shaketime).." / Status: "..(tStats.Thread.SSL.Status or "unknown").." / Data: "..(tStats.Thread.SSL.Data or "none"))
				tFuncs:SafeDebug("SSL Handshake thread running too long: "..(ctime - shaketime).." / Status: "..(tStats.Thread.SSL.Status or "unknown").." / Data: "..(tStats.Thread.SSL.Data or "none"))
				if tStats.Thread.SSL.Socket then
					--// Transfer socket to tKill table for handling by killer
					ThreadLockMutex(mutexkiller)
					table.insert(tKill,{["xSocket"] = tStats.Thread.SSL.Socket, ["data"] = (tStats.Thread.SSL.Data or "unknown") })
					ThreadUnlockMutex(mutexkiller)
					local res, err = pcall(tFuncs.FreezeKiller, 1, ctime, true)
					if not res then OnError(err) end
				end

				--// Kill SSL thread
				ThreadDestroy(tThreads.SSL.Handshake)
				tThreads.SSL.Handshake = nil

				--// Restart SSL thread
				tFuncs:UnlockMutexes()
				tThreads.SSL.Handshake = ThreadCreate(tThread.TCPSSL,{1})
			end
		end
	end
end,

["CoreDump"] = function(self, err)	-- make a forced dump of tables on request
	tFuncs:Debug("[CoreDump] "..(err or ""), "FI", -1)
	local ctime = socket.gettime()
	if (not tStats.Thread.Listen.LogsDumped or ctime > tStats.Thread.Listen.LogsDumped+600) then
		local userinfo = ""
		local sid = tStats.Thread.Listen.User
		if sid and sid ~= "" and tClient[sid] then
			userinfo = userinfo.."\n\t\tSID: "..sid..
			"\n\t\tNick: "..(tClient[sid].sNick or "<nick unknown>")..
			"\n\t\tIP: "..(tClient[sid].sIP or "<ip unknown>")..
			"\n\t\tProtocol: "..(tClient[sid].sProtocol or "<protocol unknown")..
			"\n\t\tState: "..(tClient[sid].iState or "unknown")..
			"\n\t\tHandler: "..(tClient[sid].xHandler or "unknown")..
			"\n"..
			((tCurrent and tCurrent.Main and "\n\tLast data [Main]: "..tCurrent.Main.Line) or "\n\tLast data [Main]: <data unknown>") ..
			((tCurrent and tCurrent.Login and "\n\tLast data [Login]: "..tCurrent.Login.Line) or "")
		end

		local ssl = ""
		if tStats.Thread.SSL.Status and tStats.Thread.SSL.Status ~= "Not running" then
			ssl = "\n\tSSL thread status:"..(tStats.Thread.SSL.Status or "unknown")..
			"\n\tSSL thread duration: "..((tStats.Thread.SSL.Time and (ctime-tStats.Thread.SSL.Time)) or "Finished")..
			"\n\tSSL thread client data: "..(tStats.Thread.SSL.Data or "unknown")
		end

		local res = "\n"..string.rep("=", 60)..
--		"\n"..Clock().."Error: Listenthread froze for longer than "..tHubSettings.tThreads.ListenTimeout.." seconds."..
		"\n\tListenthread status:"..(tStats.Thread.Listen.Status or unknown)..
		"\n\tListenthread starttime: "..(tStats.Thread.Listen.Time and ToDate(tStats.Thread.Listen.Time) or "unknown")..
		"\n\tListenthread client data: "..userinfo..
		"\n\n"..ssl..
		"\n"..string.rep("=", 60).."\n"

		--// Dump stuff
		tFuncs:Save_File(_LOGDIR.."Dump-tHubSettings.log", tHubSettings, "tHubSettings")
		tFuncs:Save_File(_LOGDIR.."Dump-tStats.log", tStats, "tStats")
		tFuncs:Save_File(_LOGDIR.."Dump-tLogs.log", tLogs, "tLogs")
		tFuncs:Save_File(_LOGDIR.."Dump-tListenBuffer.log", tListenBuffer, "tListenBuffer")
		tFuncs:Save_File(_LOGDIR.."Dump-tBuffer.log", tBuffer, "tBuffer")
		tFuncs:Save_File(_LOGDIR.."Dump-tTimers.log", tTimers, "tTimers")
		tFuncs:Save_File(_LOGDIR.."Dump-tThreads.log", tThreads, "tThreads")
		tFuncs:Save_File(_LOGDIR.."Dump-tServer.log", tServer, "tServer")
		tFuncs:Save_File(_LOGDIR.."Dump-tClient.log", tClient, "tClient")

		tStats.Thread.Listen.LogsDumped = ctime

		Test("CoreDump performed: "..res)
		tFuncs:SafeDebug("An error has occured. Dumped core logs to ".._LOGDIR.." STATUS:\n"..res, nil, -1)
	end
end,

["StoreUserinfo"]  = function(self, nicks, ips)
	if nicks then
		tFuncs:Save_File(_LOGDIR.."Hub-userinfo.log", tLogs.Nick, "tLogs.Nick")
--~ 		Test("Stored Hub-Userinfo.log")
	end
	if ips then
		tFuncs:Save_File(_LOGDIR.."Hub-ipinfo.log", tLogs.IP, "tLogs.IP")
--~ 		Test("Stored Hub-Ipinfo.log")
	end
end,

["FileSize"] = function(self, name) --// Relative to program folder
	local logfile = io.open(name, "r")
	if not logfile then
		return false, "File "..name.. " couldn't be opened for reading."
	end

	local len = logfile:seek("end")
	return len
end,

["FileBiggerThan"] = function(self, name, maxsize)
	local size = tFuncs:FileSize(name)
	if size and size > maxsize then
		return true, "Filesize for file: "..name.." = "..tFuncs:ConvertShare(size), size
	else
		return false, "Filesize for file: "..name.." = "..tFuncs:ConvertShare(size), size
	end
end,

["LoadUserinfo"] = function(self, nicks, ips)

	local userfile = _LOGDIR.."Hub-userinfo.log"
	local ipfile = _LOGDIR.."Hub-ipinfo.log"

	if tHubSettings.Main.bUserdataWarning then
		local bigger, msg = tFuncs:FileBiggerThan(userfile,(50*1048576))
		if bigger then
			Notify("Warning: "..msg..". Loading this takes a lot of memory. Close the hub and remove or rename this file to prevent it from loading, or  !set userdata_persistent false  and restart the hub.")
		end
		local bigger, msg = tFuncs:FileBiggerThan(ipfile,(50*1048576))
		if bigger then
			Notify("Warning: "..msg..". Loading this takes a lot of memory. Close the hub and remove or rename this file to prevent it from loading, or  !set userdata_persistent false  and restart the hub.")
		end
	end

--//	if nicks and tFuncs:CheckFileExists(_APPPATH..userfile) then
	if nicks and tFuncs:CheckFileExists(userfile) then
		local res, err = tFuncs:TryFile(userfile)
		if not res then Notify("Loading stored userinfo failed: "..err, "LoadUserinfo") else Test("Loaded previously stored userinfo.") end
	else
		Test("File '"..userfile.."' not found, skipping.")
	end
--//	if ips and tFuncs:CheckFileExists(_APPPATH..ipfile) then
	if ips and tFuncs:CheckFileExists(ipfile) then
		local res, err = tFuncs:TryFile(ipfile)
		if not res then Notify("Loading stored ipinfo failed: "..err, "LoadUserinfo") else Test("Loaded previously stored ipinfo.") end
	else
		Test("File '"..ipfile.."' not found, skipping.")
	end
	return true
end,

["CheckFolder"] = function(self, parent, name)
	local curdir = lfs.currentdir()
	local path = (parent or "")..(name or "")
	local res, err = lfs.chdir(path)
	if not res then
		local res2, err2 = lfs.mkdir(path)
		if not res2 then
			lfs.chdir(curdir)
			return false, "Directory "..path.." doesn't exist and couldn't be created: "..err2
		else
			lfs.chdir(curdir)
			return true, "Directory "..path.." didn't exist, successfully created."
		end
	else
		lfs.chdir(curdir)
		return true, "Directory "..path.." exists."
	end
end,

["GetKey"] = function(self, stable, key)
	if type(stable) ~= "table" or not key then return false end
	key = key:lower()
	for name, val in pairs(stable) do
		if name:lower() == key then
			return name
		end
	end
	return false
end,

["FlatTable"] = function(self, stable, pre, sep, suf, num)
	local reply = ""
	if stable and #stable > 0 then
		for a=0, math.floor(#stable/num) do
			if a*num+1 > #stable then return reply end
			local xuntil = a*num+num
			if xuntil > #stable then xuntil = #stable end
			reply = reply..pre..(table.concat(stable, sep, (a*num+1), xuntil))..suf
		end
	else
		return ""
	end
	return reply
end,

["TableCopy"] = function(self, object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end,

["AddDisconnect"] = function(self, sid, reason, rreason, address)
	local client = tClient[sid]
	local redirect = ""
	if address then redirect = " \tRedirected: "..address end
	if not rreason then rreason = (reason or "") end

	for num, data in ipairs(tLogs.Disconnects) do
		if data[1] == client.sIP then
			tLogs.Disconnects[num] = {client.sIP, ToDate(os.time()).." \t"..tUser:GetDataLine(sid).."\tReason: "..tostring(rreason)..redirect, data[3]+1}
			return
		end
	end
	tLogs.Disconnects[#tLogs.Disconnects+1] = {client.sIP, ToDate(os.time()).." \t"..tUser:GetDataLine(sid).."\tReason: "..tostring(rreason)..redirect, 1}
	if #tLogs.Disconnects > 50 then table.remove(tLogs.Disconnects, 1) end
end,

["SetLocale"] = function(self, locale)
	-- // DO NOT change os.setlocale(xx, "numeric"), it needs to be set to "C" at all times!!
	local current = os.setlocale(nil, "collate")
	local res = os.setlocale(locale, "collate")
	if not res then
--~ 		tFuncs:Debug("System doesn't support locale: "..locale, "FI", -1)
		return false, "System doesn't support locale: "..locale
	end
	res = os.setlocale(locale, "ctype")
	res = os.setlocale(locale, "time")
	return true, "Changed locale from: "..(current or "unknown").." to: "..(os.setlocale(nil, "collate") or "unknown")
end,

["SetCharset_old"] = function(self, charset)
	local oldcharset = tHubSettings.Main.sCharset or "windows-1252"
	local charset = charset or tHubSettings.Main.sCharset

	xConvToUTF8Test = iconv.new("utf-8", "utf-8")
	xConvToUTF8NC   = iconv.new("utf-8//TRANSLIT//IGNORE", charset)
	xConvToLocaleNC = iconv.new(charset.."//TRANSLIT//IGNORE", "utf-8")
	if not xConvToUTF8NC or not xConvToLocaleNC then
		xConvToUTF8NC = iconv.new("utf-8//TRANSLIT//IGNORE", oldcharset) or iconv.new("utf-8", "utf-8")		--// Linux missing charsets workaround
		xConvToLocaleNC = iconv.new(oldcharset.."//TRANSLIT//IGNORE", "utf-8") or iconv.new("utf-8", "utf-8")	--// Linux missing charsets workaround
		return false, "Function 'iconv' doesn't support charset: "..charset
	else
--~ 		SetManager:Set("char_set", charset)
		return true
	end
end,

["SetCharset"] = function(self, charset)
	local oldcharset = tHubSettings.Main.sCharset or "windows-1252"
	local charset = charset or tHubSettings.Main.sCharset

	xConvToUTF8Test = iconv.new("utf-8", "utf-8")
	if not xConvToUTF8Test then Notify("Error creating test charset descriptor utf-8 to utf-8") end

	xConvToUTF8NC = iconv.new("utf-8//TRANSLIT", charset)
	if not xConvToUTF8NC then Notify("Error creating charset descriptor "..charset.." to utf-8") end

	xConvToLocaleNC = iconv.new(charset.."//TRANSLIT", "utf-8")
	if not xConvToLocaleNC then Notify("Error creating charset descriptor: utf-8 to "..charset) end

	if not xConvToUTF8NC or not xConvToLocaleNC then
		xConvToUTF8NC = iconv.new("utf-8//TRANSLIT", oldcharset) or iconv.new("utf-8", "utf-8")		--// Linux missing charsets workaround
		xConvToLocaleNC = iconv.new(oldcharset.."//TRANSLIT", "utf-8") or iconv.new("utf-8", "utf-8")	--// Linux missing charsets workaround
		return false, "Function 'iconv' doesn't support charset: "..charset
	else
--~ 		SetManager:Set("char_set", charset)
		return true
	end
end,

["MakeSalt"] = function(self)
	local salt = ""
	for i=1,24 do
		local num = math.random(1,32)
		salt = salt..string.sub("ABCDEFGHIJKLMNOPQRSTUVWXYZ234567", num,num)
	end
	return salt
end,

["Find"] = function(self, ttable, pattern, from, to)
    local from = from or 1
    local to = to or #ttable
	if #ttable and #ttable > 0 then
		for a = from,to do
		local value = ttable[a]
			--local b,c, result = string.find(value, pattern)
			local b, c, result = value:find(pattern)
			if b then
				if result then
					return result, a	--// Return capture and tablenumber
				else
					return value, a	--// Return tablevalue and tablenumber
				end
			end
		end
	elseif ttable == {} then
		return false, "Table is empty"
	else
		return false, "Can't use table:find() on indexed tables"
	end
    return false, "Value not found"
end,

["Split"] = function(self, txt, delimiter)
	local delimiter = delimiter or " "
	local result = { }
	local from = 1
	local txt = tostring(txt) or ""
	local delim_from, delim_to = string.find(txt, delimiter, from)
	while delim_from do
		table.insert(result, string.sub( txt, from , delim_from-1 ))
		from = delim_to + 1
		delim_from, delim_to = string.find(txt, delimiter, from)
	end
	table.insert(result, string.sub( txt, from ))
	return result
end,

["SaveSettings"] = function(self, name)		--// Example: tFuncs:SaveSettings("accounts")
	if not name then return false, "[tFuncs:SaveSettings] [ERROR] No filename specified." end
	local hubaddy = tHubSettings and tHubSettings.Main and tHubSettings.Main.sAddress or "unconfigured"
	local charset = tHubSettings and tHubSettings.Main and tHubSettings.Main.sCharset or "unconfigured"
	local comment = "svn:".._REV.." / hub:"..hubaddy.." / charset:"..charset.." / date:"..ToDate(os.time())

	local files = {
		["accounts"] = {"HubAccounts.ini", "tAccounts"},
		["bans"] = {"HubBans.ini", "tBans"},
		["blocks"] = {"HubBlocks.ini", "tBlocks"},
		["chatrooms"] = {"HubChatRooms.ini", "tChatRooms"},
		["privaterooms"] = {"HubPrivateRooms.ini", "tPrivateRooms"},
		["commands"] = {"HubCommands.ini", "tCommands"},
		["filters"] = {"HubFilters.ini", "tFilters"},
		["patterns"] = {"HubPatterns.ini", "tPatterns"},
		["profiles"] = {"HubProfiles.ini", "tProfiles"},
		["protects"] = {"HubProtects.ini", "tProtects"},
		["settings"] = {"HubSettings.ini", "tHubSettings"},
		["security"] = {"HubSecurity.ini", "tSecurity"},
	}

	if files[name] or files[name:lower()] then
		name = name:lower()
		local filename = files[name][1]
		local tablename = files[name][2]
		tFuncs:Save_File(_CONFIGDIR..filename, _G[tablename], tablename, comment)
		tFuncs:Debug("[SaveSettings] Settings "..filename.." saved.", nil, 2)
	elseif name == "*" then
		--// When flags are used, * is the exception, this requires an immediate rewrite, no timer here!
		for fname, data in pairs(files) do
			local filename = data[1]
			local tablename = data[2]
			tFuncs:Save_File(_CONFIGDIR..filename, _G[tablename], tablename, comment)
			tFuncs:Debug("[SaveSettings] Settings "..filename.." saved.", nil, 2)
		end
		tFuncs:Debug("Saved all settings.")
	elseif name then		--// Try to find the file by tablename (files[name][2])
		local found = false
		for fname, data in pairs(files) do
			if name == data[2] then
				found = true
				local filename = data[1]
				local tablename = data[2]
				tFuncs:Save_File(_CONFIGDIR..filename, _G[tablename], tablename, comment)
				tFuncs:Debug("[SaveSettings] Settings "..filename.." saved.", nil, 2)
			end
		end
		if not found then
			OnError("[tFuncs:SaveSettings] [ERROR] Can't find file to save for tablename: "..name..".")
		else
			return true
		end
	else
		return false, "[tFuncs:SaveSettings] [ERROR] Nothing to save."
	end
end,

--// Logging functions
["OpenLogFile"] = function (self, name)
	local forbidden = "*?\"()"
	if not _NOGUI then
		forbidden = wx.wxFileName.GetForbiddenChars()
	end
	local filename = string.gsub(name, "["..forbidden.."]", "_")

	local file = io.open(_LOGDIR..filename..".log", "a+")
	if not file then
		local err = "Couldn't open file '"..filename.." for logging."
		Notify("[tFuncs:OpenLogFile] [ERROR] "..err)
--~ 		tFuncs:Debug("[tFuncs:OpenLogFile] [ERROR] "..err, "FI", -1)
		tFiles[name] = { ["Open"] = false, ["File"] = false }
		return false, err
	else
		tFiles[name] = { ["Open"] = true, ["File"] = file }
		return true
	end
end,

["ValidFilename"] = function(self, name, noext, fix)
	if not name or name=="" then return false, "No filename specified." end
	local allowed = "%w_-"
	for char in name:gmatch(".") do
		if not char:find(allowed) then return false, "Wrong filename format, use alphanumeric characters and underscore only." end
	end
	return true, name
end,

["WriteFile"] = function (self, name, data)
	if not tFiles[name] or not tFiles[name].Open then
		if not tFuncs:OpenLogFile(name) then
			return false
		end
	end
	if tFiles[name] and tFiles[name].Open then
		tFiles[name].File:write("["..ToDate(os.time()).."] "..data)
		tFiles[name].File:flush()
		return true
	else
		return false
	end
end,

["CloseFile"] = function(self, name)
	if tFiles[name] and tFiles[name].Open then
		tFiles[name].File:close()
		tFiles[name] = { ["Open"] = false, ["File"] = false }
		return true
	else
		return false
	end
end,

["CloseFiles"] = function(self)
	for name, data in pairs(tFiles) do
		if data.Open then
			tFuncs:CloseFile(name)
		end
	end
	return true
end,

["GetHash"] = function(self, name, value)
	if name and value then
		tSerial = ""
		return tth.buffer(tFuncs:Build_SerialTable(value, name))
	else
		return 0
	end
end,

["GetHashFromFile"] = function(self, fullpath, bFlag)
	local buff = ""
	local a = io.open(fullpath, "rb")
	if a then
		if bFlag then
			buff = a:read("*a")
		end
		a:close()
		local attr = lfs.attributes(fullpath).mode
		if attr and attr == "file" then
			return tth.file(fullpath), buff
		end
	end
	return false
end,

["GetSerialized"] = function(self, tablename, vtable)
	if tablename and tablename ~= "" and vtable then
		tSerial = ""
		return tFuncs:Build_SerialTable(vtable, tablename)
	else
		return ""
	end
end,

["GetTable"]  = function(self, name)
	if name then
		if name == "BA" then return tBans
			elseif name == "BL" then return tBlocks
			elseif name == "CR" then return tChatRooms
			elseif name == "CP" then return tPrivateRooms
			elseif name == "CO" then return tCommands
			elseif name == "PA" then return tPatterns
			elseif name == "PR" then return tProfiles
			elseif name == "PT" then return tProtects
			elseif name == "RE" then return tAccounts
			elseif name == "SE" then return tHubSettings
		end
	end
	return false
end,

["AddLog"] = function(self, logname, subname, msg, fromsid, fromnick, stime)
--~ 	if logname then return end -- temp disabled
	--// tFuncs:AddLog("ChatRooms", "CCCC", msg, fromsid, tosid, stime)

	--// Make log local
	local xlog = false
	if not subname then
		xlog = tLogs[logname] or { ["Lines"] = {}, ["Sids"] = {}, ["Nicks"] = {}, ["Dates"] = {} }
	elseif subname then
		if not tLogs[logname] then
			tLogs[logname] = {}
			xlog = { ["Lines"] = {}, ["Sids"] = {}, ["Nicks"] = {}, ["Dates"] = {} }
		elseif not tLogs[logname][subname] then
			xlog = { ["Lines"] = {}, ["Sids"] = {}, ["Nicks"] = {}, ["Dates"] = {} }
		else
			xlog = tLogs[logname][subname]
		end
	end

	if not xlog or xlog == {} then xlog = { ["Lines"] = {}, ["Sids"] = {}, ["Nicks"] = {}, ["Dates"] = {} } end
--~ 	tFuncs:Debug("xlog has type: "..tostring(xlog))

	--// Add log
	table.insert(xlog.Lines, 1, (msg or "<empty message>"))
	table.insert(xlog.Sids, 1, (fromsid or "????"))
	table.insert(xlog.Nicks, 1, (fromnick or "unknown nick"))
	table.insert(xlog.Dates, 1, (stime or os.time()))
	while #xlog.Lines > 500 do
		table.remove(xlog.Lines)
		table.remove(xlog.Sids)
		table.remove(xlog.Nicks)
		table.remove(xlog.Dates)
	end

	--// Determine filename
	local name = logname

	if name == "ChatRooms" and tClient[subname] then
		name = "ChatRoom-"..tClient[subname].sNick
	elseif subname then
		name = logname.."-"..subname
	end

	if subname == "CCCC" then name = "Feed" end
	if subname == "BBBB" then name = "Opchat" end

	--// Save to logfile
	if logname == "ChatRooms" and subname and subname == "CCCC" and tHubSettings.Main.Logs.feed then
--		tFuncs:WriteFile(name, "<"..(fromnick or "unknown nick").."> "..(msg or "").."\n")
		tFuncs:WriteFile(name, "<"..(fromnick or "FlexBot").."> "..(msg or "").."\n")
	elseif logname == "ChatRooms" and subname and subname == "BBBB" and tHubSettings.Main.Logs.opchat then
		tFuncs:WriteFile(name, "<"..(fromnick or "unknown nick").."> "..(msg or "").."\n")
	elseif logname == "ChatRooms" and tHubSettings.Main.Logs.rooms then
		tFuncs:WriteFile(name, "<"..(fromnick or "unknown nick").."> "..(msg or "").."\n")
	elseif logname == "Mainchat" and tHubSettings.Main.Logs.mainchat then
		tFuncs:WriteFile(name, "<"..(fromnick or "unknown nick").."> "..(msg or "").."\n")
		--// Send mainchat to webchat server >> needs to be moved to separate thread with timer
		if tHubSettings.Webchat.bEnable then
			local _,nick = tFuncs:ToUTF8(fromnick or "unknown nick")
			local _,text = tFuncs:ToUTF8(msg)
			local a,b = http.request(tHubSettings.Webchat.sAddress.."webchat_main_log.php?nick="..url.escape(nick).."&text="..url.escape(text))
			tFuncs:Debug("[WebChat] Mainchat line sent to webserver, result: "..tostring(a).." / "..tostring(b))
		end
	end

	--// Put log back in global
	if not subname then
		tLogs[logname] = xlog
	else
		tLogs[logname][subname] = xlog
	end
end,

["GetFlags"] =  function(self, num)		--// Returns an _indexed_ table containing all bitwise flags
	local h = 2^(math.ceil(math.log(num)/math.log(2)))
	local result = {}
	while h >= 1 do
		if math.mod(num, h*2) >= tonumber(h) then
			result[h] = true
		end
		h = h/2
	end
        return result
end,

["CheckFlag"] = function(self, value, test)		--// Bitwise AND
	if value and test then
		return math.mod(value, test*2) >= tonumber(test)
	else
		tFuncs:Debug("[CheckFlag] [ERROR] No valid values supplied")
		return false
	end
end,

["GetSvn"] = function(self)
	local new = 0
	for name, data in pairs(tHubFiles) do
		local _, _, val = string.find((data.sRev or ""), "(%d+)")
		--tFuncs:Debug(name.."\t"..tostring(val), "X", 2)
		if val and tonumber(val) > new then
			new = tonumber(val)
			--tFuncs:Debug("is highest", "X", 2)
		end
	end

	if tSettings.General.svnversion < new then
		tSettings.General.svnpublic = false
		_REV = new
		tSettings.General.svnversion = new
	else
		tSettings.General.svnpublic = true
		_REV = tSettings.General.svnversion
	end

	return _REV
end,

["ToUTF8_orginal"] = function(self, msg, check)
	if msg then
		local cmsg, err = false,false
		if check then
			cmsg, err = xConvToUTF8:iconv(msg)
		else
			cmsg, err = xConvToUTF8NC:iconv(msg)
		end

		if not err then
			return true, cmsg
		else
			local errmsg = "There was an unknown error: "..err
			if err == iconv.ERROR_NO_MEMORY then
				errmsg = "Failed to allocate enough memory in the conversion process"
			elseif err == iconv.ERROR_INVALID then
				errmsg = "An invalid character was found in the input sequence. "
			elseif err == iconv.ERROR_INCOMPLETE then
				errmsg = "An incomplete character was found in the input sequence."
			end
--~ 			tFuncs:Debug("[ToUTF8] Error: "..errmsg.."\nUnconverted string: \t"..msg.."\nConverted string: \t"..cmsg, "FI", 1)
			return false, "[ToUTF8] Error: "..errmsg.."\nUnconverted string: \t"..msg.."\nConverted string: \t"..cmsg
		end
	else
		tFuncs:SafeDebug("[ToUTF8] Error: No msg")
		return false, ""
	end
end,

["ToLocale_orginal"] = function(self, msg, check)
	if msg then
		local cmsg, err = false,false
		if check then
			cmsg, err = xConvToLocale:iconv(msg)
		else
			cmsg, err = xConvToLocaleNC:iconv(msg)
		end

		if not err then
			return true, cmsg
		else
			local errmsg = "There was an unknown error: "..err
			if err == iconv.ERROR_NO_MEMORY then
				errmsg = "Failed to allocate enough memory in the conversion process"
			elseif err == iconv.ERROR_INVALID then
				errmsg = "An invalid character was found in the input sequence. "
			elseif err == iconv.ERROR_INCOMPLETE then
				errmsg = "An incomplete character was found in the input sequence."
			end
--~ 			tFuncs:Debug("[ToLocale] Error: "..errmsg.."\nUnconverted string: \t"..msg.."\nConverted string: \t"..cmsg, "FI", 1)
			return false, msg, errmsg		--// returns unconverted string on error
		end
	else
		tFuncs:SafeDebug("[ToLocale] Error: No msg")
		return false, ""
	end
end,

["ToUTF8"] = function(self, msg, check)
	if msg then
		local cmsg, err = xConvToUTF8NC:iconv(msg)

		if not err then
			return true, cmsg
		else
			tLogs.Errors.ToUTF8 = tLogs.Errors.ToUTF8 + 1

			local errmsg = "There was an unknown error: "..err
			if err == iconv.ERROR_NO_MEMORY then
				errmsg = "Failed to allocate enough memory in the conversion process"
			elseif err == iconv.ERROR_INVALID then
				errmsg = "An invalid character was found in the input sequence. "
			elseif err == iconv.ERROR_INCOMPLETE then
				errmsg = "An incomplete character was found in the input sequence."
			end

			if check then
--~ 				WriteException("[ToUTF8] [Error] [check/false] "..errmsg.."\tUnconverted string: "..msg:sub(1,40))
				return false, "[ToUTF8] [Error] "..errmsg.."\tUnconverted string: \t"..msg
			else
--~ 				WriteException("[ToUTF8] [Error] "..errmsg.."\tUnconverted string: "..msg:sub(1,40))
				return true, (cmsg or msg)
			end
		end
	else
		return false, ""
	end
end,

["ToLocale"] = function(self, msg, check)
	if msg then
		local err = false
		local cmsg, test = xConvToUTF8Test:iconv(msg)
		if test then
--~ 			tFuncs:Debug("[ToLocale] [UTF8 Validation Failed] "..test.."\tConverted string: "..cmsg, "X", 3)
			cmsg, err = xConvToLocaleNC:iconv(msg.."  ")
			cmsg = cmsg:sub(1, -3)
		else
			cmsg, err = xConvToLocaleNC:iconv(msg)
		end

		if not err then
			return true, cmsg
		else
--~ 			tFuncs:SafeDebug("[ToLocale] [Test] "..err.."\tConverted string: "..cmsg, "X", 3)

			tLogs.Errors.ToLocale = tLogs.Errors.ToLocale + 1

			local errmsg = "There was an unknown error: "..err
			if err == iconv.ERROR_NO_MEMORY then
				errmsg = "Failed to allocate enough memory in the conversion process"
			elseif err == iconv.ERROR_INVALID then
				errmsg = "An invalid character was found in the input sequence. "
			elseif err == iconv.ERROR_INCOMPLETE then
				errmsg = "An incomplete character was found in the input sequence."
			end

			if check then
--~ 				tFuncs:SafeDebug("[ToLocale] [Error] [check/false] "..errmsg.."\tConverted string: "..cmsg:sub(1,40), "X", 3)
				return false, cmsg, errmsg
			else
--~ 				tFuncs:SafeDebug("[ToLocale] [Error] "..errmsg.."\tConverted string: "..cmsg:sub(1,40), "X", 3)
				return true, (cmsg or msg) .."\n"
			end
		end
	else
		tFuncs:SafeDebug("[ToLocale] Error: No msg")
		return false, ""
	end
end,

["NewUser"] = function(self)
	--// define user
	local user = {
		["sNick"] = false,				-- User nick.
		["sIP"] = false,					-- User ip address.

		--removing 'uptr' this for now, when / if we need it just put it back again /fredo
		--["uptr"] = false,					-- Memory address to original user data structure, for internal PtokaX use. Don't modify it, else functions with user table parameter will not work.
		["iProfile"] = -1,				-- User profile.

				-- Data available after GetUserAllData or GetUserData call:

--~ 		["sMode"] = false,				-- User mode (from tag) or nil when user don't have tag or mode in tag.
--~ 		["sMyInfoString"] = false,		-- User MYINFO string or nil when user don't send MyINFO yet.
		["sDescription"] = "",				-- User description or nil when user don't have description.
--~ 		["sTag"] = false,				-- User tag or nil when user don't have tag.
--~ 		["sConnection"] = false,		-- User connection or nil when user don't have connection.
		["sEmail"] = "",					-- User email or nil when user don't have email.
--~ 		["sClient"] = false,			-- User client (from tag) or nil when user don't have tag.
--~ 		["sClientVersion"] = false,		-- User client version (from tag) or false when user don't have tag.
--~ 		["sVersion"] = false,			-- User version (from $Version) or false when user don't send Version.
--~ 		["sCountryCode"] = false,		-- User country code or false when ip-to-country database is not loaded.

--~ 		["bConnected"] = false,			-- User is added in hub (visible for other users, added is after NewUser/OpConnected).
--~ 		["bActive"] = false,			-- true when user is active (from tag or is sending active commands) or false when is not active.
--~ 		["bOperator"] = false,			-- User have operator status.
--~ 		["bUserCommand"] = false,		-- User support UserCommands protocol extension.
--~ 		["bQuickList"] = false,			-- User support QuickList protocol extension.
--~ 		["bSuspiciousTag"] = false,		-- User have suspicious tag.

--~ 		["iShareSize"] = false,			-- User share size.
--~ 		["bShareHidden"] = false,
--~ 		["iHubs"] = false,				-- User hubs count (from tag).
--~ 		["iNormalHubs"] = false,		-- User hubs without registration count (from tag) or false if user don't have tag or have old-style (only H:x) tag.
--~ 		["iRegHubs"] = false,			-- User hubs with registration count (from tag) or false if user don't have tag or have old-style (only H:x) tag.
--~ 		["iOpHubs"] = false,			-- User hubs with operator status (from tag) or false if user don't have tag or have old-style (only H:x) tag.
--~ 		["iSlots"] = false,				-- User slots count (from tag).
--~ 		["iLlimit"] = false,			-- User L or B limit (from tag).
--~ 		["iDefloodWarns"] = false,		-- User deflood warns count.
--~ 		["iMagicByte"] = false,			-- Number of ascii char after connection in myinfo.
		["iLoginTime"] = os.time(),			-- User login time in seconds from 1.1.1970 = os.time()

		--// Flexhub additions to usertable, these need to be placed after all Ptokax Uservariables.
--~ 		["bProxy"] = false,

		["sProtocol"] = false,
		["sProfile"] = "User",
		["sSupports"] = "",

		["tSupports"] = {},
		["tSupportsInf"] = {},
		["tHandledMsg"] = {},

		["sCID"] = false,
		["bBot"] = false, -- if it's a bot, set to true
		["xSocket"] = false,
		["xHandler"] = false,
		["sReverseDNS"] = false,  --// will be moved to tBuffer[sid].Dns, only stored in ClientLog
		["iPort"] = false,
		["iActive"] = 0,
		["iState"] = 1,
		["sStatus"] = "New user",

		["tInfo"] = {
			["tOrig"] = {},
			["tFiltered"] = {},
			["tNew"] = {},
			["tNMDC"] = {
				["sOrig"] = "",
				["sUser"] = "",
				["sOperator"] = "",
			},
			["tADC"] = {
				["sUser"] = "",
				["sUserUpdate"] = "",
				["tUserUpdate"] = {},
				["sOperator"] = "",
				["sOperatorUpdate"] = "",
				["tOperatorUpdate"] = {},
				["sGui"] = "",
				["sGuiUpdate"] = "",
				["tGuiUpdate"] = {},
			},
		},

		["tStats"] = { ["Counters"] = {},},

		["iTcpSent"] = 0,
		["iTcpReceived"] = 0,
		["iTcpAge"] = 0,

		["iLogoutTime"] = 0,
	}
	return user
end,

["AddStats"] = function(self, sid, protocol, stype, number, msg)			--// Use negative number to signal a blocked message
	local number = number or 1
	local protocolstats = tStats.Counters[protocol]
	local userstats = tClient[sid].tStats.Counters

	if number > 0 then
		--// Add stats to hub:
		if not protocolstats[stype] then protocolstats[stype] = {} end
		protocolstats[stype].Received = (protocolstats[stype].Received or 0) + number
		protocolstats[stype].DataIn = (protocolstats[stype].DataIn or 0)+(msg and #msg or 0)

		--// Add stats to user:
		if not userstats[stype] then userstats[stype] = {} end
		userstats[stype].Received = (userstats[stype].Received or 0) + number
		userstats[stype].DataIn = (userstats[stype].DataIn or 0)+(msg and #msg or 0)
	else
		number = math.abs(number)
		--// Add stats to hub:
		if not protocolstats[stype] then protocolstats[stype] = {} end
		protocolstats[stype].Blocked = (protocolstats[stype].Blocked or 0) + number
		protocolstats[stype].BlockedDataIn = (protocolstats[stype].BlockedDataIn or 0)+(msg and #msg or 0)

		--// Add stats to user:
		if not userstats[stype] then userstats[stype] = {} end
		userstats[stype].Blocked = (userstats[stype].Blocked or 0) + number
		userstats[stype].BlockedDataIn = (userstats[stype].BlockedDataIn or 0)+(msg and #msg or 0)
	end
end,

["CreateSid"] = function()	--// create new sid
	ThreadLockMutex(mutexcreatesid)
	local sid
	repeat
		sid = ""

		for i=1,4 do
			local num = math.random(1,32)
			sid = sid..string.sub("ABCDEFGHIJKLMNOPQRSTUVWXYZ234567", num, num)
		end
	until not tClient[sid] and sid ~= "AAAA" and sid ~= "BBBB" and sid ~= "CCCC"  --// AAAA is reserved for the hub and BBBB for OpChat, and CCCC for feed

	ThreadUnlockMutex(mutexcreatesid)
	return sid
end,

["ConfirmCode"] = function(self, confirm)
	if not confirm or confirm ~= tHubSettings.Main.bConfirmCode then
		return false
	else	--// Change confirmcode and return true
		tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()
		return true
	end
end,

["CalcCid"] = function(self, name, ip)
	local pid = th.hash(tHubSettings.Main.sHubID..(name or "")..(ip or ""))
	local cid = dc.hashpid(pid)
	return cid, pid
end,

["CreateBot"] = function(self,nsid,stype,botname,desc,profilelevel,protocol, hidden, ctype)
	local bot = tFuncs:NewUser()	--// create user
	bot.sCID, bot.sPID = tFuncs:CalcCid(botname, "0.0.0.0")
	bot.sNick = botname
	bot.sDescription = desc or ""
	bot.sProtocol = protocol		--// for bots use "ChatRoom" or "Bot"
	bot.sProfile = "Bot"
	bot.iProfile = profilelevel
	bot.sIP = ""
	bot.iPort = 0
	bot.bBot = true
	bot.bHasKey = true
	bot.xSocket = nil
	bot.iActive = 1
	bot.iState = 7
	bot.bHidden = hidden
	bot.iType = ctype or 5

	bot.tInfo.tOrig = {["ID"] = bot.sCID, ["NI"] = bot.sNick, ["DE"] = bot.sDescription, ["VE"] = _APPNAME.." ".._APPVER, ["CT"] = bot.iType}
	bot.tInfo.tFiltered = {["ID"] = bot.sCID, ["NI"] = bot.sNick, ["DE"] = bot.sDescription, ["VE"] = _APPNAME.." ".._APPVER, ["CT"] = bot.iType}

--~ 	bot.tInfo.tNMDC.sOperator = "$MyINFO $ALL "..botname.." "..desc.."<".._APPNAME.." ".._APPVER..">$ $\1$$$"
--~ 	bot.tInfo.tNMDC.sUser = "$MyINFO $ALL "..botname.." "..desc.."<".._APPNAME.." ".._APPVER..">$ $\1$$$"
--~ 	bot.tInfo.tADC.sOperator = "BINF "..nsid.." ID"..cid.." NI"..ADC_E(botname).." CT"..bot.iType.." OP1 DE"..ADC_E(desc).." VE"..ADC_E(_APPNAME.." ".._APPVER).."\n"    -- Bot
--~ 	bot.tInfo.tADC.sUser = "BINF "..nsid.." ID"..cid.." NI"..ADC_E(botname).." CT"..bot.iType.." OP1 DE"..ADC_E(desc).." VE"..ADC_E(_APPNAME.." ".._APPVER).."\n"    -- Bot

	tClient[nsid] = bot
	tInfo:CreateNMDC(tClient[nsid], nsid)
	tInfo:CreateADC(tClient[nsid], nsid)

	--tNicks[botname] = nsid
	tLNicks[botname:lower()] = {["sid"] = nsid, ["sNick"] = botname}
	return nsid, bot
	--// Send the new bot to all users
--~ 	tUser:Announce(nsid)
end,

["GetCtrl"] = function(self,sid,owner)
	if not _NOGUI then
		if not owner then owner = tCtrls["app"] end
		if not sid then tFuncs:Debug("Missing sid in tFuncs:GetCtrl call") return end
		local ctrl = owner:FindWindow(tCtrls["xmlResource"].GetXRCID(sid))
		if not ctrl then tFuncs:Debug("Missing control in tFuncs:GetCtrl: "..tostring(sid)) return end
		if string.find(sid, "ID_ga_", 1, true) then
			return ctrl:DynamicCast("wxGauge")
		else
			return ctrl:DynamicCast(ctrl:GetClassInfo():GetClassName())
		end
	end
end,

["GetToolbar"] = function(self,number,owner)
	if not _NOGUI then
		if not owner then owner = tCtrls["app"] end
		local ctrl = owner:FindWindow(number)
		if not ctrl then tFuncs:Debug("Missing control in tFuncs:GetCtrl: "..tostring(sid)) return end
		if string.find(sid, "ID_ga_", 1, true) then
			return ctrl:DynamicCast("wxGauge")
		else
			return ctrl:DynamicCast(ctrl:GetClassInfo():GetClassName())
		end
	end
end,

["SendToDialog"] = function(self, msg, flags)
	if not _NOGUI then
		if not flags then local flags = "" end
		if tSettings.Debug.EnableConsole then
			ThreadLockMutex(mutexconsole)
			local evt = wx.wxCommandEvent(wx.wxEVT_NULL, tCtrls["wxID_EVENT_SHOWDIALOG"])
			evt:SetString(msg)
			tCtrls["app"]:AddPendingEvent(evt)
			ThreadUnlockMutex(mutexconsole)
		end
	end
end,

["UpdateConsole"] = function(self, ename, msg, flags)
	if not _NOGUI then
		local flags = flags or ""
		tFuncs:AddConsoleEvent(ename, tostring(flags), tostring(msg))
	end
end,

--~ ["UpdateTitle"] = function(self, msg)
--~ 	if not _NOGUI then
--~ 		local evt = wx.wxCommandEvent(wx.wxEVT_NULL, tCtrls["wxID_EVENT_GUI"])
--~ 		evt:SetString("#wt_title##"..tostring(msg)
--~ 		tCtrls["app"]:AddPendingEvent(evt)
--~ 	end
--~ end,

["SendToConsole"] = function(self, msg, flags)
	if not _NOGUI and tSettings.Debug.EnableConsole then
		local flags = flags or ""
		tFuncs:AddConsoleEvent("ID_tc_debug", tostring(flags), "["..ToDate().."] "..tostring(msg).."\n")
	end
end,

["AddConsoleEvent"] = function(self, name, flags, msg)
	ThreadLockMutex(mutexconsole)
	table.insert(tConsoleBuffer.Messages, { ["Name"]=name, ["Flags"] = flags, ["Message"] = msg })
	ThreadUnlockMutex(mutexconsole)
end,

["SendToConsoleChat"] = function(self, msg, flags)
	if not _NOGUI and tSettings.Debug.EnableConsole then
		local flags = flags or ""
		tFuncs:AddConsoleEvent("ID_tc_mainchat", tostring(flags), tostring(msg).."\n")
	end
end,

["SendRawToGui"] = function(self, msg)	-- this is raw, complete string expected with correct line ending
	if tSettings.Gui.Enable and tServer.GuiSID and msg and msg ~= "" then
		if tClient[tServer.GuiSID] and tClient[tServer.GuiSID].sProfile == "Gui" then
			tSend:Raw(tServer.GuiSID, msg)
--		else
--~ 			tServer.GuiSID = false
		end
	end
end,

["SendToGui"] = function(self, msg)
	if tSettings.Gui.Enable and tServer.GuiSID and msg and msg ~= "" then
		if tClient[tServer.GuiSID] and tClient[tServer.GuiSID].sProfile == "Gui" then
			tSend:Raw(tServer.GuiSID, "IGUI "..msg.."\n")
--		else
--~ 			tServer.GuiSID = false
		end
	end
end,

["UpdateStatusBar"] = function(self)	--// Called from the GUI timer
    local c1, c2 = tGui:Gc()        -- garbage collection
    local control = tFuncs:GetCtrl("ID_ss_statusbar")
    control:SetStatusText("U:"..(tStats.Hub.Users.Total or "0"), 1)
    control:SetStatusText("H:"..(tLogs.Counters.Ddos.Total or "0"), 2)
    control:SetStatusText("M:"..tFuncs:ConvertShare(math.ceil(c2*1024)), 3)
    c1,c2,control = nil
end,

["SetStatusText"] = function(self, msg, pos)
	if not pos then pos = 0 end
	msg = tostring(msg) or ""
--	if not statuslog then statuslog = {} end
--	table.insert(statuslog,os.date("[%H:%M:%S] ")..text)
--	if #statuslog > 10 then table.remove(statuslog,1) end
--	tCtrls["statusbar"]:SetToolTip(table.concat(statuslog,"\n"))
	tFuncs:SendToGui("SB"..pos.." VS"..ADC_E(msg))
end,

["SafeDebug"] = function(self, msg, debugtype, debuglevel, cat)
	if msg then
		local debugtype = debugtype or false
		local debuglevel = debuglevel or false
		local cat = cat or false
		ThreadLockMutex(mutexlistenbuffer)
		table.insert(tListenBuffer.Messages, {["Target"] = "D", ["msg"] = msg, ["debugtype"] = debugtype, ["debuglevel"] = debuglevel, ["cat"] = cat})
		ThreadUnlockMutex(mutexlistenbuffer)
	end
end,

["Debug"] = function(self, msg, debugtype, debuglevel, cat)
	local debuglevel = debuglevel or 0
	local debugtype = debugtype or "FI"

	if tHubSettings then
		if debugtype == "X" and (not tHubSettings.Main or not tHubSettings.Main.bExtraDebug) then return true end
		if debugtype == "G" and (not tHubSettings.Main or not tHubSettings.Main.bGuiDebug) then return true end
	end

	local debugchar = debugtype:sub(1,1) or "F"
	local debugsub = debugtype:sub(2) or "I"	-- if no subtype was given, use I (internal/general)

	msg = "["..debugchar..debugsub.."] ["..debuglevel.."] "..(cat and "["..cat.."] " or "")..tostring(msg)

 	if not tStats.Hub.SettingsLoaded then	-- we cant use Debug before settings are loaded ( try with SafeDebug instead, it will come trough here later)
		if debuglevel <= 0 or (tHubSettings and tHubSettings.Main and tHubSettings.Main.bExtraDebug) then	-- only save critical debugmessages or when xdebug is enabled
			Notify(msg)
		end
		return true
	end

	--// Send to Feedroom
	if (tHubSettings.Main.iFeedLevel or 0) >= debuglevel then
		local ttype = {
			["LNJ"] = "[LOGIN] [N] ",
			["LAJ"] = "[LOGIN] [A] ",
			["TI"] = "[TRAFFIC] ",
			["TN"] = "[TRAFFIC] [N] ",
			["TA"] = "[TRAFFIC] [A] ",
			["BI"] = "[BAN] ",
		}
		tSend:Feedroom((ttype[debugtype] or "")..msg, "AAAA", "CCCC")
	end

	--// Send to FlexGui
	if tSettings.Debug.GlobalLevel > debuglevel and tSettings.Debug.EnableGui then
		if tSettings.Debug.Gui[debugchar] and tSettings.Debug.Gui[debugchar][debugsub] and debuglevel <= tSettings.Debug.Gui[debugchar].Level then
			tFuncs:SendToGui("DB"..debugtype..debuglevel.." VS"..ADC_E(msg))
		end
	end

		--// If console isn't loaded yet, return with only a popup or print msg
	if tStats.Console.Status ~= "Running" then
		if debuglevel < 0 then
			Notify(msg)
		end
		return false
	elseif not _NOGUI and tSettings.Debug.EnableConsole and (tHubSettings.Console.DebugLevel or 0) >= debuglevel then
		tFuncs:SendToConsole(msg)
--~ 		print(msg)
	end

	return true
end,

["GetHubIP"] = function(self, address)	-- thread start, runs from timer
--	local ip, err = tFuncs:DLPage("http://whatismyip.org/")
	local ip, err = tFuncs:DLPage("http://flexhub.org/update/version.php")
	if not ip then
-- 		tFuncs:SafeDebug("External IP couldn't be retrieved: "..err, nil, 3)
	else

		local _,_,hubip = ip:find("(%d+%.%d+%.%d+%.%d+)")
		if not hubip then
-- 			tFuncs:SafeDebug("External IP couldn't be detected: "..ip, nil, 3)
		else
			if tHubSettings.Main.sHubIp and tHubSettings.Main.sHubIp ~= hubip then
				tFuncs:SafeDebug("External IP of the hub changed to : "..hubip..". This is now a protected IP.", nil, 0)
				tHubSettings.Main.sHubIp = hubip
				tFuncs:SaveSettings("settings")
				tFuncs:SafeExecute("tSend:UpdateExternalIP()")
			elseif not tHubSettings.Main.sHubIp then
				tFuncs:SafeDebug("External IP of the hub is set to: "..hubip..". This is now a protected IP.", nil, 0)
				tHubSettings.Main.sHubIp = hubip
				tFuncs:SaveSettings("settings")
				tFuncs:SafeExecute("tSend:UpdateExternalIP()")
			else
				--// IP didn't change
			end
		end
	end

--[[
				--// Check if hubaddress IP matches external IP
				if tHubSettings.Main.sHubIp and address and address~= "" then
					local address = address:gsub(":%d+", "")
					local dnsip = socket.dns.toip(address)
					if dnsip and dnsip ~= tHubSettings.Main.sHubIp then
						tFuncs:SafeDebug(
							"Your current external IP doesn't match the IP for the address in !set hub_address: "..address..
							"\nYour current external IP is: "..(tHubSettings.Main.sHubIp or "unknown")..
							"\nThe IP for address "..address.." is: "..dnsip..
							"\nMake sure you have set the right address in !set hub_address and check that that DNS points to your external IP."
						)
					end
				end
			end
		end
	end
]]--
	tThreads.HubIp = nil
	ThreadDestroy()
end,

["GetUserDNS"] = function(self, sid, ip)
	local dns = socket.dns.tohostname(ip) or "Not found"
	if tBuffer[sid] then
		tBuffer[sid].Dns = dns
	end
	ThreadDestroy()
end,

["LockHub"] = function(self)
	tStats.Hub.Status = "Locked"
	for port,data in pairs(tServer.Port) do
		local res, err = tFuncs:StopPort(port)
--~ 		tFuncs:SendToConsole(err)
--~ 		tFuncs:Debug(err)
	end
	tFuncs:SendToConsole("Hub is locked.")
	tFuncs:Debug("Hub is locked.")
	return true, "Hub is locked, no new users will be allowed to connect."
end,

["UnlockHub"] = function(self)
	tStats.Hub.Status = "Online"
	for port,data in pairs(tServer.Port) do
		local res, err = tFuncs:StartPort(port, data.stype)
--~ 		tFuncs:SendToConsole(err)
--~ 		tFuncs:Debug(err)
	end
	tFuncs:SendToConsole("Hub is unlocked.")
	tFuncs:Debug("Hub is unlocked.")
	return true, "Hub is unlocked, ports are restarted and new users will be allowed to connect."
end,

["CreateRooms"] = function(self)
	--// tFuncs:CreateBot(sid,stype,name,desc,prof,protocol,visible)
	tFuncs:Debug("The FlexHub SecurityBot is now known as "..tHubSettings.Main.sBotName ,"FI", -1)
	tFuncs:CreateBot("AAAA", "Bot", tHubSettings.Main.sBotName, tHubSettings.Main.sBotDescription, 10, "Bot")
	tFuncs:CreateBot("BBBB", "OpChat", tHubSettings.Main.sOpChat, tHubSettings.Main.sOpChatDescription, (tHubSettings.Main.iOpChatProfileLevel or 5), "ChatRoom", true)
	tFuncs:Debug("The FlexHub OpChat is now known as "..tHubSettings.Main.sOpChat ,"FI", -1)
	tFuncs:CreateBot("CCCC", "FeedRoom", tHubSettings.Main.sFeed, tHubSettings.Main.sFeedDescription, (tHubSettings.Main.iFeedProfileLevel or 8), "ChatRoom", true)
	tFuncs:Debug("The FlexHub FeedRoom is now known as "..tHubSettings.Main.sFeed ,"FI", -1)

	--// Create ChatRoom from Settings
	if tChatRooms ~= {} then
		for name, data in pairs(tChatRooms) do
			if not tNick:GetSid(name) then
				tFuncs:CreateBot(tFuncs:CreateSid(),"ChatRoom",name, data.sDesc, (data.iLevel or 5), "ChatRoom", data.bHidden)
			else
				tFuncs:Debug("[CreateRooms] Skipping "..name..", chatroom already exists.", nil, 0)
			end
		end
	end

	--// ReCreate PrivateRoom from Settings
	if tPrivateRooms ~= {} then
		for name, data in pairs(tPrivateRooms) do
			if not tNick:GetSid(name) then
				local sid, bot = tFuncs:CreateBot(tFuncs:CreateSid(), "ChatRoom", name, data.sDesc, (data.iLevel or -1), "ChatRoom", data.bHidden)
				tClient[sid].bPrivate = true
				tClient[sid].tMembersTemp = {}
				tClient[sid].sOwner = data.sOwner
			else
				tFuncs:Debug("[CreateRooms] Skipping "..name..", privateroom already exists.", nil, 0)
			end
		end
	end
end,

["StartHub"] = function(self, portlist)
	tFuncs:Debug("Hub started.", "FI", -1)

	tFuncs:CreateRooms()

	tStats.Hub.Started = os.time()

	--// Start the serverports to listen for clientconnects --// TODO: Check if port is valid and hub is started
	local onlineports = 0
	for _,stype in ipairs({"Regular", "Secure", "Gui"}) do
		if portlist[stype] and type(portlist[stype]) == "table" then
			for port,_ in pairs(portlist[stype]) do
				port = tonumber(port)
				if port and (not tServer.Port[port] or tServer.Port[port].active ~= 1) then
					local res, err = tFuncs:StartPort(port, stype)
					if res then
						onlineports = onlineports + 1
					end
				end
			end
		else
			Notify("OOPS : portlist[stype] does" ..(portlist[stype] and "" or " NOT").. " exist. Type is "..(portlist[stype] and type(portlist[stype]) or "nada") )
		end
	end

	--// Start Listening thread
	tThreads.Listen.Main = ThreadCreate(tThread.TCPListen,{1})		-- thread for each port?

	if onlineports > 0 then
		tStats.Hub.Status = "Online"
		if not _NOGUI then
			tFuncs:GetCtrl("ID_bu_startstop"):SetLabel("Stop Hub")
			local toolid = tCtrls["xmlResource"].GetXRCID("ID_to_start")
			tFuncs:GetCtrl("ID_tb_main"):ToggleTool(toolid, true)
		end
	else
		if not _NOGUI then
			tFuncs:GetCtrl("ID_bu_startstop"):SetLabel("Start Hub")
			local toolid = tCtrls["xmlResource"].GetXRCID("ID_to_stop")
			tFuncs:GetCtrl("ID_tb_main"):ToggleTool(toolid, true)
			tFuncs:Debug("No ports listening!", "FI", -1)
		else
			--// If there's no GUI, the application needs to quit since there's no way to connect
			print("No ports listening, closing FlexHub.")
			tFuncs:StopHub()
		end
	end
end,

["UpdatePorts"] = function(self, portlist)
	for _,stype in ipairs({"Regular", "Secure", "Gui"}) do
		for port,_ in pairs(portlist[stype]) do
			port = tonumber(port)
			if not tServer.Port[port] or not tServer.Port[port].server then
				tServer.Port[port] = nil
				local res, err = tFuncs:StartPort(port, stype)
			end
		end
	end
	for port,_ in pairs(tServer.Port) do
		port = tostring(port)
		if not portlist.Regular[port] and (not portlist.Secure[port]) and (not portlist.Gui[port]) then
			port = tonumber(port)
			local res, err = tFuncs:StopPort(port)
			local res, err = tFuncs:KillPort(port)
			tServer.Port[port] = nil
		end
	end
end,

["GetPorts"] = function(self, sid, filter)
	local pList = {}
	local filter = type(filter) == "table" and filter or {["Regular"] = "(R)", ["Secure"] = "(S)", ["Gui"] = false,}

	for port, dta in pairs(tServer.Port) do
		if dta.active == 1 and filter[dta.stype] then
			table.insert(pList, port..filter[dta.stype])
		end
	end

	if #pList > 1 then
		table.sort(pList)
	elseif #pList == 0 then
		return false, "No ports running"
	end

	return true, table.concat(pList, ", ")
end,

["GetAddressPort"] = function(self, sid, target)
	local addy = string.match(tHubSettings.Main.sAddress, "(.+):") or ""
	local port = tonumber(string.match(tHubSettings.Main.sAddress, ":(%d+)") or 0)
	local target = target or "adc"
	local kp = ""
	if tSettings.TLS.Keyprint then
		kp = "/?kp="..tSettings.TLS.Keyprint
	end

	if target == "nmdc" then
		if tHubSettings.tRules.bAllowNMDC then
			if tServer.Port[port] and tServer.Port[port].stype == "Regular" then
				return true, "dchub://"..tHubSettings.Main.sAddress
			else
				for p, dta in pairs (tServer.Port) do
					if dta.active == 1 and dta.stype == "Regular" then
						return true, "dchub://"..addy..":"..p
					end
				end
				return false, "No running NMDC port found"
			end
		else
			return false, "NMDC protcol disabled in settings"
		end
	elseif target == "nmdcs" then
		if tHubSettings.tRules.bAllowNMDC then
			if tServer.Port[port] and tServer.Port[port].stype == "Secure" then
				return true, "nmdcs://"..tHubSettings.Main.sAddress..kp
			else
				for p, dta in pairs (tServer.Port) do
					if dta.active == 1 and dta.stype == "Secure" then
						return true, "nmdcs://"..addy..":"..p..kp
					end
				end
				return false, "No running NMDCS port found"
			end
		else
			return false, "NMDC protcol disabled in settings"
		end
	elseif target == "adc" then
		if tHubSettings.tRules.bAllowADC then
			if tServer.Port[port] and tServer.Port[port].stype == "Regular" then
				return true, "adc://"..tHubSettings.Main.sAddress
			else
				for p, dta in pairs (tServer.Port) do
					if dta.active == 1 and dta.stype == "Regular" then
						return true, "adc://"..addy..":"..p
					end
				end
				return false, "No running ADC port found"
			end
		else
			return false, "ADC protcol disabled in settings"
		end
	elseif target == "adcs" then
		if tHubSettings.tRules.bAllowADC then
			if tServer.Port[port] and tServer.Port[port].stype == "Secure" then
				return true, "adcs://"..tHubSettings.Main.sAddress..kp
			else
				for p, dta in pairs (tServer.Port) do
					if dta.active == 1 and dta.stype == "Secure" then
						return true, "adcs://"..addy..":"..p..kp
					end
				end
				return false, "No running ADCS port found"
			end
		else
			return false, "ADC protcol disabled in settings"
		end
	else
		return false, ""
	end
end,

["StopHub"] = function(self, trigger)
	tStats.Hub.Status = "Stopping"

	--// Terminate the listenthread:
	tFuncs:Debug("Stopping hub, please wait...", "FI", -1)
	tServer.Main.tcp_listening = 0
	ThreadWait(500)

	--// Terminate webchat
	if tThreads.WebChat then
		tFuncs:Debug("WebChat: Thread terminated. Stopped listening for incoming webchat.", "FI", -3)
		ThreadDestroy(tThreads.WebChat)
		tThreads.WebChat = nil
	end

	if tThreads.Listen.Main then
		tFuncs:Debug("TCPListen: Thread terminated. Stopped listening on clientsockets.", "FI", -3)
		ThreadDestroy(tThreads.Listen.Main)
	end

	for port,data in pairs(tServer.Port) do
		if data.stype ~= "Gui" then
			local res, err = tFuncs:StopPort(port)
			local res, err = tFuncs:KillPort(port, true)
			tServer.Port[port] = nil
		else
			--// TODO: make an exception for the GUI-port
			local res, err = tFuncs:StopPort(port)
			local res, err = tFuncs:KillPort(port, true)
			tServer.Port[port] = nil
		end
	end

--~     tServer.Main.tcp_listening = 0
	if tThreads.SSL.Handshake then
		tThreads.SSL.Running = false
		ThreadDestroy(tThreads.SSL.Handshake)
		tThreads.SSL.Handshake = nil
	end

	if not _NOGUI then
		tFuncs:GetCtrl("ID_bu_startstop"):SetLabel("Start Hub")
		local toolid = tCtrls["xmlResource"].GetXRCID("ID_to_stop")
		tFuncs:GetCtrl("ID_tb_main"):ToggleTool(toolid, true)  --// Temp disabled for linux
	end

	--// If there's still a GUI-port running, don't exit
	for port, data in pairs(tServer.Port) do
		--// If GUI port is still running:
		if data.stype == "Gui" and data.active ~= 1 then
			tFuncs:Debug("Hub stopped, but GUI port ("..port..") still running.", "FI", -1)
			tFuncs:SetStatusText("Hub stopped.")
			return false, "Hub stopped, but GUI port ("..port..") still running."
		end
	end

	tFuncs:Debug("Hub stopped.", "FI", -1)
	tFuncs:SetStatusText("Hub stopped.")

	tFuncs:ResetUserCount(true)
--~ 	tFuncs:CloseFiles()

	tStats.Hub.Status = "Offline"

	--// If there's no GUI, the application needs to quit since there's no way to connect
	if _NOGUI and not trigger then	-- if we're here because of a restart or shutdown command we dont close here
		print("All ports stopped, closing FlexHub.")
		OnStop()
	end
end,

["StartPort"] = function(self, port, stype)
	if not tServer.Port[port] or not tServer.Port[port].server then
		if stype == "Secure" and _OS ~= "Linux" and not tHubSettings.Main.bForceTLS then 	--// !set force_tls allows enabling of TLS in Windows, for testing only
			local err = "Failed to open port: "..port..", secure connections are currently only supported on Linux systems."
			tFuncs:Debug(err, "FI", -1)
			return false, err
		end

		if stype == "Secure" and not tGlobals.SecureSocketContext then
			local cert = io.open("certificates/cert.pem", "r")
			if not cert then
				local err = "Failed to open secure port: "..port..", the certificate for SSL hasn't been generated yet, use !createcert -h  for info on how to generate it."
				tFuncs:Debug(err, "FI", -1)
				return false, err
			else
				cert:close()
			end

			local ch, er = SecureSocketInit()
			if not ch then
				local err = "Failed to create TLS context for port"..port..": "..tostring(er)
				tFuncs:Debug(err, "FI", -1)
				return false, err
			end
		end

		local err = false
		tServer.Port[port] = { ["server"] = false, ["active"] = 0, ["status"] = "clean", ["stype"] = stype, ["loginsok"] = 0, ["loginsfailed"] = 0, ["connects"] = 0, }
		tServer.Port[port].socket, err = socket.tcp()     -- creates master object
		if not tServer.Port[port].socket then
			return false, "Failed to open port: "..port..", "..err..". ("..stype..")"
		end

		tServer.Port[port].socket:setoption("reuseaddr", true)

		local res,err = tServer.Port[port].socket:bind((tHubSettings.Main.sBindInterface or "*"),port) -- binds master object to address and port
		if not err then
				res,err = tServer.Port[port].socket:listen(32)  -- transforms master object to server object. What's the optimal backlog value???
				if not err then
					tServer.Port[port].status = "Started listening."
					tServer.Port[port].server = true
					tServer.Port[port].active = 1
					tLogs.Counters.Ddos.LastAverageTime = os.time()
					tLogs.Counters.Ddos.Port[port] = { ["Login"] = 0, ["Count"] = 0, ["LastCount"] = 0, ["NewCount"] = 0, ["LastNewCount"] = 0, ["Average"] = 0, ["NewAverage"] = 0,  ["AveragePeak"] = 0, ["NewAveragePeak"] = 0,}
				else
					tServer.Port[port].status = "Failed to start listening on port: "..port.. ", "..err
				end
		else
			tServer.Port[port].status = "Failed to bind serversocket for port: "..port.. ", "..err
		end

		if not err then
			tStats.Thread.Port[port] = tStats.Thread.Port[port] or { ["Status"] = "Thread starting" }
			tThreads.Wait[port] = ThreadCreate(tThread.TCPWaitForUsers,{1, port, stype})	-- thread for each port
--~ 			tServer.Port[port].thread = ThreadCreate(tThread.TCPWaitForUsers,{1, port, stype})	-- thread for each port
			if stype == "Secure" and not tThreads.SSL.Handshake then
				tThreads.SSL.Running = true
				tThreads.SSL.Handshake = ThreadCreate(tThread.TCPSSL,{1})
			end
			tFuncs:Debug( "Started listening on port: "..port..". ("..stype..")", "FI", -1)
			return true, "Started listening on port: "..port..". ("..stype..")"
		else
			tServer.Port[port].socket:close()
			tServer.Port[port] = nil
			tFuncs:Debug("Failed to listen on port: "..port..", "..err..". ("..stype..")", "FI", -1)
			return false, "Failed to listen on port: "..port..", "..err..". ("..stype..")"
		end
	else
		return false, "Already listening on port: "..port..". ("..stype..")"
	end
end,

["CreateCertificate"] = function(self, path, force)
	if _OS == "Windows" and not tHubSettings.Main.bForceTLS then
		return false, "Secure connections aren't supported in FlexHub for Windows yet."
	end
	if not tHubSettings.Main.sAddress or tHubSettings.Main.sAddress == "" or string.find(tHubSettings.Main.sAddress, "localhost", 1, true) then
		return false, "The hubs address must be set, it is used to generate the certificate. Use !set hub_address to set your real hub address and try again."
	end

	--// Check for existing certificate
	local cert = io.open("certificates/cert.pem", "r")
	local key = io.open("certificates/key.pem", "r")
	if cert or key then
		if cert then cert:close() if force then os.remove("certificates/cert.pem") end end
		if key then key:close() if force then os.remove("certificates/key.pem") end end
		if not force then return false, "The file 'cert.pem' or 'key.pem' is found in the folder 'certificates'.\nIf you want to generate a new certificate you have to remove those files from that folder or use the -force option.\nMake sure you make a backup in case you need them again." end
	end

	--// Certificate configuration
	local makecert = "openssl req -x509 -nodes -days 3650 -subj /CN="..string.gsub(tHubSettings.Main.sAddress, ":.+$", "").." -newkey rsa:1024 -keyout certificates/key.pem -out certificates/cert.pem >logs/Certificate.log 2>&1"

	--// Linux:
	if _OS == "Linux" then
		--// Write batch file
		local linscript = ""
		if _DEVICE == "synox86" then
			linscript = "#!/bin/sh\nopenssl req -config certificates/openssl.cnf -x509 -nodes -days 3650 -subj /CN="..string.gsub(tHubSettings.Main.sAddress, ":.+$", "").." -newkey rsa:1024 -keyout certificates/key.pem -out certificates/cert.pem >logs/Certificate.log 2>&1\n"
		else
			linscript ="#!/bin/sh\nsudo "..makecert.."\n"
		end
		local o = io.open("MakeCert.sh", "w")
		if o then
			o:write(linscript)
			o:close()
			os.execute("chmod 700 ./MakeCert.sh")
		end

		--// Create certificate
		os.execute("./MakeCert.sh")
--~ 		os.remove("MakeCert.sh")
	--// Windows
	else
		--// Get path to openssl.exe
		if path then
			path:gsub("\\%s-$", "")
			--// Check for openssl.exe
			local prog = io.open(path.."\\openssl.exe")
			if prog then prog:close() else return false,"The file "..path.."\\openssl.exe can't be found. Make sure OpenSSL is installed and you specify the correct path to openssl.exe\n" end
		else
			return false, "You need to have OpenSSL installed and specify the path to openssl.exe.\n"
		end

		--// Write batch file
		local winscript = "set OPENSSL_CONF="..path.."\\openssl.cfg\n"..path.."\\"..makecert.."\n"
		local o = io.open("MakeCert.bat", "w")
		if o then
			o:write(winscript)
			o:close()
		end

		--// Create certificate
		os.execute("MakeCert.bat")
--~ 		os.remove("MakeCert.bat")
	end

	--// Verify and load
	local cert = io.open("certificates/cert.pem", "r")
	if cert then
		cert:close()
		local ch, er = SecureSocketInit()
		if not ch then return false, "Error initializing certificate after generation: "..tostring(er) end
		return true, "Selfsigned TLS certificate created and loaded, files 'cert.pem' and 'key.pem' are placed in folder 'certificates'."
	else
		return false, "Creating a selfsigned TLS certificate failed, check Certificate.log in the logs folder."
	end
end,

["GetAddress"] = function(self, sid)
	local addy = SetManager:Get("hub_address") or "unknown"
	local port = tonumber(string.match(addy, ":(%d+)") or 411)
	local stype = tServer.Port[port] and tServer.Port[port].stype or "Port not running!"
	local address = ""

	if sid then
		if tClient[ssid].sProtocol == "NMDC" then
			if stype == "Secure" then address = "nmdcs://"..addy else address = "dchub://"..addy end
		elseif tClient[ssid].sProtocol == "ADC" then
			if stype == "Secure" then address = "adcs://"..addy else address = "adc://"..addy end
		end
	else
		if stype == "Secure" then
			address = (tHubSettings.tRules.bAllowNMDC and "NMDCS: nmdcs://"..addy.." " or "")..(tHubSettings.tRules.bAllowADC and "ADCS: adcs://"..addy.." " or "")
		else
			address = (tHubSettings.tRules.bAllowNMDC and "NMDC: dchub://"..addy.." " or "")..(tHubSettings.tRules.bAllowADC and "ADC: adc://"..addy.." " or "")
		end
	end
	return address
end,

["StopPort"] = function(self, port)
	if tServer.Port[port] and tServer.Port[port].socket then
		tServer.Port[port].server = false
		tServer.Port[port].status = "Stopped"
		if tThreads.Wait[port] then
			ThreadDestroy(tThreads.Wait[port])
		end
		if tServer.Port[port].socket then
			tServer.Port[port].socket:close()
			tServer.Port[port].socket = nil
		end
		tFuncs:Debug("Stopped listening on port: "..port..".", "FI", -1)
		return true, "Stopped listening on port: "..port.."."
	else
		return false, "Wasn't listening on port: "..port.."."
	end
end,

["KillPort"] = function(self, port, noquit)
	local counter = 0
	if noquit then	--// Don't send $Quit or IQUI - hub is stopped
		for sid,client in pairs(tClient) do
			if client.iPort == port then
				counter = counter + 1
				tFuncs:ClientQuit(client, sid)
				tClient[sid] = nil	--// Tricky: this changed the tClient over which is looped...
			end
		end
	else
		for sid,client in pairs(tClient) do
			if client.iPort == port then
				counter = counter + 1
				tUser:Disconnect(sid)
			end
		end
	end
	tFuncs:Debug("Disconnected "..counter.." users on port "..port..".", "FI", -1)
	return true, "Disconnected "..counter.." users on port "..port.."."
end,

["AttackReport"] = function(self)
	local ctime = os.time()
	local dtime = ctime - tLogs.Counters.Ddos.LastAverageTime
	tLogs.Counters.Ddos.LastAverageTime = ctime

	--// New IP connections:
	local ncount = 0
	for port, data in pairs(tLogs.Counters.Ddos.Port) do
		--// New autobanned IPs
		local ncount = data.NewCount - data.LastNewCount
		data.LastNewCount = data.NewCount
		tLogs.Counters.Ddos.Total = tLogs.Counters.Ddos.Total + ncount
		local naverage = ncount/(dtime)  --// per second
		data.NewAverage = naverage
		data.NewAveragePeak = ((naverage > data.NewAveragePeak and naverage) or data.NewAveragePeak)

		--// Blocked connections from earlier autobans
		local bcount = data.Count - data.LastCount
		data.LastCount = data.Count
		tLogs.Counters.Ddos.Total = tLogs.Counters.Ddos.Total + bcount
		local baverage = bcount/(dtime)   --// per second
		data.Average = baverage
		data.AveragePeak = ((baverage > data.AveragePeak and baverage) or data.AveragePeak)

		--// Send stats to feed:
		if ncount+bcount > tSecurity.Ddos.iAttackMinRate then
			if not data.Attack then
				data.Attack = true
				tSettings.Security.UnderAttack = true
				data.Started = ctime
				tFuncs:Debug("[DDOS] Attack detected on port: "..port..".", "FI", -2)
			end
			tFuncs:Debug("[DDOS] Attack report for port: "..port.." / New IP's blocked: "..ncount..", rate: "..math.ceil(naverage).." per sec. / New conn. blocked: "..bcount..", rate: "..math.ceil(baverage).." per sec. / Total blocked: "..data.Count+data.NewCount, "FI", -1)
		else
			if data.Attack then
				data.Attack = false
				tSettings.Security.UnderAttack = false
				data.Stopped = ctime
				tFuncs:Debug("[DDOS] Attack stopped on port: "..port..".", "FI", -2)
			end
		end
	end

end,

["GetBlocks"] = function(self, vip)
	local btypes = {
		["mute"] = {"tMute", "User is muted"},
		["mutestealth"] = {"tMuteStealth", "User is silently muted"},
		["kenny"] = {"tKenny",  "User is kennylized"},
		["muppet"] = {"tMuppet", "User is muppetized"},
		["connects"] = {"tConnects", "Connections and searches are blocked"},
		["pm"] = {"tPm", "Private messages from this user are blocked"},
	}

	local blocks = {}
	for name, data in pairs(btypes) do
		if tBlocks[data[1]] and tBlocks[data[1]][vip] then
			table.insert(blocks, data[2])
		end
	end
	return blocks
end,

["GetBans"] = function(self, banlists, start, cmdname)
	local start = tonumber(start) or 0
	local bans = ""
	local cmdname = cmdname or "show..."
	local ctime = os.time()

	local bantable = {}
	for num, bantype in ipairs(banlists) do
		local banlist = tBans[bantype]
		for value, data in pairs(banlist) do
			local expires = 0
			local ip = false
			local nick = false

			if bantype ~= "autoban" then
				ip = (data.sIP and data.sIP ~= "" and data.sIP) or "<IP unknown>"
				nick = (data.sNick and data.sNick~="" and data.sNick) or "<nick unknown>"
				expires = data.iExpires or 0
			else
				expires = data or 0
			end

			if expires <= 0 then
				expires = "Never"
			elseif expires <= ctime then
				expires = "Expired"
			else
				expires = ToDate(expires)
			end

			if bantype == "ipban" then
				if data.sType ~= "kick" or data.iExpires > ctime then	--// Hide expired kicks
					table.insert(bantable, "\t"..bantype.."\t"..value.."\t\t"..nick.."\t\t"..expires.."\t\t"..data.sBy.."\t\t"..data.sReason.."\n")
				end
			elseif bantype == "nickban" then
				--// Multiple nicks in nickban aren't needed, only needed in IP bans, unless a !kill should ban all nicks connected to that IP, then this can be used to show related nickbans
--[[
				local nicklist = ""
				local truenick = value
--~ 				local start, stop, truenick = string.find(data.sNick:lower(), value)	-- start is used for offset in function params
--~ 				local truenick = string.sub(data.sNick, start, stop)
				if string.find(data.sNick, truenick..", ") then
					nicklist = truenick .." ("..string.gsub(data.sNick, truenick..", ", "") ..")"
				else
					nicklist = truenick .." ("..string.gsub(data.sNick, truenick, "") ..")"
				end
				if string.find(nicklist, ", )") then nicklist = string.gsub(nicklist, ", %)", "%)") end
				table.insert(bantable, "\t"..bantype.."\t"..nicklist.."\t\t"..ip.."\t\t"..expires.."\t\t"..data.sBy.."\t\t"..data.sReason.."\n")
]]--

				table.insert(bantable, "\t"..bantype.."\t"..value.."\t\t"..ip.."\t\t"..expires.."\t\t"..data.sBy.."\t\t"..data.sReason.."\n")
			elseif bantype == "warn" then
				table.insert(bantable, "\t"..bantype.." "..data.iCount.."\t"..value.."\t\t"..ip.."\t\t\t\t\t\t"..data.sBy.."\t\t"..data.sReason.."\n")
			elseif bantype == "rangeban" then
				table.insert(bantable, "\t"..bantype.."\t"..nick.."\t\t\t\t\t"..expires.."\t\t"..data.sBy.."\t\t"..data.sReason.."\n")
			elseif bantype == "autoban" then
				table.insert(bantable, "\t"..bantype.."\t"..value.."\t\t\t\t\t"..expires.."\n")
			end
		end
	end

	table.sort(bantable, SortLowercase)

	for num, value in ipairs(bantable) do
		if num > start and num < (start+50) then
			bans = bans..value
		elseif num == (start+50) then
			return bans.."\n\tShowing items "..(start+1).."-"..(start+49).." of "..(#bantable)..", use !"..cmdname.." "..(start+50).." to show more.\n"
		end
	end
	return bans
end,

["TrueOrFalse"] = function(self, value)
	if not value and type(value) == "Boolean" then
		return false
	elseif not value and type(value) ~= "Boolean" then	--// No value (nil) will return true!!
		return true
	elseif value and (value == "false" or value == "0" or value == 0) then
		return false
	elseif value and (value == "true" or value == "1" or value == 1) then
		return true
	else
		return false,  "Invalid value, use true or false"
	end
end,

["TableChunk"] = function(self, ttable, start, size)
	local start = tonumber(start or 1) or 1
	local size = (tonumber(size or 50) or 50)-1
	local res = {}
	if ttable and type(ttable) == "table" then
		if start <= #ttable then
			for i=start,start+size do
				table.insert(res, ttable[i])
				if i>#ttable then
					return res
				end
			end
		end
		return res, { start, start+size, #ttable, start+size+1 }	--// showing items [1] to [2] of [3], use !blabla [4] to show more
	else
		return false
	end
end,

["HubCommand"] = function(self, sid, line, tosid)
	local toroom = "Mainchat"
	if tosid then
		toroom = tUser:GetNick(tosid) or "Unknown chatroom"
--~ 		tFuncs:Debug("[HI] "..tUser:GetDataLine(sid).."["..toroom.."] "..line, "HI")
	else
--~ 		tFuncs:Debug("[HI] "..tUser:GetDataLine(sid)..line, "HI")
	end

	local _,_,prefix, cmd = string.find(line, "^(["..tHubSettings.Main.sCmdPrefixes.."])(%S+)")
	cmd = cmd or ""
	cmd = string.lower(cmd)
	local origcmd = cmd

-- // TEMP scroogemode
--	if cmd == "scroogemode" and tUser:IsAllowed(sid, "setcmd") then
--		if tTimers.Christmas and tTimers.Christmas.bEnabled then
--			tTimers.Christmas.bEnabled = false
--			tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Text Command successful: "..line, "HI")
--			return true, "FlexHub Christmas spammer disabled"
--		else
--			tTimers.Christmas.bEnabled = true
--			tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Text Command successful: "..line, "HI")
--			return true, "FlexHub Christmas spammer enabled"
--		end
--	end

--~ 	local cmdset = {}
--~ 	if tHubSettings.Main.sCommandSet == "FlexHub" then
--~ 		cmdset = { ["h"] = "help", ["hi"] = "hubinfo", ["rs"] = "restartscripts", ["tban"] = "tempban", ["tbanip"] = "tempbanip", ["ui"] = "userinfo", ["ii"] = "ipinfo", ["ar"] = "addreg",}
--~ 	elseif tHubSettings.Main.sCommandSet == "Ynhub" then
--~ 		cmdset = { ["h"] = "help", ["hi"] = "hubinfo", ["rs"] = "restartscripts", ["tban"] = "tempban", ["tbanip"] = "tempbanip", ["ui"] = "userinfo", ["ii"] = "ipinfo", ["ar"] = "addreg",}
--~ 	elseif tHubSettings.Main.sCommandSet == "Ptokax" then
--~ 		cmdset = { ["h"] = "help", ["hi"] = "hubinfo", ["rs"] = "restartscripts", ["getinfo"] = "userinfo", ["reginfo"] = "getinfo",}
--~ 	elseif tHubSettings.Main.sCommandSet == "Verli" then
--~ 		cmdset = { ["h"] = "help", ["hi"] = "hubinfo", ["rs"] = "restartscripts", ["getinfo"] = "userinfo", ["reginfo"] = "getinfo",}
--~ 	else
--~ 		cmdset = { ["h"] = "help", ["hi"] = "hubinfo", ["rs"] = "restartscripts", ["ui"] = "userinfo", ["ii"] = "ipinfo", }
--~ 	end
	if tHubSettings.Main.bCmdAlias and tCmdAlias[cmd] then cmd = tCmdAlias[cmd] end

	if HubCmd[cmd] then
		local allowed, reason = HubCmd[cmd]["allowed"](sid, line, cmd)	--// check if cmd is allowed
		if allowed then
			local result, ok, msg = false, false, "Error executing hubcommand: "..line
			if string.find(line, "^%S+ %-[Hh]%s-$") then
				if HubCmd[cmd]["xhelp"] then
					result, ok, msg = pcall(HubCmd[cmd]["xhelp"], sid, line, prefix..origcmd, tosid) -- filterline is the result of the command when filtered by mainchat/pm-filter for use in +me for example.
				else
					ok = true
					result, msg = pcall(HubCmd[cmd]["help"], sid, line, prefix..origcmd, tosid) -- filterline is the result of the command when filtered by mainchat/pm-filter for use in +me for example.
				end
			else
				result, ok, msg = pcall(HubCmd[cmd]["cmd"], sid, line, prefix..origcmd, tosid) -- filterline is the result of the command when filtered by mainchat/pm-filter for use in +me for example.
			end

			--// Remove passwords from Feed output
			line = tFuncs:FilterPass(line, cmd)

--~ 			local traceback = ShowTrace(15)
			if result then
				if not ok then
--					tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command failed: "..line, "HI")
					tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command failed: "..line.." MSG: "..msg, "HI")
				else
					tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command successful: "..line, "HI")
				end

				local output = "MC"
				if HubCmd[cmd]["outputtype"] then
					local otype = HubCmd[cmd]["outputtype"]
					if otype == "info" then
						output =  tHubSettings.Main.InfoOutput
					elseif otype == "trigger" then
						output = tHubSettings.Main.TriggerOutput
					elseif otype == "command" then
						output =  tHubSettings.Main.CommandOutput
					else
						output = otype
					end
				else
					output =  tHubSettings.Main.CommandOutput
				end

				if ok then	--// if a command failed, send help back to user
					--// Add hubcommand to cmdhistory
					table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, true, line})

					return true, msg, output
				else
					--// Add hubcommand to cmdhistory
					table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

					return true, msg .. " Use !"..cmd.." -h for more details.", output
				end
--				return true, msg, output
			elseif not result then
--~ 				local traceback = ShowTrace(15)
				tFuncs:Debug(tUser:GetDataLine(sid).."Error while executing function: HubCmd[\""..cmd.."\"] : "..line.." / err: "..ok)
				tFuncs:UnlockMutexes()
				tSend:ChatToProfLevel("*** An error occured during execution of a hubcommand from <"..tUser:GetNick(sid)..">, check feedroom for more info.", "AAAA", 10) 	--// Sends alert in main to all hubowners

				--// Add hubcommand to cmdhistory
				table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

				return false, "Error executing command, debug information sent to Feedroom."
			end
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command failed, no permission: "..line, "HI")
			return false, reason or Lang(sid, 0, "You are not allowed to use this command.")
		end
	elseif SetCmd[cmd] then
		local allowed, reason = SetCmd[cmd]["allowed"](sid, line, origcmd)
		if allowed then
			local result, ok, msg = false, false, "Error executing setcommand: "..line
			if string.find(line, "^%S+ %-[Hh]%s-$") then
				if SetCmd[cmd]["xhelp"] then
					result, ok, msg = pcall(SetCmd[cmd]["xhelp"], sid, line, prefix..origcmd, tosid) -- little change
				else
					ok = true
					result, msg = pcall(SetCmd[cmd]["help"], sid, line, prefix..origcmd, tosid) -- little change
				end
			else
				result, ok, msg = pcall(SetCmd[cmd]["cmd"], sid, line, prefix..origcmd, tosid) -- little change
			end
--~ 			local traceback = ShowTrace(15)

			--// Remove passwords from Feed output
			line = tFuncs:FilterPass(line, cmd)

			if result and msg then
				if not ok then
--					tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Setcommand failed: "..line, "HI")
					tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Setcommand failed: "..line.." MSG: "..msg, "HI")
				else
					tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Setcommand successful: "..line, "HI")
				end

				local output = tHubSettings.Main.CommandOutput

				--// Redirect results to PM if output is to MC and string contains 'is kicking'
				if cmd == "set" and toroom == "Mainchat" and output == "MC" and string.find(msg, "is kicking") then
					tSend:PmToUser(msg, "AAAA", sid)
					msg = "The results of !set have been sent in PM from "..tHubSettings.Main.sBotName
				end

				if ok then	--// if a command failed, send help back to user
					--// Add hubcommand to cmdhistory
					table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, true, line})

					return true, msg, output
				else
					--local hresult, hok, hmsg
					--if SetCmd[cmd]["xhelp"] then
					--	hresult, hok, hmsg = pcall(SetCmd[cmd]["xhelp"], sid, line, prefix..origcmd, tosid) -- little change
					--else
					--	hok = true
					--	hresult, hmsg = pcall(SetCmd[cmd]["help"], sid, line, prefix..origcmd, tosid) -- little change
					--	hmsg = "Help text :\t"..hmsg
					--end
					--return true, msg .. "\n\n"..hmsg, output
					--// Add hubcommand to cmdhistory
					table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

					return true, msg .. " Use !"..cmd.." -h for more details.", output
				end
			elseif not result then
				tFuncs:Debug(tUser:GetDataLine(sid).."Error while executing function: SetCmd[\""..cmd.."\"] :"..line.." / err: "..ok)
				tFuncs:UnlockMutexes()
				tSend:ChatToProfLevel("*** An error occured during execution of a hubcommand from <"..tUser:GetNick(sid)..">, check feedroom for more info.", "AAAA", 10) 	--// Sends alert in main to all hubowners
				local msg =  "Error executing command, debug information sent to Feedroom."

				--// Add hubcommand to cmdhistory
				table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

				return false, msg
			end
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command failed, no permission: "..line, "HI")
			--// Add hubcommand to cmdhistory
			table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

			return false, reason or Lang(sid, 0, "You are not allowed to use this command.")
		end
	elseif tCommands[cmd] and not tFuncs:CheckFlag((tCommands[cmd].iOptions or 0), 8) then
	--	if tClient[sid].iProfile and tClient[sid].iProfile >= tCommands[cmd].iProfile then
	--	if tClient[sid].iProfile and tClient[sid].iProfile >= (tCommands[cmd].iUseProfile or tCommands[cmd].iProfile or 10) then
		if tClient[sid].iProfile and tClient[sid].iProfile >= (tCommands[cmd].iUseProfile or tCommands[cmd].iProfile) then
			local res, reply = false, "Error executing text command: "..line
			if string.find(line, "^%S+ %-[Hh]%s-$") then
				local help = (tCommands[cmd].sHelpText and tCommands[cmd].sHelpText~="" and tCommands[cmd].sHelpText or "No help available for this textcommand.")

				--// Add hubcommand to cmdhistory
				table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, true, line})

				return true, help
			else
				res, reply = tUser:DoTrigger(sid, cmd, line)
			end

			if res then
				tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Text Command successful: "..line, "HI")
			else
				tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Text Command failed: "..line, "HI")
			end

			--// Add hubcommand to cmdhistory
			table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, res, line})

			return res, reply
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Text Command failed, no permission: "..line, "HI")

			--// Add hubcommand to cmdhistory
			table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

			return false, Lang(sid, 0, "You are not allowed to use this command")
		end
	elseif tHubSettings.Main.bIgnoreUnknownCmds then
		tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command failed: "..line, "HI", 3)
		--// Add hubcommand to cmdhistory
		table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

		-- returning nil for 'output' and true as fourth value to tell the calling function that it should keep handling the message
		return true, line, nil, true
	else
		tFuncs:Debug(tUser:GetDataLine(sid).."["..toroom.."] Command failed: "..line, "HI")
		--// Add hubcommand to cmdhistory
		table.insert(tLogs.HubCommands.History, {socket.gettime(), tClient[sid].sNick, tClient[sid].sIP, false, line})

		return false, Lang(sid, 0, "Unknown command: %s", cmd)
	end
end,

["FilterPass"] = function(self, line, cmd)
	if cmd == "addreg" then
		line = line:gsub("^(%S+ %S+ )(%S+)(.*)", "%1 ****** %3")
	elseif cmd == "setpass" then
		if string.find(line, "^%S+ (%S+) (%S+)")  then
			line = line:gsub("^(%S+ %S+ )(%S+)(.*)", "%1 ****** %3")
		else
			line = line:gsub("^(%S+ )(%S+)(.*)", "%1 ****** %3")
		end
	elseif cmd == "setmypass" then
		line = line:gsub("^(%S+)%s+(%S+)(.*)", "%1 ****** %3")
	elseif cmd == "regme" then
		line = line:gsub("^(%S+)%s+(%S+)(.*)", "%1 ****** %3")
	elseif cmd == "webstats_reg" then
		line = line:gsub("^(%S+%s+%S+)%s+(%S+)(.*)", "%1 ****** %3")
	end
	return line
end,

["CalcSpeeds"] = function(self)
	local ctime = socket.gettime()

	if ((tStats.Tcp.LastSpeedMeasure or 0) + 10) <= ctime then 	--// Recalculate speeds every x seconds
		local receivedtotal = tStats.Tcp.Received		--// read them once to prevent loss during other threads writing to this var
		local senttotal = tStats.Tcp.Sent			--// read them once to prevent loss during other threads writing to this var
		local interval = ctime - (tStats.Tcp.LastSpeedMeasure or 0)
		local received = receivedtotal - (tStats.Tcp.ReceivedInterval or 0)
		local sent = senttotal - (tStats.Tcp.SentInterval or 0)
		local receivedspeed = received / interval
		local sentspeed = sent / interval

		tStats.Tcp.ReceivedSpeed = receivedspeed
		tStats.Tcp.SentSpeed = sentspeed
		if receivedspeed > tStats.Tcp.ReceivedPeak then tStats.Tcp.ReceivedPeak = receivedspeed end
		if sentspeed > tStats.Tcp.SentPeak then tStats.Tcp.SentPeak = sentspeed end

		tStats.Tcp.ReceivedInterval = receivedtotal
		tStats.Tcp.SentInterval = senttotal
		tStats.Tcp.LastSpeedMeasure = ctime
	end

	local gc = collectgarbage("count")
	local max = tStats.Hub.MaxMem or 0
	if gc > max then
		tStats.Hub.MaxMem = gc
	end

	if tServer.GuiSID then
		if (ctime - tSettings.Gui.UpdateInterval) > tStats.Tcp.LastShown or 0 then
			tFuncs:SendToGui("ST"..
				" TTI".. tStats.Tcp.Received..
				" TTO".. tStats.Tcp.Sent..
				" TIS".. string.format("%.3f", tStats.Tcp.ReceivedSpeed)..
				" TOS".. string.format("%.3f", tStats.Tcp.SentSpeed)..
				" TIP".. string.format("%.3f", tStats.Tcp.ReceivedPeak)..
				" TOP".. string.format("%.3f", tStats.Tcp.SentPeak)..
				" CTT".. string.format("%.3f", tStats.Timers.Thread)..
				" CTR".. string.format("%.3f", tStats.Timers.ReadSockets)..
				" CTS".. string.format("%.3f", tStats.Timers.SendSockets)..
				" CTP".. string.format("%.3f", tStats.Timers.ParseCmds)..
				" MMC"..string.format("%.3f", collectgarbage("count"))..
				""
			)
			tStats.Tcp.LastShown = ctime
		end
	end
end,

["TCPParse"] = function(self, client, sid, parsetxt)
	--// Store debugging info
	local handler = client.xHandler or "none"
	if not tCurrent[handler] then tCurrent[handler] = {} end
	tCurrent.Handler = handler
	tCurrent[handler].Line = parsetxt
	tCurrent[handler].Sid = sid
	tCurrent[handler].Protocol = client.sProtocol or "none"
	tCurrent[handler].Nick = client.sNick or "unknown"

--Hub: [Incoming][81.16.170.188:7777] $Search Hub:BadWolfGrim883 F?T?0?1?Terry$Pratchett$Wintersmith
--Hub: [Outgoing][81.16.170.188:7777] $SR Jewbel -=|[ Share 2 ]|=-\E-Books\Fiction\Terry Pratchett\Terry Pratchett - Discworld 35 - Wintersmith.lit*548031 2/3*TTH:OUPC7ZFZYAWJ5YSLSNMCUASI4U4PNLL63HXEPGI (81.16.170.188:7777)*BadWolfGrim883|

	--// Prepare for parsing	-- always parse even if socket gives an error, partialtext can contain last commands from client


	if client.sProtocol == "NMDC" then
		if not tBuffer[sid].Skip then

			local parsetxtlength = parsetxt:len()
			local cmdbuffer = false
			local curlength = 0

			for line in parsetxt:gmatch("(.-)|") do
				tCurrent[handler].Line = line
				curlength = curlength + line:len() + 1 	--// Count handled characters from parsetxt (including the endpipe)

				--// handle $SR endpipes
				local firstchar = line:sub(1,1)
				if line:sub(1,3) == "$SR" and curlength < parsetxtlength then --// Skip execution of $SR until second proper command comes in, and if it isn't end of total parsetxt
					cmdbuffer = line
					line = false	--// skip parsing
				elseif cmdbuffer and firstchar ~= "$" and firstchar ~= "<" then	--// Found more text to add to $SR
					line = cmdbuffer..line
					cmdbuffer = false	--// parse combined line
				elseif cmdbuffer then														--// New line doesn't belong to previous $SR so parse cmdbuffer first
					if client and cmdbuffer ~= "" and not tBuffer[sid].bDisconnect then
						tUser:Spy(client, sid, "[RCV][NMDC] "..line)
						local result, ret = pcall(NMDC_DataArrival, client, sid, cmdbuffer)
						if not result then
							tFuncs:Debug("[ERROR] [NMDC_DataArrival] "..ret.."\n"..tUser:GetDataLine(sid).."Data: "..cmdbuffer, "FI", -2) --// Force error to console
							tFuncs:UnlockMutexes()
						end
					end
					cmdbuffer = false --// now parse the line
				end

				--// Send to parser or parsetable
				if client and line and line ~= "" and not tBuffer[sid].bDisconnect and not cmdbuffer then
					--tFuncs:Debug(tUser:GetDataLine(sid)..line, "TNR", 9)
					tUser:Spy(client, sid, "[RCV][NMDC] "..line)
					local result, ret = pcall(NMDC_DataArrival, client, sid, line)
					if not result then
						tFuncs:Debug("[ERROR] [NMDC_DataArrival] "..ret.."\n"..tUser:GetDataLine(sid).."Data: "..line, "FI", -2) --// Force error to console
							tFuncs:UnlockMutexes()
					end
				end
			end

		else
			tFuncs:Debug(tUser:GetDataLine(sid).."[TCPParse] Skipping datablock because of conversion problem: "..parsetxt:sub(1,40).."...", "CN", 1)
		end
		if parsetxt:find("(.+|)") then
			tBuffer[sid].Skip = nil
		end
		tBuffer[sid].Receive = parsetxt:gsub("^.+|", "") or ""
	elseif client.sProtocol == "ADC" then

		if not tBuffer[sid].Skip then
			for line in parsetxt:gmatch("(.-)\n") do
				tCurrent[handler].Line = line

				--// Send to parser or parsetable
				--msg = ADC:Escape(msg)
				if client and line and line ~= "" and not tBuffer[sid].bDisconnect then -- disregard empty messages
				--	tFuncs:Debug(tUser:GetDataLine(sid)..line, "TAR", 9)
					tUser:Spy(client, sid, "[RCV][ADC] "..line)
					local result, ret = pcall(ADC_DataArrival, client, sid, line)
					if not result then
						tFuncs:Debug("[ERROR] [ADC_DataArrival] "..ret.."\n"..tUser:GetDataLine(sid).."Data: "..line, "FI", -2) --// Force error to console
						tFuncs:UnlockMutexes()
					else
--~ 						client = ret	--// Redundant? client will always be referenced, no separate copy!
					end
				end
			end
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."[TCPParse] Skipping datablock because of conversion problem: "..parsetxt:sub(1,40).."...", "CA", 1)
		end

		if parsetxt:find("(.+\n)") then
			tBuffer[sid].Skip = nil
		end
		tBuffer[sid].Receive = parsetxt:gsub("^.+\n", "") or ""
	elseif client.sProtocol == "HTML" then
		local line = parsetxt
		tCurrent[handler].Line = line

		--// Send to parser or parsetable
		--msg = ADC:Escape(msg)
		if client and line and line ~= "" and not tBuffer[sid].bDisconnect then -- disregard empty messages
			tUser:Spy(client, sid, "[HTML RCV] "..line)
			local result, ret = pcall(HTML_DataArrival, client, sid, line)
			if not result then
--~ 					local traceback = ShowTrace(15)
				tFuncs:Debug(tUser:GetDataLine(sid).."[TCPParse] [ERROR] HTML_DataArrival: "..ret, "FI", -2) --// Force error to console
				tFuncs:UnlockMutexes()

			else
				client = ret
			end
		else
			tFuncs:Debug("[TCPParse] Failed execution", "EI", 1)
		end
		tBuffer[sid].Receive = ""
	elseif #parsetxt > 2 then
		if #parsetxt > 40 then parsetxt = parsetxt:sub(1,100).."..." end
		tFuncs:Debug(tUser:GetDataLine(sid).."Received data without protocol: "..parsetxt..", user disconnected")
        	tUser:Disconnect(sid, "Received data without proper protocol", "Received invalid data, no protocol: "..parsetxt)
	end
--~ 	collectgarbage("collect")
	return client
end,

--// Modified TCPSend to drop the Chat buffer and use Data buffer only, separating data from chat could lead to messages being sent in the wrong order (possibly causing the ignoring CID mainchat messages)
--[[ TCPSend
["TCPSend"] = function(self, client, sid)
	--// TODO: Check if there are clients where nothing can be sent, try for a number of seconds then disconnect the user
	--// Unresponsive users might be the cause of current (big) lag with higher usercounts
	local ctime = socket.gettime()

	--// Concatenate the Sendbuffer tables to one string
	local sendchattotal = ""
	local senddatatotal = ""

	--// Chatmessages will be queued until handler is set to "Main"
	if client.xHandler == "Login" then
--~ 		for _,loginmsg in ipairs(tBuffer[sid].Send.Login) do
--~ 			senddatatotal = senddatatotal..loginmsg
--~ 		end
		senddatatotal = table.concat(tBuffer[sid].Send.Login)
		tBuffer[sid].Send.Login = {}
	else
		--local restchat = {}
--~ 		for num,chatmsg in ipairs(tBuffer[sid].Send.Chat) do
--~ 			if num < 250 and #tBuffer[sid].Send.Buffer < 300 then	--// Will add max. 250 lines at once if sendbuffer is almost empty
--~ 				sendchattotal = sendchattotal..chatmsg
--~ 			else
--~ 				restchat[#restchat+1] = chatmsg
--~ 			end
--~ 		end
--~ 		tBuffer[sid].Send.Chat = restchat

		--// Replacing the concatenation of strings in the code above with on table.concat saves memory doubling for string on stack, huge performance improvement due to less garbage collecting
		sendchattotal = table.concat(tBuffer[sid].Send.Chat)
		tBuffer[sid].Send.Chat = {}

		--// Added a timer/counter here to reduce wait for new data instead of sending each cycle
		--local restdata = {}

		if tUser:Supports(sid, "FGUI") then --// Exception for GUI, immediate data sending
--~ 			for num,datamsg in ipairs(tBuffer[sid].Send.Data) do
--~ 				senddatatotal = senddatatotal..datamsg
--~ 			end
			senddatatotal = table.concat(tBuffer[sid].Send.Data)
			tBuffer[sid].Send.Data = {}

			tBuffer[sid].Send.Last = ctime	--// Can only prioritize chat, if chat is muted for first seconds because chat needs BINF in ADC
		else
			if (tBuffer[sid].Send.Last + tHubSettings.tTcp.DataSkipTime) <= ctime then
--~ 				for num,datamsg in ipairs(tBuffer[sid].Send.Data) do
--~ 					if num < 250 and #tBuffer[sid].Send.Buffer < 300 then	--// Make sure buffer is almost empty before adding new lines
--~ 						senddatatotal = senddatatotal..datamsg
--~ 					else
--~ 						restdata[#restdata+1] = datamsg
--~ 					end
--~ 				end
--~ 				tBuffer[sid].Send.Data = restdata

				senddatatotal = table.concat(tBuffer[sid].Send.Data)
				tBuffer[sid].Send.Data = {}

				tBuffer[sid].Send.Last = os.time()	--// Can only prioritize chat, if chat is muted for first seconds because chat needs BINF in ADC
			end
		end
	end

	--// Chat will be send with higher priority than data
	local sendtotal = sendchattotal..senddatatotal

	tStats.Tcp.SentRaw = (tStats.Tcp.SentRaw or 0)+#sendtotal
	local protocol = client.sProtocol or "OTHER"
	tStats.Tcp["Sent"..protocol] = (tStats.Tcp["Sent"..protocol] or 0) + #sendtotal

--~ 	if client.sProtocol == "ADC" then
--~ 		tStats.Tcp.SentADC = (tStats.Tcp.SentADC or 0)+#sendtotal
--~ 	elseif client.sProtocol == "NMDC" then
--~ 		tStats.Tcp.SentNMDC = (tStats.Tcp.SentNMDC or 0)+#sendtotal
--~ 	elseif client.sProtocol == "HTML" then
--~ 		tStats.Tcp.SentHTML = (tStats.Tcp.SentHTML or 0)+#sendtotal
--~ 	end

	--// Output if monitored
--~ 	if #sendtotal > 0 then tUser:Spy(client, sid, "[SND] "..sendtotal) end

	--// Compress data
	sendtotal = tFuncs:ZPipe(client, sid, sendtotal)

	--// Add previous sendbuffer (can include Zpiped data) to new totalstring
	sendtotal = tBuffer[sid].Send.Buffer..sendtotal
	if #sendtotal == 0 then return client end

	--// Set client status
	if #tBuffer[sid].Send.Buffer > 0 then
		client.sStatus = "[SENT] <Buffered data> "..string.sub(sendtotal, 1, 50)
	elseif #sendtotal > 0 then
		client.sStatus = "[SENT] "..string.sub(sendtotal, 1, 50)
	end

	--// SEND
	if #sendtotal > 0 then
--~ 		client.xSocket:settimeout(tSettings.TcpSocketSendTimeout, tSettings.TcpSocketSendTimeoutMode)
		client = tFuncs:SendSocket(client, sid, sendtotal)
	end

	return client
end,
]]--

["TCPSend"] = function(self, client, sid)		--// Simplified, no more Login and Chat buffers, only Data (the tSend:Raw "Delay" option just determines when to send the entire data buffer)
	--// TODO: Check if there are clients where nothing can be sent, try for a number of seconds then disconnect the user
	--// Unresponsive users might be the cause of current (big) lag with higher usercounts
	local buffer = tBuffer[sid].Send
	local ctime = socket.gettime()
	local res, err, bytes = false, false, false
	local tcpstats = tStats.Tcp

	--// Only try to send if no pause is set due to earlier failing sends
	if (buffer.bSendNow or (not buffer.iBufferSent) or (ctime >= buffer.iBufferSent + tHubSettings.tTcp.SendInterval)) and (not buffer.iPauseUntil or buffer.iPauseUntil <= ctime) then

		--// Concatenate the Sendbuffer tables to one string
		local sendtotal = table.concat(tBuffer[sid].Send.Data)
		tBuffer[sid].Send.Data = {}
		tBuffer[sid].Send.Last = ctime	--// Can only prioritize chat, if chat is muted for first seconds because chat needs BINF in ADC

		tcpstats.SentRaw = (tcpstats.SentRaw or 0)+#sendtotal
		local protocol = client.sProtocol or "OTHER"
		tcpstats["Sent"..protocol] = (tcpstats["Sent"..protocol] or 0) + #sendtotal

		--// Output if monitored
	--~ 	if #sendtotal > 0 then tUser:Spy(client, sid, "[SND] "..sendtotal) end

		--// Compress data
		sendtotal = tFuncs:ZPipe(client, sid, sendtotal)

		--// Add previous sendbuffer (can include Zpiped data) to new totalstring
		sendtotal = tBuffer[sid].Send.Buffer..sendtotal
		if #sendtotal == 0 then return client end

		buffer.iBufferSent = false --// Buffer is filled again, so resetting time of last successful buffer empting

		--// Set client status
		if #tBuffer[sid].Send.Buffer > 0 then
			client.sStatus = "[SENT] <Buffered data> "..string.sub(sendtotal, 1, 50)
		elseif #sendtotal > 0 then
			client.sStatus = "[SENT] "..string.sub(sendtotal, 1, 50)
		end

		--// SEND
		if #sendtotal > 0 then
	--~ 		client.xSocket:settimeout(tSettings.TcpSocketSendTimeout, tSettings.TcpSocketSendTimeoutMode)
			tcpstats.SendNow = tcpstats.SendNow + 1
			client = tFuncs:SendSocket(client, sid, sendtotal)
		end
	else
		tcpstats.SendDelayed =tcpstats.SendDelayed + 1
	end

	return client
end,

["ZPipe"] = function(self, client, sid, sendtotal)
	local tcpstats = tStats.Tcp
	local timerstats = tStats.Timers
	--// Zpipe data
	local protocol = client.sProtocol or "OTHER"
	tcpstats.SentMergedRaw = tcpstats.SentMergedRaw +#sendtotal
	tcpstats["SentRaw"..protocol] = (tcpstats["SentRaw"..protocol] or 0)+#sendtotal

	if tHubSettings.tTraffic.ZPipeMin > 0 and #sendtotal > (tHubSettings.tTraffic.ZPipeMin or 200) then	--// Only compress large messages
		--// NMDC ZPipe compression
		if client.sProtocol == "NMDC" then
			if tUser:Supports(sid, "ZPipe0") or tUser:Supports(sid, "ZPipe") then
				tcpstats["SentZlibInput"..protocol] = (tcpstats["SentZlibInput"..protocol] or 0) + #sendtotal
				local ztimer = socket.gettime()
				local sendzlib, err = zlib.compress(sendtotal,1)
				ztimer = socket.gettime() - ztimer
				if ztimer > timerstats.SendZlibMax then
					timerstats.SendZlibMax = ztimer

	--~ 				Error: TCPSend: C:/hosting/nebularis/NCBf/./core/Hub-Funcs.lua:1691: attempt to concatenate local 'sendzlib' (a nil value)
					timerstats.SendZlibMaxId = tUser:GetDataLine(sid)
				end
				if not sendzlib then
					tFuncs:Debug("[tFuncs:ZPipe] [ERROR] "..tUser:GetDataLine(sid).. "Couldn't compress with zlib. / Error: "..tostring(err).." / Data: "..sendtotal)
				else
					sendtotal = "$ZOn|"..sendzlib
				end
				tcpstats["SentZlibOutput"..client.sProtocol] = (tcpstats["SentZlibOutput"..client.sProtocol] or 0) + #sendtotal
			else
				tcpstats["SentNoZlibSupport"..client.sProtocol] = (tcpstats["SentNoZlibSupport"..client.sProtocol] or 0) + #sendtotal
			end

		--// ADC ZPipe compression
		elseif client.sProtocol == "ADC" then
			if tUser:Supports(sid, "ZLIF") then
				tcpstats["SentZlibInput"..protocol] = (tcpstats["SentZlibInput"..protocol] or 0) + #sendtotal
	--~ 			tFuncs:Debug("Client supports ZLIF, compressing: "..sendtotal.."IZOF\n")
				local ztimer = socket.gettime()
	--~ 			local sendzlib, err = zlib.compress((sendtotal.."IZOF\n"),1)	--// Looks like ZOF\n isn't needed
				local sendzlib, err = zlib.compress((sendtotal),1)
				ztimer = socket.gettime() - ztimer
				if ztimer > timerstats.SendZlibMax then
					timerstats.SendZlibMax = ztimer
					timerstats.SendZlibMaxId = tUser:GetDataLine(sid)
				end

				if not sendzlib then
					tFuncs:Debug("[tFuncs:ZPipe] [ERROR] "..tUser:GetDataLine(sid).. "Couldn't compress with zlib. / Error: "..tostring(err).." / Data: "..sendtotal)
				else
					sendtotal = "IZON\n"..sendzlib
				end
				tcpstats["SentZlibOutput"..protocol] = (tcpstats["SentZlibOutput"..protocol] or 0) + #sendtotal
			else
				tcpstats["SentNoZlibSupport"..client.sProtocol] = (tcpstats["SentNoZlibSupport"..client.sProtocol] or 0) + #sendtotal
			end
		else
			tcpstats["SentUncompressed"..protocol] = (tcpstats["SentUncompressed"..protocol] or 0) + #sendtotal
		end
	else
		tcpstats["SentUncompressed"..protocol] = (tcpstats["SentUncompressed"..protocol] or 0) + #sendtotal
	end

	tcpstats["SentMerged"..protocol] = (tcpstats["SentMerged"..protocol] or 0)+#sendtotal
	tcpstats.SentMerged = (tcpstats.SentMerged or 0)+#sendtotal

	return sendtotal
end,

["SendSocket"] = function(self, client, sid, sendtotal)
	local buffer = tBuffer[sid].Send
	local timerstats = tStats.Timers
	local tcpstats = tStats.Tcp

	local ctime = socket.gettime()
	local res, err, bytes = false, false, false

	--// Only try to send if no pause is set due to earlier failing sends
--~ 	if (buffer.bSendNow or (buffer.iBufferSent + tHubSettings.tTcp.SendInterval >= ctime)) and (not buffer.iPauseUntil or buffer.iPauseUntil <= ctime) then
		--// WARNING!!!!  Experimental addition of socket select to check if sending on the socket is allowed    // 2011/02/01

		local rect,sendt = socket.select({}, {client.xSocket}, 0)
		if #sendt < 1 then return client end
		res, err, bytes = client.xSocket:send(sendtotal)

--~ 	else
--~ 		tFuncs:Debug("Socket select doesn't allow sending on socket for IP: "..client.sIP, "FI", 5)
--~ 		return client
--~ 	end

--~ 	if client.sMonitor then
--~ 		tUser:Spy(client, sid, "[SND] Send socket data, Bytes: "..tostring(bytes).." Error: "..tostring(err).." Result: "..tostring(res))
--~ 	end

	--// TIMERS AND STATS
	local xtimer = socket.gettime() - ctime
	timerstats.SendSocketsTemp = timerstats.SendSocketsTemp + xtimer
	if xtimer > timerstats.SendSingleSocketMax then
		timerstats.SendSingleSocketMax = xtimer
		timerstats.SendSingleSocketMaxId = tUser:GetDataLine(sid)
	end

	--// See how much is sent, and store the rest in sSendbuffer (compressed string) or tSendbuffer (uncompressed table)
	if not res and (err == "timeout" or err == "wantwrite" or err == "wantread" ) then    --// if a timeout occurs while sending, store the remaining message to be sent the next loop
--~ 		tClient[sid].sStatus = "[SENT] "..string.sub(sendtotal, 1, bytes)
		tcpstats.Sent = (tcpstats.Sent or 0) + bytes

		buffer.Buffer = string.sub(sendtotal, bytes+1)

		if bytes < 1 and not tUser:Supports(sid, "FGUI") then
			buffer.iFailed = buffer.iFailed or ctime
			buffer.iSendTry = (buffer.iSendTry or 0) + 1
			tFuncs:Debug("Failed sending data to IP: "..client.sIP.." Nick: "..(client.sNick or "<unknown>").." Error: "..err.." Tries: "..buffer.iSendTry, "FI" , 5)

			--// Flush buffer when socket isn't responding after SendFlushTime seconds
			if ctime-buffer.iFailed >= tHubSettings.tTcp.SendFlushTime then
				tcpstats.BuffersFlushed = (tcpstats.BuffersFlushed or 0) + 1
				buffer.iFailed = false
				buffer.iSendTry = 0
				buffer.iPauseUntil = false
				buffer.Buffer = ""
			--// Flush buffer when socket isn't responding after SendFlushTime seconds
			elseif buffer.iSendTry > 2 then
				buffer.iPauseUntil = ctime + tHubSettings.tTcp.NonRespondWait
				if ctime > buffer.iFailed+tHubSettings.tTcp.SendTimeout then
					tcpstats.SendBufferFull = tcpstats.SendBufferFull+1
					buffer.Buffer = "" -- couldn't send data, flush buffer
					buffer.iSendTry = 0
					buffer.iPauseUntil = false
					if client.xHandler == "Login" then
						tUser:Disconnect(sid, nil, "Timeout reached on socket during login, client disconnected.", "FI", 2)
						tLogs.Errors.LoginTimeouts = tLogs.Errors.LoginTimeouts+1
					else
						tUser:Disconnect(sid, nil, "Timeout reached on socket, unresponsive, client disconnected.", "FI", 2)  --// This might cause dropping users when bandwidth limit is reached
						tLogs.Errors.Socket = tLogs.Errors.Socket+1
					end
				end
			end
		else
			buffer.iFailed = false
			buffer.iSendTry = 0
			buffer.iPauseUntil = false
		end
	elseif not res then
		tUser:Disconnect(sid, nil, "Unknown error on socket: "..err, "FI", 2)
	elseif res then
		--// Sent complete buffer to user, can use sendinterval now until buffer.bSendNow is true (is set when PM or MC is added to buffer)
		buffer.iBufferSent = ctime
		buffer.bSendNow = false

		tcpstats.Sent = (tcpstats.Sent or 0) + res
		buffer.iFailed = false
		buffer.iSendTry = 0
		buffer.tPauseUntil = false
		buffer.Buffer = ""
	else
		tUser:Disconnect(sid, nil, "Sending to socket failed. Error: "..err, "FI", 2)
	end

--~ 	tBuffer[sid].Send = buffer
	return client
end,

["DetectProtocol"] = function(self, client, sid, parsetxt)
	if string.find(parsetxt, "^%$Supports ") then
		client.sProtocol = "NMDC"       --// and local
		tFuncs:Debug(tUser:GetDataLine(sid).."NMDC Protocol detected.", "LNJ",2)
		return true, client
	elseif string.find(parsetxt, "^HSUP ") then
		client.sProtocol = "ADC"    --// and local
		tFuncs:Debug(tUser:GetDataLine(sid).."ADC Protocol detected.", "LAJ",2)
		return true, client
	end
	return false, client
end,

["ResetUserCount"] = function(self, recount)
	tStats.Hub.Users.Total = 0
	tStats.Hub.Users.ADC = 0
	tStats.Hub.Users.NMDC = 0
	tStats.Hub.Users.HTTP = 0
	if recount then
		--// Do recount here
		for sid, data in pairs(tClient) do
			if not data.bBot and data.sProtocol then
				tStats.Hub.Users[data.sProtocol] = (tStats.Hub.Users[data.sProtocol]  or 0) +1
			end
		end
	end
end,

["UpdateUserCount"] = function(self, amount, protocol)
	tStats.Hub.Users.Total = ((tStats.Hub.Users.Total or 0) + (amount or 0))
	if tStats.Hub.Users.Total > tStats.Hub.Users.TotalMax then tStats.Hub.Users.TotalMax = tStats.Hub.Users.Total end
	tStats.Hub.Users[protocol] = ((tStats.Hub.Users[protocol] or 0) + amount)
	if tStats.Hub.Users[protocol] > tStats.Hub.Users[protocol.."Max"] then tStats.Hub.Users[protocol.."Max"] = tStats.Hub.Users[protocol] end

	tFuncs:SendToGui("ST UCA"..tStats.Hub.Users.ADC.." UCN"..tStats.Hub.Users.NMDC.." UCT".. tStats.Hub.Users.Total)
end,

["ClientQuit"] = function (self, client, sid)
	local ctime = socket.gettime()
	--// Check if it's GUI
	if tServer.GuiSID == sid then
		tServer.GuiSID = false
	end

	--// Change last data
	client.iActive = -1  --// closed
	if sid ~= "AAAA" and client.xSocket then
		client.xSocket:close()
		client.xSocket = nil
	end
	client.tSendbuffer = nil
	client.sReceivebuffer = nil
	client.iLogoutTime = ctime

	--// Store seconds online for this protocol
	if tStats.Users[(client.sProtocol or "Unknown")] then
		tStats.Users[(client.sProtocol or "Unknown")].OnlineTime = (tStats.Users[(client.sProtocol or "Unknown")].OnlineTime or 0) + (client.iLogoutTime - client.iLoginTime)
	end

	--// Update usercount
	if client.iState == 7 then

		client.iTotalOnlineTime = (client.iTotalOnlineTime or 0) + (ctime - client.iLoginTime)
		--// Update total onlinetime if user is registered
		local lnick = client.sNick:lower()
		if tAccounts[lnick] then
			tAccounts[lnick].iTotalOnlineTime = client.iTotalOnlineTime		--// will be retrieved again from accounts when logging in
			tAccounts[lnick].iLastLogout = ctime
		end

		if tStats.Hub.Status ~= "Offline" then
			tUser:Quit(sid)
--~ 			tFuncs:UpdateUserCount(-1, client.sProtocol)
		end
	end

	--// Move info to Nicklog
	if client.sNick then

		--// Cleanup
		client.sNickTemp = nil
		client.iProfile = nil
		client.sProfile = nil
		client.bQuit = nil
		client.iLoginDuration = nil
		client.tHandledMsg = nil
		client.bHasKey = nil
		client.sPID = nil
		client.iProfileTemp = nil
		client.sPassSalt = nil
		client.iTcpAge = nil
		client.tSupports = nil
		client.sStatus = nil
		client.bBot = nil
		client.iState = nil
		client.iActive = nil
		client.sIPEx = nil
		client.bOperator = nil
		client.xHandler = nil
--~ 		client.tStats = nil		--// Offline !userstats aren't available atm, but maybe added

		local nick = string.lower(client.sNick)
		if not tLogs.Nick[nick] then
			tLogs.Nick[nick] = client
		end
		client.tIPHistory = tLogs.Nick[nick].tIPHistory or {}
--~ 		table.insert(client.tIPHistory, {["iLoginTime"] = client.iLoginTime, ["sProtocol"] = client.sProtocol, ["sNick"] = client.sNick, ["sIP"] = client.sIP, ["iLogoutTime"] = ctime })	--// sNick needed here??
		table.insert(client.tIPHistory, {["iLoginTime"] = client.iLoginTime, ["sProtocol"] = client.sProtocol, ["sIP"] = client.sIP, ["iLogoutTime"] = ctime })
		if #client.tIPHistory > 10 then table.remove(client.tIPHistory, 1) end
		tLogs.Nick[nick] = client      --// Store userdata in tLogs
		tLogs.Nick[nick].tInfo = nil
		--tNicks[client.sNick] = nil		--// Remove nick from current nicktable
		tLNicks[nick] = nil
	end

	--// Move info to IPlog
	ThreadLockMutex(mutexiplog)
	if tLogs.IP[client.sIP] then
		local nick = client.sNick or "pre-login / nick unknown"
		if tLogs.IP[client.sIP][nick] then
			local counter = tLogs.IP[client.sIP][nick].iCounter + 1
			tLogs.IP[client.sIP][nick] = { ["iCounter"] = counter, ["iLoginTime"] = client.iLoginTime, ["iLogoutTime"] = os.time() }
		else
			tLogs.IP[client.sIP][nick] = { ["iCounter"] = 1, ["iLoginTime"] = client.iLoginTime, ["iLogoutTime"] = os.time() }
		end
	else
	end
	ThreadUnlockMutex(mutexiplog)

	--// Cleanup
	tFlood[sid] = nil
	tBuffer[sid] = nil
	return nil
end,

["DateToEpoch"] = function(self, datestring)
	-- Assuming a date pattern like: yyyy-mm-dd hh:mm:ss
	if not datestring then return false end
	local pattern = "(%d+)-(%d+)-(%d+) (%d+):(%d+):(%d+)"
	local runyear, runmonth, runday, runhour, runminute, runseconds = datestring:match(pattern)
	if runyear and runseconds then
		regdate = os.time({year = runyear, month = runmonth, day = runday, hour = runhour, minute = runminute, seconds = runseconds})
		return regdate
	else
		return false
	end
end,

["ConvertShare"] = function(self,value, raw)
	if raw then return string.format("%.0f", (tonumber(value or 0) or 0)) end	--// Workaround for easier coding in FillVars
	if not value or value == 0 then return "0 B" end
	local iBytes = tonumber(value)
	if not iBytes then return false, "Unconvertable size: "..tostring(value) end

	local i, tUnits = 1, {"B","kB","MB","GB","TB","PB","EB","YB"}
	while iBytes >= 1024 do iBytes = iBytes / 1024 i = i + 1 end
	return string.format("%.2f",iBytes).." "..(tUnits[i] or "??")
end,

["ParseFeatures"]  = function(self, msg)
	local has = {}
	local hasnot = {}
	if msg then
		for word in string.gmatch(msg, "%+(%w+)") do
			table.insert(has, word)
		end
		for word in string.gmatch(msg, "%-(%w+)") do
			table.insert(hasnot, word)
		end
	end
	return has, hasnot
end,

["IsIP"] = function(self, str, partial)
	if not str or type(str) ~= "string" then return false end
	if not partial then
		return tFuncs:iff(str:find("^[12]?[%d]?%d%.[12]?[%d]?%d%.[12]?[%d]?%d%.[12]?[%d]?%d$"),true,false)
	elseif tFuncs:iff(str:find("^[12]?[%d]?%d%.[12]?[%d]?%d%.[12]?[%d]?%d%.[12]?[%d]?%d$"),true,false) then
		return true, 4
	elseif tFuncs:iff(str:find("^[12]?[%d]?%d%.[12]?[%d]?%d%.[12]?[%d]?%d$"),true,false) then
		return true, 3
	elseif tFuncs:iff(str:find("^[12]?[%d]?%d%.[12]?[%d]?%d$"),true,false) then
		return true, 2
	elseif tFuncs:iff(str:find("^[12]?[%d]?%d$"),true,false) then
		return true, 1
	end

	return false
end,

["IP2DEC"] = function(self,sIP)
	local iDec,iNum = 0,0
	for num in sIP:gmatch("(%d+)") do
		iNum = iNum + 1
		if tonumber(num) < 0 or tonumber(num) > 255 then return -1 end
		iDec = iDec + num*256^(4-iNum)
	end
--~ 	collectgarbage("collect")
	if iNum ~= 4 then return -1 end
	return iDec
end,

["GetImage"] = function(self,img)
	dofile(_APPPATH.."res/Images.xpm")
	local ret = false
	local wximg = wx.wxImage()
	if tIcons[img] then
		ret = wx.wxBitmap(tIcons[img])
	elseif wximg:LoadFile("./res/img/"..img) then
		ret = wx.wxBitmap(wximg)
	else
		OnError("Unable to load image: "..tostring(img), true)
	end
	tIcons = nil
	wximg = nil
--~ 	collectgarbage("collect")
	return ret
end,

["LoadIcons"] = function(self)
	local imageList = wx.wxImageList(16, 16)
	imageList:Add(tFuncs:GetImage("tb-stats.png"))
	imageList:Add(tFuncs:GetImage("tb-mainchat.png"))
	imageList:Add(tFuncs:GetImage("tb-pm.png"))
	imageList:Add(tFuncs:GetImage("tb-userspy.png"))
	imageList:Add(tFuncs:GetImage("flexhub"))
	imageList:Add(tFuncs:GetImage("tb-scripts.png"))
	imageList:Add(tFuncs:GetImage("tb-login.png"))
	imageList:Add(tFuncs:GetImage("tb-search.png"))
	imageList:Add(tFuncs:GetImage("tb-hubcmds.png"))
	imageList:Add(tFuncs:GetImage("tb-clientcmds.png"))
	imageList:Add(tFuncs:GetImage("tb-tcp.png"))
--~ 	collectgarbage("collect")
	return imageList
end,

["LoadGeoIP"] = function(self, enable)
	local err = false
	if tHubSettings.Main.bUseGeoIP or enable then
		tGeoIP.loaded, err = webnet77.new(_APPPATH..tGeoIP.file)
		if not tGeoIP.loaded then
			local msg = "Couldn't load geoip file '"..(_APPPATH..tGeoIP.file or "").."' because: "..err
			tFuncs:Debug(msg, "FI",-2)
			return false, msg
		else
			local msg = "GeoIP file loaded, countrycodes will be added to userinfo on each login now."
			return true, msg
		end
	else
		local msg = "GeoIP disabled, countrycodes will no longer be added to userinfo on login."
		return true, msg
	end
end,

["EscapePercent"] = function(self, msg)
	local msg = tostring(msg):gsub("%%", "%%%%")
	return msg
end,

["FillVarFunctions"] = {

	--// Hub related variables

	["%[hubinfo]"] = function(self, vsid, user) return tFuncs:FillVars(tHubSettings.Main.sHubInfo, vsid) end,	--// Special case, can contain other FillVars, may not contain user_reginfo or hubinfo!!
	["%[address]"] = function(self) return (tHubSettings.Main.sAddress or "") end,
	["%[address_nmdc]"] = function(self)
		local res, msg = tFuncs:GetAddressPort(sid, "nmdc")
		return (res and msg or "")
		end,
	["%[address_nmdcs]"] = function(self)
		local res, msg = tFuncs:GetAddressPort(sid, "nmdcs")
		return (res and msg or "")
		end,
	["%[address_adc]"] = function(self)
		local res, msg = tFuncs:GetAddressPort(sid, "adc")
		return (res and msg or "")
		end,
	["%[address_adcs]"] = function(self)
		local res, msg = tFuncs:GetAddressPort(sid, "adcs")
		return (res and msg or "")
		end,
	["%[webstats]"] = function(self) return tHubSettings.Webstats.sPage end,
	["%[bot]"] = function(self) return tHubSettings.Main.sBotName end,
	["%[date]"] = function(self, _, _, raw) return (raw and os.time()) or (os.date(tHubSettings.Main.sDateFormat, os.time())) end,
	["%[desc]"] = function(self) return tHubSettings.Main.sDescription end,
	["%[email]"] = function(self) return tHubSettings.Main.sEmail end,
	["%[hub]"] = function(self) return tHubSettings.Main.sName end,
	["%[lang]"] = function(self) return tHubSettings.Main.sLanguage end,
	["%[minsharedownload]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tHubSettings.tRules.iMinShareDownload or 0, raw) end,
	["%[minslots]"] = function(self) return tHubSettings.tRules.iMinSlots end,
	["%[minshare]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tHubSettings.tRules.iMinShare or 0, raw) end,
	["%[network]"] = function(self) return tHubSettings.Main.sNetwork end,
	["%[website]"] = function(self) return tHubSettings.Main.sWebsite end,
	["%[opchat]"] = function(self) return tHubSettings.Main.sOpChat end,
	["%[owner]"] = function(self) return tHubSettings.Main.sOwner end,
	["%[platform]"] = function(self) return (_OS or "Unknown") end,
	["%[svn]"] = function(self) return (tSettings.General.svnversion or "Not available") end,
	["%[time]"] = function(self) return os.date(tHubSettings.Main.sTimeFormat, os.time()) end,
	["%[topic]"] = function(self) return tHubSettings.Main.sTopic end,
	["%[totalshare]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tFuncs:GetTotalShare(), raw) end,
	["%[uptime]"] = function(self, _, _, raw) return tFuncs:HubUpTime(raw) end,
	["%[usercount]"] = function(self) return (tStats.Hub.Users.Total or 0) end,
	["%[userpeak]"] = function(self) return (tStats.Hub.Users.TotalMax or 0) end,
	["%[usersadc]"] = function(self) return (tStats.Hub.Users.ADC or 0) end,
	["%[usersnmdc]" ] = function(self) return (tStats.Hub.Users.NMDC or 0) end,
	["%[version]"] = function(self) return _APPVER end,
	["%[history]"] = function(self)
		local chatlog = tLogs.Mainchat
		local numlines = tHubSettings.Main.iHistoryLines
		if chatlog and #chatlog.Lines < numlines then
			numlines = #chatlog.Lines
		elseif not chatlog or #chatlog.Lines == 0 then
			return ""
		end
		local msgt = {}
		for a=numlines,1,-1 do
			table.insert(msgt, "- ["..ToDate(chatlog.Dates[a]).."] <"..chatlog.Nicks[a].."> "..chatlog.Lines[a]..(a > 1 and "\n" or ""))
		end
		return string.gsub(table.concat(msgt, ""), "because:", "because")
	end,
	["%[quote]"] = function(self)
		if not _LASTQUOTE then _LASTQUOTE = false end
		local num = 1

		if tFilters.Quotes and #tFilters.Quotes > 0 then
			if #tFilters.Quotes > 1 then
				while true do
					num = math.random(#tFilters.Quotes)
					if num ~= _LASTQUOTE then break end
				end
			end

			if tFilters.Quotes[num] and tFilters.Quotes[num] ~= "" then
				_LASTQUOTE = num
				return tFuncs:FillVars(tFilters.Quotes[num])
			elseif tFilters.Quotes[num] == "" then
				table.remove(tFilters.Quotes, num)
				return ""
			end
		else
			return "<invalid quote: no quotes added>"
		end
	end,
	["%[parm]"] = function(self, vsid, user, dummy, tbl)
		if tbl then
			local ret = tbl[1] or ""
			table.remove(tbl, 1)
			return ret, tbl
		else
			return ""
		end
	end,

	["%[mem]"] = function(self, _, _, raw) return tFuncs:ConvertShare(math.ceil(collectgarbage("count")*1024), raw) end,
	["%[mempeak]"] = function(self, _, _, raw) return tFuncs:ConvertShare(math.ceil((tStats.Hub.MaxMem or 0)*1024), raw) end,
	["%[speedsent]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tStats.Tcp.SentSpeed or 0 , raw)..(raw and "" or "/s") end,
	["%[speedsentpeak]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tStats.Tcp.SentPeak or 0, raw)..(raw and "" or "/s") end,
	["%[speedreceived]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tStats.Tcp.ReceivedSpeed or 0, raw)..(raw and "" or "/s") end,
	["%[speedreceivedpeak]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tStats.Tcp.ReceivedPeak or 0, raw)..(raw and "" or "/s") end,
	["%[datasent]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tStats.Tcp.Sent, raw) end,
	["%[datareceived]"] = function(self, _, _, raw) return tFuncs:ConvertShare(tStats.Tcp.Received, raw) end,

	--// User related variables
	["%[user_reginfo]"] = function(self, vsid, user) return (vsid and not tUser:IsReg(vsid) and tFuncs:FillVars(tHubSettings.Main.sRegInfo, vsid) or "") end,	--// Special case, can contain other FillVars, may not contain user_reginfo or hubinfo!!
	["%[user_ip]"] = function(self, vsid, user) return user.sIP end,
	["%[user_name]"] = function(self, vsid, user) return user.sNick end,
	["%[user_comment]"] = function(self, vsid, user)
		local reg = tNick:GetReg(user.sNick)
		return reg and reg.sComment or ""
	end,
	["%[user_countrycode]"] = function(self, vsid, user) return ((tHubSettings.Main.bUseGeoIP and user.sCountryCode) or (not tHubSettings.Main.bUseGeoIP and "<GeoIP disabled>") or "<unknown>") end,
	["%[user_desc]"] = function(self, vsid, user) return (user.sDescription or "") end,
	["%[user_email]"] = function(self, vsid, user) return (user.sEmail or "") end,
	["%[user_hubport]"] = function(self, vsid, user) return user.iPort end,
	["%[user_iprofile]"] = function(self, vsid, user) return user.iProfile end,
	["%[user_isactive]"] = function(self, vsid, user) return (user.bActive and "Active" or "Passive") end,
	["%[user_isop]"] = function(self, vsid, user) return (user.bOperator and "Yes" or "No") end,
	["%[user_login]"] = function(self, vsid, user) return ToDate(user.iLoginTime) end,		--// only useful to show current logintime
	["%[user_normalhubs]"] = function(self, vsid, user) return (user.iNormalHubs or 0) end,
	["%[user_ophubs]"] = function(self, vsid, user) return (user.iOpHubs or 0) end,
	["%[user_port]"] = function(self, vsid, user) return (user.sActivePort or "not available") end,
	["%[user_profile]"] = function(self, vsid, user) return (user.sProfile or "unknown") end,
	["%[user_protocol]"] = function(self, vsid, user) return (user.sProtocol or "unknown") end,
	["%[user_reghubs]"] = function(self, vsid, user) return (user.iRegHubs or 0) end,
	["%[user_share]"] = function(self, vsid, user, raw) return tFuncs:ConvertShare(user.iShareSize or 0, raw) end,
	["%[user_slots]"] = function(self, vsid, user) return (user.iSlots or 0) end,
	["%[user_speed]"] = function(self, vsid, user, raw) return tFuncs:ConvertShare(user.iLlimit or 0, raw) end,
	["%[user_sprofile]"] = function(self, vsid, user) return user.sProfile end,
	["%[user_tag]"] = function(self, vsid, user) return (user.sTag or "") end,
	["%[user_totalhubs]"] = function(self, vsid, user) return ((user.iRegHubs or 0) + (user.iOpHubs or 0) + (user.iNormalHubs or 0)) end,
	["%[user_downloadinfo]"] = function(self, vsid, user)
		if not vsid then return "" end
		local downloadinfo = ""
		local candownload, downloadinfo = tUser:CanDownload(vsid)
		if candownload then
			downloadinfo = "Downloading and searching is enabled, max. "..(user.bActive and tSecurity.SA.iMaxCount1 or tSecurity.SP.iMaxCount1).." searches in "..(user.bActive and tSecurity.SA.iTime1 or tSecurity.SP.iTime1).." seconds."
		end
		return downloadinfo
	end,
	["%[user_profilemessage]"] = function(self, vsid, user)
		local msg = ""
		local prof = tClient[vsid].sProfile or "User"
		if prof then
			msg = tProfiles[prof].sMessage or ""
		end

		if msg and msg ~= "" then
			return tFuncs:FillVars(msg, vsid, user)
		else
			return ""
		end
	end,
},

["FillVars"] = function(self, msg, vsid, params, data)
	-- // 2011-10-28
	-- // params is a string to use to pass a parameter or a trigger to fillvars, data is extra table or data as needed
	-- // added for %[parm] handling
	-- // 2012-02-04
	-- // added set:

	--// All vars changed to functions, to avoid having to populate all vars with values even when not used
	if not msg or msg == "" then return "" end
	local user = (vsid and vsid ~= "AAAA" and vsid ~= "BBBB" and vsid ~= "CCCC" and tClient[vsid])
	local retmsg = msg

--[[
	--// "user[^_]%S+" and "my%S+"
	if params and #params > 0 then
		for _, param in ipairs(params) do
			retmsg = retmsg:gsub("%%%[param"..word,tFuncs:EscapePercent(fill))

	end
]]--

	for word in msg:gmatch("(%%%[.-%])") do
		local raw = false
		local cword = word:lower()
		local param = cword:match(":(.-)%]")
		if param then cword = cword:gsub(":.-%]", "]") end

		if tFuncs.FillVarFunctions[cword] then
			local fill = ""

			if cword:find("^..user_") then
				if user then
					fill = tFuncs.FillVarFunctions[cword](true, vsid, user, param)
				else
					fill = "<invalid: no user>"
				end
			else
				fill, data = tFuncs.FillVarFunctions[cword](true, vsid, user, param, (data or nil))
			end
			fill = fill and tostring(fill) or false

			if fill and params and params == "single" then
				retmsg = retmsg:gsub("%%"..word,tFuncs:EscapePercent(fill), 1)
			elseif fill then
				retmsg = retmsg:gsub("%%"..word,tFuncs:EscapePercent(fill))
			else	-- this shouldnt happen now, changed it to return empty string instead of false to prevent errors
				--tFuncs:Debug("Fillvar sets retmsg to false", "X")
				retmsg = ""
			end
		elseif param and word:sub(3,6) == "set:" then
		-- retreive settings values
			local value, test, txt, data = SetManager:Get(param)
			if test then
				--tFuncs:Debug("value: "..tostring(value), "X")
				retmsg = retmsg:gsub("%%"..word, tostring(value))
			else
				retmsg = retmsg:gsub("%%"..word, "")
				--tFuncs:Debug("no value: ", "X")
			end
		elseif param and word:sub(3,6) == "cmd:" then
			--// For internal use only, will stay undocumented
			if tClient[vsid].iProfile == 10 then
				if param ~= "" then
					local res, fill = tFuncs:HubCommand(vsid, param)
					if not res then fill = "Command: "..param.." failed, "..(fill or "no reason given.") end
					fill = fill or ""
					retmsg = retmsg:gsub("%%"..word,tFuncs:EscapePercent(fill))
				end
			else
				retmsg = retmsg:gsub("%%"..word,tFuncs:EscapePercent("*** You're not allowed to use %[cmd:] ***"))
			end
		elseif param and word:sub(3,7) == "file:" then
		-- readfile
			local res, txt = tFuncs:GetTXT("text/"..param, vsid)
			if res then
				retmsg = retmsg:gsub("%%"..word, tostring(txt))
			else
				retmsg = retmsg:gsub("%%"..word, "")
			end
		--else
			--tFuncs:Debug("Fillvar has no pattern to match: "..word, "X")
		end
	end
	return retmsg
end,

["GetTXT"] = function(self, file, sid, sformat)
	file = _APPPATH..file
	local sid = sid or "AAAA"
	local txt, err = io.open(file, "r")
	if txt then
		local msg = txt:read("*a")
		txt:close()
--~ 		msg = tFuncs:ToLocale(msg)
		msg = tFuncs:FillVars(msg, sid)
		return true, msg
	else
--~ 		tFuncs:Debug("GetTXT: File: "..file.." wasn't found.")
		return false, " Error: File: "..file.." wasn't found."
	end
end,

["GetTotalShare"] = function(self, protocol)
	local sharesize = 0
	for vsid,data in pairs(tClient) do
		if not protocol or data.sProtocol == protocol then
			sharesize = sharesize + (tonumber(data.iShareSize or 0) or 0)
		end
	end
	return sharesize
end,

["Duration"] = function(self, stime)
	local stime = tonumber(stime) or 0
	local s = os.difftime(stime, 0)
	local t = os.date("!*t", s)
	local retmsg = string.format( "%i day(s) %02i:%02i:%02i", s / 86400, t.hour, t.min, t.sec )
	if stime < 0 then retmsg = "forever" end
	return retmsg
end,

["ConvertToBytes"] = function(self, value)
	if value and value ~= "" then
		value = string.lower(value)
		local t = { ["b"] = 1, ["kb"] = 1024, ["mb"] = 1024^2, ["gb"] = 1024^3, ["tb"] =1024^4, ["pb"] =1024^5}
		local _,_,num = string.find(value, "^(%d+%.*%d*)")
		local _,_,tail = string.find(value, "^%d+%.*%d*%s*(%a+)$")

		if not num then return false end
		num = tonumber(num)
		if not tail then return num end

		local multiplier = 1
		if num and tail and t[tail] then
			multiplier = t[tail]
		elseif not num or (tail and not t[tail]) then
			return false
		end
		return (tonumber(num) * multiplier)
	else
		return false
	end
end,

["ConvertToHours"] = function(self, stime)
	if stime and stime ~= "" then
		local t = { ["H"] = 1, ["h"] = 1, ["M"] = 720, ["m"] = (1/60), ["D"] = 24, ["d"] = 24, ["Y"] = 8760, ["y"] = 8760, ["w"] = 168, ["W"] = 168, ["s"] = (1/3600), ["S"] = (1/3600)}
		local _,_,num, tail = string.find(stime, "^(%d+)(%a-)$")
		if num and tail and t[tail] then
			return (tonumber(num) * t[tail])
		elseif num and tail == "" then
			return tonumber(num)
		else
			return false
		end
	else
		return false
	end
end,

["ConvertToSeconds"] = function(self, stime, base)		--// Assuming seconds as input, unless tail specifies otherwise
	if stime and stime ~= "" then
		local t = { ["H"] = 3600, ["h"] = 3600, ["M"] = 3600*720, ["m"] = 60, ["D"] = 24*3600, ["d"] = 24*3600, ["Y"] = 365*24*3600, ["y"] = 365*24*3600, ["w"] = 7*24*3600, ["W"] = 7*24*3600, ["s"] = 1, ["S"] = 1, ["ms"]=1/1000, ["MS"] = 1/1000,}
		local _,_,num, tail = string.find(stime, "^(%-?%d+)(%a-)$")
		if num and tail and t[tail] then
			return (tonumber(num) * t[tail])
		elseif num and tail == "" then
			local multiplier = (base and t[base]) or 1
			return tonumber(num*multiplier)
		else
			return false
		end
	else
		return false
	end
end,

["ConvertOsTime"] = function(self,stime)
	local s = os.difftime( os.time(), tonumber(stime) )
	local t = os.date( "!*t", s )
	return string.format( "%i day(s) %02i:%02i:%02i", s / 86400, t.hour, t.min, t.sec )
end,

["HubUpTime"] = function(self, raw)
	if raw then
		return os.time()-tStats.Hub.Started
	else
		return tFuncs:ConvertOsTime(tStats.Hub.Started)
	end
	--local old = tStats.Hub.Started
	--return tFuncs:ConvertOsTime(old)
end,

["GetUserByIP"] = function(self, ip)
	local ipsid = {}
	for tsid, tuser in pairs(tClient) do
		if tClient[tsid].sIP == ip then
			ipsid[#ipsid+1] = tsid
		end
	end
	if ipsid ~= {} then
		return ipsid
	else
		return false
	end
end,

["Escape_XML"] = function(self, s)
	s = s:gsub("&amp;", "&")	-- this one goes first if the rest is double escaped
	s = s:gsub("&apos;", "'")
	s = s:gsub("&gt;", ">")
	s = s:gsub("&lt;", "<")
	s = s:gsub("&#8482;", "")
	s = s:gsub("&#8364;", "")
	return s
end,

["Save_Serialize"] = function(self, tTable, sTableName, hFile, sTab)
	sTab = sTab or "";
	hFile:write(sTab..sTableName.." = {\n" )
	for key, value in pairs(tTable) do
		local sKey = (type(key) == "string") and string.format("[%q]",key) or string.format("[%d]",key)
		if(type(value) == "table") then
			tFuncs:Save_Serialize(value, sKey, hFile, sTab.."\t")
		else
			local sValue = (type(value) == "string") and string.format("%q",value) or tostring(value)
			hFile:write( sTab.."\t"..sKey.." = "..sValue)
		end
		hFile:write( ",\n")
	end
	hFile:write( sTab.."}")
end,

["Save_File"] = function(self, filename, tTable, tablename, comment)	--// Needs an absolute path
	local comment = (comment and "--[[ "..comment.." ]]--\n") or ""
--~ 	filename = _APPPATH..filename
	local hFile, err = io.open (filename , "wb")
	if hFile then
		hFile:write(comment)
		tFuncs:Save_Serialize( tTable, tablename, hFile)
		hFile:flush()
		hFile:close()
	else
		local err = "Error while saving file, can't open "..filename.." for writing. Error: "..err
		tFuncs:Debug(err)
	end
--~ 	collectgarbage("collect")
end,

["Build_SerialTable"] = function(self, tTable, sTableName, sTab, tSerial, uCount, fCount)
	uCount = uCount or 0
	fCount = fCount or 0
	sTab = sTab or "";
	local tSerialTotal = (tSerial or "")..sTab..sTableName.." = {\n"
	for key, value in pairs(tTable) do
		local sKey = (type(key) == "string") and string.format("[%q]",key) or string.format("[%d]",key)
		if(type(value) == "table") then
			tFuncs:Build_SerialTable(value, sKey, sTab.."\t", tSerialTotal, uCount, fCount)
		elseif(type(value) == "userdata") then
			uCount = uCount + 1
		else
			if (type(value) == "function") then fCount = fCount + 1 end
			local sValue = (type(value) == "string") and string.format("%q",value) or tostring(value)
			tSerialTotal = tSerialTotal..sTab.."\t"..sKey.." = "..sValue
		end
		tSerialTotal = tSerialTotal..",\n"
	end
	tSerialTotal = tSerialTotal..sTab.."}"
	return tSerialTotal, uCount, fCount
end,

--~ ["Build_SerialTable"] = function(self,tTable, sTableName)
--~ 	tSerial = tSerial..sTableName.."={"
--~ 	for key, value in pairs(tTable) do
--~ 		local sKey = (type(key) == "string") and string.format("[%q]",key) or string.format("[%d]",key)
--~ 		if(type(value) == "table") then
--~ 			tFuncs:Build_SerialTable(value, sKey)
--~ 		else
--~ 			local sValue = (type(value) == "string") and string.format("%q",value) or tostring(value)
--~ 			tSerial = tSerial..sKey.."="..sValue
--~ 		end
--~ 		tSerial = tSerial..","
--~ 	end
--~ 	tSerial = tSerial.."}"
--~ 	return tSerial
--~ end,

["Send_Table"] = function(self, tTable, tablename )
--~ 	tSerial = ""
--~ 	tSerial = tFuncs:Build_SerialTable( tTable, tablename)
--~ 	tSendToGui()
--~ 	collectgarbage("collect")
end,

["BackUp"] = function(self,source,target)	-- not being used for anything? /fredo
	local text
	local f,e = io.open(source)
	if f then
		text = f:read("*a")
		f:close()
	else
		return e:sub(1,-2)
	end
	if text then
		local f,e = io.open(target,"w")
		if f then
			f:write(text)
			f:flush()
			f:close()
		else
			return e:sub(1,-2)
		end
	end
end,

["RunCmd"] = function(self, cmd)
	cmd = string.gsub(cmd, "\n$", "")
	local f = loadstring(cmd)
	local val, var, valo, res, reso, ret, reto, answer
	if string.find(cmd, "^%w.*=.*") and not string.find(cmd, "\n") then        --// If it looks like setting a value, first show the current value before changing
		var = string.gsub(cmd, "=.*", "")
		var = string.gsub(var, "%s", "")
		val = string.gsub(cmd, ".-=", "")
		val = string.gsub(val, "%s", "")
		if var and val then
			local g = loadstring("return "..var)
			reso,reto = pcall(g)
			res,ret = pcall(f)
		end
	else
		res,ret = pcall(f)
	end

	if res and reso and reto then
		answer = "Variable '"..var.."' changed from: '"..tostring(reto).."' to: '"..tostring(val).."'"
	elseif res and reso then
		answer = "New variable '"..var.."' set to: '"..tostring(val).."'"
	elseif res then
		if ret then
			answer = cmd.."\nCommand executed, returned: '"..tostring(ret).."'"
		else
			answer = cmd.."\nCommand executed: OK"
		end
	else
		answer = cmd.."\nCommand error: "..tostring(ret)..""
	end
	if res then
		return true, answer
	else
		return false, answer
	end
end,

["RunScript"] = function(self, path,filename)	-- // not currently used
--     tFuncs:SendToGui("ID_tc_scripts_debug", "> Program "..filename.." is running\n")
	lfs.chdir (path)
	if _OS == "Windows" and not _NOGUI then
		local proc = wx.wxProcess()
		proc:Redirect()
		scriptPID = wx.wxExecute(programName.."/lua "..path..filename, wx.wxEXEC_NOHIDE,proc)
		local outstd = proc:GetInputStream()
		while (outstd:Eof() == false) do
			local data = outstd:Read(1024)
--            tFuncs:SendToGui("ID_tc_scripts_debug", data)
		end
		wx.wxKill(scriptPID)
		proc:CloseOutput()
		outstd = nil
		proc:Detach()
		proc = nil
	else
	if programName ~= "/usr/share/luaaio" then
--        tFuncs:SendToGui("ID_tc_scripts_debug", "To use this editor, lua executable shall be installed in /usr/share/luaaio and not in "..programName..".\nMoreover, the executable shall be accessible in your path eg. ln -s /usr/share/luaaio/lua /usr/bin/lua.\n")
	else
		local handle,err = io.popen("lua "..path..filename)
		local data = handle:read("*l")
		_,_,scriptPID = data:find("pid (%d+)")
		if scriptPID then scriptPID = tonumber(scriptPID) end
		while ( data ~= nil ) do
			if not data:find("INFO:") then
--                tFuncs:SendToGui("ID_tc_scripts_debug", data.."\n")
			end
			if data:find("INFO: Stopping") then
				handle:close()
				data = nil
			else
				data = handle:read("*l")
			end
		end
		wx.wxKill(scriptPID)
	end
	end
--    tFuncs:SendToGui("ID_tc_scripts_debug", "> Program "..filename.." is stopped\n")
--    collectgarbage()
	scriptPID = nil
	scriptID = nil
end,

["JulianDate"] = function(self)
	local tTime = os.date("*t")
	return os.time({year = tTime.year, month = tTime.month, day = tTime.day,
			hour = tTime.hour, min = tTime.min, sec = tTime.sec})
	--// ex tFuncs:JulianDate()
end,

["CheckChat"] = function(self,sid,msg)	--// Deprecated
	-- check mainchat for several things like advertising, kennylizing, gagging etc
	local nick = tUser:GetNick(sid)
	local ip = tUser:GetIp(sid)
	--// check the nick for what they done to him^^
	--// maybe in case of forgetting..add a timer to it 8-|
	--// lets start with gagging
	--if tGag[ip] then gmsg = "" return gmsg end
	local kenny = {"*umfl* *uuffum*", "*lluu* *mlmlff* *umfl* *lfumfl* *umfl*",
					"*lmmf* *uullu* *mmmm*", "*ommlu* *uullu* *lmmf* *ommlu* *mflf*",
					"*olomum* *lmmf* *mhhhmmlm*", "*Mhhl* *mujm* *umfl*", }
	--/ and offcourse kennylizing
	--if tKen[ip] then kmsg = kenny[math.random(1, #kenny)] return kmsg end
	--if not tGag[ip] or tKen[ip] then return msg end
	--// check the msg for advertising and stuff
	--if msg:match("") then
	return msg
	--end
end,

["DLPage"] = function(self, url)
	require "modules.http"
	if not url then
		return "Error: no url given in function"
	else
		local msg, err = http.request(url)
		if msg then
			return msg
		else
			return err
		end
	end
end,

["DLfile"] = function(self,url,loc)
	local file = url:gsub(".*%/","")
	local t = socket.gettime()
	local b,c,h = http.request(url)
	if b and b:len() > 0 then
		local f,e = io.open(loc..file,"w")
		if f then
			f:write(b) f:flush() f:close()
		else
			return e:sub(1,-2)
		end
	end
	s = socket.gettime() - t
	tFuncs:Debug("File: "..file.." downloaded and saved in "..tostring(s).." seconds")
end,

["Lock2Key"] = function(self,lock)
	local aKey = {}
	table.insert(aKey,bit.bxor(bit.bxor(bit.bxor(lock:byte(1),lock:byte(-1),lock:byte(-2),5))))
	for i=2,#lock do
		table.insert(aKey,bit.bxor(lock:byte(i),lock:byte(i-1)))
	end
	local t = {["0"]=1,["5"]=1,["36"]=1,["96"]=1,["124"]=1,["126"]=1}
	for i=1,#aKey do
		local b = bit.bor(bit.band(aKey[i]*(2^4),240),bit.band(math.floor(aKey[i]/(2^4)),15))
		aKey[i] = (t[tostring(b)] and string.format("/%%DCN%03d%%/",b) or string.char(b))
	end
	return table.concat(aKey)
end,

["AddHubFiles"] = function(self, check)
	if check then
		-- compare previously loaded TTH with current, if they dont match, return a file name
		for name, data in pairs(tHubFiles) do
			if data.bRestart then
				if not tFuncs:CheckFileExists(data.sPath) or (data.sTTH and data.sTTH ~= tth.file(data.sPath)) then
					--tFuncs:Debug("[HubFiles] testing "..data.sPath..", it is NOT a match", "X")	-- fredo
				--	return name..".lua"
					return name
				--else
					--tFuncs:Debug("[HubFiles] testing "..data.sPath..", its a match", "X")	-- fredo
				end
			end
		end
	else
		for name, data in pairs(tHubFiles) do
			if tFuncs:CheckFileExists(data.sPath) then
				data.sTTH = tth.file(data.sPath)
				--tFuncs:Debug("[HubFiles] loaded "..data.sPath.."\t"..tostring(data.bRestart).."\t"..data.sRev.." :\t"..(data.sTTH or "oops"), "X")
			end
		end
	end
end,

["LoadScripts"] = function(self, hubcmd)
-- // this function is used first time before hub is started, tFuncs:Debug wont work here until after SetManager:Initialize()
-- // hubcmd is true if initiated from a hubcmd, else its being called from Hub-Start.lua

	if hubcmd and hubcmd ~= "-force" then
		-- check if any of the bRestart files have changed, if so do not allow rs
		local ret = tFuncs:AddHubFiles(true)
		if ret then

		local a = ""
		if not tGlobals.bNoRestart then
			a = " To restart the hub use !restart "..tHubSettings.Main.bConfirmCode
		end

		return false, Lang(0, 0, "The file %s has been changed. !rs is not allowed until the hub is restarted.%s", ret, a)
		end
		hubcmd = nil
	end

	local result, output, err = true, "Reloading scripts: \n\n", false

	--// Initialize - Hub-SetVars.lua needs to be loaded first, tSettings.Scripts.InternalFiles in it is used to determine which scripts to reload
	local filename = "core/Hub-SetVars.lua"
	local res, err = tFuncs:TryFile(filename)
	if res then
		output = output..Lang(0, 0, "\t*** Script: ./%s loaded.", filename).."\n"
	elseif not res and err then
		return false, "", output..Lang(0, 0, "\t*** Script: %s failed to load: %s", filename, (err or "unknown error"))
	else
		return false, "", output..Lang(0, 0, "\t*** Script: %s failed to load: file can't be found.", filename)
	end

	--// Base scripts
	local err
	for filename in pairs(tSettings.Scripts.InternalFiles) do
		local res, err = tFuncs:TryFile(filename)
		if res then
			output = output..Lang(0, 0, "\t*** Script: %s loaded.", filename).."\n"
		else
			if err and err ~= "" then
				err =  "\t*** "..Lang(0, 0, "Script: %s failed to load: %s", filename, (err or "unknown error")).."\n"
			else
				err = "\t*** "..Lang(0, 0, "Script: %s failed to load: file can't be found.", filename).."\n"
			end
			output = output..err
			return false, output, err
		end
	end

	output = output.."\n"

	--// Optional scripts
	for filename in pairs(tSettings.Scripts.OptionalFiles) do
		local res, err = tFuncs:TryFile(filename)
		if res then
			output = output..Lang(0, 0, "\t*** Optional script: %s loaded.", filename).."\n"
		elseif err and err ~= "" then
			err = "\t*** "..Lang(0, 0, "Optional script: %s failed to load: %s", filename, (err or "unknown error")).."\n"
			output = output..err
		else
			--// If file wasn't found it will fail silently
		end
	end

	--// Load all settingsfiles or default values if they don't exist
	local res, err = SetManager:Initialize()
	if not res then return false, err end

	--// Set charset to initialize conversions
	tFuncs:SetCharset()

	--	--// Check accounts
	local res, err = AccountManager:Check()
	if not res then tFuncs:SafeDebug("[LoadScripts:AccountCheck] "..err , "FI", 0) end

	--// Update windowtitle if svn changed
	local oldrev = _REV
	tFuncs:GetSvn()

	if not _NOGUI and oldrev ~=_REV and tStats.Console.Status == "Running" then
		tCtrls["app"]:SetTitle(tHubSettings.Main.sName.." on ".._APPNAME.." ".._APPVER.." svn:".._REV.." - by FlipFlop, en_dator and Daywalker")
	end

	--// Mark settings as loaded successfully, this is needed to be able to save settings on close
	tStats.Hub.SettingsLoaded = true

	--// after script reloads we now also needs to update user command menus
	tUserCmds:Build()

	--// Create the internal list of redirects
	tRedirect:MakeList()

	--// Priority-sort all patterns
	for stype in pairs(SubSet.Patterntypes) do
		tCheck:PatternPriority(stype)
	end

	--// Store TTH values for all files loaded
	if not hubcmd then
		tFuncs:AddHubFiles()
	end

	return true, output, err
end,

["TryFile"] = function(self, name)
	local name = _APPPATH..name
	file = io.open(name, "r")
	if file then
		file:close()
		file = nil
		local result, err = pcall(dofile, name)
		if not result then
			return false, err
		else
			return true
		end
	else
		file = nil
		return false, ""
	end
end,

["CheckFileExists"] = function(self, fullpath)
	local a = io.open(fullpath, "rb")
	if a then
		a:close()
		return true
	else
		return false
	end
end,

["ConnectHublist"] = function(self, address, protocol, info)	-- runs in HubListReg thread
	local ctime = os.time()
	local port = string.gsub(address, "%S+:", "")
	port = tonumber(port) or 2501
	local addy = string.gsub(address, ":%d+", "")
	local online, ok, output = false, false, ""
	tHubSettings.tHubLists[protocol][address].LastRegTime = ctime

	local hubip = socket.dns.toip(addy)
	if hubip and tLogs.Hublists.Registration.LastConnects[protocol][hubip] then
		return false, "DOUBLE (cancelled): IP is the same as for "..tLogs.Hublists.Registration.LastConnects[protocol][hubip]
	elseif hubip then
		tLogs.Hublists.Registration.LastConnects[protocol][hubip] = address
	else
		return false, "DNS didn't resolve to an IP"
	end

	local err
	local sock, err = socket.tcp()

	if not sock then
		return false, "Error: "..err
	else
		sock:settimeout(5)
		online, err = sock:connect(addy, port)
		if not online then
			output = "Error: Can't connect: "..err
		else
			sock:settimeout(2)
			local hubip, hubport = sock:getpeername()

			if protocol == "NMDC" then
				local locky, err, data = sock:receive()
				if not locky then locky = data end
				if locky then
					local _,_,lock = locky:find("^%$Lock (.*)$")
					if lock then
						tFuncs:SafeDebug("Hublist Regserver sends: "..locky, "X")
						local key = tFuncs:Lock2Key(lock)
						local ok, err = sock:send("$Key "..key.."|"..info.."|")
						if not ok then
							output = "Error: "..err
						else
							tHubSettings.tHubLists[protocol][address].LastReg = true
							tHubSettings.tHubLists[protocol][address].LastRegOk = ctime
							output = "Hub registered successfully."
						end
					else
						output = "Error: Didn't receive $Lock: "..err.." "..tostring(locky)
					end
				else
					output = "Error: No data received: "..err
				end
			elseif protocol == "ADC" then
				local ok, err = sock:send("HSUP ADBASE ADPING\n")
				if not ok then
					output = "Error: Couldn't send HSUP: "..err
				else
					--// Receive ISUP and IINF
					local hublistinfo, err, data = sock:receive()
					local hublistinfo2, err2, data2 = sock:receive()
					if not hublistinfo then hublistinfo = data end
					if not hublistinfo2 then hublistinfo2 = data2 end
					if hublistinfo and hublistinfo2 then
						local ok, err = sock:send("HINF"..info.."\n")
						if not ok then
							output = "Error: Couldn't send hubinfo (HINF): "..err
						else
							tHubSettings.tHubLists[protocol][address].LastReg = true
							tHubSettings.tHubLists[protocol][address].LastRegOk = ctime
							local _,_,name = string.find(hublistinfo2, " NI(%S+)")
							local _,_,addy = string.find(hublistinfo2, " WS(%S+)")
							if name then
								name = ADC_E(name, 1)
								tHubSettings.tHubLists[protocol][address].sName = name
							end
							if addy then
								addy = ADC_E(addy, 1)
								tHubSettings.tHubLists[protocol][address].sWebsite = addy
							end
							if name and addy then
								output = "Hub registered successfully at: "..name.." / website: "..addy
								ok = true
							else
								output = "Hub registration failed, didn't receive IINF from hublist registration server."
							end
						end
					else
						output = "Error: No hublistinfo (ISUP/IINF) received: "..err.." "..err2
					end
				end
			end
		end
		sock:close()
		sock = nil
	end
--~ 	collectgarbage("collect")
	return online, output
end,

["SafeMessage"] = function(self, target, tosid, fromsid, msg)
	--// tFuncs:SafeMessage(target, tosid, fromsid, msg)
	--// target:  "M" = Mainchat, "R" = Room, "P" = PM
	ThreadLockMutex(mutexlistenbuffer)
	table.insert(tListenBuffer.Messages, {["Target"] = target, ["To"] = tosid, ["From"] = (fromsid or "AAAA"), ["Message"] = msg})
	ThreadUnlockMutex(mutexlistenbuffer)
end,

["SafeExecute"] = function(self, line)
	--// tFuncs:SafeMessage(target, tosid, fromsid, msg)
	--// target:  "M" = Mainchat, "R" = Room, "P" = PM
	ThreadLockMutex(mutexlistenbuffer)
	table.insert(tListenBuffer.Commands, line)
	ThreadUnlockMutex(mutexlistenbuffer)
end,

["RegHub"] = function(self, ssid)	-- thread start, runs from timer or from setcmd
	local result, err = pcall(tFuncs.RegHubPrepare, nil, ssid)
	if not result then
		tFuncs:SafeDebug("[THREAD] [RegHub] Error: "..(err or "unknown error"))
		tFuncs:UnlockMutexes()
	end
	tThreads.HubListReg = nil
	ThreadDestroy()
end,

["RegHubPrepare"] = function(self, ssid)
	local toutput = {}

	--// Count total hublists to register the hub with and sort the lists
	local count = 0
	local checklist = { ["NMDC"] = {}, ["ADC"] = {}}
	for _,protocol in pairs {"NMDC", "ADC"} do
		for address, data in pairs(tHubSettings.tHubLists[protocol]) do
			if data.Enabled then
				count = count+1
				table.insert(checklist[protocol], address)
			end
		end
		table.sort(checklist[protocol])
	end

	--// Start registration
	local counter = 0
	tLogs.Hublists.Registration.LastRun = os.time()
	tLogs.Hublists.Registration.LastConnects = { ["NMDC"] = {}, ["ADC"] = {},}

	local info = {}
	info.NMDC = tHubSettings.Main.sName.."|"..tHubSettings.Main.sAddress.."|"..tHubSettings.Main.sDescription.."|"..tStats.Hub.Users.Total.."|"..tFuncs:GetTotalShare("NMDC")
	info.ADC = " NI"..ADC_E(tHubSettings.Main.sName)..
		" DE"..ADC_E(tHubSettings.Main.sDescription)..
		" HH"..ADC_E(tHubSettings.Main.sAddress)..
		" OW"..ADC_E(tHubSettings.Main.sOwner)..
		" WS"..ADC_E(tHubSettings.Main.sWebsite)..
		" NE"..ADC_E(tHubSettings.Main.sNetwork)..

		" VE"..ADC_E(_APPNAME.._APPVER)..
		" UP"..os.time()-tStats.Hub.Started..
		" UC"..ADC_E(tStats.Hub.Users.Total)..
		" SS"..ADC_E(tFuncs:GetTotalShare("ADC"))..

		" ML"..ADC_E(tHubSettings.tRules.iMinSlots)..
		" XL"..ADC_E(tHubSettings.tRules.iMaxSlots)..
		" MS"..ADC_E(tHubSettings.tRules.iMinShare)..
		" XS"..ADC_E(tHubSettings.tRules.iMaxShare)..
		" XU"..ADC_E(tHubSettings.tRules.iMaxHubs)..
		" MC"..ADC_E(tHubSettings.tRules.iMaxUsers)

	for _,protocol in pairs {"NMDC", "ADC"} do
		table.insert(toutput, "\n\n"..protocol.." hublists:")
		for num, address in ipairs(checklist[protocol]) do
			counter = counter + 1
			tLogs.Hublists.Registration.Status = counter.."/"..count
			tFuncs:SafeDebug("[THREAD][RegHub] Hublists registration status: "..tLogs.Hublists.Registration.Status.. ",adress: "..address, "X")

			local online, retmsg = tFuncs:ConnectHublist(address, protocol, info[protocol])
			if online then online = "ONLINE" else online = "offline" end
			table.insert(toutput, "\t"..online.."\t\t"..Padding(address, 40).."\t"..retmsg)
		end
	end

	local output = "\n\n\tHublist registration results:\n\t"..string.rep("=",80)
	for num,line in ipairs(toutput) do
		output = output..line.."\n"
	end

	tLogs.Hublists.Registration.LastResult = output

	--// Output to mainchat of the one that started the command
	if ssid then
		tFuncs:SafeMessage("M", ssid, "AAAA", output)
	end

end,

["ValidateAddress"] = function(self, str) --a string as input and returns true/false, addy:port, addy, port
	str = str:match("(%S+)")

	local port = str:match(":(%d+)%s-$")
	if port then
		port = tonumber(port)
		str = str:gsub(":%d+%s-$", "", 1)
	end

	local protocol = str:match("^%s-(%a%a-://)")
	if protocol then
		str = str:gsub("^%s-%a%a-://", "", 1)
	end

	local addy = str:match("(%S+)")
	if addy then
		str = str:gsub("%S+", "", 1)
	end

	if #str > 0 then
		tFuncs:SafeDebug("[ValidateAddress] characters left in string ("..#str.."): "..str, "X")
	end

	tFuncs:SafeDebug("[ValidateAddress] Protocol : "..(protocol or "no protocol part found").."\tAddress : "..(addy or "no address found").."\tPort : "..(port or "no port found"), "X")

	if not addy then
		return false, "No address part found in string"
	else
		return true, (protocol or "")..addy..((port and ":"..port) or ""), addy, (port or nil)
	end
end,

["AddBlocks"] = function(self, ssid, vnick, silent, blocktype)
-- based on the muppetize function

	local check, res = tUser:IsProtected(nil, vnick, vnick)
	if check then return true, res end

	local vsid, realnick = tNick:GetSid(vnick)
	local vip = tUser:GetIp(vsid)

	local btypes = {
		["mute"]	= {"muted", },
		["kenny"]	= {"kennylized", },
		["muppet"]	= {"muppetized", },
		["connects"]= {"blocked from downloading", },
		["pm"]		= {"blocked from making PM's", },
	--	["mc"]		= {"blocked", },
	}

	if not vsid then
		local lNick = vnick:lower()

		-- offline regged user
		if tAccounts[lNick] then
			if (tClient[ssid].iProfile or -2) > (tAccounts[lNick].iLevel or -2) or tClient[ssid].iProfile == 10 then

				local check, res = tUser:IsProtected(nil, lNick, tAccounts[lNick].sLastIP)
				if check then return false, res end

				local res, retmsg = tBlock:AddBlock(ssid, blocktype, lNick, tAccounts[lNick].sLastIP, silent)
				tFuncs:Debug(retmsg)
				if res and blocktype == "connects" then tSend:ADCUserIps(vsid, 1) end

				if res and not silent then
					tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now ", realnick) ..btypes[blocktype][1], "AAAA")
					return res
				end
				return res, retmsg
			end

		-- offline unregged user thats been in the hub before
		elseif tLogs.Nick[lNick] and tLogs.Nick[lNick].sIP and tLogs.Nick[lNick].sNick then
			if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then

				local check, res = tUser:IsProtected(nil, tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP)
				if check then return false, res end

				local res, retmsg = tBlock:AddBlock(ssid, blocktype, tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP, silent)
				tFuncs:Debug(retmsg)
				if res and not silent then
					tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now ", realnick) ..btypes[blocktype][1], "AAAA")
					return res
				end
				return res, retmsg
			end

		-- ip provided instead of nick
		elseif tFuncs:IsIP(vnick) then
			local check, res = tUser:IsProtected(nil, nil, vnick)
			if check then return false, res end

			local res, retmsg = tBlock:AddBlock(ssid, blocktype, "<unknown>", vnick, silent)
			tFuncs:Debug(retmsg)
			return res, retmsg
		end

		-- if user not been in the hub and no ip we end up here
		return false, Lang(ssid, 0, "User can't be found.")
	elseif not tUser:IsHigher(ssid, vsid) then
		return false, Lang(ssid, 0, "This user is too powerful for you.")
	elseif tNick:GetSid(vnick) == ssid then
		return false, Lang(ssid, 0, "You can't do that to yourself.")

	-- online users
	else
		local res, retmsg = tBlock:AddBlock(ssid, blocktype, vnick, vip, silent)
		tFuncs:Debug(retmsg)

		if res then
			if blocktype == "muppet"  then
				if silent then
					--return res, retmsg
				else
					tSend:ChatToAll( (tFilters.Muppets.BroadcastOn or Lang(ssid, 0, "User <%s> is now ", realnick) ..btypes[blocktype][1]), "AAAA")
					--return ret
				end
			elseif blocktype == "connects" then
				tSend:ADCUserIps(vsid, 1)
				return res, retmsg
			else
				if silent then
					--return res, retmsg
				else
					tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now ", realnick) ..btypes[blocktype][1], "AAAA")
					--return ret
				end
			end
		end

		return res, (silent and retmsg or nil)
	end
end,

--// Stuff that is used for running stuff in own threads

["RevDNSStart"] = function(self, ssid, ip)
	local result, err = pcall(tFuncs.RevDNSRun, nil, ssid, ip)
	if not result then
		tFuncs:SafeDebug("[THREAD] [REVDNS] Error: "..err)
		tFuncs:UnlockMutexes()
	end
	tThreads.RevDNS = nil
	ThreadDestroy()
end,

["RevDNSRun"] = function(self, ssid, ip)
	local msg, xinfo = "", ""
	local rev, err = socket.dns.tohostname(ip)
	if rev then
		for name, value in pairs(err) do
			if type(value) == "table" then
				value = tFuncs:GetSerialized(name, value)
			end
			xinfo = xinfo.."\t"..name.." \t"..tostring(value).."\n"
		end
	end

	if not rev then rev = "Error: "..err end

	msg = msg.."\n\n\t\t"..Lang(ssid, 0, "Dns information:").."\n"
	msg = msg.."\n\t\t"..string.rep("_",25)
	msg = msg.."\n\t\t"..Lang(ssid, 0, "Requested").."    \t: "..ip
	msg = msg.."\n\t\t"..Lang(ssid, 0, "Reversed to").."  \t: "..rev

	--// Get countrycode
	if tHubSettings.Main.bUseGeoIP and tGeoIP.loaded then
		local cc, cn = tGeoIP.loaded:lookup(ip)
		msg = msg.."\n\t\t"..Lang(ssid, 0, "Countrycode").."  \t: "..(cc or "Not found").."\n\t\t"..Lang(ssid, 0, "Countryname").." \t: "..(cn or "Not found")
	end

	tFuncs:SafeDebug("Extra info on !revdns: \n"..xinfo, "X", 0)

	msg = msg.."\n\t\t"..string.rep("_",25).."\n\n"
	tFuncs:SafeMessage("M", ssid, nil, msg)
	return true, Lang(ssid, 0, msg)
end,


}
--// FlipFlop --//-- en_dator --//-- Daywalker
