--// $Id: Hub-SetCmds.lua 1467 2013-10-19 17:54:25Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tHubFiles["Hub-SetCmds.lua"] = {
	["sRev"] = "$Rev: 1467 $",
	["bRestart"] = false,
	["sPath"] = "./core/Hub-SetCmds.lua",
	}

SetCmd = {

--// commands all in lowercase for matching
--// don't use endpipes or newlines at the end of commands, tFuncs:Send will use an escapefunction to match the line to the receiving protocol.

--// warn drop kick ban kill banip nickban tempban tempbanip tempnickban rangeban temprangeban

--// text = tLang(0, "Voorbeeldtekst met nick %1 en profiel %2", { tClient[sid].sNick, tClient[sid].sProfile })

	["examplecommand"] = {
		["allowed"] = function(ssid, line, cmd)					--// Determines if the user is allowed to use this command
--~ 			return tUser:IsAllowed(ssid, "setcmd")
			return false
		end,
		["cmd"] = function(ssid, line, cmd)
			return true, "Command is executed here"				--// Actual command
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tThis is what shows in !help", cmd)		--// Text to show in !help or !sethelp
		end,
		["xhelp"] = function(ssid, line, cmd)
			return true, "Detailed help, for !<cmd> -h If there's no [''xhelp''] then, it will use the content from [''help'']"			--// Text to show on !dummy -h
		end,
		["ucmd"] = {
			{1,3,"This is the rightclick command","!%s"},		--// Rightclicks here
		},
	},

	["webstats_reg"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local nick, pass = string.match(line, "^%S+%s+(%S+)%s+(%S+)")
			if not nick or not pass then return false, Lang(ssid, 0, "Error, use: %s <forumnick> <forumpass>", cmd) end
			if nick == "<forumnick>" or pass == "<forumpass>" then return false, "You have to use your own forum loginname and password." end

			if not tThreads.WebStatsReg then
				local data = tWebStats:GetData()	--// Gets this data during the normal listenloop, not inside the thread to prevent crashes
				tThreads.WebStatsReg = ThreadCreate(tWebStats.Register, {1, "AAAA", nick, pass, "reg", false, false, data})
				return true, "Registration started, results will be shown in feedroom."				--// Actual command
			else
				return false, "Registration thread is still running, try again later."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <forumnick> <forumpass>\t\t\tRegister this hub to enable web statistics, use your http://flexhub.org forum account details.", cmd)
		end,
		["ucmd"] = {
			{1,3,"Register this hub to enable web statistics","!%s %[line:Enter your http://flexhub.org forum nickname] %[line:Forum password]"},
		},
	},

	["createcert"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			--// Force removal of existing certificate?
			local force = false
			if line:find(" -force", 1, true) then
				force = true
				line = line:gsub(" %-force", "")
			end

			--// Get path to openssl.exe	(only used for Windows)
			local path = string.match(line, "^%S+%s+(.+)")

			return tFuncs:CreateCertificate(path, force)
		end,
		["help"] = function(ssid, line, cmd)
			if _OS == "Linux" then
				return Lang(ssid, 0, "%s [-force]\t\t\t\tCreate a selfsigned certificate for SSL, use -force to replace an existing certificate.", cmd, cmd)
			else
				return Lang(ssid, 0, "%s <path to openssl.exe> [-force]\t\tCreate a selfsigned certificate for SSL. Use %s -h for more information.", cmd, cmd)
			end
		end,
		["xhelp"] = function(ssid, line, cmd)
			if _OS == "Linux" then
				return true, Lang(ssid, 0, "%s [-force]\t\t\t\tCreate a selfsigned certificate for SSL, use -force to replace an existing certificate.", cmd)
			else
				return true, Lang(ssid, 0, "%s <path> [-force]\t\t\tCreate a selfsigned certificate for SSL.\n"..
				"To generate the certificate you need OpenSSL, can be downloaded here: http://www.slproweb.com/download/Win32OpenSSL_Light-1_0_0f.exe\n"..
				"After installation, you need to specify the path to openssl.exe. For example: !createcert C:\\Program Files\\OpenSSL\\bin\n"..
				"Use -force to replace an existing certificate.\n", cmd)
			end
		end,
		["ucmd"] = {
			{1,3,"Create a selfsigned certificate for encrypted connections","!%s"},
		},
	},

	["startprofiler"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "debug")
		end,
		["cmd"] = function(ssid, line, cmd)
			if tSettings.Debug.Profiler.Listen then
				return false, "Profiler already running"
			else
				tSettings.Debug.Profiler.Listen = true
				return true, "Profiler now running during "..tSettings.Debug.Profiler.Count.." iterations of the ListenThread.\nThis can cause some serious lag, be patient, once profiling is done the hub will respond properly again."
			end
		end,
--~ 		["help"] = function(ssid, line, cmd)
--~ 			return Lang(ssid, 0, "%s\t\t\t\t\t\tStart Debug Profiler", cmd)
--~ 		end,
--		["ucmd"] = {
--			{1,3,"Profiler Start","!%s"},
--		},
	},

	["stopprofiler"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "debug")
		end,
		["cmd"] = function(ssid, line, cmd)
			if not tSettings.Debug.Profiler.Listen then
				return false, "Profiler not running"
			else
				tSettings.Debug.Profiler.Listen = false
				return true, "Profiler stopped."
			end
		end,
--~ 		["help"] = function(ssid, line, cmd)
--~ 			return Lang(ssid, 0, "%s\t\t\t\t\t\tStop Debug Profiler", cmd)
--~ 		end,
--		["ucmd"] = {
--			{1,3,"Profiler Stop","!%s"},
--		},
	},

	["import"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local hubsoft, filename = string.match(line, "^%S+%s+(%S+)%s+(%S+)")
			if not hubsoft or not filename then return false, Lang(ssid, 0, "Error, use: %s <hubsoft> <filename>", cmd) end

			if not tImport.Hubs[hubsoft] then return false, Lang(ssid, 0, "Invalid hubsoft, choose: ADCH, Aquila, Luadch, Ptokax, Verlihub or Ynhub. (names are case sensitive)") end
			filename = filename:gsub("^.+/", "")
			filename = filename:gsub("^.+\\", "")
			filename = _IMPORTDIR..filename

			local f,err = io.open(filename, "r")
			if not f then return false, Lang(ssid, 0, "Couldn't open file '"..filename.."': "..(err or "unknown error")..".\nMake sure the file to import is in the folder: ".._IMPORTDIR) end

--			local ok, res, status, importlog = pcall(tImport.Hubs[hubsoft], true, filename)
			local ok, res, status = pcall(tImport.Hubs[hubsoft], true, filename)
			if ok and res then tFuncs:SaveSettings("accounts") end

			ImportProfiles = nil
			tFuncs:CloseFile("Import")

			local out = "\nCheck the file: ".._LOGDIR.."Import.log for details, or use !showlog import."
			if not ok then return false, Lang(ssid, 0, "Error during %s accounts import: %s", hubsoft, res) end

			tFuncs:Debug(status..out)
			return res, status..out
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <hubsoft> <filename>\t\t\t\tImport accounts from ADCH, Aquila, Luadch, Ptokax, Verlihub or Ynhub, the file must be placed in the import folder (names are case sensitive)", cmd)
		end,
		["ucmd"] = {
			{1,3,"Import accounts from ADCH++","!%s ADCH %[line:Filename (without path, uses import folder)]"},
			{1,3,"Import accounts from Aquila","!%s Aquila %[line:Filename (without path, uses import folder)]"},
			{1,3,"Import accounts from Luadch","!%s Luadch %[line:Filename (without path, uses import folder)]"},
			{1,3,"Import accounts from Ptokax","!%s Ptokax %[line:Filename (without path, uses import folder)]"},
			{1,3,"Import accounts from Verlihub","!%s Verlihub %[line:Filename (without path, uses import folder)]"},
			{1,3,"Import accounts from Ynhub","!%s Ynhub %[line:Filename (without path, uses import folder)]"},
		},
	},

	["sethelp"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local prof = tUser:GetProfile(ssid)
			local reply = "\r\n\r\n\t"..Lang(ssid, 0, "Hubsettings available for your profile [%s]:", prof).."\r\n\r\n\t"..Lang(ssid, 0, "Command").."\t\t\t\t\t\t"..Lang(ssid, 0, "Description").."\r\n\t"..string.rep("_",80).."\r\n"

			local totallist = {}	--// build total commandlist
			for name, func in pairs(SetCmd) do
				totallist[name] = true
			end

			local cmdlist = SubSet.SetCommands
			local cmdprefix = string.find(tHubSettings.Main.sCmdPrefixes, "!") and "!" or tHubSettings.Main.sCmdPrefixes:sub(1,1)

			--// Build reverse list of tCmdAlias
			local cmdalias = {}
			if tHubSettings.Main.bCmdAlias then
				for short,long in pairs(tCmdAlias) do
					if not cmdalias[long] then
						cmdalias[long] = cmdprefix..short
					else
						cmdalias[long] = cmdalias[long]..", "..cmdprefix..short
					end
				end
			end

			--// Display categorized commands
			for num, cat in ipairs(cmdlist) do
				local hits = ""
				for _, name in ipairs(cat.cmds) do
					totallist[name] = nil
					if SetCmd[name] and SetCmd[name]["allowed"](ssid, line) and SetCmd[name]["help"] then
						hits = hits.."\t"..SetCmd[name]["help"](ssid, line, cmdprefix..name)..(cmdalias[name] and " ("..cmdalias[name]..")" or "").."\n"
					end
				end
				if hits ~= "" then
					reply = reply.."\n\t"..cat.text.."\n\t"..string.rep("_",18).."\n"..hits
				end
			end

			--// Display miscellaneous commands
			local hits = ""
			local restlist = {}
			for name, res in pairs(totallist) do table.insert(restlist, name) end
			table.sort(restlist, SortLowercase)
			for _,name in ipairs(restlist) do
				if SetCmd[name]["allowed"](ssid, line) and SetCmd[name]["help"] then
					hits = hits.."\t"..SetCmd[name]["help"](ssid, line, cmdprefix..name)..(cmdalias[name] and " ("..cmdalias[name]..")" or "").."\n"
				end
			end
			if hits ~= "" then
				reply = reply.."\n\t"..Lang(ssid, 0, "Miscellaneous commands").."\n\t"..string.rep("_",18).."\n"..hits
			end

			return true, reply.."\n\t"..string.rep("_",80).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tShow help for hubsettings", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1,3,"Setup Help","!%s"},
		},
	},

	["set"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local params = tFuncs:Split(line, " ")
			local name = false
			local _,_,nn = string.find(line, "^%S+%s+(%S+)")

			if #params > 1 and nn then
				name = nn:lower()
			elseif #params > 1 then
				name = params[2]:lower()
			end

			local _,_,value = string.find(line, "^%S+%s+%S+%s+(.+)$")

			--// !set used without arguments, show main categories
			if not name then

				--// Build categories
				local total, help = {},{}

				for varname, data in pairs(Set.Settings) do
					if not data[2] then
						help[data[1]] = data[8]
					end
					local split = tFuncs:Split(data[1])
					if not data[2] then
						if split[1] and not total[split[1]] then total[split[1]] = {} end
						if split[2] and not total[split[1]][split[2]] then total[split[1]][split[2]] = {} end
						if split[3] and not total[split[1]][split[2]][split[3]] then total[split[1]][split[2]][split[3]] = {} end
					end
				end

				local reply = "\n\n\tAvailable categories:\n"..string.rep("_",60).."\n\n"
				for varname, data in pairs(total) do
					if varname ~= "" then
						reply = reply.."\t!set "..varname..string.rep(" ", (32-#varname)).."\t\t<category>\t\t"..(help[varname] or "").."\n"
					end
				end
				return true, reply

			--// !set * used, show all settings in all categories
			elseif name == "*" then
				local setlist = {"Main", "Rules", "Traffic", "Hublists", "Redirects", "Advanced", "Console", "Feed",  "Tcp", "Thread", "Webstats", "Security", "Security BW", "Security CT", "Security HC", "Security MA", "Security MC", "Security MN", "Security NI", "Security NL", "Security PM", "Security PR", "Security RC", "Security SA", "Security SP", "Security SR", "Security XX"}
				local _, sub
				local tbl = {}
				for _,sub in ipairs(setlist) do
					local ok, subreply = SetCmd["set"]["cmd"](ssid, "!set "..sub, "!set")
					if ok then table.insert(tbl, "\t!set "..sub..subreply.."\n****************************************************************************************************************\n") end
				end
				return true, "\n\nShowing all !set settings:\n\n"..table.concat(tbl)

			--// !set used with an existing name, but without value
			elseif Set.Settings[name] and Set.Settings[name][2] and (not value or value == "-h") then
				local getval, res, help, data = SetManager:Get(name)
				local readonly = ""
--				if data[3] == "t" or data[7] == "GUI" then readonly = " (read only / use GUI)" end
				if data[7] == "GUI" then readonly = " (read only / use GUI)" end
				if data[3] == "t" and data[7] and data[7] ~= "" then
					readonly = " (read only / use "..data[7].." to modify)"
					getval = "not available"
				elseif data[3] == "t" and (not data[7] or data[7] == "") then
					readonly = ""
					getval = "not available"
				end
				if data[3] == "f" then getval = tFuncs:ConvertShare(getval) end
				if data[3] == "d" then getval = getval.." ( "..tFuncs:Duration(getval).." )" end

				local allowed = ""
				if data[3] == "b" then allowed = allowed.."\nChoose: true or false" end
				if data[3] == "f" and data[4] then allowed = allowed.."\nMin: "..tFuncs:ConvertShare(tFuncs:ConvertToBytes(data[4])) end
				if data[3] == "f" and data[5] then allowed = allowed.."\nMax: "..tFuncs:ConvertShare(tFuncs:ConvertToBytes(data[5])) end
				if data[3] == "i" and data[4] then allowed = allowed.."\nMin: "..data[4] end
				if data[3] == "i" and data[5] then allowed = allowed.."\nMax: "..data[5] end
				if data[3] == "s" and data[4] then
					if type(data[4]) == "string" then allowed = allowed.."\nMust match pattern: "..data[4] end
					if type(data[4]) == "table" then allowd = "\nChoose: "..table.concat(data[4], ", ") end
					if type(data[5]) == "string" then allowed = allowed.."\nCan't match pattern: "..data[5] end
					if type(data[5]) == "table" then allowd = "\nNot allowed: "..table.concat(data[5], ", ") end
				end

				if res then
					return true, "\t!set "..name.."\t"..(help or "").."\tValue: "..tostring(getval)..readonly..allowed
				else
					return false, "Error retrieving value for setting '"..name.."'."
				end

			--// !set used to show categories
			elseif Set.Settings[name] and not Set.Settings[name][2] then
				local results = {}
				local resultssub = {}
				local deep = #params-1
				local catname = table.concat(params, " ", 2)
		--		local subs = {}
				for varname, data in pairs(Set.Settings) do
					local split = tFuncs:Split(data[1], " ")
					--// Direct item in (sub)category, skip toplevel category
					if data[2] and data[1]:lower() == catname:lower() and data[3] ~= "t" then
						local getval, res, help, data = SetManager:Get(varname)
						if data[3] == "f" then getval = tFuncs:ConvertShare(getval) end
						if data[3] == "d" then getval = getval.." ( "..tFuncs:Duration(getval).." )" end

						if res then
							table.insert(results, "\t!set "..varname..string.rep(" ", (36-#varname)).."\tValue: "..tostring(getval)..string.rep(" ", (36-#tostring(getval))).."\t\t"..(data[8] or "").."\n")
						else
							table.insert(results, "\t!set "..varname.." \tError retrieving value.\n")
						end
					elseif #split==(deep+1) and not data[2] and string.find(data[1]:lower(), catname:lower()) then	--// Found a level deeper than specified
						table.insert(resultssub, "\t!set "..data[1].."\t\t<subcategory>\t\t"..(data[8] or "").."\n")
					end
				end

				if #results == 0 then return false, "This isn't a valid setting or category" end

				table.sort(resultssub)
				table.sort(results)
				return true, "\n\n\tAvailable settings:\n\n"..string.rep("_",60).."\n"..table.concat(results).."\n"..table.concat(resultssub)

			--// !set used to assign a value
			elseif Set.Settings[name] and value then

				--// Check current value
				local getval, ok, help, data = SetManager:Get(name)

				if not ok then return false, "Error retrieving value for "..tostring(name) end

				if data[3] == "s" and (value == "\"\"" or value == "''" or value == "false") then value = "" end
				if data[3] == "t" or data[7] == "GUI" then
					local a = (data[7] and data[7] ~= "" and ". Use "..data[7].." to modify" or "")
					return false, "You can't set this value with commandline"..a
				end
				if data[3] == "f" then getval = tFuncs:ConvertShare(getval) end

				local res, setval, msg = SetManager:Set(name, value, default)

				if res then
					if data[3] == "f" and setval then setval = tFuncs:ConvertShare(setval) end
					if data[3] == "d" and setval then setval = setval.." ( "..tFuncs:Duration(setval).." )" end

					if string.find(tostring(getval), "\n") or string.find(tostring(setval), "\n") then
						--// Multiline reply needed:
						return true, msg or "*** Successfully set '"..name.."' from:\n"..tostring(getval).."\n*** to:\n"..tostring(setval)
					else
						return true, msg or "*** Successfully set '"..name.."' from: "..tostring(getval).." to: "..tostring(setval)
					end
				else
					return false, msg or "Failed to change setting '"..name.."': "..tostring(setval)
				end
			--// Variable doesn't exist, assume categories
--			elseif #params > 1 then
--~ 				tFuncs:Debug("Params > 1: "..table.concat(params, " / "))
--~ 				for name in pairs(total) do
--~ 				for name in pairs(total) do
--			else
--~ 				tFuncs:Debug("Nothing: "..table.concat(params, " / "))
			end
			return false, "Setting '"..name.."' doesn't exist."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s \t\t\t\t\t\tShow hubsettings", cmd).."\n\t"..Lang(ssid, 0, "%s <settingname> <value>\t\t\t\tChange hubsetting (use %s * to display all settings)", cmd, cmd)
		end,
		["ucmd"] = {
			{1,3,"Set variable","!%s %[line:Variable name] %[line:Value]"},
			{1,3,"Show all settings categories","!%s"},
			{1,3,"Show settings category Main","!%s Main"},
			{1,3,"Show settings category Rules","!%s Rules"},
			{1,3,"Show settings category Advanced","!%s Advanced"},
			{1,3,"Show settings category Feed","!%s Feed"},
			{1,3,"Show settings category Traffic","!%s Traffic"},
			{1,3,"Show settings category Thread","!%s Thread"},
			{1,3,"Show settings category Tcp","!%s Tcp"},
			{1,3,"Show settings category Hublists","!%s Hublists"},
			{1,3,"Show settings category Console","!%s Console"},
			{1,3,"Show settings category Redirects","!%s Redirects"},
			{1,3,"Show settings category Security","!%s Security"},
			{1,3,"Show settings sub-category Security/BW - Bandwidth limiting (for unknown ADC traffic)","!%s Security BW"},
			{1,3,"Show settings sub-category Security/CT - Normal CTM (active) security and anti-flood settings","!%s Security CT"},
			{1,3,"Show settings sub-category Security/HC - Hubcommand security and anti-flood settings","!%s Security HC"},
			{1,3,"Show settings sub-category Security/MA - ADC Myinfo security and anti-flood settings","!%s Security MA"},
			{1,3,"Show settings sub-category Security/MC - Mainchat anti-flood settings","!%s Security MC"},
			{1,3,"Show settings sub-category Security/MN - NMDC Myinfo security and anti-flood settings","!%s Security MN"},
			{1,3,"Show settings sub-category Security/NI - Nick security anti-flood settings","!%s Security NI"},
			{1,3,"Show settings sub-category Security/NL - Userlist refreshing limits (NMDC only)","!%s Security NL"},
			{1,3,"Show settings sub-category Security/PM - PM anti-flood settings","!%s Security PM"},
			{1,3,"Show settings sub-category Security/PR - Passive results limits to user","!%s Security PR"},
			{1,3,"Show settings sub-category Security/RC - Reverse CTM (passive) security and anti-flood settings","!%s Security RC"},
			{1,3,"Show settings sub-category Security/SA - Active search anti-flood settings","!%s Security SA"},
			{1,3,"Show settings sub-category Security/SP - Passive search anti-flood settings","!%s Security SP"},
			{1,3,"Show settings sub-category Security/SR - Search results anti-flood settings","!%s Security SR"},
			{1,3,"Show settings sub-category Security/XX - Unspecified messages anti-flood settings","!%s Security XX"},
		},
	},

	["restartscripts"] = {	-- hidden option, use !rs -force to bypass the check of changed files
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			if not tSettings.Update.bNeedRestart then
				local force = string.match(line, "^%S+%s+(%S+)") or true
				return tFuncs:LoadScripts(force)
			else
				return false, "The hub was recently updated but hasn't been restarted yet. First close the hub completely and restart it. ('Stop hub/Start hub' won't work for this)"
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s \t\t\t\t\tRestart hubscripts", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Restart hubscripts (core)", "!%s"},
		},
	},

	["showprofile"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local profname = string.match(line, "^%S+%s+(%S+)")

			if profname then
				local data, err = ProfileManager:GetData(ssid, profname)
				if not data then return false, err end

				local sAllow = ""
				if data.tCmds and data.tCmds.all then
					sAllow = "all"
				else
					--// Get allows
					local tAllow = {}
					if data.iLevel == 10 then
						tAllow = {"all"}
					elseif data.tCmds then
						for name,value in pairs(data.tCmds) do
							if HubCmd[name] or name == "all" or SubSet.ExtraAllows[name] and value then	--// This filters any unknown allows in HubProfiles.ini
								table.insert(tAllow, name)
							end
						end
						table.sort(tAllow)
					end
					sAllow = tFuncs:FlatTable(tAllow, "", ",", "\n\t\t", 15)
				end

				--// Get overrides
				local sOverride = ""
				if data.tOverrides and data.tOverrides.all then
					sOverride = "all"
				else
					local tOverride = {}
					if data.iLevel == 10 then
						tOverride = {"all"}
					elseif data.tOverrides then
						for name,value in pairs(data.tOverrides) do
							if name == "all" or tFuncs:Find(SubSet.Overrides, name) and value then
								table.insert(tOverride, name)
							end
						end
						table.sort(tOverride)
					end
					sOverride = tFuncs:FlatTable(tOverride, "", ",", "\n\t\t", 15)
				end

				local reply = "\n\n\n\t"..Lang(ssid, 0, "Showing profile information for: ")..ProfileManager:GetName(ssid, profname).."\n\t"..string.rep("_",40).."\n"..
					"\n\tProfile level: \t\t"..data.iLevel..
					"\n\tDescription: \t\t"..tostring(data.sDescription or "")..
					"\n\tLogin message: \t\t"..tostring(data.sMessage or "")..
					"\n\tOperator: \t\t\t"..(data.bOperator and "Yes" or "No")..
					"\n\tHas Key: \t\t\t"..(data.bHasKey and "Yes" or "No")..
					"\n\tHas Fake Key: \t\t"..(data.bHasFakeKey and "Yes" or "No")..
					"\n\tForce HideShare: \t\t"..(data.bShareHiddenForced and "Yes" or "No")..
					"\n\tIs external bot:\t\t"..(data.bExtBot and "Yes" or "No")..
					"\n\tAllowed to use ORLY:\t"..(data.bAllowedToRelay and "Yes" or "No")..
					"\n\tProfile name shown to ops:\t"..(data.bShowprofileToOps and "Yes" or "No")..
					"\n\tProfile Min Share: \t\t"..((data.tRules.iMinShare and tFuncs:ConvertShare(data.tRules.iMinShare)) or "disabled, main hubsetting is used.")..
					"\n\tProfile Max Share: \t\t"..((data.tRules.iMaxShare and tFuncs:ConvertShare(data.tRules.iMaxShare)) or "disabled, main hubsetting is used.")..
					"\n\tProfile Min Slots: \t\t"..(data.tRules.iMinSlots or "disabled, main hubsetting is used.")..
					"\n\tProfile Max Slots: \t\t"..(data.tRules.iMaxSlots or "disabled, main hubsetting is used.")..
					"\n\tProfile Max Hubs: \t\t"..(data.tRules.iMaxHubs or "disabled, main hubsetting is used.")..
					"\n\tProfile Max Public Hubs:\t"..(data.tRules.iMaxPublicHubs or "disabled, main hubsetting is used.")..
					"\n\tProfile Max Reg Hubs:\t"..(data.tRules.iMaxRegHubs or "disabled, main hubsetting is used.")..
					"\n\tProfile Max Op Hubs:\t"..(data.tRules.iMaxOpHubs or "disabled, main hubsetting is used.")..
					"\n\n\tAllowed to use the following commands:\n\t\t"..sAllow..
					"\n\n\tCan override the following:\n\t\t"..sOverride.."\n"
				return true, reply
			else
				return false, Lang(ssid, 0, "Error, use: %s <profilename>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> \t\t\t\t\tShow profile settings", cmd)
		end,
		["xhelp"] = function(ssid, line, cdm)
			local cmdlist = {}
			for name,val in pairs(SubSet.ExtraAllows) do
				table.insert(cmdlist, name)
			end
			for name, func in pairs(HubCmd) do
				table.insert(cmdlist, name)
			end
			table.sort(cmdlist)
			local overridelist = SubSet.Overrides
			table.sort(overridelist)

			local reply = "\n\n\n\t"..Lang(ssid, 0, "Showing options for profile settings: ").."\n\t"..string.rep("_",40).."\n"..
				"\n\tAllowed to use:\t(use with !allow <profilename> <commandname> [false])\n"..tFuncs:FlatTable(cmdlist, "\t\t", ",", "\n", 15)..
				"\n\tPossible overrides:\t(use with !override <profilename> <overridename> [false])\n"..tFuncs:FlatTable(overridelist, "\t\t", ",", "\n", 15)..
				"\n\tSpecial cases: instead of commandnames or overridenames the words 'all' and 'none' can be used to allow or to override all or nothing."..
				"\n\n\tFor more info about how to modify a profile, use: !modprofile -h"..
				"\n\tFor more info about allowing certain commands for a profile, use: !allow -h"..
				"\n\tFor more info about overriding certain checks for a profile, use: !override -h\n"
			return true, reply
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show profile settings", "!%s %[line:Profilename]"},
		},
	},

	["allow"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local profname = string.match(line, "^%S+%s+(%S+)")
			local cmdname = string.match(line, "^%S+%s+%S+%s+(%S+)")
			local value = string.match(line, "^%S+%s+%S+%s+%S+%s+(%S+)")

			if profname and cmdname then
				if profname == "*" then
					local ok, output, err = true, "\n\n", ""
					for name in pairs(tProfiles) do
						local res,txt = ProfileManager:Allow(ssid, name, cmdname, value)
						output = output..name..": "..txt.."\n"
						if not res then
							ok = false
							err = txt
						end
					end
					if not ok then
						return false, "Error: "..err..output
					else
						return true, output
					end
				else
					return ProfileManager:Allow(ssid, profname, cmdname, value)
				end
			else
				return false, Lang(ssid, 0, "Error, use: %s <profilename> <commandname>[,<commandname2>] [false]\t\tUse !allow -h for possible commands/options", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> <commandname>[,<commandname2>] [false]\tAllow a profile to use certain commands/options", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local cmdlist = {}
			for name,val in pairs(SubSet.ExtraAllows) do
				table.insert(cmdlist, name)
			end
			for name, func in pairs(HubCmd) do
				table.insert(cmdlist, name)
			end
			table.sort(cmdlist)
			local reply = "\n\n\n\t"..Lang(ssid, 0, "Showing commandnames for !allow <profilename> <commandname>[,<commandname2>]").."\n\t"..string.rep("_",80).."\n\n"..tFuncs:FlatTable(cmdlist, "\t", ",", "\n", 10).."\n"
			return true, reply
		end,
		["ucmd"] = {
			{1,3,"Profile, allow profile to use certain commands","!%s %[line:Profile name] %[line:Command name (or multiple names comma-separated, no spaces)] %[line:Value (true or false, default= true)]"},
		},
	},

	["override"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local profname = string.match(line, "^%S+%s+(%S+)")
			local cmdname = string.match(line, "^%S+%s+%S+%s+(%S+)")
			local value = string.match(line, "^%S+%s+%S+%s+%S+%s+(%S+)")

			if profname and cmdname then
				if profname == "*" then
					local output = "\n\n"
					for name in pairs(tProfiles) do
						local res,txt = ProfileManager:Override(ssid, profname, cmdname, value)
						output = output..name..": "..txt.."\n"
					end
					return true, output
				else
					return ProfileManager:Override(ssid, profname, cmdname, value)
				end
			else
				return false, Lang(ssid, 0, "Error, use: %s <profilename> <setting> [false]\t\tUse !override -h for possible settings", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> <setting> [false]\t\t\t\tOverride a certain setting for a profile", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local overridelist = SubSet.Overrides
			table.sort(overridelist)
			local reply = "\n\n\n\t"..Lang(ssid, 0, "Showing settings for !override <profilename> <setting> [false]").."\n\t"..string.rep("_",80).."\n\n"..tFuncs:FlatTable(overridelist, "\t", ",", "\n", 10).."\n"
			return true, reply
		end,
		["ucmd"] = {
			{1,3,"Profile, override certain checks for profile","!%s %[line:Profile name] %[line:Override name (or multiple names comma-separated, no spaces)] %[line:Value (true or false, default= true)]"},
		},
	},

	["modprofile"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local profname = string.match(line, "^%S+%s+(%S+)")
			local cmdname = string.match(line, "^%S+%s+%S+%s+(%S+)")
			local value = string.match(line, "^%S+%s+%S+%s+%S+%s+(.*)$")

			if profname and not ProfileManager:GetName(ssid, profname) then
				return false, Lang(ssid, 0, "There is no profile with name : "..profname)
			elseif not ProfileManager.Settings[cmdname] then
				return false, Lang(ssid, 0, "No valid settingname specified")
			elseif profname and cmdname and profname ~= "-h" then
				return ProfileManager:Set(ssid, profname, cmdname, value)
			else
				return false, Lang(ssid, 0, "Error, use: %s <profilename> <setting> [value]", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> <setting> <value>\t\t\tChange a profilesetting or use !modprofile -h for more help", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local info = {	-- use the numbers to define view order - made to match the order in showprofile
				["level"] = {1, "Which level should this profile have, a number between -1 and 10." },
				["desc"] = {2, "Profile description, a text describing the profile."},
				["message"] = {3, "Profile login message, a text that can be displayed with a text command when the user logs in."},
				["operator"] = {4, "Is Operator, can be set to true or false, the operator switch controls numerous behaviour in the hub, in general an operator will have precedence before any non operator."},
				["key"] = {5, "Has key, can be set to true or false."},
				["fakekey"] = {6, "Has 'fake' key, can be set to true or false."},
				["hideshare"] = {7, "Force hideshare, can be set to true or false."},
				["extbot"] = {8, "This profile is used for external bots, can be set to true or false. When set to true replies from the bot will only be seen by the user sending the command, even when used in mainchat."},
				["relay"] = {9, "Enable support for the ORLY ADC extension for this profile, can be set to true or false. When set to true it allows the users of this profile to send messages to the hub that are relayed to other users as if the hub sent them."},
				["showprofiletoops"] = {10, "If enabled, the profile name will be added to the description of all users, only shown to operators, can be set to true or false."},
				["minshare"] = {11, "Minimum share to be allowed to join the hub (in B, kB, MB, GB, TB), set to 'false' to use the normal hub setting."},
				["maxshare"] = {12, "Maximum share to be allowed to join the hub (in B, kB, MB, GB, TB), set to 'false' to use the normal hub setting."},
				["minslots"] = {13, "Minimum amount of slots to be allowed to join the hub, set to 'false' to use the normal hub setting."},
				["maxslots"] = {14, "Maximum amount of slots to be allowed to join the hub, set to 'false' to use the normal hub setting."},
				["maxhubs"] = {15, "Maximum amount of hubs a user is allowed to be connected to, set to 'false' to use the normal hub setting."},
				["maxpublichubs"] = {16, "Maximum amount of public hubs a user is allowed to be connected to, set to 'false' to use the normal hub setting."},
				["maxreghubs"] = {17, "Maximum amount of hubs where user is registered to be allowed to join the hub, set to 'false' to use the normal hub setting."},
				["maxophubs"] = {18, "Maximum amount of hubs where user is operator to be allowed to join the hub, set to 'false' to use the normal hub setting."},
			}

			local slist = {}
			for name, _ in pairs(ProfileManager.Settings) do
				table.insert(slist, (info[name] and {name, info[name][1], info[name][2],} or {name, 99, "info is missing",} ) )
			end

			local sorter = function(a,b)
				if a[2] < b[2] then return true else return false end
			end
			table.sort(slist, sorter)

			local reply = "\n\n\t"..Lang(ssid, 0, "Showing avaliable settings for %s <profilename> <setting> <value>", cmd).."\n\t"
			.."Show current setting with !showprofile <profilename>\n\n"
			.."\tSetting\t\tDescription\n\t"..string.rep("_",80).."\n"

			for i, d in ipairs(slist) do
				reply = reply.."\t"..Padding(d[1], 20).."\t".. (d[3] or "") .."\n"
			end

			return true, reply.."\n"
		end,
		["ucmd"] = {
			{1,3,"Profile, change level","!%s %[line:Profilename] level %[line:Profilelevel (between -1 and 10)]"},
			{1,3,"Profile, change description","!%s %[line:Profilename] desc %[line:Description]"},
			{1,3,"Profile, change login message","!%s %[line:Profilename] message %[line:Message]"},
			{1,3,"Profile, change key","!%s %[line:Profilename] key %[line:Does this profile have a key (true or false, empty = true)]"},
			{1,3,"Profile, change fake key","!%s %[line:Profilename] fakekey %[line:Does this profile have a fake key (true or false, empty = true)]"},
			{1,3,"Profile, change operator","!%s %[line:Profilename] operator %[line:Is this a profile for operator or higher? (true or false, empty = true)]"},
			{1,3,"Profile, change hideshare","!%s %[line:Profilename] hideshare %[line:Force hideshare for this profile (true or false, empty = true)]"},
			{1,3,"Profile, change extbot","!%s %[line:Profilename] extbot %[line:Is this profile used for external bots (true or false, empty = true)]"},
			{1,3,"Profile, change relay","!%s %[line:Profilename] relay %[line:Enable support for the ORLY ADC extension for this profile (true or false, empty = true)]"},
			{1,3,"Profile, change show profile to Ops","!%s %[line:Profilename] showprofiletoops %[line:Show the name of the profile in description to ops (true or false, empty = true)]"},
			{1,3,"Profile, change profile minshare","!%s %[line:Profilename] minshare %[line:Min share for this profile (value or false]"},
			{1,3,"Profile, change profile maxshare","!%s %[line:Profilename] maxshare %[line:Max share for this profile (value or false]"},
			{1,3,"Profile, change profile minslots","!%s %[line:Profilename] minslots %[line:Min slots for this profile (value or false]"},
			{1,3,"Profile, change profile maxslots","!%s %[line:Profilename] maxslots %[line:Max slots for this profile (value or false]"},
			{1,3,"Profile, change profile maxhubs","!%s %[line:Profilename] maxhubs %[line:Max hubs for this profile (value or false]"},
			{1,3,"Profile, change profile maxpublichubs","!%s %[line:Profilename] maxpublichubs %[line:Max public hubs for this profile (value or false]"},
			{1,3,"Profile, change profile maxreghubs","!%s %[line:Profilename] maxreghubs %[line:Max reg hubs for this profile (value or false]"},
			{1,3,"Profile, change profile maxophubs","!%s %[line:Profilename] maxophubs %[line:Max op hubs for this profile (value or false]"},
		},
	},

	["update"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd") and not tGlobals.bNoUpdate
		end,
		["cmd"] = function(ssid, line, cmd)
			if tThreads.Update then
				return false, "Update already running."
			elseif tSettings.Update.bNeedRestart then
				return false, "The hub was recently updated but hasn't been restarted yet. First close the hub completely and restart it. ('Stop hub/Start hub' won't work for this)"
			else
				tThreads.Update = ThreadCreate(tUpdate.Thread, {1, ssid, line})
				return true, "Hub update started in a separate thread, output will appear in mainchat when it's finished."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [now]\t\t\t\t\tUpdate FlexHub, use %s to check if an update is avaliable and %s now to download and install updates", cmd, cmd, cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			return true , "Avaliable options for the Update command in FlexHub\n\n"
			.."\t!update\t\tTo check if there is a new version avaliable (also runs on timer every second day if option !set update_check is enabled)\n"
			.."\t!update now\tTo download and install an update\n"
			.."\t!update compare\tCompares the local files with whats on the update server and lists the ones that are not identical.\n"
			.."\t!update force\tTo force an installation of update even if the automated check tells the hub its uptodate\n"
		end,
		["ucmd"] = {
			{1,3,"Check for FlexHub updates","!%s"},
			{1,3,"Perform FlexHub update","!%s %[line:To start the update, type: now]"},
		},
	},


	["setredirects"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local which,address = string.match(line, "^%S+%s+(%S+) (%S+)")

			if not which or not address then
				return false, Lang(ssid, 0, "Error, use: %s * <address> (Or use: flood,full,main,maxhubs,maxshare,maxslots,minshare,minslots,adc_disabled,nmdc_disabled,failover,passive instead of *)", cmd)
			end

			local allowed = {
				["flood"] = "rd_flood",
				["full"] = "rd_full",
				["main"] = "rd_main",
				["maxhubs"] = "rd_maxhubs",
				["maxshare"] = "rd_maxshare",
				["maxslots"] = "rd_maxslots",
				["minshare"] = "rd_minshare",
				["minslots"] = "rd_minslots",
				["adc_disabled"] = "rd_adc_disabled",
				["nmdc_disabled"] = "rd_nmdc_disabled",
				["failover"] = "rd_failover",
				["passive"] = "rd_passive",
				["regonly"] = "rd_regonly",
			}

			local result = ""
			if which == "*" then
				for rdname, rdtype in pairs(allowed) do
					SetManager:Set(rdtype, address)
					result = result..rdname.." "
				end
			else
				for rdname in string.gfind(which, "%a") do
						if allowed[rdname] then
							SetManager:Set(allowed[rdname], address)
							result = result..rdname.." "
						end
				end
			end

			if result ~= "" then
				local _,overview = SetCmd["showredirects"]["cmd"](ssid,"!showredirects","showredirects")
				return true, Lang(ssid, 0, "Redirect address set to: %s for the following types: %s", address, result..overview)
			else
				return false, Lang(ssid, 0, "Error, use: %s * <address> (Or use: flood,full,main,maxhubs,maxshare,maxslots,minshare,minslots,adc_disabled,nmdc_disabled,failover,passive instead of *)", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tSet redirect addresses for multiple types", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			return true, Lang(ssid, 0, "%s * <address> \t(Or use: flood,full,main,maxhubs,maxshare,maxslots,minshare,minslots,adc_disabled,nmdc_disabled,failover,passive instead of *)", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1,3,"Set multiple redirect addressess","!%s %[line:List types (flood,full,main,maxhubs etc. or * for all types] %[line:Redirect address]"},
		},
	},


	["showredirects"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local opt = string.match(line, "^%S+%s+(%S+)")

			if opt and opt == "clean" then
				-- added mutex here because redirecting uses while its running makes a mess
				ThreadLockMutex(mutextmploop)
					for name, dta in pairs(tStats.Redirects.List) do
						if dta.Counter then dta.Counter = 0 end
					end
				ThreadUnlockMutex(mutextmploop)
				return true, Lang(ssid, 0, "Redirect count cleared", cmd)
			elseif opt then
				return false, Lang(ssid, 0, "Unknown option", cmd)
			else
				local rules = {"main", "failover", "regonly", "full", "maxhubs", "minslots", "maxslots", "minshare", "maxshare", "passive", "flood", "adc_disabled", "nmdc_disabled", "deadend", "deadendfail"}
				local output = "\n\n\tRedirects:\n\t"..string.rep("_", 80)

				for num,name in ipairs(rules) do
					output = output.."\n\trd_"..Padding(name,26).." \t"..tStats.Redirects.List[name].Counter.." \tAddress(es): "..((SetManager:Get("rd_"..name)) or "No address set, will use disconnect")
				end
				output = output.."\n\t"..string.rep("_", 80).."\n"
				local addysort = {}
				for address, data in pairs(tStats.Redirects.Addresses) do
					table.insert(addysort, address)
				end
				table.sort(addysort, SortLowercase)
				for num, address in ipairs(addysort) do
					local verified, status = tRedirect:GetStatus(address)
					output = output.."\n\t"..tStats.Redirects.Addresses[address].Counter.." \t"..Padding(address, 56).." \tStatus: "..status
				end
				return true, output.."\n\n"
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow redirect statistics", cmd) .."\n\t"..
					Lang(ssid, 0, "%s clean\t\t\t\tClear redirect statistics", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1,3,"Show redirect statistics","!%s"},
		},
	},

	["addprofile"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local newprofile, level = string.match(line, "^%S+%s+(%S+)%s+(%--%d+)")
			local sourceprofile = string.match(line, "^%S+%s+%S+%s+%S+%s+(%S+)") or "Reg"

			if newprofile and level then
				return ProfileManager:Add(ssid, newprofile, level, sourceprofile)
			elseif not level then
				return false, Lang(ssid, 0, "Error, no profilelevel specified, use: %s <profilename> <profilelevel> [<source profile>] (If no source profile is specified, the Reg profile settings will be used)", cmd)
			else
				return false, Lang(ssid, 0, "Error, use: %s <profilename> <profilelevel> [<source profile>] (If no source profile is specified, the Reg profile settings will be used)", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> <profilelevel> [<source profile>]\t\tAdd a profile. If no source profile is specified, the Reg profile settings will be used", cmd)
		end,
		["ucmd"] = {
			{1,3,"Add a new profile","!%s %[line:New profile name] %[line:Profile level (between -1 and 10)] %[line:Source Profile (if not supplied Reg will be used)]"},
		},
	},

	["delprofile"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local profilename = string.match(line, "^%S+%s+(%S+)")
			local replaceprofile = string.match(line, "^%S+%s+%S+%s+(%S+)")

			if profilename then
				return ProfileManager:Delete(ssid, profilename, replaceprofile)
			else
				return false, Lang(ssid, 0, "Use: %s <profilename> [<replacement profile>]\t\t\tDelete a userprofile, if no replacement profile is specified, users will be moved to 'Reg' profile", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> [<replacement profile>]\t\t\tDelete a userprofile, if no replacement profile is specified, users will be moved to 'Reg' profile", cmd)
		end,
		["ucmd"] = {
--~ 			{1,3,"Dummy","!%s"},
		},
	},

	["addhublist"] = {
		--// !addhublist NMDC addy:port <nick> <pass>
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			if tThreads.HubListReg then
				return false, "Hublist registration is currently running, can't add or delete hublists until it's finished."
			end
			local stype, address = string.match(line, "^%S+%s+(%S+)%s+(%S+)")
			if address and stype then
				address = address:lower()
				stype = stype:upper()
				if stype == "NMDC" or stype == "ADC" then
					if not tHubSettings.tHubLists[stype][address] then
						tHubSettings.tHubLists[stype][address] = { ["Enabled"] = true, ["LastRegTime"] = 0, ["LastPingTime"] = 0, ["LastReg"] = false }
						tFuncs:SaveSettings("settings")
						return true, Lang(ssid, 0, "Added %s-hublist with address: %s \nUse !reghub to manually register to all hublists", stype, address)
					else
						return true, Lang(ssid, 0, "%s-hublist entry for address: %s already exists.", stype, address)
					end
				else
					return false, Lang(ssid, 0, "Wrong protocol, use ADC or NMDC")
				end
			else
				return false, Lang(ssid, 0, "Error: use %s <protocol> <hublist-address:port>\t\t(protocol = ADC or NMDC)", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <protocol> <address:port>\t\t\tAdd a hublist (protocol = ADC or NMDC)", cmd)
		end,
		["ucmd"] = {
			{1,3,"Add hublist for registration","!%s %[line:Add hublist for ADC or NMDC?] %[line:Hublist Regserver address:port]"},
		},
	},

	["delhublist"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			if tThreads.HubListReg then
				return false, "Hublist registration is currently running, can't add or delete hublists until it's finished."
			end
			local stype, address = string.match(line, "^%S+%s+(%S+)%s+(%S+)")
			if address and stype then
				address = address:lower()
				stype = stype:upper()
				if stype == "NMDC" or stype == "ADC" then
					if tHubSettings.tHubLists[stype][address] then
						tHubSettings.tHubLists[stype][address] = nil
						tFuncs:SaveSettings("settings")
						return true, Lang(ssid, 0, "Removed "..stype.."-hublist with address: %s", address)
					else
						return false, Lang(ssid, 0, "Couldn't find "..stype.."-hublist with address: %s", address)
					end
				else
					return false, Lang(ssid, 0, "Wrong protocol, use ADC or NMDC")
				end
			else
				return false, Lang(ssid, 0, "Error: use %s <protocol> <hublist-address>\t(protocol = ADC or NMDC)", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <protocol> <address:port>\t\t\tDelete a hublist (protocol = ADC or NMDC)", cmd)
		end,
		["ucmd"] = {
			{1,3,"Delete hublist for registration","!%s %[line:Delete hublist for ADC or NMDC?] %[line:Hublist Regserver address:port]"},
		},
	},

	["showhublists"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local output = "\n\n\tHublist overview:\n\t"..string.rep("_",80)

			--// Sort hublists:
			local hublists = { ["NMDC"] = {}, ["ADC"] = {}}
			for _,protocol in pairs {"NMDC", "ADC"} do
				for address, data in pairs(tHubSettings.tHubLists[protocol]) do
					if data.Enabled then
						table.insert(hublists[protocol], address)
					end
				end
				table.sort(hublists[protocol], SortLowercase)
			end

			for _,protocol in pairs {"NMDC", "ADC"} do
				output = output.."\n\n"..protocol.." hublists:"
				for _, address in ipairs(hublists[protocol]) do
					local data = tHubSettings.tHubLists[protocol][address]
					local line = "\t"..Padding(address,40).."\tLast connection attempt: "
					if data.LastRegTime and data.LastRegTime > 0 then
						line = line..ToDate(data.LastRegTime).." Result: "
						if data.LastReg then
							line = line.."Succesful"
						elseif data.LastRegOk and data.LastRegOk > 0 then
							line = line.."Failed, last successful connection at: "..ToDate(data.LastRegOk)
						else
							line = line.."Failed, no previous successful connection was made."
						end
					else
						line = line.."No connection has been attempted yet"
					end
					output = output.."\n"..line
				end
			end
			return true, output.."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow hublists", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1,3,"Show hublists overview","!%s"},
		},
	},

	["reghub"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			if tThreads.HubListReg then
				return false, "Hublist registrations currently running: "..tLogs.Hublists.Registration.Status
			elseif tLogs.Hublists.Registration.LastRun >= (os.time() - tHubSettings.tHubLists.MinInterval) then
				return false, "Hublist registration executed at: "..ToDate(tLogs.Hublists.Registration.LastRun)..", wait until "..ToDate((tLogs.Hublists.Registration.LastRun+tHubSettings.tHubLists.MinInterval)).." before registering again."
			else
				tThreads.HubListReg = ThreadCreate(tFuncs.RegHub, {1, ssid})
				return true, "Manual hublist registration started, output will appear in mainchat when it's finished."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tRegister the hub to available hublists", cmd)
		end,
		["ucmd"] = {
			{1,3,"Register hub to hublists","!%s"},
		},
	},

	["checkredirects"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			if tThreads.CheckRedirects then
				return false, "Redirect check is currently running, try again later"
			else
				tThreads.CheckRedirects = ThreadCreate(tRedirect.Check, {1, ssid})
				return true, "Manual redirect check started, output will appear in mainchat when it's finished."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tCheck online status of redirect addresses", cmd)
		end,
		["ucmd"] = {
			{1,3,"Check online status of redirect addresses","!%s"},
		},
	},

	["addstats"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local rest = line

			--// Get commandname
			local name, seconds, content = rest:match("^%S+%s+(%S+)%s+%-[Tt](%S+)%s+(.+)")
			if not name or not content then return false, "Error: use "..cmd.." <filename> -t<time> <text>" end
			name = name:lower()

 			if name then name = name:gsub("%..*$", "") end	--// Remove everything after a dot
 			if name and not tFuncs:ValidFilename(name) then return false, "Invalid filename, use alphanumeric characters and underscore only, don't specify an extention." end

			if tHubSettings.SaveStats[name] then
				return false, "Filename already exists, use !modcmd to change it, or use a new name"
			else
				seconds = (seconds and tFuncs:ConvertToSeconds(seconds)) or false
				if not seconds then return false, "Invalid time (-t) specified, use time in seconds, or use m for minutes, h for hours" end

				tHubSettings.SaveStats[name] = {
					["sContent"] = content,
					["iTimer"] = seconds,
				}
				tFuncs:SaveSettings("settings")

				return true, "Saving statistics to file: "..name.."', runs every "..seconds.." seconds:"..content
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <filename> -t<time> <text> \t\tAdd saving statistics", cmd)
		end,
		["ucmd"] = {
			{1,3,"Add saving statistics","!%s %[line:Filename] -t%[line:Interval in seconds] %[line:Text to save]"},
		},
	},

	["modstats"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local name, rest = line:match("^%S+%s+(%S+)(.+)")

			if name and tHubSettings.SaveStats[name] then
				local data = tHubSettings.SaveStats[name]
				local changed = false

				--// Get time
				local seconds = rest:match("%s+%-[tT](%S+)")
				seconds = (seconds and tFuncs:ConvertToSeconds(seconds)) or (not seconds and (data.iTimer or 0)) or false
				if not seconds then return false, "Invalid time (-t) specified, use time in seconds, or use m for minutes, h for hours" end
				if seconds ~= data.sTimer then
					data.iTimer = seconds
					changed = true
				end

				rest = rest:gsub("^.-%-[tT]%S+", "")

				--// Get content
				local content = rest:match("%s+(.+)")

				--// Change values
				if content and content ~= data.sContent then
					data.sContent = content
					changed = true
				end

				if changed then
					tFuncs:SaveSettings("settings")
					return true, "Modified saving statistics to file: "..name.."', runs every "..data.iTimer.." seconds: "..data.sContent
				else
					return false, "Statistics didn't change"
				end
			elseif name then
				return false, "Filename not found."
			else
				return false, Lang(ssid, 0, "Use: %s <filename> [-t<time>] [<text>]", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <filename> [-t<time>] [<text>]\t\tModify a statistics file", cmd)
		end,
		["ucmd"] = {
			{1,3,"Modify saving statistics","!%s %[line:Textcommand name] %[line:-o<options>] %[line:-p<profilelevel>] %[line:-m<max profilelevel>] %[line:-u<use profilelevel>] %[line:-t<time>] %[line:-h\"<helptext>\"] %[line:\"<text in quotes>\"]"},
		},
	},

	["delstats"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local name = string.match(line, "^%S+%s+(%S+)")
			if tHubSettings.SaveStats[name] then
				tHubSettings.SaveStats[name] = nil
				tFuncs:SaveSettings("settings")

				return true, "Removed saving statistics to file: "..name
			else
				return false, "Statistics file doesn't exist, use !showstats to show available filenames."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name>\t\t\t\t\t\t\tDelete a statistics save", cmd)
		end,
		["ucmd"] = {
			{1,3,"Delete saving statistics","!%s %[line:Filename]"},
		},
	},

	["showstats"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local name = string.match(line, "^%S+%s+(%S+)")

			if name and tHubSettings.SaveStats[name] then
				local timer = "Disabled"
				if tHubSettings.SaveStats[name].iTimer and tHubSettings.SaveStats[name].iTimer > 0 then
					timer = tHubSettings.SaveStats[name].iTimer .." seconds"
					--timer = os.date("%Hh %Mm %Ss", tCommands[name].iTimer)
				end
				local reply = "\n\n\tSettings for saving statistics to file:\n"..string.rep("_",80).."\n"..
					"\tFilename: \t\t"..name.."\n"..
					"\tTimer: \t\t"..timer.."\n"..
					"\n\tText: \t\t"..tHubSettings.SaveStats[name].sContent.."\n"
				return true, reply
			elseif name then
				return false, "There's no saving statistics to filename: "..name
			else
				local reply = "\n\n\tAvailable statistics saving:\n\n\t"..Lang(ssid, 0, "Filename").."\t".."\tTimer\t".."\tContent\n\t"..string.rep("_",80).."\n"
				local result = {}
				for name, data in pairs(tHubSettings.SaveStats) do
					table.insert(result, "\t"..Padding(name, 20).."\t"..Padding(((data.iTimer and data.iTimer > 0 and data.iTimer.." seconds") or "Disabled"), 20).."\t"..data.sContent.."\n")
				end
				table.sort(result)
				reply = reply..table.concat(result)
				return true, reply
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\t\tShow saving statistics", cmd).."\n\t"..
				Lang(ssid, 0, "%s <filename>\t\t\t\t\t\t\tShow saving statistics filename", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show saving statistics","!%s"},
			{1,3,"Show saving statistics filename","!%s %[line:Filename]"},
		},
	},

	["addcmd"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			--// Get commandname
			local name = line:match("^%S+%s+([^%-]-%S+)")
			if not name then return false, "Error: no valid name specified" end
			local rest = line:gsub("^%S+%s+%S+", "")
			name = name:lower()
			--tFuncs:Debug("name = "..name, "X")
			--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")

			if tCommands[name] then
				return false, "TextCommand already exists, use !modcmd to change it"
			elseif HubCmd[name] or SetCmd[name] or tCmdAlias[name] then
				return false, "Command with name "..name .." already exists, use another name"
			elseif #name > 35 then
				return false, "Command name is too long, use a shorter name please (max 35 characters allowed)"
			elseif name =="-h" or name == "-showvars" then
				return false, "The command name matches an option name, you can't use that"
			end

			--// Get helptxt
			local helptxt = rest:match("%s+%-[hH]%s*\"(.-)\"")
			if helptxt then
				rest = rest:gsub("%s+%-[hH]%s*\".-\"", "", 1)
				--tFuncs:Debug("helptxt = "..helptxt, "X")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
		--	elseif rest:match(" %-[hH]") then
		--		return false, "Error: -h specified, but no \"<helptext>\" given, make sure to use quotationmarks"
			end

			--// Get content
			local content = rest:match("\"(.+)\"%s*$")
			if content then
				rest = rest:gsub("%s+\".+\"%s*$", "", 1)
			elseif not content then
				return false, "Error: no text specified"
			elseif not helptxt and rest:match(" %-[hH]") then
				return false, "Error: -h specified, but no \"<helptext>\" given, make sure to use quotationmarks"
			end

			-- parsing content for vars
			local VarTrue, ParmTrue = false, false
			if string.find(content, "%[", 1, true) then
				VarTrue = true
				if string.find(content, "%[parm]", 1, true) then
					local _, count = string.gsub (content, "%%%[parm%]", "")
					ParmTrue = count
				end
			end

			--tFuncs:Debug("content = "..content, "X")
			--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")

			--// Get parameters
			local profilelevel = rest:match("%s+%-[pP](%S+)")
			local profilelevel = (profilelevel and tonumber(profilelevel)) or (not profilelevel and -1) or false
			if profilelevel and (profilelevel < -1 or profilelevel > 10) then
				return false, "Invalid profilelevel (-p) specified, only numbers between -1 and 10 are valid"
			elseif not profilelevel and rest:match(" %-[pP]") then
				return false, "Error: -p specified, but no valid profilelevel given"
			else
				rest = rest:gsub("%s+%-[pP]%S+", "", 1)
				--tFuncs:Debug("profilelevel = "..profilelevel, "X")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
			end

			local maxlevel = rest:match("%s+%-[mM](%S+)")
			local maxlevel = (maxlevel and tonumber(maxlevel)) or (not maxlevel and 10) or false
			if maxlevel and (maxlevel < -1 or maxlevel > 10) then
				return false, "Invalid max profilelevel (-m) specified, only numbers between -1 and 10 are valid"
			elseif not maxlevel and rest:match(" %-[mM]") then
				return false, "Error: -m specified, but no valid max profilelevel given"
			else
				rest = rest:gsub("%s+%-[mM]%S+", "", 1)
				--tFuncs:Debug("maxlevel = "..maxlevel, "X")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
			end

			local uselevel = rest:match("%s+%-[uU](%S+)")
			local uselevel = (uselevel and tonumber(uselevel)) or (not uselevel and profilelevel) or false
			if uselevel and (uselevel < -1 or uselevel > 10) then
				return false, "Invalid use profilelevel (-u) specified, only numbers between -1 and 10 are valid"
			elseif not uselevel and rest:match(" %-[uU]") then
				return false, "Error: -u specified, but no valid use profilelevel given"
			else
				rest = rest:gsub("%s+%-[uU]%S+", "", 1)
				--tFuncs:Debug("uselevel = "..uselevel, "X")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
			end

			local option = rest:match("%s+%-[oO](%d+)")
			local option = (option and tonumber(option)) or (not option and 0) or false
			if option and (option < 0 or option > 512) then
				return false, "Invalid option (-o) specified, only numbers between 0 and 512 are valid"
			elseif not option and rest:match(" %-[oO]") then
				return false, "Error: -o specified, but no valid option given"
			else
				rest = rest:gsub("%s+%-[oO]%d+", "", 1)
				--tFuncs:Debug("option = "..option, "X")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
			end

			local seconds = rest:match("%s+%-[tT](%S+)")
			local seconds = (seconds and tFuncs:ConvertToSeconds(seconds)) or (not seconds and 0) or false
			if seconds and seconds > 0 and not tFuncs:CheckFlag(option, 2) then
				return false, "Error: You can only use the timer (-t) on messages to all (option 2 must be included)"
			elseif not seconds and rest:match(" %-[tT]") then
				return false, "Error: -t specified, but no valid time given"
			else
				rest = rest:gsub("%s+%-[tT]%S+", "", 1)
				--tFuncs:Debug("seconds = "..seconds, "X")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
			end

			-- test if we captured everything
			rest = rest:gsub("%s+", "")
			--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
			if #rest ~= 0 then
				return false, "Error: Bad or missing command"
			elseif maxlevel < profilelevel then
				return false, "Error: profilelevel can't be larger than max profilelevel"
			elseif uselevel < profilelevel then
				return false, "Error: profilelevel can't be larger than use profilelevel"
			end
			tCommands[name] = {
				["sContent"] = content,
				["iProfile"] = profilelevel,
				["iOptions"] = option,
				["iMaxProfile"] = maxlevel,
				["iUseProfile"] = uselevel,
				["iTimer"] = seconds,
				["bHasFillVars"] = VarTrue,
				["bHasParm"] = ParmTrue,
				["sHelpText"] = (helptxt or ""),
			}
			tFuncs:SaveSettings("commands")

			local ret = ""
			if seconds and seconds > 0 then
				ret = " Executes on timer each "..seconds.." seconds."
			end

			--// recreate UC when modifying the tCommands table
			tUserCmds:Build()

			local res, reply = SetCmd["showcmd"]["cmd"](ssid, "!showcmd "..name, "!showcmd")
			if res and reply then
				return true, "Textcommand '"..name.."' added: "..reply
			else
				return false, reply
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name> [-o<options>] [-p<profilelevel>] [-m<max profilelevel>] [-u<use profilelevel>] [-t<time>] [-h\"<helptext>\"] \"<text>\"\tAdd a textcommand", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local cmdhelp = "Help for "..cmd..": \n\n\t "..cmd.." <name> [-o<options>] [-p<profilelevel>] [-m<max profilelevel>] [-u<use profilelevel>] [-t<time>] [-h\"<helptext>\"] \"<text>\"\n\n"..
				"\t\t<options>\t\t\tA number created by adding the values of the following options:\n\n"..
				"\t\t\t0\t\tOutput to mainchat (default)\n"..
				"\t\t\t1\t\tDisplay text in PM instead of mainchat\n"..
				"\t\t\t2\t\tShow text to all (when this is used, you can use -p and -m to limit output to certain profilelevels)\n"..
				"\t\t\t4\t\tShow text when user logged in (welcome messages/rightclicks)\n"..
				"\t\t\t8\t\tDon't allow the command to be activated in mainchat (useful in combination with 4)\n"..
				"\t\t\t16\t\tDo not create an item in !help for this command (forced when 8 is used)\n"..
				"\t\t\t32\t\tDo not create a rightclick item for this command (forced when 8 is used)\n"..
				"\t\t\t64\t\tText should be send in raw format instead of mainchat/pm (useful to send custom rightclicks)\n"..
				"\t\t\t128\t\tOnly send to ADC user(s)\n"..
				"\t\t\t256\t\tOnly send to NMDC user(s)\n\n"..
				"\t\t<profilelevel>\t\t(for use with option 2) A number in the range from -1 (User) to 10 (Hubowner) that is the lowest profile that will see the command output (default is -1)\n"..
				"\t\t<max profilelevel>\t\t(for use with option 2) A number in the range from -1 (User) to 10 (Hubowner) that is the highest profile that will see the command output (default is level 10)\n"..
				"\t\t<use profilelevel>\t\tA number in the range from -1 (User) to 10 (Hubowner) that is the lowest profile that can use the command (default is -1)\n"..
				"\t\t<time>\t\t\t(for use with option 2) Time in seconds between broadcasts, or use m for minutes, h for hours\n"..
				"\t\t<helptext>\t\tHelptext to show in !help and display in rightclick usercommands\n"..
				"\t\t<text>\t\t\tText to send, must be the last parameter on the line\n\n"
			return true, cmdhelp
		end,
		["ucmd"] = {
			{1,3,"Show help for adding text commands","!%s -h"},
			{1,3,"Add a textcommand","!%s %[line:Textcommand name] %[line:-o<options>] %[line:-p<profilelevel>] %[line:-m<max profilelevel>] %[line:-u<use profilelevel>] %[line:-t<time>] %[line:-h\"<helptext>\"] %[line:\"<text in quotes>\"]"},
		},
	},

	["modcmd"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			--// Get commandname
			local rName = line:match("^%S+%s+([^%-]-%S+)")
			if not rName then return false, Lang(ssid, 0, "Error: no valid name specified. ") end
			local rest = line:gsub("^%S+%s+%S+", "")
			local name = ""

			if tCommands[rName] then
				name = rName
			else
				name = rName:lower()
			end
	--		tFuncs:Debug("name = "..rName, "X")

			if tCommands[name] then
				local data = tCommands[name]
				local changed = false

				--// Get helptxt
				local helptxt = rest:match("%s+%-[hH]%s*\"(.-)\"")
				if helptxt then
					rest = rest:gsub("%s+%-[hH]%s*\".-\"", "", 1)
					--tFuncs:Debug("helptxt = "..helptxt, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				elseif rest:match(" %-[hH]") then
					return false, "Error: -h specified, but no \"<helptext>\" given, make sure to use quotationmarks"
				end

				--// Get content
				local content = rest:match("\"(.+)\"%s*$")
				if content then
					rest = rest:gsub("%s+\".+\"%s*$", "", 1)
					--tFuncs:Debug("content = "..content, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				end

			-- parsing content for vars
				local VarTrue, ParmTrue = false, false
				if content and string.find(content, "%[", 1, true) then
					--tFuncs:Debug("VarTrue : True ", "X")
					VarTrue = true
					if string.find(content, "%[parm]", 1, true) then
						local _, count = string.gsub (content, "%%%[parm%]", "")
						ParmTrue = count
					--	tFuncs:Debug("ParmTrue "..ParmTrue, "X")
					end
				end

				--// Get parameters
				local profilelevel = rest:match("%s+%-[pP](%S+)")
				local profilelevel = (profilelevel and tonumber(profilelevel)) or (not profilelevel and (data.iProfile or -1)) or false
				if profilelevel and (profilelevel < -1 or profilelevel > 10) then
					return false, "Invalid profilelevel (-p) specified, only numbers between -1 and 10 are valid"
				elseif not profilelevel and rest:match(" %-[pP]") then
					return false, "Error: -p specified, but no valid profilelevel given"
				else
					rest = rest:gsub("%s+%-[pP]%S+", "", 1)
					--tFuncs:Debug("profilelevel = "..profilelevel, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				end

				local maxlevel = rest:match("%s+%-[mM](%S+)")
				local maxlevel = (maxlevel and tonumber(maxlevel)) or (not maxlevel and (data.iMaxProfile or 10)) or false
				if maxlevel and (maxlevel < -1 or maxlevel > 10) then
					return false, "Invalid max profilelevel (-m) specified, only numbers between -1 and 10 are valid"
				elseif not maxlevel and rest:match(" %-[mM]") then
					return false, "Error: -m specified, but no valid max profilelevel given"
				else
					rest = rest:gsub("%s+%-[mM]%S+", "", 1)
					--tFuncs:Debug("maxlevel = "..maxlevel, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				end

				local uselevel = rest:match("%s+%-[uU](%S+)")
				local uselevel = (uselevel and tonumber(uselevel)) or (not uselevel and (data.iUseProfile or profilelevel)) or false
				if uselevel and (uselevel < -1 or uselevel > 10) then
					return false, "Invalid use profilelevel (-u) specified, only numbers between -1 and 10 are valid"
				elseif not uselevel and rest:match(" %-[uU]") then
					return false, "Error: -u specified, but no valid use profilelevel given"
				else
					rest = rest:gsub("%s+%-[uU]%S+", "", 1)
					--tFuncs:Debug("uselevel = "..uselevel, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				end

				local option = rest:match("%s+%-[oO](%d+)")
				local option = (option and tonumber(option)) or (not option and (data.iOptions or 0)) or false
				if option and (option < 0 or option > 512) then
					return false, "Invalid option (-o) specified, only numbers between 0 and 512 are valid"
				elseif not option and rest:match(" %-[oO]") then
					return false, "Error: -o specified, but no valid option given"
				else
					rest = rest:gsub("%s+%-[oO]%d+", "", 1)
					--tFuncs:Debug("option = "..option, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				end

				local seconds = rest:match("%s+%-[tT](%S+)")
				local seconds = (seconds and tFuncs:ConvertToSeconds(seconds)) or (not seconds and (data.iTimer or 0)) or false
				if seconds and seconds > 0 and not tFuncs:CheckFlag(option, 2) then
					return false, "Error: You can only use the timer (-t) on messages to all (option 2 must be included)"
				elseif not seconds and rest:match(" %-[tT]") then
					return false, "Error: -t specified, but no valid time given"
				else
					rest = rest:gsub("%s+%-[tT]%S+", "", 1)
					--tFuncs:Debug("seconds = "..seconds, "X")
					--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				end

				-- test if we captured everything
				rest = rest:gsub("%s+", "")
				--tFuncs:Debug("rest : ("..#rest ..")"..rest, "X")
				if #rest ~= 0 then
					return false, "Error: Bad or missing command"
				elseif maxlevel < profilelevel then
					return false, "Error: profilelevel can't be larger than max profilelevel"
				elseif uselevel < profilelevel then
					return false, "Error: profilelevel can't be larger than use profilelevel"
				end

				--// Change values
				if content and content ~= data.sContent then
					data.sContent = content
					data.bHasFillVars = VarTrue
					data.bHasParm = ParmTrue
					changed = true
					--tFuncs:Debug("changed is true on content", "X")
				end

				if helptxt and helptxt ~= data.sHelpText then
					data.sHelpText = helptxt
					changed = true
					--tFuncs:Debug("changed is true on helptext", "X")
				end

				if profilelevel and data.iProfile ~= profilelevel then
					data.iProfile = profilelevel
					changed = true
					--tFuncs:Debug("changed is true on profilelevel", "X")
				end

				if maxlevel and data.iMaxProfile ~= maxlevel then
					data.iMaxProfile = maxlevel
					changed = true
					tFuncs:Debug("changed is true on maxlevel", "X")
				end

				if uselevel and data.iUseProfile ~= uselevel then
					data.iUseProfile = uselevel
					changed = true
					--tFuncs:Debug("changed is true on uselevel", "X")
				end

				if option and data.iOptions ~= option then
					data.iOptions = option
					changed = true
					--tFuncs:Debug("changed is true on option", "X")
				end

				if seconds and data.iTimer ~= seconds then
					data.iTimer = seconds
					changed = true
					--tFuncs:Debug("changed is true on seconds", "X")
				end

				if changed then
					tFuncs:SaveSettings("commands")

					--// recreate UC when modifying the tCommands table
					tUserCmds:Build()

					local res, reply = SetCmd["showcmd"]["cmd"](ssid, "!showcmd "..name, "!showcmd")
					if res and reply then
						return true, "Textcommand modified: "..reply
					else
						return false, reply
					end
				else
					return false, "Textcommand didn't change"
				end
			elseif name then
				return false, "Textcommand not found."
			else
				return false, Lang(ssid, 0, "Error: no valid name specified. ")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name> [-o<options>] [-p<profilelevel>] [-m<max profilelevel>] [-u<use profilelevel>] [-t<time>] [-h\"<helptext>\"] \"<text>\"\tModify a textcommand", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local _,res = SetCmd.addcmd.xhelp(ssid, line, cmd)
			res = res.."\n\tYou only have to specify the parameters you want to change. For example if you only want to change the options, use: !modcmd <name> -o<optionnumber>\n"
			return true, res
		end,
		["ucmd"] = {
			{1,3,"Modify a textcommand","!%s %[line:Textcommand name] %[line:-o<options>] %[line:-p<profilelevel>] %[line:-m<max profilelevel>] %[line:-u<use profilelevel>] %[line:-t<time>] %[line:-h\"<helptext>\"] %[line:\"<text in quotes>\"]"},
		},
	},

	["renamecmd"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local rName, newname = line:match("^%S+%s+(%S+)%s+(%S+)")

			if not rName or not newname then
				return false, Lang(ssid, 0, "Use: %s <current cmd name> <new cmd name>", cmd)
			end

			local name = ""
			if tCommands[rName] then
				name = rName
			elseif tCommands[rName:lower()] then
				name = rName:lower()
			else
				return false, "Textcommand '"..rName.."' not found."
			end
			tFuncs:Debug("name = "..name, "X")

			newname = newname:lower()

			if tCommands[newname] then
				return false, "The new name already exists as a textcommand."
			elseif HubCmd[newname] or SetCmd[newname] or (tCmdAlias and tCmdAlias[newname]) then
				return false, "A command with name "..newname .." already exists, use another name"
			elseif #newname > 35 then
				return false, "The command name is too long, use a shorter name please (max 35 characters allowed)"
			elseif newname =="-h" or newname == "-showvars" then
				return false, "The command name matches an option name, you can't use that"
			end

			tCommands[newname] = tCommands[name]
			tCommands[name] = nil
			tFuncs:SaveSettings("commands")

			--// recreate UC when modifying the tCommands table
			tUserCmds:Build()

			local res, reply = SetCmd["showcmd"]["cmd"](ssid, "!showcmd "..newname, "!showcmd")
			if res and reply then
				return true, "Textcommand modified: "..reply
			else
				return false, reply
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <current cmd name> <new cmd name>\tRename a textcommand (remember to reconnect to get the updated rightclicks)", cmd)
		end,
		["ucmd"] = {
			{1,3,"Rename a textcommand","!%s %[line:Current Textcommand name] %[line:New Textcommand name]"},
		},
	},

	["delcmd"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local rName = string.match(line, "^%S+%s+(%S+)")

			if tCommands[rName] then
				tCommands[rName] = nil
			elseif tCommands[rName:lower()] then
				tCommands[rName:lower()] = nil
			else
				return false, "Textcommand does not exist, use !showcmd to show available commands."
			end

			--// recreate UC when modifying the tCommands table
			tUserCmds:Build()
			tFuncs:SaveSettings("commands")
			return true, "Textcommand removed."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name>\t\t\t\t\tDelete a textcommand", cmd)
		end,
		["ucmd"] = {
			{1,3,"Delete a textcommand","!%s %[line:Textcommand name]"},
		},
	},

	["showcmd"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local name = string.match(line, "^%S+%s+(%S+)")
			local options = {
				"Display text in PM instead of mainchat",
				"Show text to all",
				"Show text when user logged in",
				"Doesn't allow the command to be activated in mainchat",
				"Doesn't create an item in !help for this command",
				"Doesn't create a rightclick item for this command",
				"Text will be send in raw format",
				"Will only be sent to ADC user(s)",
				"Will only be sent to NMDC user(s)",
			}

			if name and tCommands[name] then
				local timer = "Disabled"
				if tCommands[name].iTimer and tCommands[name].iTimer > 0 then
					timer = tCommands[name].iTimer .." seconds"
					--timer = os.date("%Hh %Mm %Ss", tCommands[name].iTimer)
				end

				local reply = "\n\n\tSettings for textcommand: "..name.."\n\t"..string.rep("_",80).."\n"..
					"\tProfilelevel:\tFrom "..tCommands[name].iProfile .." to "..(tCommands[name].iMaxProfile or 10) .." (that will see the output of the command)\n"..
					"\tUse level:\t\t"..(tCommands[name].iUseProfile or tCommands[name].iProfile) .." and higher (that will be able to use the command)\n"..
					"\tHelp/RC:\t\t"..(tCommands[name].sHelpText and "\""..tCommands[name].sHelpText.."\"" or "No help text stored").."\n"..
					"\tTimer: \t\t"..timer.."\n"..
					"\tOptions:\t\t"..tCommands[name].iOptions.."\n"
				if tCommands[name].iOptions > 0 then
					for num, txt in ipairs(options) do
						if tFuncs:CheckFlag(tCommands[name].iOptions, (2^(num-1))) then
							reply = reply.."\t\t\t* "..options[num].."\n"
						end
					end
				else
					reply = reply.."\t\t\t* Output to mainchat (default)\n"
				end
				reply = reply.."\n\tText: \""..tCommands[name].sContent.."\"\n"
				return true, reply
			elseif name and name == "-showvars" then
				local tbl, res = {}, "Showing all avaliable variables usable in text commands\n\tVariable\t\t\tReturn string\n\t"..string.rep("-", 100).."\n"
				for name in pairs(tFuncs.FillVarFunctions) do
					table.insert(tbl, name)
					--tFuncs:Debug(name)
				end
				table.sort(tbl)

				for indx, name in ipairs(tbl) do
					res = res .. "\t"..Padding(name, 21).."\t\t"..(tFuncs.FillVarFunctions[name](true, ssid, tClient[ssid]) or "").."\n"
				end

				return true, res
			elseif name then
				return false, "Textcommand not found."
			else
				local reply = "\n\n\tAvailable textcommands:\n\n\t"..Lang(ssid, 0, "Command").."\t\t"..Lang(ssid, 0, "Shown to Levels").."\t"..Lang(ssid, 0, "Options").."\t\t"..Lang(ssid, 0, "Usage level").."\tTimer\r\n\t"..string.rep("_",80).."\n"
				local result = {}
				for name, data in pairs(tCommands) do
					table.insert(result, "\t"..Padding(name, 20).."\t"..(data.iProfile or "-1") .." to "..(data.iMaxProfile or "10") .."\t\t"..(data.iOptions or "0") .."\t\t"..(data.iUseProfile or data.iProfile or "-1").."\t\t"..((tFuncs:CheckFlag((tCommands[name].iOptions or 0), 2) and data.iTimer and data.iTimer > 0 and data.iTimer.." seconds") or "Disabled").."\n")
				end
				table.sort(result)
				reply = reply..table.concat(result)
				return true, reply
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tList all textcommands", cmd).."\n\t"..
				Lang(ssid, 0, "%s -showvars\t\t\t\tShow all avaliable variables for textcommands", cmd).."\n\t"..
				Lang(ssid, 0, "%s <name>\t\t\t\t\tShow a specific textcommand", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show all textcommands","!%s"},
			{1,3,"Show all avaliable variables for textcommands","!%s -showvars"},
			{1,3,"Show a specific textcommand","!%s %[line:Textcommand name]"},
		},
	},

	["addpattern"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local stype, rest = line:match("^%S+%s+(%S+)%s+(.+)$")
			if not stype or not rest then
				return false, Lang(ssid, 0, "Use: %s <type> \"<pattern>\" [-p<profilelevel>] [-a<actions>] [-r<replace> \"<replacetxt>\"]", cmd)
			else
				stype = stype:upper()
			end
			if stype and not SubSet.CheckList[stype] then return false, "Error: Wrong type used, use !addpattern -h for help" end
			if not tPatterns[stype] then tPatterns[stype] = {} end

			--// Replaceaction and text
			local rplmsg = ""
			local replaction, repltxt = rest:match("%s+%-[Rr](%d)%s*\"(.*)\"")
			if replaction and repltxt then
				replaction = tonumber(replaction)
				if replaction == 1 then
					rplmsg = ", replacing match with: "..repltxt
				elseif replaction == 2 then
					rplmsg = ", replacing line with: "..repltxt
				else
					return false, "Error: -r specified with an invalid number, use 1 (replace match) or 2 (replace line)"
				end
				rest = rest:gsub(" %-[Rr]%d%s*\".*\"", "")
			elseif rest:match(" %-[Rr]") then
				return false, "Error: -r specified, but no \"<text>\" given, use quotationmarks."
			end

			--// Pattern
			local pat = rest:match("\"(.-)\"")
			if pat then
				--tFuncs:Debug("[addpattern] found pattern : "..pat, "X" )
				if tPatterns[stype][pat] then return false, "Pattern already exists, use !modpattern or !delpattern." end

				if string.find(pat, "%f", 1, true) and not ( string.find(pat, "%f[", 1, true) and string.find(pat, "]", 1, true)) then
					return false, "Invalid pattern detected, please read this http://lua-users.org/wiki/FrontierPattern for proper use of %f in patterns"
				end

				rest = rest:gsub("\".-\"", "", 1)
			else
				return false, "Error: No pattern found, use !addpattern -h for help"
			end

			--// Action
			local action = tonumber(rest:match("%s+%-[Aa](%d+)"))
			if action and (action < 0 or action > 1023) or not action then
				return false, "Error: Wrong actionnumber: "..tostring(action)..", use !addpattern -h for help"
			elseif action and action == 0 and rplmsg == "" then
				return false, "Error: Action 0 (Allow) can only be used together with replace , use !addpattern -h for help"
			else
				rest = rest:gsub("%s+%-[Aa]%d+", "", 1)
				--tFuncs:Debug("action : "..rest, "X")
			end

			--// Profilelevel
			local level = tonumber(rest:match("%s+%-[Pp](%-?%d+)")) or -1
			if level and (level < -1 or level > 10) then
				return false, "Wrong profilelevel: "..(level or "nil")..", use !addpattern -h for help"
			elseif tClient[ssid].iProfile <= level then
					return false, "You can't add patterns affecting the same or a higher level than your own profile."
			else
				rest = rest:gsub("%s+%-[Pp]%-?%d+", "", 1)
				--tFuncs:Debug("profile : "..rest, "X")
			end

			-- test if we captured everything
			rest = rest:gsub("%s+", "")
			if #rest ~= 0 then
				tFuncs:Debug("rest : "..rest, "X")
				return false, "Error: Bad or missing command, use !addpattern -h for help"
			end

			--// Execution
			local newpattern = {["iProfile"] = level, ["iActions"] = action}
			if replaction and repltxt then
				newpattern["iReplace"] = replaction
				newpattern["sReplace"] = repltxt
			end
			tPatterns[stype][pat] = newpattern
			tCheck:PatternPriority(stype)

			tFuncs:SaveSettings("patterns")
			return true, "Successfully added a new "..stype.." pattern: \""..pat.."\" for profilelevel: "..level.." and below, with action: "..action..rplmsg
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <type> \"<pattern>\" [-p<profilelevel>] [-a<actions>] [-r<replace> \"<replacetxt>\"] \tAdd textpattern, affects users of <profilelevel> and below", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local cmdhelp = "Help for "..cmd..":\tRead more here http://www.flexhub.org/forum/index.php?topic=50.0 \n\n\t "..cmd.." <type> \"<pattern>\" [-p<profilelevel>] [-a<actions>] [-r<replace> \"<text>\"]\n"..
				"\n\t\t<type>\t\tWhat to check:\n"..
				"\t\t\tMC = Mainchat\n"..
				"\t\t\tPM = Personal messages\n"..
				"\t\t\tRE = Reports\n"..
				"\t\t\tSE = Searches\n"..
				"\t\t\tSR = Search results\n\n"..
				"\t\t\tNI = Nickname\n"..
				"\t\t\tDE = Description\n"..
				"\t\t\tEM = Email\n"..
				"\t\t\tVE = Client version\n"..
				"\t\t\tSS = Share size\n"..
				"\n\t\t\"<pattern>\"\t\tA Lua pattern for matching the string\n"..
				"\n\t\t[-p<profilelevel>]\tThe maximum profile level to check, a number in the range from -1 (User) to 10 (Hubowner), defaults to -1 if not set\n"..
				"\n\t\t[-a<actions>]\t\tA number defining what action to take:\n"..
				"\t\t\t0\tAllow (can only be used in combination with replace)\n"..
				"\t\t\t1\tIgnore\n"..
				"\t\t\t2\tNotify user\n"..
				"\t\t\t4\tWarn user\n"..
				"\t\t\t8\tMute user\n"..
				"\t\t\t16\tDisconnect user\n"..
				"\t\t\t32\tKick user\n"..
				"\t\t\t64\tBan user\n"..
				"\t\t\t128\tRedirect user (uses the redirect address from: !set rd_flood)\n"..
				"\n\t\t\tOne of the values above (except 0) can be combined with one or more of the following:\n\n"..
				"\t\t\t256\tNotify Opchat\n"..
				"\t\t\t512\tLock Mainchat (use with caution)\n"..
				"\n\t\t[-r<number> \"<text>\"]\tWhat to replace <number> followed by the replacing \"<text>\":\n"..
				"\t\t\t1\tReplace match\n"..
				"\t\t\t2\tReplace entire message\n"..
				"\n"
			return true, cmdhelp
		end,
		["ucmd"] = {
			{1,3,"Show help for adding text patterns","!%s -h"},
			{1,3,"Add a pattern","!%s %[line:Pattern type] %[line:Pattern (in quotes)] %[line:-p<profilelevel>] %[line:-a<action>] %[line:-r1\"<replace word>\" or -r2\"<replace message>\"]"},
		},
	},

	["modpattern"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local stype, name, rest = line:match("^%S+%s+(%S+)%s+\"(.-)\"(.+)$")

			if not stype or not tPatterns[stype:upper()] then
				local reply = "Valid pattern-types:\n\n"
				for name, data in pairs(SubSet.CheckList) do
					reply = reply.."\t"..name.."\t("..data[2]..")\n"
				end
				return false, "Incorrect or missing data. ".. Lang(ssid, 0, "Use: %s <type> \"<pattern>\" [-p<profilelevel>] [-a<action>] [-r<replace> \"<replacetxt>\"]", cmd) .."\n\n"..reply
			else
				stype = stype:upper()
			end

			if name and tPatterns[stype][name] then
				if string.find(name, "%f", 1, true) and not ( string.find(name, "%f[", 1, true) and string.find(name, "]", 1, true)) then
					return false, "Invalid pattern detected, please read this http://lua-users.org/wiki/FrontierPattern for proper use of %f in patterns"
				end

				local rplmsg = false
				local pattern = tPatterns[stype][name]

				--// Replacement
				local replaction = rest:match(" %-[Rr](%d)")
				local repltxt = rest:match(" %-[Rr]%d*%s*\"(.*)\"")
				if replaction then replaction = tonumber(replaction) end

				if replaction and replaction == 0 and not pattern.iReplace then
					return false, "Error: -r0 specified, but stored pattern does not contain a replace action."
				elseif string.find(rest," %-[Rr]") and not repltxt and not replaction then
					return false, "Error, -r specified, but no replace action number or text were found."
				elseif replaction and repltxt and (replaction == 1 or replaction == 2) then
					rplmsg = true
				elseif repltxt and not replaction then
					return false, "Error: -r specified with missing or invalid number, use 1 (replace match), 2 (replace line) or 0 (remove replace action)."
				elseif rest:match(" %-[Rr]") then
					return false, "Error: -r specified, but no \"<text>\" given, use quotationmarks."
				end
				rest = rest:gsub("%-[Rr]%d%s*", "", 1)
				rest = rest:gsub("\".*\"", "", 1)

				--// Action
				local action = rest:match(" -[Aa](%d+)")
				if action then action = tonumber(action) end

				if action and (action < 0 or action > 1023) then
					return false, "Error: Wrong actionnumber: "..tostring(action)
				elseif action and action == 0 and not rplmsg and (not pattern.iReplace or replaction == 0) then
					return false, "Error: Action 0 (Allow) can only be used together with replace , use !addpattern -h for help"
				else
					rest = rest:gsub("%s+%-[Aa]%d+", "", 1)
				end

				--// Profilelevel
				local level = tonumber(rest:match(" -[Pp](%S+)"))
				if level and (level < -1 or level > 10) then
					return false, "Wrong profilelevel: "..(level or "nil")
				elseif level and tClient[ssid].iProfile <= level then
					return false, "You can't add patterns affecting the same or a higher level than your own profile."
				else
					rest = rest:gsub("%s+%-[Pp]%-?%d+", "", 1)
				end

				-- test if we captured everything
				rest = rest:gsub("%s+", "")
				if #rest ~= 0 then
					--tFuncs:Debug("clean : "..rest)
					return false, "Error: Bad or missing command."
				end

				--// Execution
				local changed = false
				if replaction and replaction == 0 and pattern.iReplace then
					pattern.iReplace = nil
					pattern.sReplace = nil
					changed = true
				elseif replaction and repltxt and (replaction ~= pattern.iReplace or repltxt ~= pattern.sReplace) then
					changed = true
					pattern.iReplace = replaction
					pattern.sReplace = repltxt
				end

				if level and level ~= pattern.iProfile then
					changed = true
					pattern.iProfile = level
				end

				if action and action ~= pattern.iActions then
					changed = true
					pattern.iActions = action
				end

				if changed then
					tCheck:PatternPriority(stype)
					tFuncs:SaveSettings("patterns")
					local res, reply = SetCmd["showpattern"]["cmd"](ssid, "!showpattern "..stype.." \""..name.."\"", "!showpattern")
					if res and reply then
						return true, "Pattern modified: "..reply
					else
						return false, reply
					end
				else
					return false, "Pattern didn't change."
				end
			elseif name then
				return false, "Textcommand not found."
			else
				return false, Lang(ssid, 0, "Use: %s <type> \"<pattern>\" [-p<profilelevel>] [-a<action>] [-r<replace> \"<replacetxt>\"]", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <type> \"<pattern>\" [-p<profilelevel>] [-a<action>] [-r<replace> \"<replacetxt>\"] \tModify a textpattern", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local _,res = SetCmd.addpattern.xhelp(ssid, line, cmd)
			res = res.."\t\t0\tRemove replace string\n\tYou only have to specify the parameters you want to change. For example if you only want to change the actions, use: !modpattern <type> \"<pattern>\" -a<actions>\n"
			return true, res
		end,
		["ucmd"] = {
			{1,3,"Show help for modifying text patterns","!%s -h"},
			{1,3,"Modify a textpattern","!%s %[line:Pattern type] %[line:Pattern (in quotes)] %[line:-p<profilelevel>] %[line:-a<action>] %[line:-r1\"<replace word>\" or -r2\"<replace message>\"]"},
		},
	},

	["showpattern"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local clist = SubSet.CheckList
			local alist = SubSet.ActionList
			local stype, pat = string.match(line, "^%S+%s+(%S+)(.*)")
			if stype then stype = stype:upper() end
			if pat then pat = pat:sub(3,-2) end
			--tFuncs:Debug("[ShowPattern] "..tostring(pat))

			if stype and tPatterns[stype] and tPatterns[stype][pat] then
				local reply = "\n\n\tShowing pattern for: "..stype.." "..(clist[stype][2]).." \n\n\tProfile\tActions\tPattern\t\t\t\tActions\n\t"..string.rep("_",80).."\n\n"
				local data = tPatterns[stype][pat]
				local actions = ""
				for num,action in ipairs(alist) do
					if tFuncs:CheckFlag(data.iActions, (2^(num-1))) then
						actions = actions..alist[num].." "
					end
				end
				reply = reply.."\t"..data.iProfile.."\t"..data.iActions.."\t"..Padding("\""..pat.."\"", 32).."\t"..actions
				if data.iReplace and data.iReplace == 1 then reply = reply.."\t\tReplace match with:\t"..(data.sReplace or "") end
				if data.iReplace and data.iReplace == 2 then reply = reply.."\t\tReplace line with:\t\t"..(data.sReplace or "") end
				reply = reply.."\n"
				return true, reply
			elseif stype and tPatterns[stype] and clist[stype] and tPatterns[stype] ~= {} then
				local reply = "\n\n\tShowing patterns for: "..stype.." "..(clist[stype][2]).." (priority ordered)\n\n\tProfile\tActions\tPattern\t\t\t\tActions\n\t"..string.rep("_",80).."\n\n"

				local patorder = tPatterns["Sorted"] and tPatterns["Sorted"][stype] or tCheck:PatternPriority(stype)
				for num, pat in ipairs(patorder) do
					local data = tPatterns[stype][pat]
					local actions = ""
					for num,action in ipairs(alist) do
						if tFuncs:CheckFlag(data.iActions, (2^(num-1))) then
							actions = actions..alist[num].." "
						end
					end
					reply = reply.."\t"..data.iProfile.."\t"..data.iActions.."\t"..Padding("\""..pat.."\"", 32).."\t"..actions
					if data.iReplace and data.iReplace == 1 then reply = reply.."\t\tReplace match with:\t"..(data.sReplace or "") end
					if data.iReplace and data.iReplace == 2 then reply = reply.."\t\tReplace line with:\t\t"..(data.sReplace or "") end
					reply = reply.."\n"
				end
				return true, reply
			elseif stype and clist[stype] then
				return false, "Found no patterns with type: "..stype.." ("..clist[stype][2]..")"
			elseif stype == "*" then
				local reply = ""
				for stype in pairs(clist) do
					local text, hit = "", false
					text = "\n\n\tShowing patterns for: "..stype.." "..(clist[stype][2]).." (priority ordered)\n\n\tProfile\tActions\tPattern\t\t\tActions\n\t"..string.rep("_",80).."\n\n"
					local patorder = tPatterns["Sorted"] and tPatterns["Sorted"][stype] or tCheck:PatternPriority(stype)
					for num, pat in ipairs(patorder) do
						local data = tPatterns[stype][pat]
						local actions = ""
						for num,action in ipairs(alist) do
							if tFuncs:CheckFlag(data.iActions, (2^(num-1))) then
								actions = actions..alist[num].." "
								hit = true
							end
						end
						text = text.."\t"..data.iProfile.."\t"..data.iActions.."\t"..Padding("\""..pat.."\"", 32).."\t"..actions
						if data.iReplace and data.iReplace == 1 then text = text.."\t\tReplace match with:\t"..(data.sReplace or "") end
						if data.iReplace and data.iReplace == 2 then text = text.."\t\tReplace line with:\t\t"..(data.sReplace or "") end
						text = text.."\n"
					end
					if hit then reply = reply .. text end
--~ 					reply = reply..text		--// Shows all categories even if empty
				end

				if reply == "" then reply = "There are no patterns set. Use !addpattern -h for information on how to add a pattern." end

				return true, reply
			else
				local reply = "\n\nValid pattern-types:\n\n"
				for name, data in pairs(clist) do
					reply = reply.."\t"..name.."\t("..data[2]..")\n"
				end
				return false, Lang(ssid, 0, "Use: %s <type> \"<pattern>\"", cmd).. reply
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <type>\t\t\t\t\t\t\tShow patterns of a certain type", cmd).."\n\t"..
				Lang(ssid, 0, "%s <type> \"<pattern>\"\t\t\t\t\t\tShow a specific pattern", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show textpatterns","!%s %[line:Type (MC, PM, SE, SR, NI, DE, EM, VE, SS, RE)]"},
			{1,3,"Show a specific textpattern","!%s %[line:Type (MC, PM, SE, SR, NI, DE, EM, VE, SS, RE)] %[line:Textpattern (between quotes)]"},
		},
	},

	["delpattern"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local stype, pat = string.match(line, "^%S+%s+(%S+)%s+\"(.+)\"")
			if not stype or not pat then
				return false, Lang(ssid, 0, "Error: use %s <type> \"<pattern>\", or use !showpattern <type> to show available patterns", cmd)
			end
			stype = stype:upper()
			if tPatterns[stype] then
				if tPatterns[stype][pat] then
					tPatterns[stype][pat] = nil
					tCheck:PatternPriority(stype)
					tFuncs:SaveSettings("patterns")
					return true, "Pattern removed."
				else
					return false, "Pattern not found, use !showpattern <type> to show available patterns."
				end
			else
				return false, "Patterntype not found, use !showpattern to show available types."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <type> \"<pattern>\"\t\t\t\t\t\t\tDelete a textpattern", cmd)
		end,
		["ucmd"] = {
			{1,3,"Delete a textpattern","!%s %[line:Type (MC, PM, SE, SR, NI, DE, EM, VE, SS, RE)] %[line:Textpattern (between quotes)]"},
		},
	},

	["testpattern"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local stype, text = string.match(line, "^%S+%s+(%S+)%s+\"(.+)\"")
			if not stype or not text then
				return false, Lang(ssid, 0, "Error: use %s <type> \"<pattern>\", or use !showpattern <type> to show available patterns", cmd)
			end

			stype = stype:upper()
			if not tPatterns[stype] then
				return false, Lang(ssid, 0, "Error: Patterntype not found, use !showpattern to show available types.")
			elseif not tPatterns.Sorted or not tPatterns.Sorted[stype] or not tPatterns.Sorted[stype][1] then
				return false, Lang(ssid, 0, "The pattern type has no stored patterns, use !showpattern to show available types or !showpattern <type> to show avaliable patterns for a type.")
			end

			local result = {}
			local match = false
			for indx, pattern in ipairs(tPatterns.Sorted[stype]) do
				if string.find(text, pattern) then
					match = true
					table.insert(result, pattern)
				end
			end

			if not match then
				return false, Lang(ssid, 0, "The text does not match any pattern of type "..stype)
			end

			return true, "The text '"..text.."' matches the following patterns of type "..stype.."\n\n\t"..table.concat(result, "\n\t")
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <type> \"<text>\"\t\t\t\t\	t\t\tTest if a text matches a stored pattern", cmd)
		end,
		["ucmd"] = {
			{1,3,"Test a textpattern","!%s %[line:Type (MC, PM, SE, SR, NI, DE, EM, VE, SS, RE)] %[line:Textpattern (between quotes)]"},
		},
	},

	["setbotname"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,nick = line:find("^%S+%s+(%S+)")
			if nick then
				local res, err = tCheck:NickChars(nick)
				if not res then return false, (err or Lang(ssid, 0, "Error: invalid botname")) end

				local bsid = "AAAA" -- hub sid
				tHubSettings.Main.sBotName = nick
				tFuncs:SaveSettings("settings")

				tUser:Quit(bsid, true)

				tClient[bsid].sNick = nick
				tClient[bsid].tInfo.tOrig.NI = nick
				tClient[bsid].tInfo.tFiltered.NI = nick
				tInfo:CreateNMDC(tClient[bsid], bsid)
				tInfo:CreateADC(tClient[bsid], bsid)

				tUser:Announce(bsid)

				return true, Lang(ssid, 0, "Botname is changed to: %s",nick)
			else
				return false, Lang(ssid, 0, "Error: Use !setbotname <botname>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <botname>\t\t\t\tChange securitybot name", cmd)
		end,
		["ucmd"] = {
			{1,3,"Change bot name","!%s %[line:Bot name]"},
		},
	},

	["setbotdesc"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,desc = line:find("^%S+%s+(.-)$")
			local desc = desc or ""
			local bsid = "AAAA"

			tHubSettings.Main.sBotDescription = desc
			tFuncs:SaveSettings("settings")

			tClient[bsid].sDescription = desc
			tClient[bsid].tInfo.tOrig.DE = desc
			tClient[bsid].tInfo.tFiltered.DE = desc
			tClient[bsid].tInfo.tNew = {["DE"] = desc}
			tInfo:CreateNMDC(tClient[bsid], bsid)
			tInfo:CreateADC(tClient[bsid], bsid)
			tClient[bsid].tInfo.tNew = {}

			tSend:MyInfUpdate(bsid)

			return true, Lang(ssid, 0, "Bot description is changed to: %s",desc)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <description>\t\t\t\tChange security bot description", cmd)
		end,
		["ucmd"] = {
			{1,3,"Change bot description","!%s %[line:Bot description]"},
		},
	},

	["setroomname"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,old,newname = line:find("^%S+%s+(%S+)%s+(%S+)")
			if old and newname and newname ~= "" then
				local rsid, oldname = tNick:GetSid(old)
				local xsid = tNick:GetSid(newname)
				local res, _, reg, nick = tCheck:Nick(nil, nil, newname, 10)
				if not res then return false, nick end

				if not rsid then
					return false, Lang(ssid, 0, "Error: chatroom %s not found.", old)
				elseif xsid then
					return false, Lang(ssid, 0, "This name is already in use.")
				elseif reg then
					return false, (nick or Lang(ssid, 0, "Error: This is a registered nickname."))
				elseif not res then
					return false, (nick or Lang(ssid, 0, "Error: New chatroom name is invalid, use a proper name."))
				else
					if rsid == "BBBB" then
						tHubSettings.Main.sOpChat = newname
						tFuncs:SaveSettings("settings")
					elseif rsid == "CCCC" then
						tHubSettings.Main.sFeed = newname
						tFuncs:SaveSettings("settings")
					elseif tClient[rsid].sProtocol == "ChatRoom" and tPrivateRooms[oldname] then
						local chatroom = tPrivateRooms[oldname]
						tPrivateRooms[oldname] = nil
						tPrivateRooms[newname] = chatroom
						tFuncs:SaveSettings("privaterooms")
					elseif tClient[rsid].sProtocol == "ChatRoom" and tChatRooms[oldname] then
						local chatroom = tChatRooms[oldname]
						tChatRooms[oldname] = nil
						tChatRooms[newname] = chatroom
						tFuncs:SaveSettings("chatrooms")
					else
						return false, Lang(ssid, 0, "The specified name is not a chatroom.")
					end

					tUser:Quit(rsid, true)

					tClient[rsid].sNick = newname
					tClient[rsid].tInfo.tOrig.NI = newname
					tClient[rsid].tInfo.tFiltered.NI = newname
					tInfo:CreateNMDC(tClient[rsid], rsid)
					tInfo:CreateADC(tClient[rsid], rsid)

					tUser:Announce(rsid)

					return true, Lang(ssid, 0, "Chatroom %s is changed to: %s", oldname, newname)
				end

			else
				return false, Lang(ssid, 0, "Error: Use !setroomname <oldname> <newname>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <oldname> <newname>\t\t\tChange Chatroom name", cmd)
		end,
		["ucmd"] = {
			{1,3,"Change chatroom name","!%s %[line:Current chatroom name] %[line:New chatroom name]"},
			{2,3,"Change chatroom name (selected)","!%s %[userNI] %[line:New chatroom name]"},
		},
	},

	["setroomdesc"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,name = line:find("^%S+%s+(%S+)")
			local _,_,desc = line:find("^%S+%s+%S+%s+(.-)$")
			local desc = desc or ""
			local rsid, name = tNick:GetSid(name)

			if not name or not rsid then
				return false, Lang(ssid, 0, "Chatroom not found, specify a valid name.")
			elseif tClient[rsid].sProtocol ~= "ChatRoom" then
				return false, Lang(ssid, 0, "The specified name is not a chatroom.")
			else
				if rsid == "BBBB" then
					tHubSettings.Main.sOpChatDescription = desc
					tFuncs:SaveSettings("settings")
				elseif rsid == "CCCC" then
					tHubSettings.Main.sFeedDescription = desc
					tFuncs:SaveSettings("settings")
				elseif tPrivateRooms[name] then
					tPrivateRooms[name].sDesc = desc
					tFuncs:SaveSettings("privaterooms")
				elseif tChatRooms[name] then
					tChatRooms[name].sDesc = desc
					tFuncs:SaveSettings("chatrooms")
				else
					return false, Lang(ssid, 0, "An unknown error occured, please report this to the FlexHub crew.")
				end

				tClient[rsid].sDescription = desc
				tClient[rsid].tInfo.tOrig.DE = desc
				tClient[rsid].tInfo.tFiltered.DE = desc
				tClient[rsid].tInfo.tNew = {["DE"] = desc}
				tInfo:CreateNMDC(tClient[rsid], rsid)
				tInfo:CreateADC(tClient[rsid], rsid)
				tClient[rsid].tInfo.tNew = {}

				tSend:MyInfUpdate(rsid)

				return true, Lang(ssid, 0, "Chatroom description set to: %s", desc)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name> <description>\t\t\tChange chatroom description", cmd)
		end,
		["ucmd"] = {
			{1,3,"Change chatroom description","!%s %[line:Chatroom name] %[line:New description]"},
			{2,3,"Change chatroom description (selected)","!%s %[userNI] %[line:New description]"},
		},
	},

	["sethubname"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,bhub = line:find("^%S+%s+(.-)%s*$")
			if bhub and #bhub>0 then
				bhub = bhub:gsub("\n", "")
				bhub = bhub:gsub("\t", "")
				tHubSettings.Main.sName = bhub
				tFuncs:SaveSettings("settings")
				tFuncs:UpdateConsole("ID_tc_hubname", bhub, 1)
				tFuncs:UpdateConsole("ID_wt_title", bhub, 1)

				tSend:Raw("P_NMDC", "$HubName "..NMDC_E(tHubSettings.Main.sName).." - "..NMDC_E(tHubSettings.Main.sTopic).."|")
				tSend:Raw("P_ADC", "IINF NI"..ADC_E(tHubSettings.Main.sName).."\n")

				return true, Lang(ssid, 0, "Hubname changed to: %s", bhub)
			else
				return false, Lang(ssid, 0, "Error: Use !sethubname <name>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name>\t\t\t\tChange hubname", cmd)
		end,
		["ucmd"] = {
			{1,3,"Change hubname","!%s %[line:Hub name]"},
		},
	},

	["sethubdesc"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local s,e,bhubdesc = line:find("^%S+%s+(.-)$")
			if bhubdesc then
				tHubSettings.Main.sDescription = bhubdesc
				tFuncs:SaveSettings("settings")
				return true, Lang(ssid, 0, "Hubdescription is changed to: %s",bhubdesc)
			else
				return false, Lang(ssid, 0, "Error: Use !sethubdesc <desc>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name>\t\t\t\tChange hubdescription", cmd)
		end,
		["ucmd"] = {
			{1,3,"Change hubdescription","!%s %[line:Hub description]"},
		},
	},

	["purgeaccounts"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,profile = line:find("^%S+%s+(%S+)")
			local _,_,days = line:find("^%S+%s+%S+%s+(%S+)")
			local _,_,confirm = line:find("^%S+%s+%S+%s+%S+%s+(%S+)")
			if not profile or not days then return false,  Lang(ssid, 0, "Error: Use %s <profilename or *> <days of inactivity>", cmd) end
			return AccountManager:Purge(ssid, profile, days, confirm)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename or *> <days of inactivity>\t\t\t\tDelete unused and expired accounts", cmd)
		end,
	},

	["addroom"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,name,level = line:find("^%S+%s+(%S+)%s+(%S+)")
			local _,_,desc = line:find("^%S+%s+%S+%s+%S+%s+(.+)$")

			if not name or not level then
				return false, Lang(ssid, 0, "Error: use %s <name> <profilelevel> [<description>]", cmd)
			end
			if tonumber(level) then
				level = tonumber(level)
				if level < -2 or level > 10 then
					return false, Lang(ssid, 0, "Invalid profilelevel, valid range: from -1 to 10.")
				end
			else
				return false, Lang(ssid, 0, "Invalid profilelevel, valid range: from -1 to 10.")
			end

			if tNick:GetSid(name) or tNick:GetReg(name) then
				return false, Lang(ssid, 0, "Error: name is already in use.")
			else
				local res, _, reg, cnick = tCheck:Nick(nil,nil, name, 10)
				if not res then return false, (cnick or Lang(ssid, 0, "Error: invalid chatroom name")) end

				local desc = desc or "ChatRoom"
				tChatRooms[name] = { ["sDesc"] = desc, ["bBot"] = true, ["iLevel"] = level, ["bHidden"] = true}
				tFuncs:SaveSettings("chatrooms")

				local rsid = tFuncs:CreateSid()
				tFuncs:CreateBot(rsid,"ChatRoom",name,desc,level,"ChatRoom", true)
				tUser:Announce(rsid)
				return true, Lang(ssid, 0, "ChatRoom %s for profilelevel %s and higher is added.", name, level)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name> <profilelevel> [<description>]\t\tAdd chatroom for <profilelevel> and higher", cmd)
		end,
		["ucmd"] = {
			{1,3,"Add chatroom","!%s %[line:Chatroom name] %[line:Minimum profile level that has access to this room] %[line:Description (optional)]"},
		},
	},

	["delroom"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local _,_,name = line:find("^%S+%s+(%S+)")
			if not name then
				return false, Lang(ssid, 0, "Error: Use %s <name>", cmd)
			end
			local rsid, rname = tNick:GetSid(name)
			if not rsid then
				return false, Lang(ssid, 0, "Error: Chatroom not found.")
			elseif rsid == "BBBB" or rsid == "CCCC" then
				return false, Lang(ssid, 0, "Error: This chatroom can't be removed.")
			elseif tClient[rsid].sProtocol ~= "ChatRoom" then
				return false, Lang(ssid, 0, "Error: This is not a chatroom.")
			elseif tChatRooms[rname] then
				local name = tClient[rsid].sNick
				tChatRooms[name] = nil
				tFuncs:SaveSettings("chatrooms")

				tUser:Quit(rsid, true)
				tClient[rsid] = nil
				return true, Lang(ssid, 0, "Removed chatroom: %s", rname)
			elseif tPrivateRooms[rname] and (tClient[ssid].sNick == tPrivateRooms[rname].sOwner or tClient[ssid].iLevel == 10) then
				local name = tClient[rsid].sNick
				tPrivateRooms[name] = nil
				tFuncs:SaveSettings("privaterooms")

				tUser:Quit(rsid, true)
				tClient[rsid] = nil
				return true, Lang(ssid, 0, "Removed private chatroom: %s", rname)
			elseif tPrivateRooms[rname] then
				return false, Lang(ssid, 0, "Error: You are not the owner of this room, you are not allowed to delete it.")
			else
				return false, "Dont do that (what did you do to end up here??)."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <name>\t\t\t\t\tDelete chatroom", cmd)
		end,
		["ucmd"] = {
			{1,3,"Delete chatroom","!%s %[line:Chatroom name]"},
--			{2,2,"Delete chatroom (selected)","!%s %[userNI]"},
		},
	},

	["showrooms"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid,line, cmd)
			local reply = "\n\n\t"..Lang(ssid, 0, "List of profile chatrooms").."\n\n"..Lang(ssid, 0, "\t"..Padding("Name", 20).."\t\tType\t\Level\t\t\tDescription").."\n"
			.."\t"..string.rep("_",80).."\n"
			.."\t"..Padding(tHubSettings.Main.sFeed, 20).."\t\tFeedRoom\tAccess level: "..tClient.CCCC.iProfile.." and higher\t"..tHubSettings.Main.sFeedDescription.."\n"
			.."\t"..Padding(tHubSettings.Main.sOpChat, 20).."\t\tOpChat\t\tAccess level: "..tClient.BBBB.iProfile.." and higher\t"..tHubSettings.Main.sOpChatDescription.."\n"

			local preply = "\n\n\t"..Lang(ssid, 0, "List of private chatrooms").."\n\n"..Lang(ssid, 0, "\tName\t\t\tOwner\t\tMembers").."\n\t"..string.rep("_",80).."\n"

			local rooms = {}
			local privrooms = {}
			for vsid, data in pairs(tClient) do
				if data.sProtocol == "ChatRoom" and vsid ~= "BBBB" and vsid ~= "CCCC" then
					if data.bPrivate then
						local room = tPrivateRooms[tClient[vsid].sNick]
						local nicks = ""
						for nick in pairs(room.tMembers) do
							nicks = nicks..nick.." "
						end
						for nick in pairs(data.tMembersTemp) do
							nicks = nicks..nick.." "
						end
						table.insert(privrooms, "\t"..Padding(data.sNick, 20).."\t\t"..data.sOwner.."\t\t"..nicks.."\n")
					else
						table.insert(rooms, "\t"..Padding(data.sNick, 20).."\t\tChatroom\t\tAccess level: "..data.iProfile.." and higher\t"..data.sDescription.."\n")
					end
				end
			end
			table.sort(rooms, SortLowercase)
			table.sort(privrooms, SortLowercase)

			return true, reply..table.concat(rooms, "")..preply..table.concat(privrooms, "").."\n\t"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow all chatrooms", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show all chatrooms","!%s"},
		},
	},

	["showlog"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local log, lines = false, false
			if line:find("\"") then
				log = line:match("^%S+%s+\"(.+)\"")
				lines = line:match("%s+(%d+)%s*$")
			else
				log = line:match("^%S+%s+(%S+)")
				lines = line:match("^%S+%s+%S+%s+(%d+)$")
			end

			local names, exts, dat = {}, {}, {}
			lines = tonumber(lines) or 25

			if lines > 500 then lines = 500 end

			for file in lfs.dir(_LOGDIR) do
				if file:sub(1, 1) ~= "." and file:sub(#file -2, #file) == "log" then
				--	tFuncs:Debug("[showlog] ".. file)
					local attr = lfs.attributes(_LOGDIR.."/"..file)
					if attr and attr.mode == "file" then
						local st, _, ext = file:find("%.([^%.]+)$")
						local filename = file:sub(1, st-1)
						if ext and filename then
							table.insert(names, filename)
							table.insert(exts, ext)
							table.insert(dat, ToDate(attr.modification).."\t"..tFuncs:ConvertShare(attr.size).."\t\t"..filename)
						end
					end
				end
			end

			if log and names[1] then
				for i, j in pairs (names) do
					if log:lower() == j:lower() or log:lower() == j:lower() .."."..exts[i]:lower() then
						local logfile = io.open(_LOGDIR..j.."."..exts[i], "r")
						if not logfile then
							return false, "[showlog] Couldn't open file '"..j.."."..exts[i].."' for reading."
						end

						local len = logfile:seek("end")
						if len > 1048576 then		-- If it's larger than 1Mb, set startpoint to 1Mb from end of the file.
							logfile:seek("set", len - 1048576)
						else
							logfile:seek("set")
						end

						local buffer = logfile:read("*a")
						logfile:close()
						local output = {}
						for k in string.gmatch(buffer, "(.-)\n") do
							table.insert(output, k)
						end

						if #output > lines then
							return true, string.gsub("[showlog] '"..j.."."..exts[i].."', last "..lines .." lines :\n\n"..table.concat(output , "\n", #output - (lines -1)),"because:", "because")
						else
							return true, string.gsub("[showlog] '"..j.."."..exts[i].."', last "..#output .." lines :\n\n"..table.concat(output , "\n"),"because:", "because")
						end
					end
				end
			elseif log then
				return false, "[showlog] no logfiles found on disc."
			end

			return true, "[showlog] usage : !showlog <log file> [<number of lines, default 25, max 100>]\tTo show filenames containing spaces, use quotes on the name\n\tAvaliable files are :\n\t\t\t"..table.concat(dat,"\n\t\t\t").."\n\t"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <log file> [<number of lines>]\t\t\tShow the last lines of a log file, type showlog without parameters to see a list of avaliable logfiles (default 25, max 100 lines).", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show the last nn lines of a log file", "!%s %[line:Log file to display] %[line:Number of lines to display (default 25, max 100)]"},
		},
	},

	["shutdown"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd") and not tGlobals.bNoShutdown
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,confirm = line:find("^%S+%s+(%w+)")
			local message = line:match("^%S+%s+%a+%s+(.-)$") or false

			if confirm and tFuncs:ConfirmCode(confirm) then
				message = message or "Hub shutdown in progress.\n"
				tSend:ChatToAll(message, "AAAA", ssid)
				tFuncs:Debug("[shutdown] Hub shutdown initiated by "..tUser:GetNick(ssid), nil , -2)
				_SHUTDOWN = "s"
				return true, "Shutting down..."
			else
				return false, Lang(ssid, 0, "Wrong or missing confirmation code. Use %s %s [optional message to users]", cmd, tHubSettings.Main.bConfirmCode)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <confirmation code> [<message to users>]\t\t\tShut down FlexHub after posting the message in main", cmd)
		end,
		["ucmd"] = {
			{1,3,"Shut down FlexHub","!%s %[line:Confirmation Code] %[line:Optional shutdown message]"},
		},
	},

	["restart"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd") and not tGlobals.bNoRestart
		end,
		["cmd"] = function(ssid, line, cmd)
			if not tHubSettings.Console.AutoStartHub and not _NOGUI then
				return false, Lang(ssid, 0, "Error: To be able to restart the hub the autostart setting must be enabled, use '!set autostart true' to enable.", cmd)
			end
			local _,_,confirm = line:find("^%S+%s+(%w+)")
			local message = line:match("^%S+%s+%a+%s+(.-)$") or false

			if confirm and tFuncs:ConfirmCode(confirm) then
				message = message or "Hub restart in progress.\n"
				tSend:ChatToAll(message, "AAAA", ssid)
				tFuncs:Debug("[restart] Hub restart initiated by "..tUser:GetNick(ssid), nil, -2)
				_SHUTDOWN = "r"
				return true, "Restarting..."
			else
				return false, Lang(ssid, 0, "Wrong or missing confirmation code, use %s %s [optional message to users]", cmd, tHubSettings.Main.bConfirmCode)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <confirmation code> [<message to users>]\t\t\t\tRestart FlexHub after posting the message in main", cmd)
		end,
		["ucmd"] = {
			{1,3,"Restart FlexHub","!%s %[line:Confirmation Code] %[line: Optional restart message]"},
		},
	},

	["showemptypass"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			return AccountManager:FindNoPassAccounts(ssid)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s \t\t\t\tShow all regs with empty passwords", cmd)
		end,
		["ucmd"] = {
			{2,2,"Show all regs with empty passwords","!%s "},
		},
	},

	["checkcid"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local t = tClient
			local a = {}
			local o = {}

			ThreadLockMutex(mutexclient)
			for sid, data in pairs(t) do
				if a[data.sCID] then
					table.insert(o, sid.." "..data.sNick.." has same CID as "..a[data.sCID].." "..t[a[data.sCID]].sNick.." CID = "..data.sCID)
				else
					a[data.sCID] = sid
				end
			end
			ThreadUnlockMutex(mutexclient)

			if o[1] then
				return true, "The following duplicates were found in the client list:\n\n"..table.concat(o,"\n")
			else
				return true, "No duplicate CID's found in client list."
			end
		end,
	--	["help"] = function(ssid, line, cmd)
	--		return Lang(ssid, 0, "%s \t\t\t\tShow all regs with empty passwords", cmd)
	--	end,
		["ucmd"] = {
	--		{2,2,"Show all regs with empty passwords","!%s "},
		},
	},

	["sendalive"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)

	--[[Can be hub or client initiated:
	Hub sends: ILIV TOtoken
	Client sends: HLIV <sid> TOtoken

	Client sends HLIV <sid> TOtoken
	Hub sends: ILIV TOtoken
	--]]

		local _, _, nick = line:find("%S+%s+(%S+)")

		if not nick then
			return false, "No nick supplied."
		end

		local sid, nnick = tNick:GetSid(nick)

		if not sid then
			return false, "No user with that nick found."
		elseif tClient[sid].sALIV then
			local token = tClient[sid].sALIV
			tClient[sid].sALIV = nil
			return true, "The token ".. token.." has been sent, but no reply has been received."
		else
			tClient[sid].sALIV = os.time()
			tSend:Raw(sid, "ILIV TO"..tClient[sid].sALIV.."\n", nil, nil, "Delay")
			return true, "Sent ILIV with token ".. tClient[sid].sALIV.." to user "..(nnick or nick)
		end
	end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\t\tsend a ILIV command to a user", cmd)
		end,
		["ucmd"] = {
		--	{1,3,"a silly command, never does what you think","!%s"},
		},
	},

	["addports"] = {
		["allowed"] = function(ssid, line, cmd)
			return not tGlobals.bNoChangePorts and tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local ports = line:match("^%S+%s+(.-)%s*$")
			local force = (line:find("-force", 1, true) and true or false)

			if not ports then
				return false, Lang(ssid, 0, "No ports found.")
			end

			ports = ports:gsub(" ", "")
			local newports = tFuncs:Split(ports:lower(), ",")
			local added = {}
			local portlist = tHubSettings.Main.tPortList
			local changed = false

			for indx, port in ipairs(newports) do
				local p = port:match("%d+")
				local p = (tostring(p) or false)

				if p then
					local t = port:match("[srg]")
					local sType = ((t and t == "s" and "Secure") or (force and t and t == "g" and "Gui") or "Regular")

					if not portlist[sType][p] then
						if portlist.Secure[p] then
							return false, Lang(ssid, 0, "Port %s is already in use as a Secure port, it can not be added.", p)
						elseif portlist.Gui[p] then
							return false, Lang(ssid, 0, "Port %s is already in use as a Gui port, it can not be added.", p)
						elseif portlist.Regular[p] then
							return false, Lang(ssid, 0, "Port %s is already in use as a Regular port, it can not be added.", p)
						elseif sType == "Secure" and _OS ~= "Linux" and not tHubSettings.Main.bForceTLS then
							return false, Lang(ssid, 0, "Secure connections are currently only supported on Linux systems.")
						else
							portlist[sType][p] = true
							changed = true
							table.insert(added, p)
						end
					else
						return false, Lang(ssid, 0, "Port %s is already in use as a %s port, it can not be added.", p, sType)
					end
				end
			end

			if changed then
				local addon = ""
				if tArgsCmd and (tArgsCmd.port or tArgsCmd.sport or tArgsCmd.gport) then
					addon = "\n\n\t"..Lang(ssid, 0, "Note! The hub was started with ports defined on the commandline. If you keep them they will overwrite the setting you just made. Remember to either remove them or change them to match").."\n"
				end

				tHubSettings.Main.tPortList = portlist
				tFuncs:UpdatePorts(portlist)
				local res, reply = HubCmd["showports"]["cmd"](ssid, "!showports", "!showports")
				tFuncs:SaveSettings("settings")

				if not _NOGUI then tGui:UpdatePorts(portlist) end

				return true, Lang(ssid, 0, "Ports added (%s), current ports are: %s\n", table.concat(added, ","), reply) ..addon
			else
				local res, reply = HubCmd["showports"]["cmd"](ssid, "!showports", "!showports")
				return true, Lang(ssid, 0, "No new ports added, current ports are: %s\n", reply)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <port>,<port>S, \t\t\t\tAdd listenports for the hub, add a S for secure port (or multiple ports comma-separated, no spaces).", cmd)
		end,
		["ucmd"] = {
			{1,3,"Add listenports for the hub","!%s %[line:port, or portS for secure port (or multiple ports comma-separated, no spaces)]"},
		},
	},

	["delports"] = {
		["allowed"] = function(ssid, line, cmd)
			return not tGlobals.bNoChangePorts and tUser:IsAllowed(ssid, "setcmd")
		end,
		["cmd"] = function(ssid, line, cmd)
			local ports = line:match("^%S+%s+(.-)%s*$")

			if not ports then
				return false, Lang(ssid, 0, "No ports found.")
			end

			ports = ports:gsub(" ", "")
			local delports = tFuncs:Split(ports:lower(), ",")

			local portlist = tHubSettings.Main.tPortList
			local changed = false
			local closed = {}
			local cPort = tonumber(string.match(SetManager:Get("hub_address"), ":(%d+)") or 411)
			local redicount = 0
			local addon = ""

			for indx, port in ipairs(delports) do
				local p = port:match("%d+")
				local p = (tostring(p) or false)

				if p then
					local t = port:match("[srg]")
					local sType = ((t and t == "s" and "Secure") or (t and t == "g" and "Gui") or "Regular")

					if tonumber(p) == cPort then
						return false, Lang(ssid, 0, "Port %s is listed as hub main port, closing it is not allowed.", p)
					elseif tClient[ssid].iPort == tonumber(p) then
						return false, Lang(ssid, 0, "Port %s is the port you are connected on, closing it is not allowed.", p)
					end

					if portlist[sType][p] then
						portlist[sType][tostring(p)] = nil
						changed = true
						local nPort = tonumber(p)
						table.insert(closed, nPort)

						--// redirect users on that port before closing it
						for vsid, data in pairs(tClient) do
							if data.iPort == nPort then
								local rAddress = (data.sProtocol == "ADC" and "adc://"..tHubSettings.Main.sAddress or "dchub://"..tHubSettings.Main.sAddress)
								local ret, res, addy = tUser:Redirect(vsid, "This port is being closed, you are redirected to another open port on the same hub", rAddress, "Port is being closed")
								if ret then
									redicount = redicount + 1
								end
							end
						end

						local check = { "port", "sport", "gport" }
						local handled = {}
						--// Check if port was part of commandline
						for _, name in ipairs(check) do
							if tArgsCmd and tArgsCmd[name] then
								for port in string.gmatch(tArgsCmd[name], "%d+") do
									if port == tostring(p) then
										addon = addon.."\n\nNote! Port "..port.." was started with commandline. Remember to modify the commandline if you don't want this port to be started on next hub start."
										handled[port] = true
									end
								end
							end
						end
						--// Check if port was part of tArgs in FlexHub.lua
						for _, name in ipairs(check) do
							if tArgs[name] then
								for port in string.gmatch(tArgs[name], "%d+") do
									if port == tostring(p) then
										if not handled[port] then
											addon = addon.."\n\nNote! Port "..port.." was specified in tArgs in ".._SOURCE.." when the hub was started. Remember to modify ".._SOURCE.." if you don't want this port to be started on next hub start."
										end
									end
								end
							end
						end
					end
				end
			end

			if changed then
--~ 				if _NOGUI and tArgs.port and (tArgs.port ~= "" or tArgs.sport ~= "" or tArgs.gport ~= "") then
--~ 					addon = "\n\n\t"..Lang(ssid, 0, "Note! The hub was started with ports defined on the commandline or stored in FlexHub.lua, If you keep them they will overwrite the setting you just made. Remember to either remove them or change them to match").."\n"
--~ 				end

				tHubSettings.Main.tPortList = portlist
				tFuncs:SaveSettings("settings")
				tFuncs:UpdatePorts(portlist)
				if not _NOGUI then tGui:UpdatePorts(portlist) end

				local res, reply = HubCmd["showports"]["cmd"](ssid, "!showports", "!showports")
				return true, Lang(ssid, 0, "Ports closed (%s), and users redirected (redirect count %s), current ports are: %s\n", table.concat(closed, ","), redicount, reply) ..addon
			else
				local res, reply = HubCmd["showports"]["cmd"](ssid, "!showports", "!showports")
				return true, Lang(ssid, 0, "No port was closed, current ports are: %s\n", reply)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <port>,<port>S, \t\t\t\tRemove listenports for the hub, add a S for secure port (or multiple ports comma-separated, no spaces).", cmd)
		end,
		["ucmd"] = {
			{1,3,"Remove listenports for the hub","!%s %[line:port, or portS for secure port (or multiple ports comma-separated, no spaces)]"},
		},
	},

--~ 	["addredirect"] = {
--~ 		--// !addhublist NMDC addy:port <nick> <pass>
--~ 		["allowed"] = function(ssid, line, cmd)
--~ 			return tUser:IsAllowed(ssid, "setcmd")
--~ 		end,
--~ 		["cmd"] = function(ssid, line, cmd)
--~ 			local address = string.match(line, "^%S+%s+(%S+)")
--~ 			if address then
--~ 				local addy = address:lower()
--~ 				local found = false
--~ 				for num, name in ipairs(tHubSettings.tRedirects.tList) do
--~ 					if name == addy then
--~ 						found = true
--~ 					end
--~ 				end
--~ 				if not found then
--~ 					table.insert(tHubSettings.tRedirects.tList, addy)
--~ 					table.sort(tHubSettings.tRedirects.tList)
--~ 					tFuncs:SaveSettings("settings")
--~ 					return true, Lang(ssid, 0, "Added %s to the redirect list.", address)
--~ 				else
--~ 					return false, Lang(ssid, 0, "Address %s is already in the redirect list, use !showredirects for an overview.", address)
--~ 				end
--~ 			else
--~ 				return false, Lang(ssid, 0, "Error: use %s <hubaddress:port>", cmd)
--~ 			end
--~ 		end,
--~ 		["help"] = function(ssid, line, cmd)
--~ 			return Lang(ssid, 0, "%s <hubaddress:port>\t\t\tAdd a hubaddress to the redirect list", cmd)
--~ 		end,
--~ 		["ucmd"] = {
--~ 			{1,3,"Add a hubaddress to the redirect list","!%s %[line:Add a hubaddress to the redirect list: address:port]"},
--~ 		},
--~ 	},

--~ 	["delredirect"] = {
--~ 		["allowed"] = function(ssid, line, cmd)
--~ 			return tUser:IsAllowed(ssid, "setcmd")
--~ 		end,
--~ 		["cmd"] = function(ssid, line, cmd)
--~ 			local address = string.match(line, "^%S+%s+(%S+)")
--~ 			if address then
--~ 				local addy = address:lower()
--~ 				local found = false
--~ 				for num, name in ipairs(tHubSettings.tRedirects.tList) do
--~ 					if name == addy then
--~ 						table.remove(tHubSettings.tRedirects.tList, num)
--~ 						found = true
--~ 						break
--~ 					end
--~ 				end
--~ 				if found then
--~ 					tFuncs:SaveSettings("settings")
--~ 					return true, Lang(ssid, 0, "Removed %s from the redirect list.", address)
--~ 				else
--~ 					return false, Lang(ssid, 0, "Address %s wasn't in the redirect list, use !showredirects for an overview.", address)
--~ 				end
--~ 			else
--~ 				return false, Lang(ssid, 0, "Error: use %s <hubaddress:port>", cmd)
--~ 			end
--~ 		end,
--~ 		["help"] = function(ssid, line, cmd)
--~ 			return Lang(ssid, 0, "%s <hubaddress:port>\t\t\t\tRemove a hubaddress from the redirect list", cmd)
--~ 		end,
--~ 		["ucmd"] = {
--~ 			{1,3,"Remove a hubaddress from the redirect list","!%s %[line:Remove a hubaddress from the redirect list: address:port]"},
--~ 		},
--~ 	},
}

