--// $Id: Hub-Start.lua 1465 2013-10-10 14:43:34Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

--~ package.path = ""
local p = "./"
package.path =  p.."lib/?.lua;"
--//package.cpath = p.."lib/?.so;./lib/?.so;./?.so;"..p.."lib/?.dll;"..p.."lib/?.dylib;"

do
--// Basics
_APPNAME = "FlexHub"
_APPVER = "Beta 0.2"
_REV = 1416		--// NOTE: Use the current svnnumber if this file changes. Only used for info on errors that happen before all files are loaded, after loading it will match tSettings.General.svnversion through tFuncs:GetSVN()

local os = package.config:sub(1,1)
	if os == '\\' then
		_OS = "Windows"
		package.cpath = p.."lib/?.dll;"..p.."lib/?.dylib;"
	elseif os == '/' then
		_OS = "Linux"
		package.cpath = p.."lib/?.so;./lib/?.so;./?.so;"
	else
		Test("Undetected operating system : "..os)
		_OS = os.getenv("%$HOME") and "Linux" or "Windows"
	end

	_SOURCE = debug.getinfo(3, "S").source:sub(2) or "FlexHub.lua"
end

tGlobals = {}	-- storage for global data that doesnt fit into tStats or tLogs
tCurrent = { ["Handler"] = "none" }
tCounters = {}
tNames = {}
tDebugArray = {}
tHubFiles = {}

	tHubFiles["Hub-Start.lua"] = {
		["sRev"] = "$Rev: 1465 $",
		["bRestart"] = true,
		["sTTH"] = false ,
		["sPath"] = "./core/Hub-Start.lua",
		}

-- making _REV update automagicly ;)
local _, _, tmp = string.find((tHubFiles["Hub-Start.lua"].sRev or ""), "(%d+)")
if tmp and tonumber(tmp) then
	_REV = tonumber(tmp)
end

-- Workaround for hubs running svn 1358:
tArgs = tArgs or tSetup or {}
tArgsFile = tArgs --// Storing original tArgs to be able to see what the content of FlexHub.lua actually is.

if arg then
	tArgsCmd = {} --// Storing parsed commandline arguments here

	-- // allowed startup parameters :
	local tParams = { ["nick"] = true, ["pass"] = true, ["profile"] = true, ["nickip"] = true, ["port"] = true, ["bind"] = true, ["sport"] = true, ["gport"] = true, ["userlimit"] = true,
		["nogui"] = true, ["noupdate"] = true, ["logdir"] = true, ["configdir"] = true, ["storagedir"] = true, ["norestart"] = true, ["noshutdown"] = true, ["nochangeports"] = true, ["notouchrestartscript"] = true, ["device"] = true,
		}

	for _, val in ipairs(arg) do
		if val:find("=", 1, true) then
			local name, value = val:match("%-%-(.-)=(.+)")
--			print("Commandline argument: "..tostring(name).." = "..tostring(value).."\n")
--			Test( "Commandline argument: "..tostring(name).." = "..tostring(value).."\n")
			if name and value and tParams[name:lower()] then
				tArgsCmd[name:lower()] = value
				tArgs[name:lower()] = value
				--print("Commandline argument: "..tostring(name).." = "..tostring(value))
				--Test( "Commandline argument: "..tostring(name).." = "..tostring(value))
			else
				print("Unknown commandline argument used: "..val)
				return
			end
		else
			name = val:match("%-%-(.+)")
			if name and tParams[name:lower()] then
				tArgs[name:lower()] = true
				tArgsCmd[name:lower()] = true
				--print("Commandline argument: "..tostring(name).." is set as TRUE ")
			else
				print("Unknown commandline argument used: "..val)
				return
			end
		end
	end
	tArgsStartup = arg
end


--// Store file paths if they were specified as commandline aguments
if tArgs.storagedir and tArgs.storagedir ~= "" and tArgs.storagedir ~= "./" then
	_STORAGEDIR = tArgs.storagedir or "./"
	_LOGDIR = tArgs.storagedir.."logs/"
	_CONFIGDIR =  tArgs.storagedir.."settings/"
	_IMPORTDIR =  tArgs.storagedir.."import/"
	_BACKUPDIR =  tArgs.storagedir.."backups/"
else
	_STORAGEDIR = tArgs.storagedir or "./"
	_LOGDIR = tArgs.logdir or "./logs/"
	_CONFIGDIR = tArgs.configdir or "./settings/"
	_IMPORTDIR = tArgs.importdir or "./import/"
	_BACKUPDIR = tArgs.backupdir or "./backups/"
end

_NOGUI = tArgs.nogui or false
_DEVICE = tArgs.device or "pc"

--// Apply startup settings
tGlobals.bNoUpdate = tArgs.noupdate or false
tGlobals.bNoRestart = tArgs.norestart or false
tGlobals.bNoShutdown = tArgs.noshutdown or false
tGlobals.bNoChangePorts = tArgs.nochangeports or false
tGlobals.bNoTouchRestartScript = tArgs.notouchrestartscript or false

--// For quick and dirty debugging
function Test(msg)
	local msg = (msg and tostring(msg)) or "empty msg"
	local file = io.open(_LOGDIR.."Debug.log","a+")
	if file then
		file:write(os.date("[%c] ")..tostring(msg).."\n")
		file:flush()
		file:close()
		print(os.date("[%c] ")..tostring(msg).."\n")
	else
		print(os.date("[%c] ")..tostring(msg).."\n")
	end
end


--// Write exceptionerrors to file
function WriteException(msg)
	local msg = msg or "unknown error"
	local file, err = io.open(_LOGDIR.."Error.log","a+")
	if file then
		file:write(os.date("[%c] ")..tostring(msg).."\n")
		file:flush()
		file:close()
		print(os.date("[%c] ")..tostring(msg).."\n")
	else
		print(os.date("[%c] ")..tostring(msg).."\n")
	end
end

function ShowTrace(num)
	local counter = 0
	local traceback = ""
	local getname = function(func)
		if tNames[func] then
			local n = tNames[func]
			if n.what == "C" then return n.name end
			local loc = string.format("[%s]:%s",n.short_src, n.linedefined)
			if n.namewhat ~= "" then return string.format("%s [%s] (%s)", loc, n.currentline,n.name)
			else return string.format("%s", loc) end
		else
			return "<No function found for "..tostring(func)..">"
		end
	end
	for i,arr in ipairs(tDebugArray) do
		if not num or counter < num then
			traceback = traceback..tostring(getname(arr)).."\n"
			counter = counter +1
		end
	end
	return traceback
end

function OnError(errorstr, continue)
	print(errorstr)
	local msg = string.rep("=",60).."\nEXCEPTION ERROR\n"..string.rep("-",60)..
		string.rep("-",60).."\nError message description: "..tostring(errorstr)..
		"\n"..string.rep("-",60)..
		"\nStack traceback: "..
		"\n"..string.rep("-",60)..
		"\n"..ShowTrace()..
		"\n"..string.rep("=",60).."\n"
	WriteException(msg)
end

function DoRequire(name)
	return require(name)
end

local required = {
	"socket", 		-- Luasocket
	"modules.socket",
	"modules.http",
	"modules.url",		--// For escaping urls
	"th",			-- Tiger Hash
	"tth",			-- Tiger Tree Hash	--// Used to hash tables for synchronizationchecks
	"iconv",		-- Converting to and from utf-8
	"sha1",		-- Pure Lua hashing with SHA1
--~ 	"geoip",		-- Get the country of an IP
	"zlib",			-- Compressing data
--~ 	"uname",		-- OS information - deprecated
	"ssl",			-- Secure sockets can block with this lib
	"profiler",		-- Writes a logfile for analysis of performance problems
	"dc",			-- For hashing/base32 and PID/CID checks
	"webnet77",	-- Replacement for geoip
	"lfs",			-- Extra file systems functions
}

if not _NOGUI then
	local res, err = DoRequire("wx")
	if not res then
		print("Package 'wx' couldn't be loaded, FlexHub will run without GUI")
		_NOGUI = true
		table.insert(required, "bit")	-- Bit library (only needed when wx isn't loaded)
		wx = {}
	end
else
	table.insert(required, "bit")		-- Bit library (only needed when wx isn't loaded)
	wx = {}
end

for _,name in ipairs(required) do
	local res, err = DoRequire(name)
	if not res then
		WriteException("Error while trying to load package: "..name..", error: "..tostring(res or "")..tostring(err or ""))
		print("Error while trying to load package: "..tostring(name)..", error: "..tostring(res or "")..tostring(err or ""))
		break
	end
end

_APPPATH = (lfs and lfs.currentdir().."/" or "")

--//Debugging temp
Test("Paths set to:\n\t_APPPATH: ".._APPPATH.."\n\t_STORAGEDIR: ".._STORAGEDIR.."\n\t_LOGDIR: ".._LOGDIR.."\n\t_CONFIGDIR : ".._CONFIGDIR.."\n\t_IMPORTDIR: ".._IMPORTDIR.."\n\t_BACKUPDIR: ".._BACKUPDIR)


--// Store PID
if _OS == "Linux" then
	local rfilename = "/proc/self/stat"
	local rfile, err = io.open(rfilename, "r")
	if rfile then
		local msg = rfile:read("*a")
		rfile:close()
		local wfile, err = io.open (_CONFIGDIR.."pidfile", "w")
		if wfile then
			wfile:write(msg)
			wfile:close()
		end
	end
end

--~ Test("dofile wx...")
--~ dofile(p.."lib/wx.so")
--~ Test("dofile wx ok.")

function OnError(errorstr, continue)
	local msg = string.rep("=",60).."\nEXCEPTION ERROR\n"..string.rep("-",60)..
		"\n".._APPNAME.." ".._APPVER..(_REV and " SVN: ".._REV or "")..
		"\nSystem information: "..
		"\n\tTime: "..os.date("%x %X")..
		"\n\tLUA version: ".._VERSION..
		"\n\tStartup path: ".._APPPATH

	--// Add info from wxlua and running hub
	if package.loaded["wx"] then
		local handler = tCurrent.Handler
		local data = tCurrent[handler] or {}
--~ 		local tUptime = os.date("*t",socket.gettime())

		msg = msg..
			"\n\tOS: "..wx.wxGetOsDescription()..
			"\n\tFree memory: "..(tFuncs and ConvertShare(wx.wxGetFreeMemory():ToDouble()) or "unknown")..
--~ 			"\n\tUptime: "..string.format("%i hour(s) %i minute(s) %i second(s)",tUptime.hour,tUptime.min, tUptime.sec)..
			"\n\twxWidgets version: "..wx.wxVERSION_STRING..
			"\n\twxLUA version: "..wxlua.wxLUA_VERSION_STRING..
			"\n"..string.rep("-",60)..
			"\nExecuting data:"..
			"\n\tHandler: "..(handler or "unknown").." \tSID: "..(data.Sid or "unknown").." \tNick: "..(data.Nick or "unknown").." \tProtocol: "..(data.Protocol or "")..
			"\n\tData: "..(data.Line or "").."\n"
	end

	msg = msg..string.rep("-",60).."\nError message description: "..tostring(errorstr)..
		"\n"..string.rep("-",60)..
		"\nStack traceback: "..
		"\n"..string.rep("-",60)..
		"\n"..ShowTrace()..
		"\n"..string.rep("=",60).."\n"

	Notify("An error occured in the application:\n"..tostring(msg), "ERROR")
	if not continue then OnStop() end
--~ 		wx.wxExit(1)
end

function OnStop()
	-- stopping the threads and closing of sockets needs to be handled properly here
--~ 	tFuncs:SaveSettings("*")	--// When this is called before tables are loaded it will empty all files...

	if tFuncs then
		tFuncs:CloseFiles()

		if tStats and tStats.Hub.Status ~= "Offline" then
			tFuncs:StopHub()
		end

		--// Save last changes to tAccounts
		tFuncs:SaveSettings("accounts")

		--// Store tLogs.Nick and tLogs.IP for reload after hub restart
		if tHubSettings and tHubSettings.Main.bUserdataPersistent then
			tFuncs:StoreUserinfo(true, true)
		end

	end

	if not _NOGUI then
		if tmr then	-- whats this for ???? Fredo
			tmr:Stop() -- always stop before exiting or deleting it
			tmr:delete()
			tmr = nil
		end
		if tCtrls and tCtrls["taskbar"] then
			tCtrls["taskbar"]:RemoveIcon()
		end
		wx.wxExit(0)
	end
	return
end

--os.setlocale(wx.wxLocale():GetCanonicalName())

hook = function()
	local f = debug.getinfo(2,"f").func
	local filter = {}
	--	local filter = { ["nil"] = true, ["GetXRCID"] = true, ["FindWindow"] = true, ["GetClassInfo"] = true, ["GetClassName"] = true, ["DynamicCast"] = true}
	--	local filter = { ["GetXRCID"] = true, ["FindWindow"] = true, ["GetClassInfo"] = true, ["GetClassName"] = true, ["DynamicCast"] = true, ["nil"] = true,
	--	["SetIcon"] = true, ["SetTitle"] = true, ["SetFieldsCount"] = true, ["SetStatusWidths"] = true, ["wxImage"] = true, ["wxBitmap"] = true,
	--	["wxIcon"] = true, ["ToggleTool"] = true, ["wxArrayString"] = true, ["CopyFromBitmap"] = true, ["SetLabel"] = true, ["GetValue"] = true, ["GetId"] = true,
	--	["Add"] = true, ["Connect"] = true,
	--	}
	if #tDebugArray > 120 then
		local fremoved = table.remove(tDebugArray)
		if tNames[fremoved] and not tDebugArray[fremoved] then tNames[fremoved] = nil end
	end
	if not tNames[f] then
		local data = debug.getinfo(2, "Snl")
		if not filter[tostring(data.name)] then
			table.insert(tDebugArray,1,f) -- Order the stack called by time - first call
			tNames[f] = data
		end
	end
end
--~ debug.sethook(hook,"c") -- Set the hook

function ConvertShare(value)
	if not value or value == 0 then return "0 B" end
	iBytes = tonumber(value)
	if not iBytes then return false, "Unconvertable size: "..tostring(value) end
	local i,tUnits = 1, {"B","kB","MB","GB","TB","PB","EB","YB"}
	while iBytes >= 1024 do iBytes = iBytes / 1024 i = i + 1 end
	return string.format("%.2f",iBytes).." "..(tUnits[i] or "??")
end

function OnBreak(errorstr)
	local traceback = ShowTrace()
	Notify("BreakPoint: "..errorstr.."\n"..traceback, "ERROR")
end

function Notify(msg, title, icon)
	if msg then
		msg = string.gsub(msg, "ERROR:", "Error:")
		msg = string.gsub(msg, "WARNING:", "Warning:")
		WriteException("["..(title or "Notify").."] "..msg)
	--~ 	if not _NOGUI then
	--~ 		wx.wxMessageBox(tostring(msg),(title or "Information"),(icon or wx.wxICON_INFORMATION))
	--~ 	else
			print(os.date("[%c] ")..tostring(msg).."\n")
	--~		end
	end
end

--// Insert write test here?

-- we do this even before we check if we can write, that makes the write test redundant ffs
do -- // parse HubSettings.ini for dot/comma errors if the wrong locale is used
	local res, err = os.setlocale(nil, "numeric")
	if not res or res ~= "C" then
		os.setlocale("C", "numeric")

		local name = "HubSettings.ini"
		local a, err = io.open(_CONFIGDIR..name, "r")
		if a then
			local changed = false
			local count = 0
			local new = {}
			for line in a:lines() do
				count = count + 1
				if line:match("%[\"i.-=.+,.+,$") then
					Notify("[Locale] ERROR on line ".. count .." \t"..tostring(line))
					line = line:gsub(",", ".", 1)
					Notify("[Locale] Line ".. count .." changed to\t"..tostring(line))
					changed = true
				end
				table.insert(new, line)
			end

			if changed then	-- save
				local hFile, err = io.open (_CONFIGDIR..name , "w")
				if hFile then
					hFile:write(table.concat(new, "\n"))
					hFile:flush()
					hFile:close()
				else
					Notify("Error while saving file, can't open "..name.." for writing. Error: "..err)
					return false, "Fatal error when trying to save settings."
				end
			end
		end
	end
end

--// Convert specific files to UTF8 for Linux
function FilesToUTF8()
	local luafiles = {"core/Hub-Check.lua", "core/Hub-Cmds.lua", "core/Hub-Funcs.lua", "core/Hub-Escape.lua"}

	local convertfile = function(name)
		local descriptor,err = iconv.new("utf-8", "windows-1252")
		if not descriptor then
			return false, "Error when creating iconv descriptor for conversion: "..tostring(err)
		end
		local file = io.open(_APPPATH..name, "r")
		local firstline = file:read("*line")
		local content = file:read("*all")
		file:close()
		if string.find(firstline, " %$$") then --// Find $ at end of line then
			firstline = string.gsub(firstline, " charset:%S+", "")
			firstline = string.gsub(firstline, " %$$", " charset:utf-8 $")
		else
			firstline = "--// charset:utf-8\n"..firstline
		end
		local converted, err = descriptor:iconv(content)
		if err then
			return false, "Error during conversion of file "..name.." to UTF-8: "..err
		else
			if _OS == "Windows" then name = name..".tst" end
			local file = io.open(name, "w")
			if file then
				file:write(firstline.."\n"..converted)
				file:close()
				return true
			else
				return false, "Can't open file "..name.." for writing UTF-8 conversion"
			end
		end
	end

	--// Lua files
	for _,name in ipairs(luafiles) do
		local file, err = io.open(_APPPATH..name, "r")
		if file then
			local line = file:read("*line")
			file:close()
			if not string.find(line, "charset:utf%-8") then
				local ok, err = convertfile(name)
				if not ok then
					Notify("Error converting file "..name.." to UTF-8: "..tostring(err))
					OnStop()
				else
					Notify("Succesfully converted file "..name.." to UTF-8.")
				end
			end
		else
			Notify("Error opening file "..name.." for conversion to UTF-8")
			OnStop()
		end
	end
end

--[[

if _OS == "Linux" then
	local res, err = FilesToUTF8()
	if not res then Notify("Error converting files to UTF8: "..tostring(err)) end
end

]]--

--// Load basic settings
dofile(_APPPATH..'core/Hub-StartVars.lua')	-- Contains basic settings

--// Load basic functions
dofile(_APPPATH..'core/Hub-Funcs.lua')		-- loads all basic functions / starts filedebugging
dofile(_APPPATH..'core/Hub-Threads.lua')	-- loads all basic functions / starts filedebugging
dofile(_APPPATH..'core/Hub-Lang.lua')		-- language functions

--// Check for existinglogdir and configdir, create if not existing
local res, err = tFuncs:CheckFolder(_LOGDIR)
if not res then Notify("[Fatal error] Logdir: "..err) return end
local res, err = tFuncs:CheckFolder(_CONFIGDIR)
if not res then Notify("[Fatal error] Configdir: "..err) return end
local res, err = tFuncs:CheckFolder(_IMPORTDIR)
if not res then Notify("[Fatal error] Importdir: "..err) return end
local res, err = tFuncs:CheckFolder(_BACKUPDIR)
if not res then Notify("[Fatal error] Backupdir: "..err) return end

--// Load all other scripts and the settingsfiles
local res, msg, err = tFuncs:LoadScripts()
if not res then
	Notify(err or msg, "Error loading scripts")
	return
end

--// Load userinfo and ipinfo if stored
if tHubSettings and tHubSettings.Main.bUserdataPersistent then
	local res, err = tFuncs:LoadUserinfo(true, true)
	if not res then Notify(err) end
end

--// initialise timers
tTimerFuncs:Initialize()

--// Set charset to initialize conversions
--tFuncs:SetCharset()		moved inside LoadScripts

--// Load GeoIP file
tFuncs:LoadGeoIP()

--// redefine print here so it will output to debugwindow
function RedefinePrint()
	oldprint = print
	function print(a,b,c,d,e,f,g)	--// this might need more arguments, print accepts any number of arguments
		local msg = tostring(a)
		if b then msg = msg.." "..tostring(b) end
		if c then msg = msg.." "..tostring(c) end
		if d then msg = msg.." "..tostring(d) end
		if e then msg = msg.." "..tostring(e) end
		if f then msg = msg.." "..tostring(f) end
		if g then msg = msg.." "..tostring(g) end
	if _NOGUI then
		oldprint(msg)
	else
		tFuncs:SendToGui("ID_tc_monitor_debug", "PRINT: "..msg.."\n")
	end
	end
end

--// scriptdebug for cage, will be called when external scripts call print
function scriptdebug(...)
	local printstr = "[scriptdebug] "
	for i,v in pairs (arg) do
		printstr = printstr..tostring(v).."\t"
	end
	tFuncs:Debug(printstr, "EI", 1)
end

--// writing restart scripts
if _OS == "Linux" then
	local make = true
	if tGlobals.bNoTouchRestartScript then
		-- check if restartscript exists, else create one
		local o = io.open(_CONFIGDIR.."RestartScript.sh", "r")
		if o then
			local msg = o:read("*a")
			if string.find(msg, _SOURCE ,1, true) then
				-- we're okay, dont need make new one
				make = false
			end
			o:close()
		end
	end

	if make then
		local linscript ="#!/bin/sh\n"
		.."ping -w 3 127.0.0.1 > NUL\n"
		.."cd ".._APPPATH.."\n"
		.."./lua ".._SOURCE..((tArgsStartup and " "..table.concat(tArgsStartup, " ")) or "")
		local o = io.open(_CONFIGDIR.."RestartScript.sh", "w")
		if o then
			o:write(linscript)
			o:close()
			os.execute("chmod 700 ".._CONFIGDIR.."RestartScript.sh")
		end
	end
elseif _OS == "Windows" and _NOGUI then	-- this script only makes sence if run as service on windows, it wont work else so need to figure out how to detect that someday
	local make = true
	if tGlobals.bNoTouchRestartScript then
		-- check if restartscript exists, else create one
		local o = io.open(_CONFIGDIR.."RestartScript.bat", "r")
		if o then
			local msg = o:read("*a")
			if string.find(msg, _SOURCE:match("^(.-)%.") ,1, true) then
				-- we're okay, dont need make new one
				make = false
			end
			o:close()
		end
	end

	if make then
		local winscript = "@ECHO OFF\n"
		.."REM ECHO Restart script for FlexHub on Windows\n"
		.."ping -w 3 127.0.0.1 > NUL\n"
		.."NET STOP " .. _SOURCE:match("^(.-)%.") .. "" .."\n"
		.."NET START " .. _SOURCE:match("^(.-)%.") .. "" .."\n"
		local o = io.open(_CONFIGDIR.."RestartScript.bat", "w")
		if o then
			o:write(winscript)
			o:close()
		end
	end
elseif _OS == "Windows" then
	local make = true
	if tGlobals.bNoTouchRestartScript then
		-- check if restartscript exists, else create one
		local o = io.open(_CONFIGDIR.."RestartScript.bat", "r")
		if o then
			local msg = o:read("*a")
			if string.find(msg, _SOURCE:match("^(.-)%.") ,1, true) then
				-- we're okay, dont need make new one
				make = false
			end
			o:close()
		end
	end

	if make then
		local winscript = "@ECHO OFF\n"
		.."REM ECHO Restart script for FlexHub on Windows\n"
		.."ping -w 3 127.0.0.1 > NUL\n"
		.."START \"Restarting FlexHub\" /B /D\"".._APPPATH:gsub("/","\\").."\" \"" .. _SOURCE:match("^(.-)%.") .. ".exe\"" .."\n"
		local o = io.open(_CONFIGDIR.."RestartScript.bat", "w")
		if o then
			o:write(winscript)
			o:close()
		end
	end
end

-- Hub-Gui.lua loads Hub-GuiFunc.lua and then starts the GUI
dofile(_APPPATH..'core/Hub-Gui.lua')

--// Daywalker and FlipFlop
