--// $Id: Hub-Threads.lua 1465 2013-10-10 14:43:34Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tHubFiles["Hub-Threads.lua"] = {
	["sRev"] = "$Rev: 1465 $",
	["bRestart"] = true,
	["sPath"] = "./core/Hub-Threads.lua",
	}

--// Major rewrite of threads, merging Listen and Login thread back to one thread again, hopefully stopping BINF / IGUI / MyINFO from ending up in the wrong place

tThread = {

["TCPWaitForALLUsers"] = function(self)
	--// Mutexes used: ..



	local portstats = tStats.Thread.Port[port] or {}
	portstats["Status"] = "Running"
	portstats["MaxDuration"] = portstats["MaxDuration"] or 0

--~ 	debug.sethook(hook,"c") -- Set the hook
	--// To do? Keep a local table for quick access to hammerlog, this can also prevent a table constant overflow when under heavy ddos

	while tServer.Port[port].server do
		local res, err = pcall(tCore.Wait, port, stype)
		if not res then
			tFuncs:SafeDebug("[tCore.Wait] [ERROR] "..err, "FI", -3)
			tFuncs:UnlockMutexes()
--~ 			break
		end
	end
	tServer.Port[port].socket:close()
	portstats["Status"] = "Finished"
	tFuncs:SafeDebug("TCPWaitForUsers: Serversocket for port: "..port.." closed.", "FI", -3)
	ThreadDestroy()
end,


["TCPWaitForUsers"] = function(self, port, stype)
	--// Mutexes used: ..
	local portstats = tStats.Thread.Port[port] or {}
	portstats["Status"] = "Running"
	portstats["MaxDuration"] = portstats["MaxDuration"] or 0

--~ 	debug.sethook(hook,"c") -- Set the hook
	--// To do? Keep a local table for quick access to hammerlog, this can also prevent a table constant overflow when under heavy ddos

	while tServer.Port[port].server do
		local res, err = pcall(tCore.Wait, port, stype)
		if not res then
			tFuncs:SafeDebug("[tCore.Wait] [ERROR] "..err, "FI", -3)
			tFuncs:UnlockMutexes()
--~ 			break
		end
	end
	tServer.Port[port].socket:close()
	portstats["Status"] = "Finished"
	tFuncs:SafeDebug("TCPWaitForUsers: Serversocket for port: "..port.." closed.", "FI", -3)
	ThreadDestroy()
end,

["TCPSSL"] = function(self)
	--// Mutexes used: ..
	tStats.Thread.SSL["Status"] = "Running"

	while tThreads.SSL.Running do
		tStats.Thread.SSL["Time"] = socket.gettime()
		local res, err = pcall(tCore.SSLHandshake)
		if not res then
			tFuncs:SafeDebug("[tCore.TCPSSL] [ERROR] "..err, "FI", -3)
			tFuncs:UnlockMutexes()
			break
		end
		tStats.Thread.SSL["Time"] = false
	end
	tThreads.SSL.Running = false
	tStats.Thread.SSL["Time"] = false
	tStats.Thread.SSL["Status"] = "Finished"
	tFuncs:SafeDebug("TCPSSL: Handshake thread stopped.", "FI", -3)
	ThreadDestroy()
end,

["TCPListen"] = function(self)      -- first only one port supported
--~ 	debug.sethook(hook,"c") -- Set the hook

	local nextcollect = os.time()+5
	local pcount = 0
	local prunning = false
	tServer.Main.tcp_listening = 1

	-- // Start of TCP Listening thread
	while tServer.Main.tcp_listening == 1 do	-- setting this to 0 will make this thread selfterminating
		--// Profiles start
		if tSettings.Debug.Profiler.Listen and not prunning then
			profiler.start("Profiler-timings.log")
			prunning = true
		end

		--// Listen to all client-sockets
		local res, err = pcall(tCore.Listen)
		if not res then
			tFuncs:Debug("[tCore.Listen] [ERROR] "..err, "FI", -3)
			tFuncs:CoreDump("TCPListen socket error")
			tFuncs:UnlockMutexes()
		end

		--// Profiler end
		if prunning then
			pcount = pcount+1
		end
		if pcount == tSettings.Debug.Profiler.Count or (prunning and not tSettings.Debug.Profiler.Listen) then
			profiler.stop()
			prunning = false
			tSettings.Debug.Profiler.Listen = false
			tFuncs:Debug("Profiler ran during "..pcount.." iterations over the TCPListen Thread, please email the file 'Profiler-timings.log' and 'Profiler-hubstats.log' to dev@flexhub.org")
			pcount = 0
		end

		if (not prunning and nextcollect < os.time()) then collectgarbage("collect") nextcollect = os.time()+5 end
	end
	tFuncs:Debug("TCPListen: Thread closed. Stopped listening on clientsockets.", "FI", -3)
	ThreadDestroy()
	tThreads.Listen.Main = false
end,

--// Disabled
--[[
["TCPLogin"] = function(self)      -- first only one port supported
--~ 	debug.sethook(hook,"c") -- Set the hook

--~ 	local nextcollect = os.time()+5
	tServer.Main.TCPLogin = 1

	-- // Start of TCP Listening thread
	while tServer.Main.TCPLogin == 1 do          -- setting this to 0 will make this thread selfterminating
		local res, err = pcall(tCore.Login)
		if not res then
			tFuncs:Debug("[tCore.Login] [ERROR] "..err, "FI", -3)
			break
		end

--~ 		if nextcollect < os.time() then collectgarbage("collect") nextcollect = os.time()+5 end
	end
	tFuncs:SafeDebug("TCPLogin: Thread closed. Stopped listening on clientsockets.", "FI", -3)
	ThreadDestroy()
	tServer.Main.TCPLogin = -1
end,
]]--

}

tCore = {
["SafeUserAdd"] = function(self, client, sid)
	ThreadLockMutex(mutexwait)
	table.insert(tClientWait, { ["sid"] = sid, ["client"] = client })
	ThreadUnlockMutex(mutexwait)
end,

["SafeUserDelete"] = function(self, client, sid)
	client.iActive = 0
	ThreadLockMutex(mutexdelete)
	table.insert(tClientDelete, { ["sid"] = sid, ["client"] = client })
	ThreadUnlockMutex(mutexdelete)
	return client
end,

["SafeSSLAdd"] = function(self, client, sid)
	ThreadLockMutex(mutexssl)
	table.insert(tClientWaitSSL, { ["sid"] = sid, ["client"] = client })
	ThreadUnlockMutex(mutexssl)
end,

--~ ["SafeSSLDelete"] = function(self, client, sid)
--~ 	client.iActive = 0
--~ 	ThreadLockMutex(mutexssl)
--~ 	tClientSSL[sid] = nil
--~ 	ThreadUnlockMutex(mutexssl)
--~ 	return client
--~ end,

["CreateBuffer"] = function(self)
	return {
		["Send"] = {		--// start with a clean send and receive buffer
			["Buffer"] = "",
			["Chat"] = {},
			["Login"] = {},
			["Data"] = {},
			["Last"] = 0,
			["iBufferSent"] = 0,
			["iSendTry"] = 0,
			["iPauseUntil"] = false,
		},
		["MyInfo"] = {
			["LastSent"] = 0,
			["SendAll"] = false,
		},
		["Receive"] = "",
		["Dns"] = false,
		["bDisconnect"] = false,
		["Stats"] = {
			["BytesSent"] = 0,	--// Actual data, possibly compressed
			["BytesSentRaw"] = 0,	--// Uncompressed data
			["BytesReceived"] = 0,	--// Actual data
		},
	}
end,

["PreLogin"] = function(client, sid)
	local ctime = socket.gettime()

	--// Check for normal hammering
	ThreadLockMutex(mutexiplog)
	local iplog = tLogs.IP[client.sIP]
	if iplog then
--~ 		iplog["$"].LastSid = sid
		iplog["$"].LastLogin = tLogs.IP[client.sIP]["$"].CurrentLogin
		iplog["$"].CurrentLogin = ctime
--~ 		iplog["$"].LastStatus = "Socket connected"
	else
		tLogs.IP[client.sIP] = { ["$"] = {
--~ 			["LastSID"] = sid,
			["LastLogin"] = 0,
			["DoubleLogins"] = 0,
			["CurrentLogin"] = ctime,
--~ 			["LastStatus"] = "Socket connected",
--~ 			["Nicks"] = {},
--~ 			["LastTraffic"] = {},
		}}
	end
	ThreadUnlockMutex(mutexiplog)

	--// Check for banned/ hammering / clones
--~ 	local res, msg = tCheck:IP(sid, client.sIP)

--~ 	--// Check for connection hammering
	local res, msg, timeout = tCheck:Hammer(sid, client.sIP)

	--// this function relies on the fact that unknown msg:s are discarded in the client so we skip checking protocol and just send both msgs to all clients
	if not res then
		local retmsg = "IINF NI"..ADC_E(not tHubSettings.tRules.bRegOnly and tHubSettings.Main.sName or " ").." CT32\n"..	-- Empty Hubname
			--	"ISTA 220 "..ADC_E(msg).."\n"
			"ISTA 232 "..ADC_E(msg).." TL"..(timeout or 20).."\n"	-- there was a space after TL, and flags have to be placed after message

		local res, err, to = client.xSocket:send(msg.."|\n"..retmsg)
		tLogs.Counters.ConnectionsBlocked = tLogs.Counters.ConnectionsBlocked + 1

		return false, client
	else
--~ 		--// Show connection
--~ 		tFuncs:SafeDebug("SID: "..sid.." IP: "..client.sIP.." <login state:1> connected to: "..hip..":"..hport, "LNJ",2)

		tLogs.Counters.ConnectionsAllowed = tLogs.Counters.ConnectionsAllowed + 1

--~ 		ThreadCreate(tFuncs.GetclientDNS,{1, sid, client.sIP})

		client.iActive = 1
		client.xHandler = "Login"

		--// Get countrycode
		if tHubSettings.Main.bUseGeoIP and tGeoIP.loaded then
			client.sCountryCode = tGeoIP.loaded:lookup(client.sIP) or "ERROR"
		end

		--// Check if IP is on Monitor
		if tMonitor[client.sIP] then
			client.sMonitor = tMonitor[client.sIP]
			tFuncs:Debug("Monitored IP is connecting [ "..client.sIP .." ], all traffic outputs to: "..tMonitor[client.sIP])
		end

		return true, client
	end
end,

["Wait"] = function(port, stype)
	local ctime = socket.gettime()
	local portstats = tStats.Thread.Port[port]
	portstats.Status = "Waiting for connection"
	portstats.Time = false
	portstats.xSocket = false

	local clientsocket, info = tServer.Port[port].socket:accept()	--// no timeout, so this is blocking the thread until a new socket is accepted

	local ptime = socket.gettime()
	portstats.Time = ptime
	portstats.Status = "Handling incoming connection"
	portstats.xSocket = clientsocket

	tServer.Port[port].connects = tServer.Port[port].connects+1

	if not clientsocket then
		tLogs.Errors.CreateSocket = tLogs.Errors.CreateSocket + 1
		tFuncs:SafeDebug("[Threads:Wait] Error creating socket : "..(info or "unknown error"), nil, 0)
		return
	end

	portstats.Status = "Get peername"
	local ip, iport = clientsocket:getpeername()

	if not clientsocket or not ip or not iport then
		tLogs.Counters.PeerNameFailed = tLogs.Counters.PeerNameFailed + 1
		return
	end

	portstats.Data = "IP: "..ip.." / Port: "..iport
	portstats.Status = "Create user"
	--// Create new user -- only locally until all checks are ok  --Could need a mutex to prevent pseudo-random generated same-SIDs by different Wait-threads (ports)
	local sid = tFuncs:CreateSid() --// generate sid
	local user = tFuncs:NewUser()
	user.xSocket = clientsocket
	user.iPort = port
	user.sIP, user.iOwnPort = ip, iport
	user.iLoginClock = ctime

	portstats.Status = "Setting timeout and options"
	user.xSocket:settimeout((tHubSettings.tTcp.SocketSendTimeout or 0), "b", "t")

	--// Experiment: this might partly solve the ghosting issue, it's likely that a ghost is detected and dropped when it doesn't respond to the keepalive tcp packets
	user.xSocket:setoption("keepalive", true)

--~ 	user.xSocket:settimeout((tHubSettings.tTcp.SocketSendTimeout or 0), (tHubSettings.tTcp.SocketSendTimeoutMode or "t"))

	--//  Check for DDOS hammering, this is kept in a separate compact table to prevent table count overflows
	portstats.Status = "Check IP bans"
	ThreadLockMutex(mutexbanlist)
	if tBans.autoban[user.sIP] and tBans.autoban[user.sIP] >= ctime then
		tLogs.Counters.Ddos.Port[port].Count = tLogs.Counters.Ddos.Port[port].Count + 1
		tBans.autoban[user.sIP] = ctime + tSecurity.General.DefaultAutoban	--// Autoban another 2 hours after each ddos-connection
		ThreadUnlockMutex(mutexbanlist)
		user.xSocket:close()
		user = nil
		portstats.xSocket = nil

		--// Waittime when hub is under attack:
		if tLogs.Counters.Ddos.Port[port].Attack then
			ThreadWait(tSecurity.Ddos.iServerWaitTime)
		end

		return
	elseif tBans.autoban[user.sIP] then
		--// Tempban expired
		tBans.autoban[user.sIP] = nil
		ThreadUnlockMutex(mutexbanlist)
	else
		ThreadUnlockMutex(mutexbanlist)
	end

--[20:35:18] <[bot]-Security> [FI] [-3] [tCore.Wait] [ERROR] ...saymon/data/etc/flexhub/FlexHub/core/Hub-Threads.lua:307: attempt to index field 'xSocket' (a nil value)
--[20:35:18] <[bot]-Security> [FI] [0] [TLS] SID: Y3ZD IP: 178.158.196.130 <> Wrapping failed: error loading certificate ((null)) (note: was client.xSocket closed properly??)

	portstats.Status = "Move to secure handler or prelogin"

	--// Secure or Regular connection?
	if stype and stype == "Secure" then
		portstats.Status = "Wrap secure socket"
		local ret, user = SecureSocket(user, sid)

		if ret then
			user.bSecure = true
			user.xHandler = "SSLHandshake"
			portstats.xSocket = user.xSocket
			portstats.Status = "Move socket to SSL handshake thread"
			tCore:SafeSSLAdd(user, sid)
		else
			--user.xSocket:close()
			user = nil
			return
		end
	else
		portstats.Status = "Regular socket prelogin"
		local ret, user = tCore.PreLogin(user, sid)
		if ret then
			--// Only the Listen thread is allowed to add or delete users, mark for adding to tClient with tClientWait
			portstats.Status = "Move socket to Listen thread"
			tCore:SafeUserAdd(user, sid)
		else
			user.xSocket:close()
			user = nil
			return
		end
	end

	portstats.LastDuration = socket.gettime() - ptime
	if portstats.LastDuration > portstats.MaxDuration then portstats.MaxDuration = portstats.LastDuration end
	portstats.Time = false
	portstats.Data = false
	portstats.Socket = false

	--// Delay ListenThread to reduce CPU and max users login/second
	if tLogs.Counters.Ddos.Port[port].Attack then	--// When under attack, use a different waittime
		ThreadWait(tSecurity.Ddos.iServerWaitTime)
	else
		ThreadWait(tHubSettings.tThreads.ServerWait)
	end
end,

["Listen"] = function()
	local socket = socket
	local ctime = socket.gettime()
	local listenstats = tStats.Thread.Listen	--// Localize global table
	local timerstats = tStats.Timers			--// Localize global table

	listenstats.Time = ctime
	listenstats.Status = "Started"
	timerstats.SendSocketsTemp = 0

	local threadtime, sendtime, parsetime, socketreadtime, socketsendtime = ctime, 0, 0, 0, 0

	--// Timers - runs only if at least a full second has passed
	if ctime >= timerstats.LastRun+1 then
		timerstats.LastRun = ctime
		local result, ret = pcall(tTimerFuncs.FlexTimer)
		if not result then
			tFuncs:Debug("[Listen] Error when running timers: "..ret)
			tFuncs:UnlockMutexes()
		end

		--// Update speed stats
		listenstats.Status = "Listen: Calc speeds"
		tFuncs:CalcSpeeds()
	end

	--// Attackstats
	if (tLogs.Counters.Ddos.LastAverageTime + tSecurity.Ddos.iAttackReportInterval) < ctime then
		tFuncs:AttackReport()
	end

	--// Combined handling of adding and deleting all new/quitting users from all other threads through mutex protected tables
	--// tClientWait / tClientSSL / tClientDelete:  wait, login, ssl, and other threads can't add or delete from tClient, only Listen can to prevent collision

	--// Add clients from tClientWait
	listenstats.Status = "Getting new logins from waitthreads"
	if #tClientWait > 0 then
		ThreadLockMutex(mutexwait)
		for num,data in pairs(tClientWait) do
			tClient[data.sid] = data.client
			tBuffer[data.sid] = tCore:CreateBuffer()
		end

		tClientWait = {}
		ThreadUnlockMutex(mutexwait)
	end

	--// Delete clients from tClientDelete
	listenstats.Status = "Deleting disconnected users"
	if #tClientDelete > 0 then
		ThreadLockMutex(mutexdelete)
		for num,data in pairs(tClientDelete) do
			listenstats.User = data.sid
			local res, err = pcall(tFuncs.ClientQuit, true, data.client, data.sid)
			if not res then
				tFuncs:UnlockMutexes()
				tFuncs:Debug(tUser:GetDataLine(sid).."Error: ClientQuit: "..err)
			end
			tClient[data.sid] = nil
		end
		tClientDelete = {}
		ThreadUnlockMutex(mutexdelete)
	end

	--// Retrieve data from other threads
	listenstats.Status = "Retrieve data from other threads"
	ThreadLockMutex(mutexlistenbuffer)
	local messages = tListenBuffer.Messages
	tListenBuffer.Messages = {}
	local commands = tListenBuffer.Commands
	tListenBuffer.Commands = {}
	ThreadUnlockMutex(mutexlistenbuffer)

	--// Handle messages from other threads
	listenstats.Status = "Handle messages from other threads"
	for num, line in pairs(messages) do
		if line.Target == "D" then
			tFuncs:Debug(line.msg, line.debugtype, line.debuglevel, line.cat)
		elseif line.Target == "R" then
			tSend:ChatRoom(line.Message, line.From, line.To)
		elseif line.Target == "M" then
			if line.To and tClient[line.To] then
				tSend:ChatToUser(line.Message, line.From, line.To)
			else
				tSend:ChatToAll(line.Message, line.From)
			end
		elseif line.Target == "P" then
			tSend:PmToUser(line.Message, line.From, line.To)
		end
	end
	messages = nil

	--// Handle commands from other threads
	listenstats.Status = "Handle commands from other threads"
	for num, line in pairs(commands) do
		local res, err, arg1 = pcall(LoadChunk, line)
		if not res then
			tFuncs:Debug("[Listen] Queued command from tListenBuffer failed: "..err..". Command: "..line, "FI", -1)
			tFuncs:UnlockMutexes()
		else
			tFuncs:Debug("[Listen] Queued command from tListenBuffer: "..line.."\nResult: "..tostring(err).." / "..tostring(arg1), "X", 0)
		end
	end
	commands = nil

	--// Client listen, transfer to parser and send queued messages loop
	listenstats.Status = "Client-cycle: started"
	for sid, client in pairs(tClient) do		-- // socketloop - receive from all active clients
		listenstats.User = sid

--~ 		if client.xHandler == "Main" and client.iActive > 0 and not client.bBot and (not client.iWaitTime or client.iWaitTime < os.time())then	--// internal Bots have no sockets, TODO: hub needs to internally handle messages from and to them
		if client.iActive > 0 and not client.bBot and ((not client.iWaitTime) or client.iWaitTime < os.time()) then

			--// Read from socket
			listenstats.Status = "Client-cycle: read from socket"
			local readtxt, err, partialtxt
			if tBuffer[sid] and not tBuffer[sid].bDisconnect then	--// don't read the socket of users that are marked for disconnection
				local timer = socket.gettime()
				listenstats.SocketTime = timer

--~ 					client.xSocket:settimeout(tSettings.TcpSocketReadTimeout, tSettings.TcpSocketReadTimeoutMode)

				listenstats.Status = "Client-cycle: read from socket : select"			-- more dbugg
				--// WARNING!!!!  Experimental addition of socket select to check if receiving on the socket is allowed    // 2011/02/01
				local rect, sendt = socket.select({client.xSocket},{}, 0)		--// this is where we crash! 2012-12-30 fredo
				listenstats.Status = "Client-cycle: read from socket : selected"			-- more dbugg
				if #rect < 1 then
					readtxt, err, partialtxt = "", nil, nil
				else
					listenstats.Status = "Client-cycle: read from socket : receive"			-- more dbugg
					readtxt, err, partialtxt = client.xSocket:receive(tHubSettings.tTcp.BufferSize)
					listenstats.Status = "Client-cycle: read from socket : received"			-- more dbugg
				end
				listenstats.SocketTime = false
				timer = socket.gettime() - timer
				socketreadtime = socketreadtime + timer
				if timer > timerstats.ReadSingleSocketMax then
					timerstats.ReadSingleSocketMax = timer
					timerstats.ReadSingleSocketMaxId = tUser:GetDataLine(sid)
				end
			else
				err = nil	--// Set err = nil to skip errorchecking - move on to sending last messages of user to be disconnected
			end

			local parsetxt
			if not err then parsetxt = readtxt else parsetxt = partialtxt end

			if not client.sProtocol then
				listenstats.Status = "Client-cycle: checking protocol"
				tCore:Protocol(client, sid, parsetxt)
			end

			listenstats.Status = "Client-cycle: check parsetxt"
			if parsetxt and parsetxt ~= "" then
				client.sLastIncoming = ctime
				client.sStatus = "[RECV] "..parsetxt:sub(1, 50)
				tStats.Tcp.Received = tStats.Tcp.Received + #parsetxt		--// This might collide with setting this same variable in TCPListen
				if client.sProtocol and client.sProtocol == "ADC" then
					local ok = false
					ok, parsetxt = tFuncs:ToLocale(parsetxt)
					if not ok then
						tFuncs:Debug("Error converting message from: "..(client.sNick or client.sIP).." / ToLocale: "..parsetxt:sub(40).."...", "FI", 1)

						--// Mark current receive buffer to discard the current command(s)
						tBuffer[sid].Skip = true

						if client.xHandler == "Login" then
							tUser:Disconnect(sid, "200 Your login-data contains characters that can't be converted from UTF8 to WIN-1252, this is needed for the NMDC side of FlexHub\nPlease use alphanumeric characters in your nick, description, and emailaddress.\n\n")
						end
					end
				end
			end

			--// Update statistics
--~ 				local received, sent, age = client.xSocket:getstats()
--~ 				local recbytes = received - client.iTcpReceived
--~ 				local sentbytes = sent - client.iTcpSent

--~ 				client.iTcpSent = sent
--~ 				client.iTcpReceived = received
--~ 				client.iTcpAge = age

--~ 				tStats.Tcp.Received = tStats.Tcp.Received + recbytes
--~ 				tStats.Tcp.Sent = tStats.Tcp.Sent + sentbytes
--~ 				tStats.Tcp.SentInterval = tStats.Tcp.SentInterval + sentbytes
			--// Protocol detection

			-- // CHECK FOR SOCKET ERRORS
			if err and err == "closed" then
				listenstats.Status = "Client-cycle: socket closed, marked inactive"
				tFuncs:Debug(tUser:GetDataLine(sid).."Socket closed.", "TI", 3)
				client.iActive = 0 --// socket closed, mark for cleaning

			elseif err and err ~= "timeout" and err ~= "wantread" and err ~= "wantwrite" then
				listenstats.Status = "Client-cycle: socket error: "..err..", marked inactive"
				if not tSettings.Security.UnderAttack then	-- turning off error reporting while port is under attach
					tFuncs:Debug(tUser:GetDataLine(sid).."Socket error: "..err..". User disconnected.", "FI", 3)
				end
				client.iActive = 0 --// socket closed, mark for cleaning

			--// SEND PENDING MESSAGES
--~ 			elseif client.iActive > 0 and tBuffer[sid].Send.Chat ~= {} or tBuffer[sid].Send.Buffer ~= "" or (tBuffer[sid].Send.Data ~= {} and tBuffer[sid].Send.Last < (socket.gettime() - tHubSettings.tTcp.DataSkipTime)) then
			elseif client.iActive > 0 and (tBuffer[sid].Send.Buffer ~= "" or (tBuffer[sid].Send.Data ~= {})) then
				listenstats.Status = "Client-cycle: sending pending messages"
				local stimer = socket.gettime()
				listenstats.SocketTime = timer

--~ 				if not tBuffer[sid].Send.iPauseUntil or tBuffer[sid].Send.iPauseUntil < stimer then	-- wait/skip if socket was unresponsive previously  << this check is done in TCPSend
					local result, ret1 = pcall(tFuncs.TCPSend, nil, client, sid)
					if not result then
						tFuncs:Debug(tUser:GetDataLine(sid).."Error: TCPSend: "..ret1)
						tFuncs:UnlockMutexes()
					end
--~ 				end

				listenstats.SocketTime = false
				stimer = socket.gettime() - stimer
				sendtime = sendtime + stimer
				if stimer > timerstats.SendSingleMax then
					timerstats.SendSingleMax = stimer
					timerstats.SendSingleMaxId = tUser:GetDataLine(sid)..Clock()
				end
			end

			--// PARSE
			if not tBuffer[sid].bDisconnect and client.iActive > 0 and parsetxt and parsetxt ~= "" then   --// Skip parsing of usermessages from users that have 'client.disconnect = true'
				listenstats.Status = "Client-cycle: parsing incoming data"
				local ptimer = socket.gettime()
				local result, ret1

				--// PARSE ONLY WHEN AN ENDPIPE OR NEWLINE IS RECEIVED IN LAST TRAFFIC
				if client.sProtocol == "NMDC" and parsetxt:find("|", 1, true) then
					result, ret1 = pcall(tFuncs.TCPParse, nil, client, sid, tBuffer[sid].Receive..parsetxt)
					if not result then tFuncs:UnlockMutexes() tFuncs:Debug(tUser:GetDataLine(sid).."Error: TCPParse: "..ret1) tUser:Disconnect(sid, ret1) end
				elseif client.sProtocol == "ADC" and parsetxt:find("\n", 1, true) then
					result, ret1 = pcall(tFuncs.TCPParse, nil, client, sid, tBuffer[sid].Receive..parsetxt)
					if not result then tFuncs:UnlockMutexes() tFuncs:Debug(tUser:GetDataLine(sid).."Error: TCPParse: "..ret1) tUser:Disconnect(sid, ret1) end
				else		--// no endpipe or newline received, check if buffer doesn't exceed maximum

					--// Extra debug to catch the EMSG bug
--~ 					tFuncs:Debug("Buffering message from "..(client.sNick or client.sIP)..": "..parsetxt, "X")
--~ 					Test("BUFFERING: "..parsetxt)

					result, ret1 = tDeflood.Check.Buffer(sid, tBuffer[sid].Receive..parsetxt)
					if not result then
						tFuncs:Debug(tUser:GetDataLine(sid).."User sending too much data for one command (commandsize), disconnected.", "FI", 0)
						tUser:Disconnect(sid, ret1)
					else
						tBuffer[sid].Receive = tBuffer[sid].Receive..parsetxt
					end
				end

--~ 				if not result then
--~ 					tFuncs:Debug(tUser:GetDataLine(sid).."Error: TCPParse: "..ret1)
--~ 					tUser:Disconnect(sid, ret1)
--~ 				else
--~ 					client = ret1
--~ 				end

				ptimer = socket.gettime() - ptimer
				parsetime = parsetime + ptimer
				if parsetime > timerstats.ParseSingleCmdMax then
					timerstats.ParseSingleCmdMax = ptimer
					timerstats.ParseSingleCmdMaxId = tUser:GetDataLine(sid)..Clock()..parsetxt:sub(1, 100)
				end

				--// SEND POOLED MYINFO OF THIS USER TO ALL
				if not tBuffer[sid].bDisconnect then
					if tBuffer[sid].MyInfo.SendAll and ctime > (tBuffer[sid].MyInfo.LastSent + (tHubSettings.tTraffic.iMyInfoInterval or 900)) then
						tBuffer[sid].MyInfo.LastSent = ctime
						tSend:MyInfUpdate(sid)
						tBuffer[sid].MyInfo.SendAll = false
					end
				end

			elseif client.iActive == 0 then		-- clientsocket is flagged inactive, close socket and delete user
				listenstats.Status = "Client-cycle: disconnecting inactive client (1)"
				tFuncs:Debug(tUser:GetDataLine(sid).."User socket was closed or will be closed now, start SafeUserDelete", "LI", 2)

				--// Put client in Delete bin
				tCore:SafeUserDelete(client, sid)
			elseif tBuffer[sid].bDisconnect then	-- this will make the client disconnect after the next clientsloop
				listenstats.Status = "Client-cycle: marked for disconnection,  marked inactive"
--~ 				if tBuffer[sid].iDisconnectTime <= os.time() then
					tFuncs:Debug(tUser:GetDataLine(sid).."User marked for disconnection, user will be disconnected after last pending messages are sent", "LI", 2)
					client.iActive = 0
--~ 				elseif tBuffer[sid].Send.Buffer == "" and tBuffer[sid].Send.Data == {}
			end

			--// CHECK IF LOGIN IS COMPLETE
			if client.xHandler == "Login" and client.iState == 7 and not tBuffer[sid].bDisconnect then
				listenstats.Status = "Client-cycle: login complete"
--~ 				local buffer = ""
--~ 				for _,login in ipairs(tBuffer[sid].Send.Login) do
--~ 					if login ~= "" then
--~ 						buffer = buffer.."\n"..login
--~ 					end
--~ 				end
--~ 				if tBuffer[sid].Send.Buffer..buffer == "" then
					client.xHandler = "Main"
--~ 				else
--~ 						tFuncs:Debug("Clientstate = 7 but login-send buffer is still filled: "..buffer, "FI", -2)
--~ 				end

			--// CHECK IF HANDSHAKE TIMEOUT IS REACHED
			elseif client.xHandler == "Login" and not tBuffer[sid].bDisconnect then
				listenstats.Status = "Client-cycle: check login timeout"
				--// check if login isn't taking too long
				local logintimeout = tSecurity.General.LoginTimeout or 600
				if client.iLoginTime and (ctime - client.iLoginTime) > logintimeout then
					tUser:Disconnect(sid, "Login handshake timeout, try again later", "Login handshake timeout, last status: "..client.sStatus, "LI", 2)
				end
			end

		elseif client.iActive == 0 then
			listenstats.Status = "Client-cycle: disconnecting inactive client (2)"
			tFuncs:Debug(tUser:GetDataLine(sid).."User socket was closed or will be closed now, start SafeUserDelete", "LI", 2)
			--// Put client in Delete bin
			tCore:SafeUserDelete(client, sid)
		end

	end

	listenstats.Status = "Client-cycle: complete"
	threadtime = math.abs(socket.gettime()- threadtime)
	timerstats.SendFunction = sendtime
	timerstats.ReadSockets = socketreadtime
	timerstats.SendSockets = timerstats.SendSocketsTemp
	timerstats.ParseCmds = parsetime
	timerstats.Thread = threadtime

	if sendtime > timerstats.SendFunctionMax then timerstats.SendFunctionMax = sendtime end
	if socketreadtime > timerstats.ReadSocketsMax then timerstats.ReadSocketsMax = socketreadtime end
	if timerstats.SendSockets > timerstats.SendSocketsMax then timerstats.SendSocketsMax =  timerstats.SendSockets end
	if parsetime > timerstats.ParseCmdsMax then timerstats.ParseCmdsMax = parsetime end
	if threadtime > timerstats.ThreadMax then timerstats.ThreadMax = threadtime end

--~ 	collectgarbage("collect") --// change this to collect garbage once every few seconds

	--// Delay ListenThread to reduce CPU
	local maxtime = tHubSettings.tThreads.ListenWait or 200
 	local waittime = math.floor(maxtime - (threadtime*1000))  --// Converts threadtime from seconds to milliseconds and calculates the remaining waittime
	waittime = waittime>0 and ((waittime<maxtime and waittime) or maxtime) or 0
	ThreadWait(waittime)
end,

["SSLHandshake"] = function()
	local threadtime = socket.gettime()
	tStats.Thread.SSL.Status = "Start handshakes"
	tStats.Thread.SSL.Time = false

	local socketreadtime = 0

	if #tClientWaitSSL > 0 then
		tStats.Thread.SSL.Status = "Get waiting clients"
		ThreadLockMutex(mutexssl)
		while #tClientWaitSSL > 0 do
			local data = table.remove(tClientWaitSSL)
			if data then tClientSSL[data.sid] = data.client end
		end
		--// tClientWaitSSL is now empty, unlock mutex
		ThreadUnlockMutex(mutexssl)
	end

	--// Client listen, transfer to parser and send queued messages loop
	tStats.Thread.SSL.Status = "Handshakes loop"
	for sid,client in pairs(tClientSSL) do	 -- // socketloop - receive from all active clients
		tStats.Thread.SSL.Time = socket:gettime()
		tStats.Thread.SSL.User = sid
		tStats.Thread.SSL.Socket = client.xSocket
		tStats.Thread.SSL.Data = client.sIP


		--// Perform SSL Handshake
		local ret, err = client.xSocket:dohandshake()

		--// SSL Handshake successful
		if ret then
			tFuncs:SafeDebug("SID: "..sid.." IP: "..client.sIP.." [SSL] Secure connection successful", "LI", 3)
			tStats.Thread.SSL.Status = "Secure socket prelogin"
			ret = tCore.PreLogin(client, sid)
--~ 			ret, client = tCore.PreLogin(client, sid)
			if not ret then
				tStats.Thread.SSL.Status = "Secure socket failed prelogin, closing"
				client.xSocket:close()
				tClientSSL[sid] = nil
			else
				tStats.Thread.SSL.Status = "Secure socket move to Listen thread"
				client.xHandler = "Login"
				tCore:SafeUserAdd(client, sid)
				tClientSSL[sid] = nil
			end
		--// Errorhandling
		elseif not ret and err == "closed" then
			tStats.Thread.SSL.Status = "Secure socket failed handshake, closing"
			tFuncs:SafeDebug("User with IP "..client.sIP.." socket closed, disconnected during SSL handshake.", "LI", 3)
			client.xSocket:close()
			tClientSSL[sid] = nil
		--// Check if timeout occured
		else
			local ssltimeout = tSecurity.General.SSLHandshakeTimeout or 5
			if client.iLoginTime and (os.time() - client.iLoginTime) > ssltimeout then
				tFuncs:SafeDebug("User with IP "..client.sIP.." SSL handshake timeout, disconnected.", "LI", 3)
				client.xSocket:close()
				tClientSSL[sid] = nil
			end
		end

		tStats.Thread.SSL.Time = false
		tStats.Thread.SSL.User = false
		tStats.Thread.SSL.Socket = false
		tStats.Thread.SSL.Data = false
	end
	tStats.Thread.SSL.Status = "Handshakes loop complete"

--~ 	--// Delay ListenThread to reduce CPU
 	threadtime = math.abs(socket.gettime() - threadtime)  --// This can cause a huge threadtime when the clock turns over, therefor use maxtime to prevent an eternal wait
	local maxtime = tHubSettings.tThreads.ServerWait or 200
 	local waittime = math.floor(maxtime - (threadtime*1000))  --// Converts threadtime from seconds to milliseconds and calculates the remaining waittime
	waittime = waittime>0 and ((waittime<maxtime and waittime) or maxtime) or 0
	ThreadWait(maxtime)
end,

["Protocol"] = function(self, client, sid, parsetxt)
	--// No traffic received and took over a second, send the NMDC lock now
	if not parsetxt or parsetxt == "" and (socket.gettime() - client.iLoginClock) >= tHubSettings.Main.iLockWait and not client.bSentLock then
		local lock = "$Lock EXTENDEDPROTOCOL_FLEXHUB_MULTIPROTOCOL_"..sid.." Pk=version0.2|$HubName "..((tHubSettings.tRules.bRegOnly and " ") or tHubSettings.Main.sName).."|"
		tSend:Raw(sid, lock)
		client.bSentLock = true
		client.sStatus = "[Sent] "..lock

		if tMonitor[client.sIP] then tFuncs:Debug("[Monitoring "..client.sIP.." ] Sending $Lock") end
	elseif parsetxt and parsetxt ~= "" then
		parsetxt = tBuffer[sid].Receive..parsetxt

		--// ADC User
		if parsetxt:sub(1,4) == "HSUP" or parsetxt:sub(1,4) == "HSTA" then
			client.sProtocol = "ADC"

			if tLogs.Counters.Ddos.Port[client.iPort].Attack then
				tLogs.Counters.Ddos.Port[client.iPort].Login = tLogs.Counters.Ddos.Port[client.iPort].Login + 1
			end

			if client.bSentLock then
				tSend:Raw(sid, "\n")	--// In case a $Lock was sent to this ADC client, end the command with a newline
			end

			--// Check for banned/ hammering / clones
			local res, msg = tCheck:IP(sid, client.sIP)

			return client

		--// NMDC User
		elseif parsetxt:sub(1,4) == "$Sup" then
			client.sProtocol = "NMDC"

			if tLogs.Counters.Ddos.Port[client.iPort].Attack then
				tLogs.Counters.Ddos.Port[client.iPort].Login = tLogs.Counters.Ddos.Port[client.iPort].Login + 1
			end
			--// Check for banned/ hammering / clones
			local res, msg = tCheck:IP(sid, client.sIP)

			return client

		--// Bot
		elseif parsetxt:sub(1,4) == "$Key" then
			--local filter = parsetxt:gsub("$Key (.-)|", "***") or parsetxt
			local filter, dummy
			if tHubSettings.Main.iFeedLevel < 1 then
				filter = parsetxt:match("(%S+).*$") .." ..." or parsetxt
			else
				dummy, filter = tCheck:ControlChars(nil, nil, parsetxt)
			end

			--if tIp:Autoban(client.sIP, "Invalid Connection", parsetxt) then
			if tIp:Autoban(client.sIP, "Invalid Connection: Spam bot or ancient client", filter) then
				client.iActive = 0
			else
				--// Protected IP's are allowed to skip $Supports and start with a $Key
				client.sProtocol = "NMDC"
				client.iState = 2
				if tLogs.Counters.Ddos.Port[client.iPort].Attack then
					tLogs.Counters.Ddos.Port[client.iPort].Login = tLogs.Counters.Ddos.Port[client.iPort].Login + 1
				end
				--// Check for banned/ hammering / clones
				local res, msg = tCheck:IP(sid, client.sIP)
			end
			return client

		--// DDOS attack NMDC
		elseif parsetxt:sub(1,4) == "$MyN" or parsetxt:find("^$%a*$MyN") then
			client.iActive = 0

			tLogs.Counters.Ddos.Port[client.iPort].NewCount = tLogs.Counters.Ddos.Port[client.iPort].NewCount + 1
--~ 			tIp:Autoban(client.sIP, false, false, true)
			tIp:Autoban(client.sIP, false, false)		--// Disabled clearlog, this might fill up the IP log when a lot of ip's are used in the attack...

			local ref = parsetxt:match("Ref=(%S-)|")
			if ref then
				ref = NMDC_E(ref, true)
				ref = "dchub://"..ref:gsub(".+://", "")
				if not tLogs.Counters.Ddos.Source[ref] then
					tLogs.Counters.Ddos.Source[ref] = socket.gettime()
					tFuncs:Debug("[DDOS] Attack source detected: "..ref.." from incoming client with IP: "..client.sIP, "FI", -1)
				end
			end
		--// DDOS attack ADC
		elseif parsetxt:sub(1,4) == "CSUP" then
			client.iActive = 0
			tLogs.Counters.Ddos.Port[client.iPort].NewCount = tLogs.Counters.Ddos.Port[client.iPort].NewCount + 1
			tIp:Autoban(client.sIP, false, false, true)
			local ref = parsetxt:match("CSTA 000 (*.)")
		--// Unknown traffic
		elseif parsetxt == "|" then
			client.sProtocol = "NMDC"

			--// Check for banned/ hammering / clones
			local res, msg = tCheck:IP(sid, client.sIP)

			return client
		elseif parsetxt == "\n" then
			client.sProtocol = "ADC"

			--// Check for banned/ hammering / clones
			local res, msg = tCheck:IP(sid, client.sIP)

			return client
		--// TLS connection attempt on non-secure port
		elseif parsetxt:sub(1,3) == "" then
			client.iActive = 0
			--parsetxt = nil
		elseif #parsetxt < 4 then
			--// Do nothing, wait for more
		elseif parsetxt:find("^GET /%S+") then
			client.iActive = 0
			client.bSentLock = true	--// Prevent the Lock from being sent
			client.iState = 2
			client.sProtocol = "HTML"

			--// Check for banned/ hammering / clones
--~ 			local res, msg = tCheck:IP(sid, client.sIP)
			return client
		elseif parsetxt:find("^GET %S+") then
			if not tHubSettings.Main.bAllowHTTP then
				local msg = parsetxt:sub(1,120):match("^GET %S+")
				tIp:Autoban(client.sIP, "HTTP Access disabled", msg)
			else
				client.iActive = 0
				client.bSentLock = true	--// Prevent the Lock from being sent
				client.iState = 2
				client.sProtocol = "HTML"
			end

			--// Check for banned/ hammering / clones
--~ 			local res, msg = tCheck:IP(sid, client.sIP)
			return client
		elseif not client.bSecure then		--// needs a fix, secure clients shouldn't be excluded from other invalid traffic
			client.iActive = 0
			tIp:Autoban(client.sIP, "Invalid Connection", parsetxt)
			return client
		elseif client.bSecure then	--// Disconnect secure clients on invalid traffic
			tFuncs:Debug("TLS User with IP: "..client.sIP.." sends invalid traffic: "..parsetxt, "X")
--~ 			client.iActive = 0
--~ 			tIp:Autoban(client.sIP, "Invalid Connection", parsetxt)
			return client
		end
	end
	return client
end,

}

