--// $Id: Hub-WebStats.lua 1467 2013-10-19 17:54:25Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tHubFiles["Hub-WebStats.lua"] = {
	["sRev"] = "$Rev: 1467 $",
	["bRestart"] = false,
	["sPath"] = "./core/Hub-WebStats.lua",
	}

tWebStats = {
["GetData"] = function(self)		--// Don't call this from a separate thread, it loops over all clients for some fillvars and gettotalshare.
	local data = ""
	--//Modified fields are ignored for now
	local fields = "%[address],%[desc],%[email],%[hub],%[network],%[website],%[owner],%[platform],%[version],%[svn],%[topic],"..
			"%[minsharedownload:raw],%[minshare:raw],%[minslots],%[totalshare:raw],%[uptime:raw],%[usercount],%[userpeak],%[usersadc],%[usersnmdc],"..
			"%[mem:raw],%[mempeak:raw],%[speedsentpeak:raw],%[speedreceivedpeak:raw],%[datasent:raw],%[datareceived:raw]"

	for word in fields:gmatch("(%%%[.-%])") do
		local name = string.match(word, "%%%[(.-)%]")
		local name = name:gsub(":.*", "")
		local value = url.escape(tFuncs:FillVars(word))
		data = data.."&"..name.."="..value
	end

	local adcshare = tFuncs:ConvertShare(tFuncs:GetTotalShare("ADC") or 0, true) or 0
	local nmdcshare = tFuncs:ConvertShare(tFuncs:GetTotalShare("NMDC") or 0, true) or 0
	local public = tHubSettings.tRules.bRegOnly and 0 or 1
	local hublist = tHubSettings.tHubLists.FlexHubList and 1 or 0
	data = data.."&shareadc="..adcshare.."&sharenmdc="..nmdcshare.."&public="..public.."&hublist="..hublist

	return data
end,

["Register"]  = function(self, ssid, nick, pass, mode, key, value, data)
	local result, err = pcall(tWebStats.RegisterRun, nil, ssid, nick, pass, mode, key, value, data)
	if not result then
		tFuncs:SafeDebug("[THREAD] [WebStats][Register] Error: "..err)
		tFuncs:UnlockMutexes()
	end
	tThreads.WebStatsReg = nil
	ThreadDestroy()
end,

["RegisterRun"]  = function(self, ssid, nick, pass, mode, key, value, data)
	if (mode == "reg") then
		local ok, reply, saltpass = tWebStats:GetSalt(ssid, nick, pass)
		if not ok then tFuncs:SafeDebug("Webstats login failed: \n\nError: "..reply) return end

		local request = "http://flexhub.org/stats/flexhub_reg.php?mode=reg&nick="..url.escape(nick).."&pass="..saltpass.."&address="..url.escape(tHubSettings.Main.sAddress)
		local ok,reply = pcall(http.request, request)
		if not ok then tFuncs:SafeDebug("[THREAD] [WebStatsRegRun] Error:  "..reply) return end
		tFuncs:SafeDebug("[THREAD] [WebStatsReg] Register reply: \n"..reply, "X")

		local ok, result = tWebStats:Results(reply)
		if ok then
			tFuncs:SafeDebug("Webstats registration successful! \n\n"..result)
			tFuncs:SafeExecute("SetCmd[\"set\"][\"cmd\"](\"AAAA\", \"!set webstats_enable true\", \"set\")")
			ThreadWait(10000)    --// Wait 10 seconds before attempting to register
			if not tThreads.WebStats then
				tThreads.WebStats = ThreadCreate(tWebStats.SendStats, {1, ssid, data})
			end
		else
			tFuncs:SafeDebug("Webstats registration failed: \n\nError: "..result)
		end
	elseif (mode == "mod") then
		key = url.escape(key)
		value = url.escape(value)
		hubid = url.escape(tHubSettings.Webstats.sId)
		token = url.escape(tHubSettings.Webstats.sToken)
		local request = "http://flexhub.org/stats/flexhub_reg.php?mode=mod&hubid="..hubid.."&token="..token.."&key="..key.."="..value
		local ok,reply = pcall(http.request, request)
		if not ok then tFuncs:SafeDebug("[THREAD] [WebStatsRegRun] Error:  "..reply) return end
		tFuncs:SafeDebug("[THREAD] [WebStatsReg] Modification reply: \n"..reply, "X")

		local ok, result = tWebStats:Results(reply)
		if ok then
			tFuncs:SafeDebug("Webstats modification successful! \n\n"..result)
		else
			tFuncs:SafeDebug("Webstats modification failed: \n\nError: "..result)
		end
	end
end,

["SendStats"]  = function(self, ssid, data)
	if data then
		local result, err = pcall(tWebStats.SendStatsRun, nil, ssid, data)
		if not result then
			tFuncs:SafeDebug("[THREAD] [WebStats] [SendStats] Error: "..err)
			tFuncs:UnlockMutexes()
		end
	else
		tFuncs:SafeDebug("Error: There's no data to send.")
	end
	tThreads.WebStats = nil
	ThreadDestroy()
end,

["SendStatsRun"] = function(self, ssid, data)
	if not tHubSettings.Webstats.sId or tHubSettings.Webstats.sId  == "" then
		tFuncs:SafeDebug("There's no webstats hub ID set, use !webstats_reg <forumnick> <forumpass> to register your hub for statistics storage.")
		return
	elseif not tHubSettings.Webstats.sToken or tHubSettings.Webstats.sToken == "" then
		tFuncs:SafeDebug("There's no webstats hub token set, use !webstats_reg <forumnick> <forumpass> to register your hub for statistics storage.")
		return
	end

	local addy = "http://flexhub.org/stats/send_flexstats.php"
	local data = "?hubid="..tHubSettings.Webstats.sId.."&token="..tHubSettings.Webstats.sToken..data

--	tFuncs:SafeDebug("Sending statistics to flexhub.org server: \n\n"..addy..data.."\n\n", "X")
	local ok, reqt = tFuncs:ToUTF8(addy..data)
	if not ok then
		tFuncs:SafeDebug("Conversion error when sending hub statistics to flexhub.org server: "..result, "FI", 1)
		return
	end

	local ok,reply = pcall(http.request, reqt)
	local result, resultdebug = "" , ""
	if ok and reply then
		ok, result, resultdebug = tWebStats:Results(reply)
		if ok then
			result = "Statistics received successfully. "..result
		else
			result = "Statistics denied by flexhub.org server: "..result
		end
	else
		if not reply then
			result = "Couldn't make contact with flexhub.org server, check your firewall settings."
		else
			result = "Thread error: "..tostring(reply)
		end
	end

	local debuglevel = 1
	if not ok then debuglevel = 0 end
	tFuncs:SafeDebug("Result of sending hub statistics to flexhub.org server: "..result, "FI", debuglevel)
end,

["Results"] = function(self, lines)
	local ok = false
	if not lines then
		return false, "No response from server."
	end
	local result, resultdebug = "", ""
	lines = lines.."\n"

	for line in lines:gmatch"[^\n]+" do
		if not line:find("^%s-%[") then
			line = line:gsub("\n", "")
			local code = line:match("^%s-(%d+)")
			if code == "0" then
				line = line:gsub("^%s-0%s+", "")
				ok = true
				--// Parse fields to see if something needs updating
				local changed = false
				for data in string.gfind(line, "(%S+)") do
					local key = data:gsub("=.+$", "")
					local value = data:gsub("^.-=", "")
					if key == "token" then tHubSettings.Webstats.sToken = value tFuncs:Debug("[WebStats] [Results] Token updated.", "X") changed = true end
					if key == "hubid" then tHubSettings.Webstats.sId = value tFuncs:Debug("[WebStats] [Results] Webstats ID updated.", "X") changed = true end
					if key == "statspage" then tHubSettings.Webstats.sPage = value result = result.."\n\nThis is your hub statistics page: "..value.."\n" changed = true end
					if key == "salt" then value = value:gsub("%s", "") return true, "salt", value end
				end
				if changed then tFuncs:SaveSettings("settings") end
			else
				result = result..line.."\n"
			end
		else
			resultdebug = resultdebug..line.."\n"
		end
	end
	if resultdebug and resultdebug ~= "" then
		tFuncs:SafeDebug("[Thread] Webstats parsed debugresult: \n"..resultdebug, "X")
	end

	return ok, result, resultdebug
end,

["GetSalt"] = function(self, ssid, nick, pass)
	local request = "http://flexhub.org/stats/flexhub_reg.php?mode=login&nick="..nick

	--// Send nick to retrieve salt
	local ok,reply = pcall(http.request, request)
	tFuncs:SafeDebug("[THREAD] [WebStats] [GetSalt] Login reply:\n"..reply, "X")
	if not ok then
		tFuncs:SafeDebug("[THREAD] [WebStats] [GetSalt] Error: \n"..reply)
		return false
	end

	local ok, reply, salt = tWebStats:Results(reply)
	if ok and reply == "salt" and salt then
		--// Salt received, now generate salted and hashed pass
		local together = string.lower(nick)..pass
		local hashpass = sha1(together)
		tFuncs:SafeDebug("[THREAD] [WebStats] [GetSalt] hashed nick & pass (forum pass): "..hashpass, "X")
		local saltpass = sha1(hashpass..salt)
		tFuncs:SafeDebug("[THREAD] [WebStats] [GetSalt] Salted and hashed hashpass: "..saltpass, "X")
		return ok, reply, saltpass
	else
		return false, reply
	end
end,

}
