--// $Id: Hub-ADC.lua 1465 2013-10-10 14:43:34Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tHubFiles["Hub-ADC.lua"] = {
	["sRev"] = "$Rev: 1465 $",
	["bRestart"] = false,
	["sPath"] = "./protocols/Hub-ADC.lua",
	}

--[[
How to simply connect to some ADC hub.

Socket Connection = Initiate_new_connection ( hostname, port) ; lets assume that we have a new ;TCP blocking socket to the hostname and port which should be the adc hub we want to connect to
if Connection=null
Abort ; oups, cant connect

Connection.write "HSUP ADBASE<cr>" ; this is the initial minimal greeting that the client should always ;send first
String Greeting = Connection.getLine ; getting the initial greeting, note that any getLine call should ;have a set timeout
; Greeting should be something like ISUP ADBASE
parsing... ; here, the string should be parsed to see what the hub supports
if not Greeting contains ADBASE
Abort ; doesnt support base

String mySID = Connection.getLine ; getting the line ISID mysid
parsing... ; get the SID from the mySID string
String Information = Connection.getLine ; getting the IINF
parsing... ; get the hubname and so on from the line, IINF NIhubname DEtopic ;VEhubsoftwareversion
;now need to send our own information
Connection.write "BINF " sid " ID" mycid " PD" mypid " NI" mynick " HN" my hubs normal " HO" my hubs op " HR" my hubs reg " VE" myclientversion " SU" whatmyclientsupports.. etc
String info=null
while info not equals myinf
newuserinfo= Connection.getLine; a line for each new user
;until you get your own info back,
;thats it, youre now in NORMAL state.

6.1. Client  Hub connection

Client 	                                       Hub
HSUP ADBASE ADTIGR 
										  ISUP ADBASE ADTIGR 
										  ISID <client-sid>
										  IINF HU1 HI1 
BINF <my-sid> ID PD
										  IGPA 
HPAS 
										  BINF <all clients>
										  BINF <Client-SID>

--~ function ADC_Join(client, sid)
--~ 	--// Send Join stuff to user
--~ 	client = tUser:Joined(client, sid)        --// send Welcome, MOTD, etc

--~ 	--// Announce user to all users
--~ 	tUser:Announce(sid)		--// Sends MyINFO / UserIP / Oplist to all users
--~
--~ 	--// Send Debug to GUI
--~ 	tFuncs:Debug(tUser:GetDataLine(sid).."BINF OK.", "LAJ", 2)
--~ 	return client
--~ end

--~ function ADC_CheckMyinf(client, sid, msg)
--~ 	if string.find(msg, " NI") then
--~ 		local _,_,nick = string.find(msg, " NI(%S+)")
--~ 		if not tNick:GetSid(nick) then
--~ 			return true
--~ 		else
--~ 			tUser:Disconnect(sid, "222 Nick <"..nick.."> is already in use, change it to something else before reconnecting.")
--~ 			return false
--~ 		end
--~ 	end
--~ end

-----------------
Protocol :
3.3. Message types

Message type specifies how messages should be routed and thus which additional fields can be found in the message header. Clients should use the most limiting type,
in terms of recipients, that makes sense for a particular message when sending it to the hub for distribution. Clients should use the message type only to aid in parsing
the message and otherwise ignore it. The following message types are defined:

B	Broadcast			Hub must send message to all connected clients, including the sender of the message.
C	Client message		Clients must use this message type when communicating directly over TCP.
D	Direct message		The hub must send the message to the target_sid user.
E	Echo message		The hub must send the message to the target_sid user and the my_sid user.
F	Feature broadcast	The hub must send message to all clients that support both all required (+) and no excluded (-) features named. The feature name is matched against the corresponding SU field in INF sent by each client.
H	Hub message			Clients must use this message type when a message is intended for the hub only.
I	Info message		Hubs must use this message type when sending a message to a client that didnt come from another client.
U	UDP message			Clients must use this message type when communicating directly over UDP.
-----------------------

]]--

function ADC_DataArrival(client, sid, msg)
	local wrong = "Malformed ADC message received."
	--tFuncs:Debug(tUser:GetDataLine(sid)..msg, "TAR", 9)

	local _,_,cmd = msg:find("^(%S%S%S%S)")

	--// Add stats to user and hub
	if not cmd then
		tUser:Disconnect(sid, wrong, "User sent ADC traffic without using a commandname, data: "..msg)
		return client
	end

	msg = msg:gsub("\n$", "")	-- remove last newline / test

	tFuncs:AddStats(sid, "ADC", cmd, 1, msg)

	local tIgnore_Commands = {
			["DUBD"] = 3,	-- Bundle upload notification (AirDC++) is not supposed to be sent to hub
			["DSCH"] = 3,	-- direct search (AirDC++)
		--	[""] = 1,
	}

	--// Block certain commands (regardless of type)
	local tBlocked_Commands = {
			["GPA"]= true,
			["ZON"] = true,
			["ZOF"] = true,
			["SID"] = true,
	}

	local cmdname = cmd:sub(2)
	if tBlocked_Commands[cmdname] then
		tUser:Disconnect(sid, wrong, "[ADC Blocked] Possible hack attempt, received "..cmd.." in state: "..client.iState.." / "..msg)
		return client
	end


	if ADC_Commands[cmd] then
		return ADC_Commands[cmd](client, sid, msg)
	else
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		tFuncs:Debug(tUser:GetDataLine(sid).."Client sent an unsupported ADC command: "..msg, "CA", (tIgnore_Commands[cmd] or 1))

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "XX", msg)
		if not res then
--~ 			tFuncs:Debug(tUser:GetDataLine(sid).."is flooding: "..msg)
			return client
		end

		-- // testing to partially support the adc rule of letting everything trough even if we do not support it
		local msgtype = string.sub(cmd, 1,1)
		if ADC_Leftovers[msgtype] and client.iState == 7 then
			return ADC_Leftovers[msgtype](client, sid, msg, cmd)
		elseif not client.iState == 7 then
		-- error on wrong state
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received "..cmd.." in state: "..client.iState.." / "..msg)
		end
	end
	return client
end



-- // this is a way to comply with the ADC protocol, while at the same time keep the FlexHub spirit of merging the two protocols
-- // ADC commands not yet added to FlexHub should now be able to work at least with all ADC users in the hub.

ADC_Leftovers = {
	["B"] = function(client, sid, msg, cmd)	-- broadcast			- DONE
		--// Validation
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", cmd, -1, msg) return client end

		-- send msg to all adc users ;)
		tSend:Raw("P_ADC", msg.."\n")
		return client
	end,
	["C"] = function(client, sid, msg, cmd)	-- client msg			- DONE
		-- not supposed to be sent to hub
		tUser:WrongProtocol(sid, "'C'-type messages should not be sent to the hub: "..msg)
		return client
	end,
	["D"] = function(client, sid, msg, cmd)	-- direct msg			- DONE
		--// Validation
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", cmd, -1, msg) return client end

		--// Does remote user exist?
		local tosid = msg:match("^%S%S%S%S %S+ (%S+) ")
		if not tosid or not tClient[tosid] or tClient[tosid].iState < 7 or not tClient[tosid].sProtocol == "ADC" then
			--tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline nick in "..cmd..": "..retmsg, "CA", 4)
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline/not ADC nick in "..cmd..": "..retmsg, "X", 2)
		else
			tSend:Raw(tosid, msg.."\n")
		end
		return client
	end,
	["E"] = function(client, sid, msg, cmd)	-- echo msg				- DONE
		--// Validation
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", cmd, -1, msg) return client end

		--// Does remote user exist?
		local _,_,fromsid, tosid, pmsid = msg:find("^%S%S%S%S (%S+) (%S+) %S+ (%S+)")

		if not tosid or not tClient[tosid] or tClient[tosid].iState < 7 or not tClient[tosid].sProtocol == "ADC" then
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline/not ADC nick in "..cmd..": "..retmsg, "X", 2)
		elseif fromsid ~= pmsid then
			tFuncs:Debug(tUser:GetDataLine(sid).."Echo msg with no matching fromsid/pmsid"..cmd..": "..retmsg, "X", 2)
		else
			tSend:Raw(tosid, msg.."\n")
			tSend:Raw(fromsid, msg.."\n")
		end
		return client
	end,
	["F"] = function(client, sid, msg, cmd)	-- feature broadcast	- Done
		--// Validation
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", cmd, -1, msg) return client end

		local features, retmsg = msg:match("^%S%S%S%S %S+ (%S+) (.+)$")

		--/ Execution
		local has,hasnot = tFuncs:ParseFeatures(features)
		for vsid, vdata in pairs(tClient) do
			if vdata.iState == 7 and vdata.sProtocol == "ADC" then
				if tUser:HasFeatures(vsid, vdata.tSupportsInf, has, hasnot) then
					tSend:Raw(vsid, msg.."\n")
				end
			end
		end
		return client
	end,
	["H"] = function(client, sid, msg, cmd)	-- hub msg				- not so much
		-- since we dont support these, lets just send them to feed for now

		tUser:WrongProtocol(sid, "'H'-type messages should not be sent to the hub: "..msg)
--//		tFuncs:Debug(tUser:GetDataLine(sid).."Unknown H-type msg received :"..cmd..", Msg: "..msg, "X", 2)
		return client
	end,
	["I"] = function(client, sid, msg, cmd)	-- info msg				- not so much
		-- since we dont support these, lets just send them to feed for now
		tUser:WrongProtocol(sid, "'I'-type messages should not be sent to the hub: "..msg)
--//		tFuncs:Debug(tUser:GetDataLine(sid).."Unknown I-type msg received :"..cmd..", Msg: "..msg, "X", 2)
		return client
	end,
	["U"] = function(client, sid, msg, cmd)	-- udp					- not so much
		-- since we dont support these, lets just send them to feed for now
		tUser:WrongProtocol(sid, "'U'-type messages should not be sent to the hub: "..msg)

--//		tFuncs:Debug(tUser:GetDataLine(sid).."Unknown U-type msg sent :"..cmd..", Msg: "..msg:sub(1,45), "X", 2)
		return client
	end,
}

ADC_Commands = {
	["DPSR"] = function(client, sid, msg)
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DPSR in state: "..client.iState.." / "..msg)
			return client
		end
--		local retmsg = msg

		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DPSR", -1, msg) return client end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "XX", msg)
		if not res then
			tFuncs:AddStats(sid, "ADC", "DPSR", -1, msg)
			return client
		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "NT", smsg) end

		--/ Execution
		if res and tClient[tosid] and tClient[tosid].iState == 7 and tClient[tosid].sProtocol == "ADC" then
			if tClient[tosid].tSupports["PFSR"] then
				tSend:Raw(tosid, retmsg.."\n", nil, nil, "Delay")
			end
		else
			tFuncs:AddStats(sid, "ADC", "DPSR", -1, msg)
			--// Remote client isn't online, skip
		end

		return client
	end,
	["HGUI"] = function(client, sid, msg)
--		--// Check for control chars
--		local res
--		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		tFuncs:Debug("[GUI] Received data: "..msg)
		if client.sProfile == "Gui" then
			local _,_,cmd,smsg = msg:find("^HGUI (..)(.*)$")
			if ADC_GuiCmds[cmd] then
				ADC_GuiCmds[cmd](client, sid, smsg)
			end
		else
			tFuncs:AddStats(sid, "ADC", "HGUI", -1, msg)
			tUser:Disconnect(sid, wrong, "Hack attempt: using GUI commands is not allowed:"..msg)
			return client
		end
		return client
	end,
	["HSTA"] = function(client, sid, msg)
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Statusmessage from client, ignored
		tFuncs:Debug(tUser:GetDataLine(sid).."sends ADC Status message: "..msg, "CA", 9)

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "XX", msg)
		if not res then
			tFuncs:AddStats(sid, "ADC", "HSTA", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."is flooding: "..msg)
			return client
		end

		return client
	end,
	["HSUP"] = function(client, sid, msg)
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

--		local DoSUPArrival = function(sid, msg) return true end     --// TODO: replace this with an actual function

		local tSup = {
		 --// supported
		["BAS0"] = 1, ["BASE"] = 1, ["TIGR"] = 1, ["UCMD"] = 1, ["UCM0"] = 1, ["PING"] = 1, ["ZLIF"] = 1, ["NAT0"] = 1,
		["DHT0"] = 1, ["FGUI"] = 1, ["PFSR"] = 1, ["ALIV"] = 1, --[""] = 1, [""] = 1, [""] = 1,

		-- // supported but not allowed at this stage
		["ORLY"] = 0,	-- the right to use ORLY/HRLY is set on profile, we dont know which profile the user has until after a complete login

		-- // not supported
		["BLO0"] = 0, ["HBRI"] = 0,-- [""] = 0, [""] = 0,
		}
		-- HBRI : (hybrid ipv4 and ipv6 support) http://www.dcbase.org/forums/viewtopic.php?f=55&t=771&p=2860&hilit=HBRI#p2860

		-- must be in state 1 to be able to start login process, this is set by TCPWait upon a new clientconnection
		if client.iState == 1 then
			if string.find(msg, "ADBASE", 1, true) or string.find(msg, "ADBAS0", 1, true) then -- must at least support this
				--if DoSUPArrival(sid, msg) then	-- enable this back again if DoSUPArrival ever makes it into a real function (not sure what its supposed to do anyway, fredo)
--~ 						tFuncs:Debug(tUser:GetDataLine(sid).."Supports: "..msg)
					--// Test sending statusmessage
					--	tSend:Raw(sid,"ISTA 000 "..ADC_E("Received HSUP: "..msg).."\n")

					--// Send back what hub supports
					local supstr = msg:gsub("HSUP ", "")
					client.sSupports = ""
					for w in supstr:gmatch("AD(%S+)") do
						if tSup[w] and tSup[w] == 1 then
							client.sSupports = client.sSupports..w..","
							client.tSupports[w] = true
						elseif not tSup[w] then
							tFuncs:Debug("[XXX] unknown HSUP received from "..(client.sNick or client.sNickTemp or "unknown nick") .." : "..w, "X")
						end
					end
					client.sSupports = client.sSupports:sub(1,-2)

					--// Remove BAS0 if BASE is supported
					if client.tSupports["BASE"] and client.tSupports["BAS0"] then
						client.tSupports["BAS0"] = nil
					end

					local supsend = {"ISUP"}
					for name, value in pairs(client.tSupports) do
						if value then table.insert(supsend, "AD"..name) end
					end
					tSend:Raw(sid, table.concat(supsend, " ").."\n")

					--// Send generated SID
					tSend:Raw(sid, "ISID "..sid.."\n")

					--// Pinger connected
					if client.tSupports["PING"] then
						if tHubSettings.tRules.bRegOnly and not tHubSettings.tRules.bAllowPinger then
							tUser:Disconnect(sid, "222 This hub is for registered users only, pingers not allowed.")
						elseif not tHubSettings.tRules.bAllowPinger then
							tUser:Disconnect(sid, "222 Pingers are not allowed.")
						else
							local inf = "IINF"..
								" NI"..ADC_E(tHubSettings.Main.sName)..
								(#tHubSettings.Main.sTopic > 0 and " DE"..ADC_E(tHubSettings.Main.sTopic) or "")..
	--~ 							" HHadc://"..ADC_E(tHubSettings.Main.sAddress)..
								(#tHubSettings.Main.sOwner > 0 and " OW"..ADC_E(tHubSettings.Main.sOwner) or "")..
								(#tHubSettings.Main.sWebsite > 0 and " WS"..ADC_E(tHubSettings.Main.sWebsite) or "")..
								(#tHubSettings.Main.sNetwork > 0 and " NE"..ADC_E(tHubSettings.Main.sNetwork) or "")..

								" VE"..ADC_E(_APPNAME.." ".._APPVER)..
								" UP"..(os.time()-tStats.Hub.Started)..
								" UC"..tStats.Hub.Users.Total..
								" SS"..(tFuncs:GetTotalShare("ADC") or 0)..

								" ML"..tHubSettings.tRules.iMinSlots..
								" XL"..tHubSettings.tRules.iMaxSlots..
								" MS"..tHubSettings.tRules.iMinShare..
								" XS"..tHubSettings.tRules.iMaxShare..
								" XU"..tHubSettings.tRules.iMaxHubs..
								" MC"..tHubSettings.tRules.iMaxUsers..
								" CT32\n"
							tSend:Raw(sid, inf)

	--~ 						tFuncs:Debug("Pinger connected, sending: "..inf, "FI", 0)
							--//[12:42:55] <FlexBot> [FI] [2] Pinger connected, sending: IINF NIFlexHub\sSupport DEImportant\ssecurity\supdate:\ssvn\s1055,\suse\s'!update\snow'\sin\syour\shub OWFlipFlop\sand\sDaywalker VEFlexHubBeta\s0.1 UP130237 UC88 SS0 ML0 XL999 MS0 XS10995116277760 XU300 MC5000 CT32 HU1 HI1
						end
					else
						local failover = (tHubSettings.tRedirects.sFailOver and tHubSettings.tRedirects.sFailOver ~= "" and " FO"..tHubSettings.tRedirects.sFailOver) or ""

						if tHubSettings.tRules.bRegOnly then
							tSend:Raw(sid, "IINF NI\\s"..failover.." CT32 HU1 HI1\n")		--// Send empty hubname first, after successful password, full IINF will be sent
						else
							tSend:Raw(sid, "IINF NI"..ADC_E(tHubSettings.Main.sName).." DE"..ADC_E(tHubSettings.Main.sTopic).." VE"..ADC_E(_APPNAME.." ".._APPVER.." ".._REV)..failover..(" CT32 HU1 HI1\n"))
						end
					end

					--// Send the hubbot first, so all other messages can use it:
					tSend:MyInf("AAAA", sid)	-- Hubbot

					--// Update to iState 2: VERIFY
					client.iState = 2
					tFuncs:Debug(tUser:GetDataLine(sid).."HSUP OK.", "LAJ", 2)
			--	else
			--		tFuncs:Debug(tUser:GetDataLine(sid).."HSUP was denied by DoSUPArrival: "..msg, nil, 1)
			--		tFuncs:AddStats(sid, "ADC", "HSUP", -1, msg)
			--		tUser:Disconnect(sid, "222 Supports failed")
			--		return client
			--	end
			else
				tUser:Redirect(sid, "Your client doesn't support both BASE and TIGR, update your client", "adc_disabled")
				return client
			end

		--// Handle new support announcements
		elseif client.iActive > 0 then
			tFuncs:Debug(tUser:GetDataLine(sid).."[HSUP] Client sent new HSUP after login: "..msg, "LAJ", 2)

			local supstr = msg:gsub("HSUP ", "")
			local sSupports = ""
			local tSupports = {}
			for w in supstr:gmatch("AD(%S+)") do
				if tSup[w] and tSup[w] == 1 then
					sSupports = sSupports..w..","
					tSupports[w] = true
				elseif not tSup[w] then
					tFuncs:Debug("[XXX] unknown HSUP received from "..(client.sNick or client.sNickTemp or "unknown nick") .." : "..w, "X")
				end
			end
			sSupports = sSupports:sub(1,-2)

			--compare
			if client.sSupports == sSupports then
				tFuncs:Debug(tUser:GetDataLine(sid).."[HSUP] No change, HSUP ignored", "LAJ", 4)
				-- no change, we can ignore this update
			else
				--loop current to check whats removed
				for i,j in pairs(client.tSupports) do
					if not tSupports[i] then
						tFuncs:Debug(tUser:GetDataLine(sid).."[HSUP] Removed :"..i, "X", 4)
						client.tSupports[i] = nil
					end
				end

				--loop new to check whats added
				for i,j in pairs(tSupports) do
					if not client.tSupports[i] then
						tFuncs:Debug(tUser:GetDataLine(sid).."[HSUP] Added :"..i, "X", 4)
						client.tSupports[i] = true
					end
				end

				--// Remove BAS0 if BASE is supported
				if client.tSupports["BASE"] and client.tSupports["BAS0"] then
					client.tSupports["BAS0"] = nil
				end

				--after update, check so we still have adbase
				if (not client.tSupports["BASE"] and not client.tSupports["BAS0"]) or not client.tSupports["TIGR"] then
					tUser:Redirect(sid, "Your client doesn't support both BASE and TIGR, update your client", "adc_disabled")
					return client
				else
					tFuncs:Debug(tUser:GetDataLine(sid).."[HSUP] Supports updated", "LAJ", 4)
					-- we already have tSupports updated so only sSupports left
					client.sSupports = sSupports

					local supsend = {"ISUP"}
					for name, value in pairs(client.tSupports) do
						if value then table.insert(supsend, "AD"..name) end
					end
					tSend:Raw(sid, table.concat(supsend, " ").."\n")
				end
			end
		end
		return client
	end,
	["DRNT"] = function(client, sid, msg)
		--// Check for control chars
		--local res
		--res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

			--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DNAT in state: "..client.iState.." / "..msg)
			return client
		end

		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DRNT", -1, msg) return client end

		local tosid, protocol, port, token = msg:match("DRNT %S+ (%S+) (%S+) (%S+) (%S*)")
		if not tosid or not protocol or not port or not token then
			tFuncs:AddStats(sid, "ADC", "DRNT", -1, msg)
			tUser:Disconnect(sid, wrong, "Received a malformed DRNT: "..msg)
			return client
		end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "RC", msg, false, 0.5)
		if not res then
			tFuncs:AddStats(sid, "ADC", "DRNT", -1, msg)
			return client
		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "NT", smsg) end

		--/ Execution
		if tClient[tosid] and tClient[tosid].iState == 7 and tClient[tosid].sProtocol == "ADC" then
			local toinf = tClient[tosid].tSupportsInf
			if tUser:HasFeatures(tosid, toinf, {"NATT"}) or tUser:HasFeatures(tosid, toinf, {"NAT0"}) then
				tSend:Raw(tosid, retmsg.."\n", nil, nil, "Delay")
			else
				tFuncs:Debug(tUser:GetDataLine(sid).."DRNT blocked, remote client doesn't support NAT0 or NATT: "..msg, "X")
				tFuncs:AddStats(sid, "ADC", "DRNT", -1, msg)
			end
		else
			--// Remote client isn't online, or is NMDC, skip
			tFuncs:AddStats(sid, "ADC", "DRNT", -1, msg)
		end

		return client
	end,
	["DNAT"] = function(client, sid, msg)
		--// Check for control chars
		--local res
		--res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

			--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DNAT in state: "..client.iState.." / "..msg)
			return client
		end

		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DNAT", -1, msg) return client end

		local tosid, protocol, port, token = msg:match("DNAT %S+ (%S+) (%S+) (%S+) (%S*)")
		if not tosid or not protocol or not port or not token then
			tFuncs:AddStats(sid, "ADC", "DNAT", -1, msg)
			tUser:Disconnect(sid, wrong, "Received a malformed DNAT: "..msg)
			return client
		end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "XX", msg)
		if not res then
			tFuncs:Debug(tUser:GetDataLine(sid).."DNAT blocked by Deflood: "..msg, "X")
			tFuncs:AddStats(sid, "ADC", "DNAT", -1, msg)
			return client
		end

		--// Does remote user exist?
		if not tClient[tosid] or tClient[tosid].iState < 7 or not tUser:CanDownload(tosid) then
			tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."DNAT blocked: target user offline or can't download: "..retmsg, "CA", 4)
			return client
		end


		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "NT", smsg) end

		--/ Execution
		if tClient[tosid] and tClient[tosid].iState == 7 and tClient[tosid].sProtocol == "ADC" then
			local toinf = tClient[tosid].tSupportsInf
			if tUser:HasFeatures(tosid, toinf, {"NATT"}) or tUser:HasFeatures(tosid, toinf, {"NAT0"}) then
				tSend:Raw(tosid, retmsg.."\n", nil, nil, "Delay")
			else
				tFuncs:Debug(tUser:GetDataLine(sid).."DNAT blocked, remote client doesn't support NAT0 or NATT: "..msg, "X")
			end
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."DNAT blocked, remote client isn't level 7 or ADC: "..msg, "X")
			tFuncs:AddStats(sid, "ADC", "DNAT", -1, msg)
			--// Remote client isn't online, skip
		end

		return client
	end,
	["DSTA"] = function(client, sid, msg)		--~ DSTA O6XD ET46 240 IP\sunknown
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Basic validation
		if client.iState ~= 7 then
			tFuncs:AddStats(sid, "ADC", "DSTA", -1, msg)
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DSTA in state: "..client.iState.." / "..msg)
			return client
		end

		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DSTA", -1, msg) return client end

		local tosid, errortotal, description = msg:match("DSTA %S+ (%S+) (%d+)%s-(%S*)")
		if not tosid or not errortotal then
			tFuncs:AddStats(sid, "ADC", "DSTA", -1, msg)
			tUser:Disconnect(sid, wrong, "Received a malformed DSTA: "..msg)
		end
		local errorcode = tonumber(string.sub((errortotal or "000"), 2))

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "ST", msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DSTA", -1, msg) return client end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

		--/ Execution
--		if res and tClient[tosid] and tClient[tosid].iState == 7 and tClient[tosid].sProtocol == "ADC" then
		if tClient[tosid] and tClient[tosid].iState == 7 and tClient[tosid].sProtocol == "ADC" then
			--// Skip any client-client communication errors (40-54) since no client <= dc++ 0.75 does anything with it except flood mainchat.
			if errorcode < 40 or errorcode > 54 then
				tSend:Raw(tosid, msg.."\n", nil, nil, "Delay")
			end
		elseif tClient[tosid] and tClient[tosid].sProtocol == "NMDC" then
			tFuncs:AddStats(sid, "ADC", "DSTA", -1, msg)
			--// Other client is NMDC, is there anything translatable?
		else
			tFuncs:AddStats(sid, "ADC", "DSTA", -1, msg)
			--// Remote client isn't online, skip
		end

		return client
	end,
	["DRCM"] = function(client, sid, msg)		--~ DCTM E33T YLFV ADC/0.10 1382 331560749
--		--// Check for control chars
--		local res
--		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Basic validation
		if client.iState ~= 7 then
			tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg)
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DRCM in state: "..client.iState.." / "..msg)
			return client
		end
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg) return client end

		local tosid, protocol,token = msg:match("DRCM %S+ (%S+) (%S+)%s-(%S*)")
		if not tosid or not protocol or not token then
			tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg)
			tUser:Disconnect(sid, wrong, "Received a malformed DRCM: "..msg)
			return client
		end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "RC", msg, false, 0.5)
		if not res then tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg) return client end

		--// Does remote user exist?
		if not tClient[tosid] or tClient[tosid].iState < 7 then
			tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline nick in DRCM: "..retmsg, "CA", 4)
			return client
		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

		--// Share hidden
		if tClient[tosid].bShareHidden then
			return client
		end

		--/ Execution
		if tClient[tosid].sProtocol == "ADC" and tClient[tosid].bActive then
--~ 			tSend:Raw(tosid, "BINF "..sid.." I4"..client.sIP.."\n")
			tSend:Raw(tosid, retmsg.."\n")
		elseif tClient[tosid].sProtocol == "ADC" then
			--// Passive ADC client: if both clients support NAT0 or NATT
			local frominf  = tClient[sid].tSupportsInf
			local toinf = tClient[tosid].tSupportsInf
			if (tUser:HasFeatures(sid, frominf, {"NATT"}) or tUser:HasFeatures(sid, frominf, {"NAT0"})) and (tUser:HasFeatures(tosid, toinf, {"NATT"}) or tUser:HasFeatures(tosid, toinf, {"NAT0"})) then
				tSend:Raw(tosid, retmsg.."\n", nil, nil, "Delay")
				tFuncs:Debug(tUser:GetDataLine(sid).."DRCM sent to NATT enabled passive user: "..msg, "CA", 4)
			else
				tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg)
				tFuncs:Debug(tUser:GetDataLine(sid).."DRCM to passive user attempted, message dropped: "..msg, "CA", 4)
				return client
			end
		else
			tFuncs:AddStats(sid, "ADC", "DRCM", -1, msg)
			tSend:Raw(sid, "DSTA "..tosid.." "..sid.." 241 "..ADC_E("You can't download from <"..tClient[tosid].sNick.."> because it's using a different protocol (NMDC).").." TO"..token.." PR"..protocol.."\n", nil, nil, "Delay")
			tFuncs:Debug(tUser:GetDataLine(sid).."DRCM to NMDC user attempted, message dropped: "..msg, "CA", 5)
			return client
		end

		return client
	end,
	["DCTM"] = function(client, sid, msg)
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--~ DCTM E33T YLFV ADC/0.10 1382 331560749

		--// Basic validation
		if client.iState ~= 7 then
			tFuncs:AddStats(sid, "ADC", "DCTM", -1, msg)
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DCTM in state: "..client.iState.." / "..msg)
			return client
		end

		local res = tCheck:Sid(sid, msg)
		if not res then return client end

		local tosid, protocol, port, token = msg:match("DCTM %S+ (%S+) (%S+) (%S+)%s-(%S*)")
		if not token or not port or not protocol or not tosid then
			tFuncs:AddStats(sid, "ADC", "DCTM", -1, msg)
			tUser:Disconnect(sid, wrong, "Received a malformed DCTM: "..msg)
			return client
		end

		client.sActivePort = port
		client.sActiveProtocol = protocol

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "CT", msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DCTM", -1, msg) return client end

		--// Does remote user exist?
		if not tClient[tosid] or tClient[tosid].iState < 7 then
			tFuncs:AddStats(sid, "ADC", "DCTM", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline nick in DCTM: "..retmsg, "CA", 4)
			return client
		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

		--// Share hidden
		if tClient[tosid].bShareHidden then
			tFuncs:AddStats(sid, "ADC", "DCTM", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Trying to download a hidden share, DCTM dropped: "..retmsg, "CN", 4)
			return client
		end

		--/ Execution
		if tClient[tosid].sProtocol == "ADC" then
--~ 			tSend:Raw(tosid, "BINF "..sid.." I4"..client.sIP.."\n")
			tSend:Raw(tosid, retmsg.."\n")
		else
			tFuncs:AddStats(sid, "ADC", "DCTM", -1, msg)
			tSend:Raw(sid, "DSTA "..tosid.." "..sid.." 241 "..ADC_E("You can't download from <"..tClient[tosid].sNick.."> because it's using a different protocol (NMDC).").." TO"..token.." PR"..protocol.."\n", nil, nil, "Delay")
			tFuncs:Debug(tUser:GetDataLine(sid).."DCTM to NMDC user attempted, message dropped: "..msg, "CA", 5)
		end

		return client
	end,
	["BINF"] = function(client, sid, msg)
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Split and assign
		local params = tFuncs:Split(msg, " +")
		local vsid = params[2]
		local retmsg = params[3]
		local rest = table.concat(params, " ",4)

		local res = tCheck:Sid(sid, msg)
		if not res then return client end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "MA", msg)
		if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) return client end

		if client.iState == 2 then
			local nick = tFuncs:Find(params, "^NI(%S+)$", 3)
			local cid = tFuncs:Find(params, "^ID(%S+)$", 3)
			local pid = tFuncs:Find(params, "^PD(%S+)$", 3)

			if not nick and tFuncs:Find(params, "^NI", 3) then
				tFuncs:AddStats(sid, "ADC", "BINF", -1, msg)
				tUser:Disconnect(sid, "Your nickname is missing from your BINF login information, please use alphanumeric characters for your nickname.", "[BINF] Nick was empty", "LI", 2)
				return client
			elseif not cid or not pid or not nick then
				tFuncs:AddStats(sid, "ADC", "BINF", -1, msg)
				tUser:Disconnect(sid, wrong, "Missing information in Login BINF: "..msg)
				return client
			end

			--// Check if nick is on Monitor
			if nick and tMonitor[nick] then
				client.sMonitor = tMonitor[nick]
		--		tFuncs:Debug("Monitored nick is connecting "..tUser:GetShortData(sid)..", all traffic outputs to: "..tMonitor[nick])
				tFuncs:Debug("Monitored nick is connecting <"..(nick or "nick unknown").."> IP: "..(client.sIP or "IP unknown").." on ADC, all traffic outputs to: "..tMonitor[nick])
			end

			--// Check nick
			local res, client, reg, err, inuse = tCheck:Nick(client, sid, nick)		--// Ban / pattern and IP-binding check
--~ 			if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) return client end			--// This caused a halted login process when res was false due to inuse, the following 'if' handles the res
		--	tFuncs:Debug("[BINF] iState : "..client.iState, "X")

			--// Registered user
			if res and reg then
				--// Set MyINFO for handling after passcheck
--~ 				tFuncs:Debug(tUser:GetDataLine(sid).."BINF: User is regged, delaying BINF parsing", "X")
				client.tInfo.tADC.sOrig = msg
				client.tInfo.tADC.sCID = cid
				client.tInfo.tADC.sPID = pid
				client.sPassSalt = tFuncs:MakeSalt() --// Store random for comparison on pass-arrival

				if tSecurity.General.AllowBlankPass and reg.sPassword == "" then	--// CHECK
					client.bNotifySetOwnPass = true
				else
					tSend:Raw(sid, "IGPA "..client.sPassSalt.."\n")
				end

				--// Set hidden and sharehidden status
				if reg.bHidden then
					client.bHidden = true
				end

				if reg.bShareHidden or (tProfiles[reg.sProfile] and tProfiles[reg.sProfile].bShareHiddenForced) then
					client.bShareHidden = true
				end

				-- // new stuff for blankpass, since a pass request wont be sent, we jump directly to HPAS from here
				if client.bNotifySetOwnPass then
					client = ADC_Commands.HPAS(client, sid, msg)
				end

				--// add support for ORLY if allowed by profile
				if tProfiles[reg.sProfile] and tProfiles[reg.sProfile].bAllowedToRelay then
					client.sSupports = client.sSupports..",ORLY"
					client.tSupports["ORLY"] = true
					tSend:Raw(sid, "ISUP ADORLY\n")
				end

			--// Unregistered user
			elseif res and not inuse then
				--// Check CID
				local res, client = tCheck:Cid(client, sid, pid, cid)
				if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) return client end
				client.iProfile = -1
			elseif inuse then
				tFuncs:AddStats(sid, "ADC", "BINF", -1, msg)
				tUser:Disconnect(sid, "Nickname <"..nick.."> is in use, change nickname before reconnecting", "Nick <"..nick.."> in use", "FI", 2)
				return client
			else
				tFuncs:AddStats(sid, "ADC", "BINF", -1, msg)
				if not tBuffer[sid].bDisconnect then tUser:Disconnect(sid, err) end
				return client
			end

			--// Login is now complete only for non-regs, full login for regs after password check
			if not reg then
				--// Set MyINFO
				local res, client, err = tInfo:ParseADC(client, sid, msg, true)	--// 4th argument set to true means a full BINF is needed
				if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) tUser:Disconnect(sid, wrong, err) return client end

				--// Check rules
				res, retmsg = tCheck:Rules(sid)
				if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) return client end

				--// Send Join stuff to user
				tUser:Joined(client, sid)	--// send Welcome, MOTD, etc

				--// Announce user to all users
				tUser:Announce(sid)		--// Sends MyINFO / UserIP / Oplist to all users
			end

			--// Send Debug to GUI
			tFuncs:Debug(tUser:GetDataLine(sid).."BINF OK.", "LAJ", 4)

			return client
		elseif client.iState == 7 then
			--// Set MyINFO
			local res, client, err = tInfo:ParseADC(client, sid, msg)	--// 4th argument set to true means a full BINF is needed
			if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) tUser:Disconnect(sid, wrong, err) return client end

			--// Check rules
			res, retmsg = tCheck:Rules(sid)
			if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) return client end

			tSend:MyInfPool(sid)

			tFuncs:Debug(tUser:GetDataLine(sid).."BINF update OK, resent to all users if myinfointerval has passed.", "TA", 4)
			return client
		else
			tFuncs:AddStats(sid, "ADC", "BINF", -1, msg)
			tUser:Disconnect(sid, wrong, "Malformed Login BINF: "..msg)
			return client
		end
		return client
	end,
	["BMSG"] = function(client, sid, msg)
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// BMSG AAAA <msg> <var1> <var2>

		--// Split and assign
		local params = tFuncs:Split(msg, " +")
		local vsid = params[2]
		local retmsg = params[3]
		local rest = table.concat(params, " ",4)

		--// Basic validation
		local res = tCheck:State(client, sid, 7, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg) return client end

		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg) return client end

		if not vsid or not retmsg then
			tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg)
			tUser:Disconnect(sid, wrong, "Invalid BMSG, user disconnected: "..msg)
			return client
		end

		if retmsg == "" then	-- ugly workaround to avoid line going completely missing
			retmsg = "?"
		--	tFuncs:Debug("[ToLocale trigger] "..#retmsg .."\t"..retmsg, "X", 3)
		end

		retmsg = ADC_E(retmsg, 1)

		local me = false
		if tFuncs:Find(params, "^ME1$", 4) then
			me = true
		end
		if retmsg:find("^[/%+][Mm][Ee] ") then
			me = true
			retmsg = retmsg:gsub("^[/%+][Mm][Ee] ", "")
		end

		--// Deflood check
		if retmsg then
			if retmsg:find("^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found
				res, retmsg = tDeflood:Main(sid, "HC", retmsg)
				if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg)  tFuncs:AddStats(sid, "ADC", "HC", -1, msg) return client end
			else
				tFuncs:AddStats(sid, "ADC", "MC", 1, msg)
				res, retmsg = tDeflood:Main(sid, "MC", retmsg)
				if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg)  tFuncs:AddStats(sid, "ADC", "MC", -1, msg) return client end
				if res then res, retmsg = tCheck:Pattern(sid, "MC", retmsg) end
				if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg)  tFuncs:AddStats(sid, "ADC", "MC", -1, msg) return client end
				if res then res, retmsg = tCheck:Filter(sid, "MC", retmsg) end
				if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg)  tFuncs:AddStats(sid, "ADC", "MC", -1, msg) return client end
			end
		else
			tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg)
			tUser:Disconnect(sid, wrong, "No content found in BMSG: "..msg)
			return client
		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end
--~ 		if not res then tFuncs:AddStats(sid, "ADC", "BMSG", -1, msg) return client end

		--// Execution
		client = tSend:Mainchat(client, sid, retmsg, me)

		return client
	end,
	["BSCH"] = function(client, sid, msg)
--		--// Check for control chars
--		local res
--		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// WTF!! :  BSCH ET46 LE367001600 ANsome ANmovie EXmpg EXmp4 EXflv EXmov EXasf EXavi EXpxp EXwmv EXogm EXmkv EXm1v EXm2v EXmpe EXmps EXmpv EXram EXvob EXrm EXdivx EXmpeg EXmp1v EXmp2v EXmpv1 EXmpv2 EXqt EXrv EXvivo EXts TO608314282
		--// additional FT (filetype) will save loads of searchtraffic: FTaudio FTvideo FTimage FTall FTdoc FTexe FTcomp FTdir FTtth

		if tHubSettings.Feed.bShowSearches then
			tFuncs:Debug(tUser:GetDataLine(sid)..msg, "SAP", 9)
		end

		--// Basic validation
		if client.iState ~= 7 then
			tFuncs:AddStats(sid, "ADC", "BSCH", -1, msg)
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received BSCH in state: "..client.iState.." / "..msg)
			return client
		end
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "BSCH", -1, msg) return client end

		local retmsg = msg:match("^BSCH %S+ (.+)$")
--~ 		local anwords = ""
--~ 		local nowords = ""

--~ 		local token = string.match(retmsg, " (TO%S+)") or ""
--~ 		local ext = string.match(retmsg, " (EX%S+)") or ""
--~ 		for word in string.gmatch(retmsg, " AN(%S+)") do
--~ 			anwords = anwords..ADC_E(word, 1).." "
--~ 		end
--~ 		for word in string.gmatch(retmsg, " NO(%S+)") do
--~ 			nowords = nowords..ADC_E(word, 1).." "
--~ 		end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "SP", retmsg)
		if not res then tFuncs:AddStats(sid, "ADC", "BSCH", -1, msg) return client end

		--// Pattern check
		local search = ""
		for word in retmsg:gmatch("%sAN(%S+)") do
			search = search..word.." "
		end
--~ 		retmsg = retmsg:gsub("(%sAN%S+)", "")
--~ 		local search = search:gsub(" $", "")	--// Remove trailing space
		local res, search = tCheck:Pattern(sid, "SE", search)
		if not res then tFuncs:AddStats(sid, "ADC", "BSCH", -1, msg) return client end

--~ 		--// Rebuild?
--~ 		for word in search:gmatch("%S+") do
--~ 			retmsg = retmsg.." AN"..word
--~ 		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

--[[		--// Rebuild searchline
		local search = "BSCH "..sid.." "
		for word in string.gmatch(anwords, "(%S+) ") do
			search = search.."AN"..ADC_E(word).." "
		end
		for word in string.gmatch(nowords, "(%S+) ") do
			search = search.."NO"..ADC_E(word).." "
		end
		search = search..ext..token.."\n"
		tFuncs:Debug(tUser:GetDataLine(sid).."Converted BSCH from: "..msg.." to: "..search, "CA", 1)
]]--

		--/ Execution
		if res then tSend:Raw("P_ADC", msg.."\n", sid, nil, "Delay") end
		return client
	end,
	["FSCH"] = function(client, sid, msg)
--~ 		[2011/08/11 01:33:35] <FlexBot> [Monitoring <AKSUNAKSU>] [RCV] FSCH QTZ4 +TCP4-NAT0 TOauto TREXWS2GKKO26BMZMO757XNFYZKRBWMQZF74RJZJA
--~ 		[2011/08/11 01:33:35] <FlexBot> [Monitoring <AKSUNAKSU>] [RCV] FSCH QTZ4 +NAT0 TOauto TREXWS2GKKO26BMZMO757XNFYZKRBWMQZF74RJZJA

--		--// Check for control chars
--		local res
--		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--//BSCH A23L AN02 ANyou ANwere ANmeant ANfor ANme TO236229712

		if tHubSettings.Feed.bShowSearches then
			tFuncs:Debug(tUser:GetDataLine(sid)..msg, "SAA", 9)
		end

		--// Basic validation
		if client.iState ~= 7 then
			tFuncs:AddStats(sid, "ADC", "FSCH", -1, msg)
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received FSCH in state: "..client.iState.." / "..msg)
			return client
		end
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "FSCH", -1, msg) return client end

		local features, retmsg = msg:match("^FSCH %S+ (%S+) (.+)$")
		if not retmsg or not features then
			tFuncs:AddStats(sid, "ADC", "FSCH", -1, msg)
			tUser:Disconnect(sid, wrong, "[FSCH] Malformed search: "..msg)
			return client
		end
		retmsg = ADC_E(retmsg, 1) --// Why?

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "SA", retmsg, false, 0.5)
		if not res then tFuncs:AddStats(sid, "ADC", "FSCH", -1, msg) return client end

		--// Pattern check
		local search = ""
		for word in retmsg:gmatch("%sAN(%S+)") do
			search = search..word.." "
		end
		if res then res, search = tCheck:Pattern(sid, "SE", search) end
		if not res then tFuncs:AddStats(sid, "ADC", "FSCH", -1, msg) return client end


		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

		--/ Execution
		if res then
			local has,hasnot = tFuncs:ParseFeatures(features)
			for vsid, vdata in pairs(tClient) do
				if vdata.iState == 7 and vdata.sProtocol == "ADC" and vdata.iShareSize and vdata.iShareSize > 0 and (not vdata.bShareHidden) then
					if tUser:HasFeatures(vsid, vdata.tSupportsInf, has, hasnot) then
						tSend:Raw(vsid, msg.."\n", nil, nil, "Delay")
					end
				end
			end
		end
		return client
	end,
	["DRES"] = function(client, sid, msg)
--		--// Check for control chars
--		local res
--		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DRES in state: "..client.iState.." / "..msg)
			return client
		end
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DRES", -1, msg) return client end

		local tosid, retmsg = msg:match("^DRES %S+ (%S+) (.+)$")
		retmsg = ADC_E(retmsg, 1)

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "SR", retmsg)
		if not res then tFuncs:AddStats(sid, "ADC", "DRES", -1, msg) return client end

		--// Check if passive results are allowed for receiver
		if tClient[tosid] then
			local res, retmsg = tDeflood:Main(tosid, "PR", retmsg)
			if not res then tFuncs:AddStats(sid, "ADC", "DRES", -1, msg) return client end
		end

		--// Pattern check
		if res then res, retmsg = tCheck:Pattern(sid, "SR", retmsg) end
		if not res then tFuncs:AddStats(sid, "ADC", "DRES", -1, msg) return client end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

		--/ Execution
		if res and tosid and tClient[tosid] and tClient[tosid].sProtocol == "ADC" then
			--// Deflood based on sid of receiver, to block too many searchresults
			local res, retmsg = tDeflood:Main(tosid, "PR", retmsg)
			if not res then tFuncs:AddStats(sid, "ADC", "DRES", -1, msg) return client end
			tSend:Raw(tosid, msg.."\n", nil, nil, "Delay")
		else
			tFuncs:AddStats(sid, "ADC", "DRES", -1, msg)
			tFuncs:Debug("Target SID in DRES doesn't exist: "..msg, "CA", 4)
			--// skip
		end
		return client
	end,
	["HMSG"] = function(client, sid, msg)	-- message to hub only
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received HMSG in state: "..client.iState.." / "..msg)
			return client
		end

		local retmsg = msg:match("^HMSG (.+)$")

		--// Deflood check
		if retmsg then
			retmsg = ADC_E(retmsg, 1)
			if retmsg:find("^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found
				res, retmsg = tDeflood:Main(sid, "HC", retmsg)
				if not res then tFuncs:AddStats(sid, "ADC", "HMSG", -1, msg) return client end

				--// Script call
		--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

				--// Execution
				client = tSend:Mainchat(client, sid, retmsg)
			end
		else
			tFuncs:AddStats(sid, "ADC", "HMSG", -1, msg)
			tUser:Disconnect(sid, wrong, "No content found in HMSG: "..msg)
			return client
		end

		return client
	end,
	["DMSG"] = function(client, sid, msg)	-- don't send back to user, message to one user, in main if no pm flag

--DMSG fromSID toSID some\smessage\shere PMfromSID\n ME1
--EMSG IV2R 4WX4 Stop,\splease\si\sneed\ssome\srest PMIV2R

		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received DMSG in state: "..client.iState.." / "..msg)
			return client
		end
		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "DMSG", -1, msg) return client end

		local _,_,fromsid, tosid, smsg, rest = msg:find("^DMSG (%S+) (%S+) (%S+)(.*)$")
		if not fromsid or not tosid then
			tFuncs:AddStats(sid, "ADC", "DMSG", -1, msg)
			tUser:Disconnect(sid, wrong, "Source or target SID missing: "..msg)
			return client
		end

		local _, _, pm, targ = rest:find("(PM)(%S%S%S%S)")
		local _, _, me = rest:find("(ME1)")

		--if pm and targ and targ == fromsid then
		--	tFuncs:Debug("we have a DMSG PM to "..targ,"X")
		--else
		--	pm = nil
		--end

		--if me then
		--	tFuncs:Debug("we have a DMSG with ME1 ","X")
		--end

		if smsg then
			smsg = ADC_E(smsg, 1)
			if smsg:find("^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found
				--// Deflood check
				tFuncs:AddStats(sid, "ADC", "HC", 1, msg)
				local res, retmsg = tDeflood:Main(sid, "HC", smsg)
				if not res then tFuncs:AddStats(sid, "ADC", "DMSG", -1, msg)  tFuncs:AddStats(sid, "ADC", "HC", -1, msg) return client end
				tFuncs:HubCommand(sid, smsg)
			else
				tFuncs:AddStats(sid, "ADC", "PM", 1, msg)
				--// Deflood check
				local res, retmsg = tDeflood:Main(sid, "PM", smsg)
				if not res then tFuncs:AddStats(sid, "ADC", "DMSG", -1, msg)  tFuncs:AddStats(sid, "ADC", "PM", -1, msg) return client end

				if not tClient[tosid] then
					tFuncs:AddStats(sid, "ADC", "DMSG", -1, msg)
					tFuncs:Debug("PM from "..client.sNick.." to user with SID: "..tosid.." not delivered, user not found.", "X", 0)
					return client
				end

				if pm then
					tSend:PmToUser(smsg, fromsid, tosid, nil, me)
				else
					tSend:ChatToUser(smsg, fromsid, tosid)	-- chattouser doesnt have ME support, ignore for now?
				end
			end
		else
			tFuncs:AddStats(sid, "ADC", "DMSG", -1, msg)
			tUser:Disconnect(sid, wrong, "No content found in DMSG: "..msg)
			return client
		end
		return client
	end,
	["EMSG"] = function(client, sid, msg)	-- echo message back to user
		--// Check for control chars
		local res
		local origmsg = msg
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		--~ 		EMSG WJNC ASGZ eh\sbla\sbla PMWJNC ME1

		--// Split and assign
		local params = tFuncs:Split(msg, " ")
		local fromsid = params[2]
		local tosid = params[3]
		local retmsg = ADC_E(params[4], 1)
		local rest = table.concat(params, " ",5)
		local pmsid = tFuncs:Find(params, "^PM(%w+)$", 5)
		local me = tFuncs:Find(params, "^ME1$", 5)

		if retmsg:find("^[/%+][Mm][Ee] ") then
			me = true
			retmsg = retmsg:gsub("^[/%+][Mm][Ee] ", "")
		end

		local original = retmsg
		local hcmd = false

		--// Basic validation
		local res = tCheck:State(client, sid, 7, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg) return client end

		local res = tCheck:Sid(sid, msg)
		if not res then tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg) return client end

		if fromsid and tosid then
			if fromsid == pmsid then
				if tClient[tosid] and tClient[tosid].iState == 7 then
					hcmd = retmsg:find("^["..tHubSettings.Main.sCmdPrefixes.."].+$")
					res = true
				else
					tDeflood:Main(sid, "PM", retmsg, tosid)
					tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg)
					return client
					--// User offline, only count for deflood
				end
			else  --// This is a big assumption, if there's no pmsid then how to handle?
				tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg)
				tUser:Disconnect(sid, wrong, "PM Sid spoofed: "..origmsg)
				return client
			end
		else
			tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg)
			tUser:Disconnect(sid, wrong, "Malformed PM: "..msg)
			return client
		end

		--// Deflood check
		if res and hcmd and (tosid == "AAAA" or tosid == "BBBB" or tosid == "CCCC") then
			tFuncs:AddStats(sid, "ADC", "HC", 1, msg)
			res, retmsg = tDeflood:Main(sid, "HC", retmsg)
			if not res then tFuncs:AddStats(sid, "ADC", "HC", -1, msg) tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg) return client end
		else
			tFuncs:AddStats(sid, "ADC", "PM", 1, msg)
			res, retmsg = tDeflood:Main(sid, "PM", retmsg, tosid)
			if not res then tFuncs:AddStats(sid, "ADC", "PM", -1, msg) tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg) return client end
			if res then res, retmsg = tCheck:Pattern(sid, "PM", retmsg) end
			if not res then tFuncs:AddStats(sid, "ADC", "PM", -1, msg) tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg) return client end
			if res then res, retmsg = tCheck:Filter(sid, "PM", retmsg) end
			if not res then tFuncs:AddStats(sid, "ADC", "PM", -1, msg) tFuncs:AddStats(sid, "ADC", "EMSG", -1, msg) return client end
		end

		--// Script call
--~ 		if res then res, sLine = tApi:Main(sid, "PM", sLine, tosid)

		--//Execution
		if res and retmsg then
			--// Send echo
			local filtered = ""
			if original ~= retmsg then
				filtered = "[Message filtered] "
			end
			if res then
				if not me then
					tSend:Raw(sid, "EMSG "..fromsid.." "..tosid.." "..ADC_E(filtered..retmsg).." PM"..pmsid.."\n")	--// Sends the echo
				else
					tSend:Raw(sid, "EMSG "..fromsid.." "..tosid.." "..ADC_E(filtered..retmsg).." PM"..pmsid.." ME1\n")	--// Sends the echo
				end
			end
			res, client = tSend:Pm(client, sid, tosid, retmsg, me)
		end

		return client
	end,
	["HPAS"] = function(client, sid, msg)	-- get the password
		--// Check for control chars
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r]")

		if client.iState == 4 then
			if not client.bNotifySetOwnPass then
				local _,_,pass = msg:find("^HPAS (%S+)")
				local res, client, err = tCheck:Pass(client, sid, pass)
				if not res then
					tFuncs:AddStats(sid, "ADC", "HPAS", -1, msg)
					if tSecurity.General.WrongPassMsg and tSecurity.General.WrongPassMsg ~= "" then
						tUser:Disconnect(sid, "223 "..tFuncs:FillVars(tSecurity.General.WrongPassMsg), err)
					else
						tUser:Disconnect(sid, "223 You provided a wrong password.", err)
					end
					return client
				end
			else
	--			tFuncs:Debug("[HPAS] blankpass", "X")
				local res, client, err = tCheck:Pass(client, sid, "")
				if not res then
					tFuncs:AddStats(sid, "ADC", "HPAS", -1, msg)
					if tSecurity.General.WrongPassMsg and tSecurity.General.WrongPassMsg ~= "" then
						tUser:Disconnect(sid, "223 "..tFuncs:FillVars(tSecurity.General.WrongPassMsg), err)
					else
						tUser:Disconnect(sid, "223 You provided a wrong password.", err)
					end
					return client
				end
			end

			--// Retrieve delayed data
			local info = client.tInfo.tADC.sOrig
			local cid = client.tInfo.tADC.sCID
			local pid = client.tInfo.tADC.sPID
			client.tInfo.tADC.sCID = nil
			client.tInfo.tADC.sPID = nil

			--// Check CID
			local res, client = tCheck:Cid(client, sid, pid, cid)
			if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, msg) return client end

			--// Set MyINFO
	--		tFuncs:Debug(tUser:GetDataLine(sid).."HPAS: Parsing delayed BINF: "..info, "X")
			local res, client, err = tInfo:ParseADC(client, sid, info, true)	--// 4th argument set to true means a full BINF is needed
			if not res then tFuncs:AddStats(sid, "ADC", "BINF", -1, info) tUser:Disconnect(sid, wrong, err) return client end

			--// Check rules
	--		tFuncs:Debug(tUser:GetDataLine(sid).."HPAS: Checking rules", "X")
			res, retmsg = tCheck:Rules(sid)
			if not res then tFuncs:AddStats(sid, "ADC", "HPAS", -1, msg) return client end

			--// Update ADC INF strings due to possibly changed CT in tCheck:Pass
	--		tFuncs:Debug(tUser:GetDataLine(sid).."HPAS: Creating ADC BINF", "X")
			tInfo:CreateADC(client, sid)	-- // fredo - if this is needed then can we remove the one in ParseADC above? now CreateADC is always run twice?

			--// Send Join stuff to user
			if tHubSettings.tRules.bRegOnly then
				local failover = (tHubSettings.tRedirects.sFailOver ~= "" and " FO"..tHubSettings.tRedirects.sFailOver) or ""
				tSend:Raw(sid, "IINF NI"..ADC_E(tHubSettings.Main.sName).." DE"..ADC_E(tHubSettings.Main.sTopic).." VE"..ADC_E(_APPNAME.." ".._APPVER.." ".._REV)..failover.." CT32 HU1 HI1\n")
			end
			client = tUser:Joined(client, sid)	--// send Welcome, MOTD, etc

			-- // Add protect for level 10 users, temporary measure
			local lvl = tUser:GetProfileLevel(sid)
			if lvl and lvl == 10 then
				local save = false
				local now = os.time()
				local reason = "Automatic level 10 protection for: "..client.sNick

				if not tProtects["Ip"][client.sIP] then
					tProtects["Ip"][client.sIP] = { ["By"] = tUser:GetNick("AAAA"),  ["Date"] = now, ["Reason"] = reason }
					save = true
					tFuncs:Debug("[Login] added to tProtects: "..client.sIP, "X")
				elseif tProtects["Ip"][client.sIP] and reason == tProtects["Ip"][client.sIP].Reason and (now - tSecurity.General.AutoProtectTime - 86400) > tProtects["Ip"][client.sIP].Date then
					save = true
					tProtects["Ip"][client.sIP].Date = now
					tFuncs:Debug("[Login] updating protection for : "..client.sIP, "X")
				end

				if save then
					tFuncs:SaveSettings("protects")
				end
			end

			--// Announce user to all users
			tUser:Announce(sid)		--// Sends MyINFO / UserIP / Oplist to all users

			--// Sent login status to GUI
			if client.sProfile == "Gui" then tSend:Raw(sid, "IGUI ST LIS2\n") end

			--// Send Debug to GUI
			tFuncs:Debug(tUser:GetDataLine(sid).."PASS OK.", "LAJ", 2)

			return client
		else
				tFuncs:Debug("[HPAS] messed up", "X")
--~ 			res, retmsg = tDeflood:Bandwidth(sid, "HPAS", msg)	--// Pass to bandwidthcheck to prevent floodattempts
--~ 			if not res then return client end
		end

		return client
	end,
	["HLIV"] = function(client, sid, msg)	-- keep alive ping
	--[[Can be hub or client initiated:
	Hub sends: ILIV TOtoken
	Client sends: HLIV <sid> TOtoken

	Client sends HLIV <sid> TOtoken
	Hub sends: ILIV TOtoken
	--]]

		--// Basic validation
		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received HLIV in state: "..client.iState.." / "..msg)
			return client
		end
		local _, _, ssid, token = msg:find("%S+ (%S+) TO(%S+)")

		if not ssid or not token or ssid ~= sid then
			tFuncs:AddStats(sid, "ADC", "HLIV", -1, msg)
			tUser:Disconnect(sid, wrong, "SID or token missing: "..msg)
			return client
		elseif client.sALIV then
			client.sALIV = nil
			tFuncs:Debug("[ADC] HLIV reply received from "..client.sNick , nil, 3)
		else
			-- reply with token
			tSend:Raw(sid , "ILIV TO"..token.."\n", nil, nil, "Delay")
			tFuncs:Debug("[ADC] HLIV received from "..client.sNick , nil, 3)
		end
		return client
	end,
	["HRLY"] = function(client, sid, msg)	-- hub act as relay for msg's
-- input from client	HRLY <targetsid> <cmd> <data>
-- output from hub to targetsid: <cmd> <data>

		if client.iState ~= 7 then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, received HRLY in state: "..client.iState.." / "..msg)
			return client
		else
			-- check if user is allowed to use HRLY
			local prof, err = ProfileManager:GetData(sid, client.sProfile)
			if not prof or not prof.bAllowedToRelay then
				tFuncs:AddStats(sid, "ADC", "HRLY", -1, msg)
				tUser:Disconnect(sid, wrong, "User not allowed to use HRLY "..msg)
				return client
			end
		end

		-- add the allowed commands for relay here
		local allowed = {["ICMD"] = true, }

		local _, _, tosid, cmd, data = msg:find("%S+ (%S+) (%S+) (.+)")

		if not tosid or not tClient[tosid] then
			tFuncs:Debug("[HRLY] trying to relay to unknown user, msg ignored : "..msg, nil, 1)
		elseif not cmd or not allowed[cmd] or not data then
			tFuncs:Debug("[HRLY] Invalid HRLY, msg ignored : "..msg, nil, 1)
			--tUser:Disconnect(sid, wrong, "Invalid HRLY, user disconnected: "..msg)
		elseif tClient[tosid].sProtocol == "ADC" then
			tSend:Raw(tosid, cmd.." "..data.."\n", nil, nil, "Delay")
		end
		return client
	end,
}

ADC_Replace = {
	["\\"] = [[\\]],
	[" "] = [[\s]],
	["\r"] = [[]],
	["\n"] = [[\n]],
}

ADC_Inverse = {
	["\\\\"] = [[\]],
	["\\s"] = [[ ]],
	["\\n"] = "\n",
}

function ADC_E(msg, inverse)
	if not msg or msg == "" then return "" end
	msg = tostring(msg)
	local ret = {}
	if inverse then
		local skip = false
		for k = 1, #msg do
			if skip then
				skip = false
			else
				local c = msg:sub(k, k + 1)
				if ADC_Inverse[c] then
					table.insert(ret, ADC_Inverse[c])
					skip = true
				else
					table.insert(ret, c:sub(1, 1))
				end
			end
		end
	else
		for k = 1, #msg do
			local c = msg:sub(k, k)
			if ADC_Replace[c] then
				table.insert(ret, ADC_Replace[c])
			else
				table.insert(ret, c)
			end
		end
	end
	return table.concat(ret)
end

ADC_GuiCmds = {
	["GH"] = function(client, sid, smsg)	-- Get Hash - GUI asking for hashes of the different tables
--~ 		local tablenames = {  ["BA"] = {}, ["BL"] = {}, ["CR"] = {}, ["CO"] = {}, ["PA"] = {}, ["PR"] = {}, ["PT"] = {}, ["RE"] = {}, ["SE"] = {}}
		local hashes, output = "", ""
		for word in string.gmatch(smsg, "(%S+)") do
			local vtable = tFuncs:GetTable(word)
			if vtable then
				local hash = tFuncs:GetHash(word, vtable)
				hashes = hashes.." "..word..hash
				output = output.." "..word
			end
		end
		if hashes ~= ""  then
			tSend:Raw(sid, "IGUI SH"..hashes.."\n")
			tFuncs:Debug("[GUI] Sending hashes for tables:"..output)
		end
	end,
	["GV"] = function(client, sid, smsg)	-- Get Hash - GUI asking for hashes of the different tables
--~ 		local tablenames = {  ["BA"] = {}, ["BL"] = {}, ["CR"] = {}, ["CO"] = {}, ["PA"] = {}, ["PR"] = {}, ["PT"] = {}, ["RE"] = {}, ["SE"] = {}}
		local vtotal, output = "", ""
		for word in string.gmatch(smsg, "(%S+)") do
			--// Send profiles to GUI
			if word == "PR" then
				--// Send Commands and Overrides lists
				local cmds = ""
				for num,name in ipairs(Set.ProfileCommands) do
					cmds = cmds.." "..ADC_E(name)
				end
				tSend:Raw(sid, "IGUI PSCM"..cmds.."\n")
--~ 				vtotal = vtotal.."IGUI PSCM"..cmds.."\n"
				local cmds = ""
				for num,name in ipairs(Set.ProfileOverrides) do
					cmds = cmds.." "..ADC_E(name)
				end
				tSend:Raw(sid, "IGUI PSOV"..cmds.."\n")
--~ 				vtotal = vtotal.."IGUI PSOV"..cmds.."\n"

				--// Send Profiles
				--[[// Send to GUI:
				IGUI PA<name> PL<profilelevel> DE<desc> KE<0/1> KF<0/1> WO<0/1 enable welcome> WT<welcometext> AO<0/1 enable announcement> AT<announcementtext to all> PR<protocol required>
					SSLO<1/0 minshare on/of> SSLV<minshare> SSLO<1/0 maxshare on/off> SSHV<maxshare> SLLO<1/0 minslots on/off> SLLV<minslots> SLHO<1/0 maxslots on/off> SLHV<maxslots> HRHO<1/0 hubration on/off> HRHV<hubratio>
					OV<010101010101> CM<01010101010>
				]]
				tSend:Raw(sid, "IGUI PR\n")

				for pname,data in pairs(tProfiles) do
					local prof = "PA NA"..ADC_E(pname).." LE"..(data.iLevel or -1).." DE"..(ADC_E(data.sDescription) or "")
					if data.bHasKey then prof = prof.." KE1" end
					if data.bHasFakeKey then prof = prof.." KF1" end

					-- added 2011-11-23, update in gui /fredo
					if data.sMessage then prof = prof.." ME"..ADC_E(data.sMessage or "") end
					if data.bShareHiddenForced then prof = prof.." HS1" end

					if data.bWelcome then prof = prof.." WO1" end
					if data.sWelcome then prof = prof.." WT"..ADC_E(data.sWelcome or "") end
					if data.bAnnounce then prof = prof.." AO1" end
					if data.sAnnounce then prof = prof.." AT"..ADC_E(data.sAnnounce or "") end
					if data.sProtocol then prof = prof.." PR"..ADC_E(data.sProtocol or "") end
					if data.bMinShare then prof = prof.." S11" end
					if data.iMinShare then prof = prof.." S2"..(data.iMinShare or 0) end
					if data.bMaxShare then prof = prof.." S31" end
					if data.iMaxShare then prof = prof.." S4"..(data.iMaxShare or 0) end
					if data.bMinSlots then prof = prof.." L11" end
					if data.iMinSlots then prof = prof.." L2"..(data.iMinSlots or 0) end
					if data.bMaxSlots then prof = prof.." L31" end
					if data.iMaxSlots then prof = prof.." L4"..(data.iMaxSlots or 0) end
					if data.bSlotRatio then prof = prof.." R11" end
					if data.fSlotRatio then prof = prof.." R2"..(data.fSlotRatio or 1.0) end
					prof = prof.." OV"
					for num, oname in ipairs(Set.ProfileOverrides) do
						if data.tOverrides and data.tOverrides[oname] then prof = prof.."1" else prof = prof.."0" end
					end
					prof = prof.." CM"
					for num, cname in ipairs(Set.ProfileCommands) do
						if data.tCmds and data.tCmds[cname] then prof = prof.."1" else prof = prof.."0" end
					end
					tSend:Raw(sid, "IGUI "..prof.."\n")
--~ 					vtotal = vtotal.."IGUI "..prof.."\n"
				end
				output = output.." "..word
			elseif word == "AL" then	--// Actionlists
				vtotal = "IGUI AL"
				for num, val in ipairs(SubSet.ActionList) do
					vtotal = vtotal.." VA"..ADC_E(val)
				end
				output = output.." "..word
			elseif word == "SE" then
				vtotal = SetManager:Send()
				output = output.." "..word
			elseif word == "RE" then
--~ 				tGuiTemp.RegNames = {"NI", "PR", "ON", "PA", "BI", "RB", "RD", "EX", "LL", "LS", "IP", "GR", "PC", "HI", "HS"}
--~     				tFuncs:FillList(listctrl, {[1]={"Nickname", "L", 100}, [2]={"Profile", "L", 0}, [3]={"Online", "C", 45}, [4]={"Password", "L", 0}, [5]={"Bound to IP", "L", 0}, [6]={"Registered by", "L", 0}, [7]={"Reg Date", "L", 0}, [8]={"Expire Date", "L", 0}, [9]={"Last Online", "L", 0}, [10]={"Last Spoke", "L", 0}, [11]={"Last IP", "L", 0}, [12]={"Greeting", "L", 0},  [13]={"WrongPass", "L", 0}, [14]={"Hide", "L", 0}, [15]={"Hide Share", "L", 0}})

				tSend:Raw(sid, "IGUI RR\n")	--// Reset regs in GUI, clears the entire reglist
				for name,data in pairs(tAccounts) do
					tSend:Raw(sid, "IGUI RA NI"..ADC_E(data.sNick).." PR"..ADC_E(data.sProfile).." ON"..(tNick:GetSid(data.sNick) and 1 or 0).." PA"..ADC_E(data.sPassword).." BI"..ADC_E(data.sBindIP or "")..
						" RB"..ADC_E(data.sReggedBy or "").." RD"..(data.iRegDate or "").." EX"..(data.iExpireDate or "")..
						" LL"..(data.iLastLogin or "").." LS"..(data.iLastSpoke or "").." IP"..(data.sLastIP or "").." GR"..ADC_E(data.sGreeting or "").." PC"..(data.tFailed and data.tFailed.iTimes or 0)..
						" HI"..(data.bHidden and 1 or 0).." HS"..(data.bShareHidden and 1 or 0).."\n"
					)
				end

			else
				local vtable = tFuncs:GetTable(word)
				if vtable and type(vtable) == "table" then
					local vt = tFuncs:GetSerialized(word, vtable)
					vtotal = vtotal.."IGUI TV "..ADC_E(vt)
					output = output.." "..word
				end
			end
		end
		if vtotal ~= "" then
			tSend:Raw(sid, vtotal.."\n")
			tFuncs:Debug("[GUI] Sending data for tables:"..output)
--~ 			tFuncs:Debug("[GUI] Sending tables content:\n"..vtotal, "G", 0)
		end
	end,
	["HR"] = function(client, sid, smsg)	-- Hub returns the result of a given HubCommand (HC), check ID to match action
		-- content here
	end,
	["SV"] = function(client, sid, smsg)     --// Set Variable
		local _,_,varcode = string.find(smsg, "^HGUI SV(%S+)")
		local _,_,varname = string.find(smsg, " NA(%S+)")
		local _,_,vartype = string.find(smsg, " TY(%S+)")
		local _,_,varvalue = string.find(smsg, " VA(%S+)")
		if varcode and varname and vartype and varvalue then
			varvalue = ADC_E(varvalue, 1)
			print(varvalue)
		end
	end,
	["RC"] = function(client, sid, smsg)     --// Reg Change
	end,
	["RD"] = function(client, sid, smsg)     --// Reg Del
	end,
	["PA"] = function(client, sid, smsg)     --// Profile Add
	end,
	["PC"] = function(client, sid, smsg)     --// Profile Change
	end,
	["PD"] = function(client, sid, smsg)     --// Profile Del
	end,
	["BA"] = function(client, sid, smsg)     --// Ban Add
	end,
	["BC"] = function(client, sid, smsg)     --// Ban Change
	end,
	["BD"] = function(client, sid, smsg)     --// Ban Del
	end,
	["SA"] = function(client, sid, smsg)     --// Script Add
	end,
	["SC"] = function(client, sid, smsg)     --// Script Change
	end,
	["SD"] = function(client, sid, smsg)     --// Script Del
	end,
	["SL"] = function(client, sid, smsg)     --// Script Load
	end,
	["SR"] = function(client, sid, smsg)     --// Script Run
	end,
	["SS"] = function(client, sid, smsg)     --// Script Stop
	end,
	["SO"] = function(client, sid, smsg)     --// Script Open
	end,
	["EX"] = function(client, sid, smsg)     --// Execute command
	end,
	["XR"] = function(client, sid, smsg)     --// Raw message
		local _,_,sid,msg = smsg:find("^ (%S+) (.+)")
		if sid and msg then
			msg = ADC_E(msg, 1)
			if sid == "PN" then
				tSend:Raw("P_NMDC", msg)
			elseif sid == "PA" then
				tSend:Raw("P_ADC", msg)
			elseif sid == "AL" then
				tSend:Raw("P_NMDC", msg)
				tSend:Raw("P_ADC", msg)
			elseif #sid == 4 then
				tSend:Raw(sid, msg)
			end
		end
	end,
}
