--// $Id: Hub-Cmds.lua 1467 2013-10-19 17:54:25Z Ruud $
--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tHubFiles["Hub-Cmds.lua"] = {
	["sRev"] = "$Rev: 1467 $",
	["bRestart"] = false,
	["sPath"] = "./core/Hub-Cmds.lua",
	}

HubCmd = {

	["help"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local prof = tUser:GetProfile(ssid)
			local setallow = tUser:IsAllowed(ssid, "setcmd") or false
			local setmsg = ""
			if setallow then setmsg = "\n\n\t"..Lang(ssid, 0, "NOTE! Help for all setup commands are displayed with !sethelp.") end
			local reply = "\n\n\t"..Lang(ssid, 0, "Hubcommands available for your profile [%s]:", prof)..setmsg.."\n\n\t"..Lang(ssid, 0, "Command").."\t\t\t\t\t\t"..Lang(ssid, 0, "Description").."\r\n\t"..string.rep("_",80).."\r\n"

			local totallist = {} --// build total commandlist
			for name, func in pairs(HubCmd) do
				totallist[name] = true
			end

			local cmdprefix = string.find(tHubSettings.Main.sCmdPrefixes, "!", 1, true) and "!" or tHubSettings.Main.sCmdPrefixes:sub(1,1)
			local cmdlist = SubSet.HubCommands

			--// Build reverse list of tCmdAlias
			local cmdalias = {}
			if tHubSettings.Main.bCmdAlias then
				for short, long in pairs(tCmdAlias) do
					if not cmdalias[long] then
						cmdalias[long] = cmdprefix..short
					else
						cmdalias[long] = cmdalias[long]..", "..cmdprefix..short
					end
				end
			end

			--// Display categorized commands
			for num, cat in ipairs(cmdlist) do
				local hits = ""
				for _, name in ipairs(cat.cmds) do
					totallist[name] = nil
					if HubCmd[name] and HubCmd[name]["allowed"](ssid, line, name) and HubCmd[name]["help"] then
						hits = hits.."\t"..HubCmd[name]["help"](ssid, line, cmdprefix..name)..(cmdalias[name] and " ("..cmdalias[name]..")" or "").."\n"
					end
				end
				if hits ~= "" then
					reply = reply.."\n\t"..cat.text.."\n\t"..string.rep("_",18).."\n"..hits
				end
			end

			--// Display miscellaneous commands
			local hits = ""
			local restlist = {}
			for name, res in pairs(totallist) do table.insert(restlist, name) end
			table.sort(restlist, SortLowercase)
			for _, name in ipairs(restlist) do
				if HubCmd[name]["allowed"](ssid, line, name) and HubCmd[name]["help"] then
					hits = hits.."\t"..HubCmd[name]["help"](ssid, line, cmdprefix..name)..(cmdalias[name] and " ("..cmdalias[name]..")" or "").."\n"
				end
			end
			if hits ~= "" then
				reply = reply.."\n\t"..Lang(ssid, 0, "Miscellaneous commands").."\n\t"..string.rep("_",18).."\n"..hits
			end

			--// Display triggerhelp
			local hits = ""
			local jointriggers = {}
			for name, data in pairs(tCommands) do
				if not tFuncs:CheckFlag(data.iOptions or 0, 16) and not tFuncs:CheckFlag(data.iOptions or 0, 8) and tClient[ssid].iProfile >= data.iProfile then
					table.insert(jointriggers, name)
				end
			end
			table.sort(jointriggers, SortLowercase)
			for num, name in ipairs(jointriggers) do
				hits = hits.."\t"..cmdprefix..name.."\t"..(tCommands[name].sHelpText or "")..(cmdalias[name] and " ("..cmdalias[name]..")" or "").."\n"
			end
			if hits ~= "" then
				reply = reply.."\n\t"..Lang(ssid, 0, "Text commands").."\n\t"..string.rep("_",18).."\n"..hits
			end

			--// Add time explanation for tempbans etc
			if tUser:IsAllowed(ssid, "kick") or tUser:IsAllowed(ssid, "tempban") then
				reply = reply.."\n\t"..Lang(ssid, 0, "Time specification").."\n\t"..string.rep("_",18).."\n\t"..
				Lang(ssid, 0, "Time in tempbans or other time-related commands is specified in hours, or using the following letters next to the number:\n\th = hours, m = minutes, s = seconds, d = days, w = weeks, M = months, y = years\n\tExample: !tempbanip 1.2.3.4 3w Spammer\n")
			end

			--// Add password restrictions explanation
			local charset = SetManager:Get("char_set")
			reply = reply.."\n\t"..Lang(ssid, 0, "Password limitations").."\n\t"..string.rep("_",18).."\n\t"..
			Lang(ssid, 0, "Passwords can only contain characters supported by the hub's local charset: http://en.wikipedia.org/wiki/%s\n\tThe following characters are not allowed: | $ \\ \" space line-break\n", charset)

			return true, reply.."\n\t"..string.rep("_",80).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s \t\t\t\t\t\tShow this help", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1,3,"Help","!%s"},
		},
	},

	["quote"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,num = line:find("^%S+%s+(%d+)")
			if not _LASTQUOTE then _LASTQUOTE = false end

			if not tFilters.Quotes or #tFilters.Quotes < 1 then
				return false, Lang(ssid, 0, "No quotes have been added yet, use !addquote <text>")
			end

			if not num or not tonumber(num) then
				if tFilters.Quotes and #tFilters.Quotes > 0 then
					if #tFilters.Quotes > 1 then
						while true do
							num = math.random(#tFilters.Quotes)
							if num ~= _LASTQUOTE then break end
						end
					end
				end
			end
			num = tonumber(num) or 1

			if not tFilters.Quotes[num] then
				return false, Lang(ssid, 0, "That quote does not exist, use a number between 1 and %s", #tFilters.Quotes)
			end

			if tFilters.Quotes[num] and tFilters.Quotes[num] ~= "" then
				tSend:ChatToAll(tFuncs:FillVars(tFilters.Quotes[num], nil, "quote"))
				_LASTQUOTE = num
				return true
			elseif tFilters.Quotes[num] == "" then
				table.remove(tFilters.Quotes, num)
				return false, "Quote number "..num .." was empty."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <num>\t\t\t\t\tShow a quote in mainchat or without number to show a random quote", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show a random quote in mainchat","!%s"},
		},
	},

	["addquote"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			if not tFilters.Quotes then tFilters.Quotes = {} end
			local _,_,quote = line:find("^%S+ (.+)")
			if quote then
				if string.find(quote, "%%%[quote%]") then
					return false, Lang(ssid, 0, "You are not allowed to use the variable for quote inside a quote.")
				end


				table.insert(tFilters.Quotes, quote)
				tFuncs:SaveSettings("filters")
				return true, Lang(ssid, 0, "Quote added.")
			else
				return false, Lang(ssid, 0, "You did not specify a quote, use !addquote <text>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <text>\t\t\t\t\tAdd a quote", cmd)
		end,
		["ucmd"] = {
			{1,3,"Add a quote","!%s %[line:Quote]"},
		},
	},

	["delquote"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			if not tFilters.Quotes then tFilters.Quotes = {} end
			local _,_, quotenum = line:find("^%S+%s+(%d+)")
			local _,_, trigger = line:find("^%S+%s+(%S+)")
			local _,_, confirm = line:find("^%S+%s+%S+%s+(%w+)")

			if trigger and trigger:lower() ~= "all" then trigger = false end

			if quotenum then
				quotenum = tonumber(quotenum)
				if tFilters.Quotes[quotenum] then
					local quote = tFilters.Quotes[quotenum]
					table.remove(tFilters.Quotes, quotenum)
					tFuncs:SaveSettings("filters")
					return true, Lang(ssid, 0, "Quote number: %s deleted: %s", quotenum, quote)
				else
					return false, Lang(ssid, 0, "Quote number %s does not exist. Use !showquotes to show the list of quotes.", quotenum)
				end
			elseif trigger and not confirm then
				return false, Lang(ssid, 0, "'all' keyword found but no confirmation code, to remove all quotes, use %s %s %s ", cmd, "all", tHubSettings.Main.bConfirmCode)
			elseif trigger and confirm and confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s %s", cmd, "all", tHubSettings.Main.bConfirmCode)
			elseif trigger and confirm and confirm == tHubSettings.Main.bConfirmCode then
				--// Change confirmcode
				tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()
				local count = #tFilters.Quotes
				tFilters.Quotes = {}
				_LASTQUOTE = false
				tFuncs:SaveSettings("filters")
				return true, Lang(ssid, 0, "Removed %s quotes, the list is now empty.", count)
			else
				return false, Lang(ssid, 0, "You didn't specify a quote number. Use !delquote <number> or !showquotes to show the list of quotes.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <number>\t\t\t\t\tDelete quote number\n\t%s all <confirmation code>\t\t\tDelete all quotes", cmd, cmd)
		end,
		["ucmd"] = {
			{1,3,"Delete a quote","!%s %[line:Delete quote number]"},
		},
	},

	["showquotes"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local quotes = tFilters.Quotes or {}
			if quotes[1] then
				local tbl = {}
				table.insert(tbl, "\n\n\t"..Lang(ssid, 0, "Quote list:").."\n\t"..string.rep("_",80).."\n")
				for num, quote in ipairs(quotes) do
					table.insert(tbl, "\n\t"..num.."\t"..quote)
				end
				table.insert(tbl, "\n")
				return true, table.concat(tbl)
			else
				return false, Lang(ssid, 0, "There are no quotes added yet, use !addquote <text>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow all quotes", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show all quotes","!%s"},
		},
	},

	["cmdhistory"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local offset, nick = 0, false
			local var1 = line:match("^%S+%s+(%S+)")
			local var2 = line:match("^%S+%s+%S+%s+(%S+)")

			if var2 then
				nick = var1:lower()
				offset = tonumber(var2) or 0
			elseif var1 then
				offset = tonumber(var1) or 0
				if not tonumber(var1) then nick = var1:lower() end
			end

			local fulllogs = tLogs.HubCommands.History
			local logs = {}
			if nick then
				for num, data in ipairs(fulllogs) do
					if data[2]:lower() == nick or data[3] == nick then
						table.insert(logs, data)
					end
				end
			else
				logs = fulllogs
			end

			local logs, left, total = tFuncs:TableChunk(logs, offset)

			local reply = "\n\n\t"..Lang(ssid, 0, "Hubcommand history:")..
				"\n\n\t"..Lang(ssid, 0, "Time \t\t\tNick \t\tIP \t\tOK \tCommand").."\n"..
				"\t"..string.rep("_",80).."\n\n"

			if #logs > 0 then
				for i = #logs, 1, -1  do
					local logdata = logs[i]
					reply = reply.."\t"..ToDate(logdata[1]).." \t"..Padding(logdata[2], 20).." \t"..Padding(logdata[3], 16).." \t"..(logdata[4] and "Yes" or "No").." \t"..tostring(logdata[5]).."\n"
				end
				if left then
					reply = reply.."\n\t"..Lang(ssid, 0, "Showing items %s-%s of %s, use %s %s to show more.", left[1],left[2],left[3],cmd,left[4]).."\n"
				end
			else
				if left then
					reply = reply.."\n\t"..Lang(ssid, 0, "No results in the range %s-%s, total number of items: %s", left[1], left[2], left[3])
				else
					if nick then
						reply = Lang(ssid, 0, "There is no command history for this nickname/IP")
					else
						reply = Lang(ssid, 0, "There is no hubcommand history.")
					end
				end
			end
			return true, reply
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick/ip>]\t\t\t\tShow command history", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show command history","!%s %[line:Nick or IP or leave empty for complete history]"},
			{2,2,"Show command history (selected)","!%s %[userNI]"},
		},
	},

	["about"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			return true,
				"\n\n\tFlexHub Beta	- Made by Daywalker, FlipFlop and en_dator"..

				"\n\n\tThis hubsoft is dedicated to the memory of Daywalker, who died way too young in July of 2009."..
				"\n\tWe started this project together just to see if we were able to create a hubsoft supporting two different protocols, NMDC and ADC. It looks like it's working..."..

				"\n\n\tFlexHub Team:"..
				"\n\tFlipFlop \t\tDeveloper"..
				"\n\ten_dator \t\t\tDeveloper"..
				"\n\tDaywalker \t\tDeveloper (RIP)"..
				"\n\tLadystardust \t\tMain bugtester / Forum admin / Wiki"..
--~ 				"\n\tMutor \t\t\tDeveloper"..
				"\n\tBlackie\t\t\tBugtester"..
				"\n\texleper\t\t\tNAS package builder"..
				"\n\tiLEGALaLIEN \t\tBugtester"..
				"\n\tMخWЙ \t\tBugtester"..
				"\n\toldindian \t\t\tBugtester"..
				"\n\tsergius_s\t\t\tBugtester and Russian support"..
				"\n\tTraN \t\t\tBugtester (Linux)"..
				"\n"..
				"\n\tSpecial thanks to: Massi (Italian support), ApexDC++ team, and all the hubowners that keep giving feedback!"..
				"\n\n\tDownload latest release: \thttp://www.flexhub.org/forum/index.php?action=downloads"..
				"\n\tWiki Manual:	\thttp://flexhub.org/wiki/index.php/Manual"..
				"\n\tWiki FAQ \t\thttp://www.flexhub.org/wiki/index.php/FAQ"..
				"\n\tForum:	\t\thttp://www.flexhub.org/forum"..
				"\n\tForum FAQ \t\thttp://www.flexhub.org/forum/index.php?board=13.0"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tAbout FlexHub", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1,3,"About FlexHub","!%s"},
		},
	},

	["showtopic"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function (ssid, line, cmd)
			local tp = SetManager:Get("hub_topic")
			if tp == "" then
				return true, "No topic set"
			else
				return true, "Topic: "..tp
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow topic", cmd)
		end,
		["ucmd"] = {
			{1,3,"Show topic","!%s"},
		},
	},

	["list"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local room = tClient[tosid]
			if not room or room.sProtocol ~= "ChatRoom" then
				return false, Lang(ssid, 0, "This command is only available inside a chatroom")

			--// Private chatroom
			elseif room.bPrivate then
				local privRoom = tPrivateRooms[room.sNick]

				local reply = {}
				for lnick, nick in pairs(privRoom.tMembers) do
					if tNick:GetSid(nick) then
						--reply = reply.."\n\t"..nick.." (online)"
						--reply = reply.."\n\t"..nick
						table.insert(reply, nick)
					--else
					--	reply = reply.."\n\t"..nick.." (offline)"
					end
				end
				for lnick, nick in pairs(room.tMembersTemp) do
					if tNick:GetSid(nick) then
						--reply = reply.."\n\t"..nick.." (temporary) (online)"
						table.insert(reply, nick.." (temporary)")
					--else
					--	reply = reply.."\n\t"..nick.." (temporary) (offline)"
					end
				end
				--return true, Lang(ssid, 0, "Memberlist: ")..table.concat(reply, ", ")
				return true, Lang(ssid, 0, "The online members of the private chat room %s are:", room.sNick).."\n\n\t"..table.concat(reply, ", ").."\n"

			 --// opchat and feed
			elseif tosid == "BBBB" or tosid == "CCCC" then
				local name = (tosid == "BBBB" and "opchat" or "feed")
				local label = (tosid == "BBBB" and "FlexHub Operator Chat" or "FlexHub FeedRoom")
				local users, profiles = {}, {}

				for label, data in pairs(tProfiles) do
					if data.tCmds and (data.tCmds.all or data.tCmds[name]) then
						table.insert(profiles, label)
					end
				end
				table.sort(profiles, SortLowercase)

				for vnick, data in pairs(tAccounts) do
					if tProfiles[data.sProfile] and (tProfiles[data.sProfile].iLevel == 10 or (tProfiles[data.sProfile].tCmds and (tProfiles[data.sProfile].tCmds.all or tProfiles[data.sProfile].tCmds[name]))) then
						--table.insert(users, data.sNick.." ("..(tNick:GetSid(data.sNick) and "Online" or "Offline (last online: "..(ToDate(data.iLastLogin) or "never"))..")")	--// Needs fixing: iLastOnline needs to be added to accountdata
						if tNick:GetSid(data.sNick) then table.insert(users, data.sNick) end
					end
				end
				table.sort(users, SortLowercase)
			--	return true, Lang(ssid, 0, "This chatroom is available for the following users")..":\n\t"..table.concat(users, "\n\t").."\n\n"
				return true, Lang(ssid, 0, "The online members of the %s %s are:", label, room.sNick).."\n\n\t"..Lang(ssid, 0, "Profiles:").."\t\t"..table.concat(profiles, ", ").."\n\t"..Lang(ssid, 0, "Online Users:").."\t"..table.concat(users, ", ").."\n"

			--// Normal chatroom based on profilelevel
			else
				local profiles = tUser:GetHighProfiles(room.iProfile)
				table.sort(profiles, SortLowercase)

				local on, off = {}, {}
				for vnick, data in pairs(tAccounts) do
					if data.iLevel and data.iLevel >= room.iProfile then
						if tNick:GetSid(data.sNick) then
							table.insert(on, data.sNick)
						--else
							--table.insert(off, data.sNick)
						end
					end
				end

				if room.iProfile == -1 then -- we also need check unregged
					for vnick, data in pairs(tClient) do
						if data.iProfile and data.iProfile == room.iProfile and not data.sProtocol == "ChatRoom" then
							if tNick:GetSid(data.sNick) then
								table.insert(on, data.sNick)
							--else
								--table.insert(off, data.sNick)
							end
						end
					end
				end

				table.sort(on, SortLowercase)
				--table.sort(off, SortLowercase)
				--return true, Lang(ssid, 0, "This chatroom is available to all users with profilelevel %s and higher", room.iProfile) ..";\n\n\t"..Lang(ssid, 0, "Profiles")..":\t\t"..table.concat(profiles, ", ").."\n\n\t"..Lang(ssid, 0, "Online Users")..":\t"..table.concat(on, ", ").."\n\n\t"..Lang(ssid, 0, "Offline Users")..:\t"..table.concat(off, ", ").."\n"
				return true, Lang(ssid, 0, "This chatroom is available to all users with profilelevel %s and higher", room.iProfile) ..";\n\n\t"..Lang(ssid, 0, "Profiles:").."\t\t"..table.concat(profiles, ", ").."\n\t"..Lang(ssid, 0, "Online Users:").."\t"..table.concat(on, ", ").."\n"
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tList the members of a chatroom (use inside room)", cmd)
		end,
		["ucmd"] = {
--~ 			{1,2,"List the members of a chatroom (use inside room)","!%s"},
		},
	},

	["invite"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local room = tClient[tosid]
			if not room or room.sProtocol ~= "ChatRoom" then
				return false, Lang(ssid, 0, "This command is only available in a chatroom.")

			--// Private chatroom
			elseif room.bPrivate then
				local privRoom = tPrivateRooms[room.sNick]
				if tClient[ssid].sNick:lower() ~= room.sOwner:lower() then
					return false, Lang(ssid, 0, "Only the owner of this room (%s) can invite users.", room.sOwner)
				end

				local _,_,nicks = string.find(line, "^%S+%s+(.+)")
				if not nicks then
					return false, Lang(ssid, 0, "Use %s <nick1> [<nick2> <nick3>..]", cmd)
				end

				local reply = ""
				local ignored, addednicks, tempnicks = {}, {}, {}
				--// Add nicks to private chatroom
				for nick in string.gmatch(nicks, "%S+") do
					local usid, unick = tNick:GetSid(nick)
					local reg, vnick = tNick:GetReg(nick)

					if usid and (tClient[usid].sProtocol == "ChatRoom" or tClient[usid].sProtocol == "Bot") then
						table.insert(ignored, unick)
					elseif reg then
						if not privRoom.tMembers[nick:lower()] then
							privRoom.tMembers[nick:lower()] = vnick
							table.insert(addednicks, vnick)
						else
							table.insert(ignored, vnick)
						end
					else
						if not room.tMembersTemp[nick:lower()] then
							room.tMembersTemp[nick:lower()] = nick
							table.insert(tempnicks, nick.."(unregged)")
						else
							table.insert(ignored, nick)
						end
					end
				end

				if addednicks[1] then
					table.sort(addednicks, SortLowercase)
					addednicks = table.concat(addednicks, ", ")

					-- updating settings
					tFuncs:SaveSettings("privaterooms")
				else
					addednicks = false
				end
				if tempnicks[1] then
					table.sort(tempnicks, SortLowercase)
					tempnicks = table.concat(tempnicks, ", ")
				else
					tempnicks = false
				end

				if addednicks or tempnicks then
					reply = Lang(ssid, 0, "%s invited user(s) to this room: ", room.sOwner)..(addednicks or "")..(tempnicks and addednicks and ", " or "")..(tempnicks or "")
					tSend:ChatRoom(reply, tosid, tosid)
					if ignored[1] then
						table.sort(ignored, SortLowercase)
						return true, "User(s) that already are member of this room or that can not be added: "..table.concat(ignored, ", ")
					else
						return true
					end
				else
					if ignored[1]then
						table.sort(ignored, SortLowercase)
						return false, "User(s) that already are member of this room or that can not be added: "..table.concat(ignored, ", ")
					else
						return false, "No users invited."
					end
				end
			else
				--// Normal chatroom based on profilelevel
				return false, "Inviting users to normal chatrooms isn't supported (yet), use !privateroom."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tInvite user(s) to a chatroom (use inside room)", cmd)
		end,
		["ucmd"] = {
--~ 			{1,3,"Invite users to a chatroom (use inside room)","!%s"},
		},
	},

	["leave"] = {
		["allowed"] = function(ssid, line, cmd)
--~ 			return tUser:IsAllowed(ssid, cmd)
			return true
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local room = tClient[tosid]
			if not room or room.sProtocol ~= "ChatRoom" then
				return false,  Lang(ssid, 0, "This command is only available in a private chatroom.", cmd)
			elseif room and room.bPrivate then
				local nick = tClient[ssid].sNick

				--// Owner leaves, delete chatroom
				if nick:lower() == room.sOwner:lower() then
					tSend:ChatRoom("The owner of the private room: "..nick.." has left, this room is deleted.", tosid, tosid)
					tUser:Quit(tosid)
					tClient[tosid] = nil
					tPrivateRooms[room.sNick] = nil
					tFuncs:SaveSettings("privaterooms")
					return true

				--// Other member leaves the room
				else
					local privRoom = tPrivateRooms[room.sNick]
					local removed = false
					if privRoom.tMembers[nick:lower()] then
						privRoom.tMembers[nick:lower()] = nil
						removed = true
						tFuncs:SaveSettings("privaterooms")
					end
					if room.tMembersTemp[nick:lower()] then
						room.tMembersTemp[nick:lower()] = nil
						removed = true
					end

					if removed then
						tSend:ChatRoom(tClient[ssid].sNick.." has left the room.", tosid, tosid)
						if tClient[ssid].sProtocol == "NMDC" then
							tSend:Raw(ssid, "$Quit "..room.sNick.."|")
						--elseif tClient[ssid].sProtocol == "ADC" then
						else
							tSend:Raw(ssid, "IQUI "..tosid.."\n")
						end
						return true
					else
						return false, "You're not a member of this room."
					end
				end

			--// Normal chatroom based on profilelevel
			else
				return false, "Leaving normal chatrooms isn't supported (yet), it only works for private chatrooms."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tLeave a private chatroom (use inside room)", cmd)
		end,
		["ucmd"] = {
--~ 			{1,3,"Leave a private chatroom (use inside room)","!%s"},
		},
	},

	["uninvite"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local room = tClient[tosid]
			local _,_,nicks = string.find(line, "^%S+ (.+)")

			if not room or room.sProtocol ~= "ChatRoom" then
				return false,  Lang(ssid, 0, "This command is only available in a chatroom.", cmd)
			elseif room and room.bPrivate then
				if tClient[ssid].sNick:lower() ~= room.sOwner:lower() then
					return false, "Only the owner of this room ("..room.sOwner..") can uninvite users."
				end

				if not nicks then
					return false, Lang(ssid, 0, "Use %s <nick1> [<nick2> <nick3>..]", cmd)
				end

				local reply, delnicks, ignored = "", "", "", ""
				local privRoom = tPrivateRooms[room.sNick]

				--// Remove nicks from private chatroom
				for nick in string.gmatch(nicks, "%S+") do
					if nick:lower() == room.sOwner:lower() then
						return false, "You can't uninvite yourself, use +leave to delete the chatroom."
					end

					if privRoom.tMembers[nick:lower()] then
						privRoom.tMembers[nick:lower()] = nil
						delnicks = delnicks.." "..nick
					end
					if room.tMembersTemp[nick:lower()] then
						room.tMembersTemp[nick:lower()] = nil
						delnicks = delnicks.." "..nick
					end
				end

				--// Send QUIT to uninvited users
				for nick in delnicks:gmatch("%S+") do
					local usersid = tNick:GetSid(nick)
					if usersid then
						if tClient[usersid].sProtocol == "NMDC" then
							tSend:Raw(usersid, "$Quit "..room.sNick.."|")
						--elseif tClient[usersid].sProtocol == "ADC" then
						else
							tSend:Raw(usersid, "IQUI "..tosid.."\n")
						end
					end
				end

				if delnicks ~= "" then
					tFuncs:SaveSettings("privaterooms")
					reply = room.sOwner.." uninvited user(s) from this room:"..delnicks
					tSend:ChatRoom(reply, tosid, tosid)
					return true
				else
					return false, "User(s) not a member of this room."
				end
			else
				--// Normal chatroom based on profilelevel
				return false, "Uninviting users from normal chatrooms isn't supported."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tUninvite a member of a chatroom (use inside room)", cmd)
		end,
		["ucmd"] = {
--~ 			{1,3,"Uninvite a member of a chatroom (use inside room)","!%s"},
		},
	},

	["privateroom"] = {
		["allowed"] = function(ssid, line, cmd)
			if tUser:GetReg(ssid) and tUser:IsAllowed(ssid, cmd) then return true else return false end
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			--// privateroom name nick1 nick2 nick3..
			local reg, nick = tUser:GetReg(ssid)
			if not tUser:IsOp(ssid) then
				--for sid, data in pairs(tClient) do
				--	if data.sOwner and data.sOwner:lower() == nick:lower() then
				--		return false, Lang(ssid, 0, "You are already using a private chatroom: <%s>, use !delroom %s before creating a new one.", data.sNick, data.sNick)
				--	end
				--end

				-- looping tPrivateRooms should be faster than looping tClient
				for nme, data in pairs(tPrivateRooms) do
					if data.sOwner and data.sOwner:lower() == nick:lower() then
						return false, Lang(ssid, 0, "You are already using a private chatroom: <%s>, use !delroom %s before creating a new one.", data.sNick, data.sNick)
					end
				end
			end

			local _,_,roomname, nicks = string.find(line, "^%S+%s+(%S+)%s*(.*)")
			if not roomname then
				return false, Lang(ssid, 0, "Use %s <roomname> <nick1> [<nick2> <nick3>..]", cmd)
			end

			local reply = ""
			local addednicks, tempnicks = "", ""
			local roomsid, room = false, false

			if not tNick:GetSid(roomname) then
				local res, _, reg = tCheck:Nick(nil,nil, roomname, 10)
				if not res then
					return false, Lang(ssid, 0, "Error: invalid chatroom-name")
				elseif reg then
					return false, Lang(ssid, 0, "This name is already registered as a chatroom or user account.")
				end

				--// Create new room
--~ 				tFuncs:CreateBot(sid,stype,botname,desc,profilelevel,protocol, hidden, ctype)
				roomsid, room = tFuncs:CreateBot(tFuncs:CreateSid(), "ChatRoom", roomname, "Private Chatroom by "..tClient[ssid].sNick, (tClient[ssid].iProfile or -1), "ChatRoom", true)		--// TODO: Level of creator should be used to filter messages, so invited persons can send the same type of messages as creator
				room.sOwner = tClient[ssid].sNick
				room.bPrivate = true

				tPrivateRooms[roomname] = {
				["tMembers"] = { [tClient[ssid].sNick:lower()] = tClient[ssid].sNick },
				["sDesc"] = "Private Chatroom by "..tClient[ssid].sNick,
			--	["bBot"] = true,
				["iLevel"] = (tClient[ssid].iProfile or -1),
			--	["bHidden"] = true,
			--	["bPrivate"] = true,
				["sOwner"] = room.sOwner,
				}

				room.tMembersTemp = {}
				reply = Lang(ssid, 0, "Private chatroom created by: %s with ", tClient[ssid].sNick)
			else
				roomsid = tNick:GetSid(roomname)
				if not tClient[roomsid].bPrivate then
					return false, Lang(ssid, 0, "This is not a private chatroom.")
				end
				room = tClient[roomsid]
				reply = Lang(ssid, 0, "User(s) added to private chatroom: ")
			end

			--// Add nicks to private chatroom
			local ignored, addednicks, tempnicks = {}, {}, {}
			local privRoom = tPrivateRooms[roomname]

			for nick in string.gmatch(nicks, "%S+") do
				local usid, unick = tNick:GetSid(nick)
				if usid and (tClient[usid].sProtocol == "ChatRoom" or tClient[usid].sProtocol == "Bot") then
					table.insert(ignored, unick)
				elseif tNick:GetReg(nick) then
					privRoom.tMembers[nick:lower()] = nick
					table.insert(addednicks, nick)
					--addednicks = addednicks.." "..nick
				else
					room.tMembersTemp[nick:lower()] = nick
					table.insert(tempnicks, nick.."(unregged)")
					--tempnicks = tempnicks.." "..nick.."(unregged)"
				end
			end

			if addednicks[1] then
				table.sort(addednicks, SortLowercase)
				addednicks = table.concat(addednicks, ", ")
			else
				addednicks = false
			end
			if tempnicks[1] then
				table.sort(tempnicks, SortLowercase)
				tempnicks = table.concat(tempnicks, ", ")
			else
				tempnicks = false
			end
			if ignored[1] then
				table.sort(ignored, SortLowercase)
				ignored = "\t"..Lang(ssid, 0, "Users NOT added: ")..table.concat(ignored, ", ")
			else
				ignored = ""
			end

			tClient[roomsid] = room
			tUser:Announce(roomsid)
			local ret = (addednicks or "")..(tempnicks and addednicks and ", " or "")..(tempnicks or "")
			tSend:ChatRoom(room.sOwner..Lang(ssid, 0, " invited user(s) to this room: ")..ret, roomsid, roomsid)
			tFuncs:Debug(tUser:GetDataLine(sid)..reply..ret)
			tFuncs:SaveSettings("privaterooms")

			return true, reply..ret..ignored
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <roomname> <nick>\t\t\tCreate a private chatroom/ Add nick(s) to chatroom, multiple nicks separated by space.", cmd)
		end,
		["ucmd"] = {
			{1,3,"Create private chatroom","!%s %[line:Room name] %[line:Invite nicks (space separated)]"},
			{2,2,"Create private chatroom with user (selected)","!%s %[line:Room name] %[userNI]"},
			{2,2,"Invite selected user to a private chatroom","!%s %[line:Room name] %[userNI]"},
		},
	},

	["bindip"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,snick = line:find("^%S+%s+(%S+)")
			local _,_,rest = line:find("^%S+%s+%S+%s+(.*)")

			if not snick then
				return false, Lang(ssid, 0, "No nick provided.")
			end

			local reg, nick, lnick = tNick:GetReg(snick)
			if not reg then
				return false, Lang(ssid, 0, "User <%s> is not registered, only registered user accounts can be bound to an IP.", snick)
			end

			local ip = ""
			local bad, good = {}, {}

			if rest then
				rest = rest:gsub(" ", "")
				rest = rest:gsub(",", ";")..";"

				for str in rest:gmatch("(%S-);") do
					local tst, cnt = tFuncs:IsIP(str, true)

					if tst then
						table.insert(good, str)
					elseif str ~= "" then
						local tbl = tFuncs:Split(str, "%.")
						local count = 0

						for indx, line in ipairs(tbl) do
							tFuncs:Debug("indx "..indx .." and line "..line, "X")
							local _, _, one = string.find(line, "^([12%?%*]?[%d%?%*]?[%d%?%*])$")
							if one then
								count = count + 1
							else
								count = 0
								break
							end
						end
						if count > 0 then
							tFuncs:Debug("good '"..str.."'", "X")
							table.insert(good, str)
						else
							tFuncs:Debug("bad '"..str.."'", "X")
							table.insert(bad, str)
						end
					end
				end

				if bad[1] then
					return false, Lang(ssid, 0, "The following parts of your string does not match an IP or an allowed pattern: ")..table.concat(bad, ";")
				elseif not good[1] then
					return false, Lang(ssid, 0, "No valid IP's found in your string: %s", rest)
				else
					ip = table.concat(good, ";")
				end
			end

			if tProfiles[reg.sProfile] then
				if tProfiles[reg.sProfile].iLevel < tUser:GetProfileLevel(ssid) or tUser:GetProfileLevel(ssid) == 10 then
					if ip == "" then
						if tAccounts[lnick].sBindIP then
							tAccounts[lnick].sBindIP = nil
							tFuncs:SaveSettings("accounts")
							return true, Lang(ssid, 0, "Removed IP-binding of account for <%s>", nick)
						else
							return true, Lang(ssid, 0, "Account wasn't bound to an IP")
						end
					else
						tAccounts[lnick].sBindIP = ip
						tFuncs:SaveSettings("accounts")
						return true, Lang(ssid, 0, "Account for <%s> is now bound to IP: %s", nick, ip)
					end
				else
					return false, Lang(ssid, 0, "You are not allowed to set an IP for this account.")
				end
			else
				return false, Lang(ssid, 0, "Profile '%s' for user <%s> doesn't exist, use !setprofile <profilename> to fix that first.", reg.sProfile, nick)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> [<ip>]\t\t\t\t\tBind an IP to an account, if no IP is specified it will remove the current binding. Multiple IPs/ranges supported (use semi colon to separate, ? and * can be used as wildcards)", cmd,cmd)
		end,
		["ucmd"] = {
			{1, 3, "Bind IP to account", "!%s %[line:Nickname] %[line:IP]"},
			{2, 2, "Bind current IP to account (selected)", "!%s %[userNI] %[userI4]"},
			{1, 3, "Remove IP Binding from account", "!%s %[line:Nickname]"},
			{2, 2, "Remove IP Binding from account (selected)", "!%s %[userNI]"},
		},
	},

	["protect"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,target = string.find(line, "^%S+ (%S+)")
			local _,_,reason = string.find(line, "^%S+ %S+ (.+)$")
			reason = reason or ""

			if target then
				target = target:lower()
				local stype = "Nick"
				if tFuncs:IsIP(target) then
					stype = "Ip"
				end
				if not tProtects[stype][target] then
					tProtects[stype][target] = { ["By"] = tUser:GetNick(ssid),  ["Date"] = os.time(), ["Reason"] = reason }
					tFuncs:SaveSettings("protects")
					return true, Lang(ssid, 0, "%s: %s is now protected.", stype, target)
				else
					return false, Lang(ssid, 0, "%s: %s is already protected.", stype, target)
				end
			else
				return false, Lang(ssid, 0, "Use %s <nick or ip> [<reason>]", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick or ip> [<reason>]\t\t\tProtect a nick or IP from kicks, bans, etc.", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Protect nick or IP", "!%s %[line:Nickname or IP] %[line:Reason (optional)]"},
			{2, 2, "Protect nick (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
			{2, 2, "Protect IP (selected)", "!%s %[userI4] %[line:Reason (optional)]"},
		},
	},

	["unprotect"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,target = string.find(line, "^%S+%s+(%S+)$")
			local orgtarget = target

			if target then
				local stype = "Nick"
				local changed = false

				if tFuncs:IsIP(target) then
					stype = "Ip"
				else
					target = target:lower()
				end

				if tProtects[stype][target] then	--// Remove direct hit on Nick or Ip
					tProtects[stype][target]  = nil
					changed = true
				end

				if stype == "Nick" then			--// Remove automatic protection for nick
					for ip,data in pairs(tProtects["Ip"]) do
						if data.Reason and string.find(data.Reason, "Automatic level 10 protection for: ") then
							tFuncs:Debug("Found autoprotected IP: "..data.Reason, "X")
							local nick = string.gsub(data.Reason, "^.-: ", "")
							tFuncs:Debug("Found autoprotected nick: "..nick.." / Target: "..target, "X")
							if nick and nick:lower() == target then
								tProtects["Ip"][ip] = nil
								changed = true
							end
						end
					end
				end

				if changed then
					tFuncs:SaveSettings("protects")
					return true, Lang(ssid, 0, "Removed protection of %s: %s", stype, orgtarget)
				else
					return false, Lang(ssid, 0, "%s: %s wasn't protected.", stype, orgtarget)
				end
			else
				return false, Lang(ssid, 0, "Use %s <nick or ip>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick or ip>\t\t\t\tRemove protection", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unprotect nick or IP", "!%s %[line:Nickname or IP]"},
			{2, 2, "Unprotect nick (selected)", "!%s %[userNI]"},
			{2, 2, "Unprotect IP (selected)", "!%s %[userI4]"},
		},
	},

	["showlogins"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local reply = "Showing all clients in login-queue:\n\n\t"..string.rep("_",70).."\n"
			for vsid, client in pairs(tClient) do
				if client.xHandler == "Login" then
					reply = reply.."\tSID: "..vsid.."\tIP: "..client.sIP.."\tSince: "..ToDate(client.iLoginTime).."\tState: "..client.iState.."\tStatus: "..client.sStatus.."\n"
				end
			end
			return true, reply
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow users currently logging in.", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show current logins", "!%s"},
		},
	},

	["showprotected"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local count = 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all protected nicks and ip's")..":\n\t"..string.rep("_",70).."\n"

			--PaddingT(txt, length)
			for nick, data in pairs(tProtects.Nick) do
				reply = reply..Lang(ssid, 0, "\tNick: %s\t By: %s\tDate: %s\tReason: %s\n", nick, data.By, ToDate(data.Date), data.Reason)
				count = count + 1
			end
			for ip, data in pairs(tProtects.Ip) do
				reply = reply..Lang(ssid, 0, "\tIP: %s\t By: %s\tDate: %s\tReason: %s\n", ip, data.By, ToDate(data.Date), data.Reason)
				count = count + 1
			end
			if count > 0 then
				return true, reply
			else
				return true, Lang(ssid, 0, "There are no protected nicks or IP's.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow all protected nicks/ip's.", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show protected nicks and IPs", "!%s"},
		},
	},

	["whisper"] = {		--// TODO: if this is allowed, this command could be used to bypass mainchat checks!!!!
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, target, msg = string.find(line, "^%S+%s+(%S+) (.*)$")
			if target and msg then
				if tNick:GetSid(target) then
					tSend:ChatToUser("[Whispers to you] *** "..msg, ssid, tNick:GetSid(target))
					return true, "Whisper sent"
				elseif (tonumber(target) or -1) >= 0 and tonumber(target) < 11 then
					tSend:ChatToProfLevel("[Whispers to level "..target.."-10] *** "..msg, ssid, tonumber(target))
					return true, "Whisper sent"
				elseif ProfileManager:GetName(ssid, target) then
					local prof = ProfileManager:GetName(ssid, target)
					tSend:ChatToProf("[Whispers to profile "..prof.."-10] *** "..msg, ssid, prof)
					return true, "Whisper sent"
				else
					return false, Lang(ssid, 0, "User or profilelevel not found.")
				end
			else
				return false, Lang(ssid, 0, "Use %s <nick or profilelevel> <message>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick or profile or level> <msg>\t\t\tSend a mainchat message to a user, to all in a profile or to all users equal to or higher than level", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Whisper a mainchat message", "!%s %[line:Nick, Profile or Profilelevel (0-10)] %[line:Message]"},
			{2, 2, "Whisper a mainchat message (selected)", "!%s %[userNI] %[line:Message]"},
		},
	},

	["history"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local _,_,numlines = string.find(line, "^%S+%s+(%d+)")
			local numlines = tonumber(numlines) or tHubSettings.Main.iHistoryLines

			if tUser:IsOp(ssid) then
				if numlines > tHubSettings.Main.iHistoryMaxOp then numlines = tHubSettings.Main.iHistoryMaxOp end --// Maximum history for operators
			else
				if numlines > tHubSettings.Main.iHistoryMaxUser then numlines = tHubSettings.Main.iHistoryMaxUser end --// Maximum history for non-operators
			end

			local chatlog = false

			if tClient[tosid] and tClient[tosid].sProtocol == "ChatRoom" and tLogs.ChatRooms[tosid] then
				chatlog = tLogs.ChatRooms[tosid]
			elseif not tosid then
				chatlog = tLogs.Mainchat
			else
				return true, "No history available."
			end

			if chatlog and #chatlog.Lines < numlines then
				numlines = #chatlog.Lines
			elseif not chatlog then
				return true, "No history available."
			end

			local msgt = {}
			for a=numlines,1,-1 do
				table.insert(msgt, "- ["..ToDate(chatlog.Dates[a]).."] <"..chatlog.Nicks[a].."> "..chatlog.Lines[a]..(a > 1 and "\n" or ""))
			end
			local msg = Lang(ssid, 0, "Showing last %s lines of chat history:", numlines).."\n\n"..table.concat(msgt, "")

			if tosid and tClient[tosid].sProtocol == "ChatRoom" then
				tSend:PmToUser(msg, tosid, ssid, tosid)
				return true
			else
				return true, msg:gsub("because:", "because")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<number of lines>]\t\t\t\tShow mainchat/chatroom history", cmd)
		end,
		["outputtype"] = "MC",
		["ucmd"] = {
			{1, 3, "Show mainchat history", "!%s %[line:Number of lines (or leave empty for default)]"},
		},
	},

	["historyclean"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local _,_,confirm = string.find(line, "^%S+%s+(%w+)")

			if not confirm then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s.", cmd, tHubSettings.Main.bConfirmCode)
			elseif tFuncs:ConfirmCode(confirm) then
				if tClient[tosid] and tClient[tosid].sProtocol == "ChatRoom" and tLogs.ChatRooms[tosid] then
					tLogs.ChatRooms[tosid] = { ["Lines"] = {}, ["Sids"] = {}, ["Nicks"] = {}, ["Dates"] = {} }
					return true, "Chatroom history for <"..tClient[tosid].sNick.."> cleared."
				else
					tLogs.Mainchat = { ["Lines"] = {}, ["Sids"] = {}, ["Nicks"] = {}, ["Dates"] = {} }
					return true,  "Mainchat history cleared."
				end
			else
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s.", cmd, tHubSettings.Main.bConfirmCode)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <confirmation code>\t\t\tClean mainchat/chatroom history", cmd)
		end,
	--	["outputtype"] = "MC",
		["ucmd"] = {
			{1, 3, "Clean mainchat history", "!%s %[line:Confirmation Code]"},
		},
	},

	["showdisconnects"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, tosid)
			local tbl = {}

			--local sorter = function(a,b)
			--	if a[2] < b[2] then return true else return false end
			--end

--			local disc = tLogs.Disconnects
			--table.sort(disc, sorter)

--			table.sort(disc, function(a,b)
--				if a[2] < b[2] then return true else return false end
--				end
--			)

			for num, data in ipairs(tLogs.Disconnects) do
				table.insert(tbl, "\t"..data[2].."\t\t("..(data[3] or 1).." times)\n")
			end
			table.sort(tbl)
			table.insert(tbl, 1, "\n\n\tDisconnection history: \n\t"..string.rep("_",80).."\n\n")
			table.insert(tbl, "\n")
			return true, table.concat(tbl,"")
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow recent disconnection reasons", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show disconnects", "!%s"},
		},
	},

	["drop"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+(.*)$")
			local reason = reason or false

			local vsid = tNick:GetSid(vnick)
			if vsid then
				return tBan:DropUser(vsid, reason, ssid)
			elseif vnick then
				return false, Lang(ssid, 0, "The user you are trying to drop can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tDisconnects a user from the hub", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Drop (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["dropsid"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vsid = string.find(line, "^%S+%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+(.*)$")
			local reason = reason or ""

			if tClient[vsid] then
				return tBan:DropUser(vsid, reason, ssid)
			elseif vsid then
				return false, Lang(ssid, 0, "The user you are trying to drop can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <sid> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <sid> <reason>\t\t\t\tDisconnect a user", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Drop SID (debug)", "!%s %[line:SID] %[line:Reason (optional)]"},
		},
	},

	["unwarn"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")

			if vnick then
				return tBan:UnWarnUser(vnick, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tResets the number of warnings of a user", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unwarn nick", "!%s %[line:Nickname]"},
			{2, 14, "Unwarn (selected)", "!%s %[nick]"},
		},
	},

	["warn"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_,reason = string.find(line, "^%S+%s+%S+%s+(.*)$")

			local vsid = tNick:GetSid(vnick)
			if vsid then
				return tBan:WarnUser(vsid, reason, ssid)
			elseif vnick then
				return false, Lang(ssid, 0, "The user you are trying to warn can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick> <reason>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tWarn a user", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Warn nick", "!%s %[line:Nickname]"},
			{2, 14, "Warn (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["resettimeout"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local ip = line:match("^%S+%s+(%d+%.%d+%.%d+%.%d+)") or tClient[ssid].sIP
			if ip and tFuncs:IsIP(ip) then
				return tIp:ResetTimeout(ip)
			else
				return Lang(ssid, 0, "Error: Use %s <ip>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<ip>]\t\t\t\t\tRemoves the reconnection timeout for an IP (or own IP when unspecified) for next login", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Disable own reconnection timeout for next login", "!%s"},
			{1, 3, "Disable reconnection timeout for IP on next login", "!%s %[line:IP]"},
			{2, 2, "Disable reconnection timeout for IP on next login (selected)", "!%s %[userI4]"},
		},
	},

	["kick"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+(.*)$")
			local reason = reason or ""

			local vsid = tNick:GetSid(vnick)
			if vsid then
				return tBan:KickUser(vsid, reason, ssid)
			elseif vnick then
				return false, Lang(ssid, 0, "The user you are trying to kick can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick> <reason>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tKicks and disconnects a user from the hub", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Kick (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["redirect"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick, address, reason = string.find(line, "^%S+%s+(%S+)%s+(%S+)%s-(.*)$")

			if vnick and address then
				if address:find("%S+%.%S+") then
					local vsid = tNick:GetSid(vnick)
					if vsid then
						if tUser:IsHigherUser(ssid, vsid) then
							return tUser:Redirect(vsid, reason, address, "(redirected by "..tClient[ssid].sNick..") "..reason, "FI", 0)
						else
							return false, "This user is too powerful, better rethink your actions."
						end
					else
						return false, Lang(ssid, 0, "User not found.")
					end
				else
					return false, Lang(ssid, 0, "Redirect failed, invalid address.")
				end
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick> <address> [<reason>]", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <address> [<reason>]\t\t\tRedirect a user", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Redirect (selected)", "!%s %[userNI] %[line:Address] %[line:Reason (optional)]"},
		},
	},

	["clearipbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, confirm = string.find(line, "^%S+%s+(%w+)")

			if confirm and confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s", cmd, tHubSettings.Main.bConfirmCode)
			end

			local count = 0
			for var in pairs(tBans.ipban) do
				count = count+1
			end

			if count == 0 then
				return false, Lang(ssid, 0, "No IP bans found")
			elseif not confirm then
				return false, Lang(ssid, 0, "Found %s IP bans. To confirm removing all IP bans, use %s %s", count, cmd, tHubSettings.Main.bConfirmCode)
			end

			--// Change confirmcode
			tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()

			--// Clear IP bans
			tBans.ipban = {}
			tFuncs:SaveSettings("bans")
			return true, "IP bans removed: "..count.."."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<confirmation code>]\t\t\tRemove all IP bans", cmd)
		end,
		["ucmd"] = {
		},
	},

	["clearautobans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, confirm = string.find(line, "^%S+%s+(%w+)")

			local count = 0
			for var in pairs(tBans.autoban) do
				count = count+1
			end

			if count == 0 then
				return false, Lang(ssid, 0, "No autobans found")
			elseif not confirm then
				return false, Lang(ssid, 0, "Found %s autobans. To confirm removing all autobans, use %s %s", count, cmd, tHubSettings.Main.bConfirmCode)
			elseif confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s", cmd, tHubSettings.Main.bConfirmCode)
			end

			--// Change confirmcode
			tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()

			--// Clear IP bans
			tBans.autoban = {}
			tFuncs:SaveSettings("bans")
			return true, "Autobans removed: "..count.."."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<confirmation code>]\t\t\tRemove all autobans", cmd)
		end,
		["ucmd"] = {
		},
	},

	["clearnickbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, confirm = string.find(line, "^%S+%s+(%w+)")

			local count = 0
			for var in pairs(tBans.nickban) do
				count = count+1
			end

			if count == 0 then
				return false, Lang(ssid, 0, "No nickbans found")
			elseif not confirm then
				return false, Lang(ssid, 0, "Found %s nickbans. To confirm removing all nickbans, use %s %s", count, cmd, tHubSettings.Main.bConfirmCode)
			elseif confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s", cmd, tHubSettings.Main.bConfirmCode)
			end

			--// Change confirmcode
			tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()

			--// Clear IP bans
			tBans.nickban = {}
			tFuncs:SaveSettings("bans")
			return true, "Nickbans removed: "..count.."."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<confirmation code>]\t\t\tRemove all nickbans", cmd)
		end,
		["ucmd"] = {
		},
	},

	["clearrangebans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, confirm = string.find(line, "^%S+%s+(%w+)")

			local count = 0
			for var,data in pairs(tBans.rangeban) do
				count = count+1
			end

			if count == 0 then
				return false, Lang(ssid, 0, "No rangebans found")
			elseif not confirm then
				return false, Lang(ssid, 0, "Found %s rangebans. To confirm removing all rangebans, use %s %s", count, cmd, tHubSettings.Main.bConfirmCode)
			elseif confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s", cmd, tHubSettings.Main.bConfirmCode)
			end

			--// Change confirmcode
			tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()

			--// Clear IP bans
			tBans.rangeban = {}
			tFuncs:SaveSettings("bans")
			return true, "Rangebans removed: "..count.."."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<confirmation code>]\t\t\tRemove all rangebans", cmd)
		end,
		["ucmd"] = {
		},
	},

	["clearexpiringbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local ctime = os.time()
			local _,_, shours = string.find(line, "^%S+%s+(%S+)")
			local _,_,confirm = string.find(line, "^%S+%s+%S+%s+(%w+)")

			local hours = tFuncs:ConvertToHours(shours or 0)

			if not hours then return false, Lang(ssid, 0, "Error: Use %s <hours> [<confirmation code>]", cmd) end
			local expires = ctime + (hours * 3600)
			local reply = "Tempban cleared for:"

			local count, ucount = 0, 0
			for _,slist in ipairs({"ipban", "nickban", "rangeban"}) do
				for var,data in pairs(tBans[slist]) do
					if data.iExpires and data.iExpires > 0 and expires > data.iExpires and (data.sType ~= "kick" or (data.iExpires+(3600*24*30)) < ctime)then		--// kicks will stay in list for 30 days even if expired, to allow ban after number of kicks
						count = count+1
						local res, retmsg = tBan:Unban(var, "Tempban expired (Banreason: "..data.sReason..")", ssid, true, "unban")	--// true = don't save banlist
						if res then
							ucount = (ucount or 0) + 1
							reply = reply.."\n\t"..data.sType.."\t"..(data.sIP or "<IP unknown>").."\t"..(data.sNick or "<nick unknown>").."\t"..ToDate(data.iExpires).."\t"..(data.sReason or "<reason unspecified>").."\t"..(data.sBy or "")
						end
					end
				end
			end

			if count == 0 then
				return true, Lang(ssid, 0, "No bans found in the last %s hours", hours), 2	-- added feed level 2 when running from timer and no bans to clear
			elseif not confirm then
				return false, Lang(ssid, 0, "Found %s expiring bans. To confirm removing these, use %s %s %s", count, cmd, (hours or 0), tHubSettings.Main.bConfirmCode)
			elseif confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s %s", cmd, (hours or 0), tHubSettings.Main.bConfirmCode)
			end

	--		for _,slist in ipairs({"ipban", "nickban", "rangeban"}) do
	--			for var,data in pairs(tBans[slist]) do
	--				if data.iExpires and data.iExpires > 0 and expires > data.iExpires and (data.sType ~= "kick" or (data.iExpires+(3600*24*30)) < ctime)then		--// kicks will stay in list for 30 days even if expired, to allow ban after number of kicks
	--					local res, retmsg = tBan:Unban(var, "Tempban expired (Banreason: "..data.sReason..")", ssid, true)	--// true = don't save banlist
	--					if res then
	--						count = (count or 0) + 1
	--						reply = reply.."\n\t"..data.sType.."\t"..(data.sIP or "<IP unknown>").."\t"..(data.sNick or "<nick unknown>").."\t"..ToDate(data.iExpires).."\t"..(data.sReason or "<reason unspecified>").."\t"..(data.sBy or "")
	--					end
	--				end
	--			end
	--		end

			--// Change confirmcode
			tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()

			if ucount > 0 then
				tFuncs:SaveSettings("bans")

				local output = ""
				if ucount < 50 then output = reply.."\n\n\t" end
				if hours > 0 then
					return true, output.."Removed "..ucount.." bans that would expire in the next "..hours..".", 1 -- level 1 feed
				else
					return true, output.."Removed "..ucount.." expired bans.", 1	-- level 1 feed
				end
			else
				if hours > 0  then
					return true, "No temporary bans expiring in the next "..hours..".", 2	-- level 2 feed
				else
					return true, "No temporary bans expired.", 2	-- level 2 feed
				end
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <hours>\t\t\t\tRemove bans that will expire in the next number of hours", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Remove expiring bans", "!%s %[line:Remove bans expiring in how many hours?] %[line:Confirmation Code]"},
		},
	},

	["clearlastbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,shours = string.find(line, "^%S+%s+(%S+)")
			local _,_,confirm = string.find(line, "^%S+%s+%S+%s+(%w+)")

			local hours = tFuncs:ConvertToHours(shours)
			if not hours or hours <= 0 then return false, Lang(ssid, 0, "Error: Use %s <hours> [<confirmation code>]", cmd) end
			local ago = os.time() - (hours * 3600)

			--// Count
			local count = 0
			for _,slist in ipairs({"ipban", "nickban", "rangeban"}) do
				for var,data in pairs(tBans[slist]) do
					if data.iDate > ago then
						count = count +1
					end
				end
			end

			if count == 0 then
				return false, Lang(ssid, 0, "No bans found in the last %s hours", shours)
			elseif not confirm then
				return false, Lang(ssid, 0, "Found %s recent bans. To confirm removing these, use %s %s %s", count, cmd, shours, tHubSettings.Main.bConfirmCode)
			elseif confirm ~= tHubSettings.Main.bConfirmCode then
				return false, Lang(ssid, 0, "Wrong or expired confirmation code, use %s %s %s", cmd, shours, tHubSettings.Main.bConfirmCode)
			end

			--// Change confirmcode
			tHubSettings.Main.bConfirmCode = tFuncs:CreateSid()

			local reply = "Bans removed:"
			for _,slist in ipairs({"ipban", "nickban", "rangeban"}) do
				for var,data in pairs(tBans[slist]) do
					if data.iDate > ago then
						local res, retmsg = tBan:Unban(var, "Last ban cleared.", ssid, true)	--// true = don't save banlist
						if res then
							if slist == "nickban" then
								reply = reply.."\n\t"..data.sType.." \t"..var.." \tIP: "..(data.sIP or "<unknown>").." \tDate: "..ToDate(data.iDate).." \tUntil: "..(ToDate(data.iExpires) or "Forever").." \t"..(data.sReason or "").." \tBy: "..data.sBy
							else
								reply = reply.."\n\t"..data.sType.." \t"..data.sIP.." \tNick(s): "..(data.sNick or "<unknown>").." \tDate: "..ToDate(data.iDate).." \tUntil: "..(ToDate(data.iExpires) or "Forever").." \t"..(data.sReason or "").." \tBy: "..data.sBy
							end
						end
					end
				end
			end
			if count > 0 then
				tFuncs:SaveSettings("bans")
				if count < 50 then
					return true, reply.."\nRemoved "..count.." bans added in the last "..shours.."."
				else
					return true, "Removed "..count.." bans added in the last "..shours.."."
				end
			else
				return true, "No bans added in the past "..shours.."."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <hours> <confirmation code>\t\tRemove bans added in the past hours", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Remove bans added in the past hours", "!%s %[line:Delete bans added in the last how many hours?] %[line:Confirmation Code]"},
		},
	},

	["baninfo"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd, direct)
			local _,_, value = string.find(line, "^%S+ (%S+)")
			if not value then return false, "Error: use !baninfo <nick>" end
			local orgvalue = value
			local value = value:lower()
		--	local wildcard = false

--[[
			--// the targetip is derived from value, not from a user in the hub

			--// if !baninfo 1.2.3.4 is used, it will first try to find a direct match in one of the bantables
			--// after that it will search in nickbans for IP's matching 1.2.3.4
			--// after that it will search in ipbans for nicks matching 1.2.3.4

			--// new
			local targetip = 0
			local targetsid = tNick:GetSid(value)
			if targetsid then
				targetip = tClient[targetsid].sIP or 0
				--tFuncs:SafeDebug("1165 targetip: "..targetip)
			else

]]--
			local loggedip = tLogs.Nick[value] and tLogs.Nick[value].sIP or false

			if value:find("%*") and value:find("%w") then
				value = value:gsub("%*", ".*")
			--	wildcard = true
			end

--~ 			local banlists = {"nickban", "ipban", "rangeban", "warn"}
			local banlists = {"nickban", "ipban", "rangeban", "autoban"}
			local reply, replyend = "", ""
			local found = false
			local foundbans = {}
			for num, bantype in ipairs(banlists) do
				foundbans[bantype] = {}
				if tBans[bantype] then
					if bantype == "autoban" then
						local autofound = false
						for val, data in pairs(tBans[bantype]) do
--~ 							if (wildcard and val:find(value)) or val == value then
							if (val==value or (not direct and val:find(value)) or (loggedip and loggedip==val)) and not foundbans[bantype][val] then
								found = true
								autofound = true
								local togo = (data - os.time())
								if togo > 0 then togo = "(Expires in "..tFuncs:Duration(togo)..")" else togo = "(Expired)" end
								reply = reply.."\t"..val.."\t\t"..togo.."\n"
							end
						end
						if autofound then
							reply = "\n\n\tThe following IP(s) have been automatically tempbanned due to Ddos or invalid connection attempts: \n\t"..string.rep("_",80).."\n"..reply.."\n\n"
						end
					else
						--// Find direct bans
						for val, data in pairs(tBans[bantype]) do
--~ 							if (wildcard and val:find(value)) or val == value then
							if val==value or (not direct and val:find(value)) or (loggedip and loggedip == val) and not foundbans[bantype][val] then
								found = true
								foundbans[bantype][val] = true
								local expiretxt = ""
								if data.iExpires and data.iExpires > 0 then
									local togo = (data.iExpires or os.time()) - os.time()
									if togo > 0 then togo = "(Expires in "..tFuncs:Duration(togo)..")" else togo = "(Expired)"	end
									expiretxt = ToDate(data.iExpires).." "..togo
								else
									expiretxt = "Never"
								end

								reply = reply.."\n\n\t"..bantype:upper().." information for: "..val.."\n\t"..string.rep("_",80).."\n"..
									"\tNick:      \t"..(data.sNick or "").."\n"..
									"\tIP:         \t"..(data.sIP or "").."\n"..
									"\tReason: \t"..(data.sReason or "").."\n"..
									"\tDate:     \t"..ToDate((data.iDate or 0)).."\n"..
									"\tBy:         \t"..(data.sBy or "").."\n"..
									"\tExpires: \t"..expiretxt.."\n\n"
							end
						end
					end

					--// Add indirect bans, matching IP's from other banlists
					if bantype ~= "ipban" and bantype ~= "rangeban" and bantype ~= "autoban" then		--// Process nickbans to search for bans for IP's matching value
						for val, data in pairs(tBans[bantype]) do
--~ 							if (wildcard and ip:find(value)) or value == data.sIP then
							local ip = data.sIP or ""
							if (ip==value or (not direct and ip:find(value)) or (loggedip and ip == loggedip)) and not foundbans[bantype][val] then
								found = true
								foundbans[bantype][val] = true
								local togo = (data.iExpires or os.time()) - os.time()
								if togo > 0 then togo = "(Expires in "..tFuncs:Duration(togo)..")" else togo = "(Expired)"	end

								replyend = replyend.."\n\t"..bantype:upper().." related to this IP with nick: "..val.."\n\t"..string.rep("_",80).."\n"..
									"\tNick:      \t"..(data.sNick or "").."\n"..
									"\tIP:         \t"..(data.sIP or "").."\n"..
									"\tReason: \t"..(data.sReason or "").."\n"..
									"\tDate:     \t"..ToDate((data.iDate or 0)).."\n"..
									"\tBy:         \t"..(data.sBy or "").."\n"..
									"\tExpires: \t"..ToDate((data.iExpires or 0)).." "..togo.."\n\n"
							end
						end
					--// Add indirect bans, matching nicks from other banlists
					elseif bantype ~= "warn" and bantype ~= "autoban" then		--// Process ipbans and rangebans to search for nicks matching value
						local nickfound = false
						for val, data in pairs(tBans[bantype]) do
							local nick = string.lower((data.sNick or ""))
--~ 							if (wildcard and nick:find(value)) or value == nick then
							if (nick==value or (not direct and nick:find(value))) and not foundbans[bantype][val] then

--[[
							--// It can display all nicks from the IP-ban/rangeban if there's a match with nick:find, no need to split

								nickfound = true
								found = true
							else
								local tNick = tFuncs:Split(data.sNick, ", ")
								for i, v in pairs (tNick) do
									if v == value then
										nickfound = true
										found = true
									end
								end
							end

							if nickfound then
]]--
								found = true
								foundbans[bantype][val] = true

								local togo = (data.iExpires or os.time()) - os.time()
								if togo > 0 then togo = "(Expires in "..tFuncs:Duration(togo)..")" else togo = "(Expired)"	end

								replyend = replyend.."\n\t"..bantype:upper().." related to this nick with IP: "..(data.sIP or "").."\n\t"..string.rep("_",80).."\n"..
									"\tNick:      \t"..(data.sNick or "").."\n"..
									"\tIP:         \t"..(data.sIP or "").."\n"..
									"\tReason: \t"..(data.sReason or "").."\n"..
									"\tDate:     \t"..ToDate((data.iDate or 0)).."\n"..
									"\tBy:         \t"..(data.sBy or "").."\n"..
									"\tExpires: \t"..ToDate((data.iExpires or 0)).." "..togo.."\n\n"
							end
						end
					end
				end
			end

			if found then
				return true, reply..replyend
			else
				return false, "No ban found."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick/IP>\t\t\t\t\tShows information about a banned nick/IP", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show baninfo on nick or IP", "!%s %[line:Nickname or IP]"},
			{2, 14, "Show baninfo (selected)", "!%s %[userNI]"},
		},
	},

	["silentban"] = {
		["allowed"] = function(ssid,line, cmd)
			return false		--// Not finished yet!!!!
--~ 			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+ (%S+)")
			local _,_, reason = string.find(line, "^%S+ %S+ (.*)$")
			local reason = reason or ""

			local vsid = tNick:GetSid(vnick)
			if vsid then
				if tUser:IsHigher(ssid,vsid) then
					local vip = tUser:GetIp(vsid)
					return tBan:BanIp(vip, reason, 0, ssid)
				else
					return false, Lang(ssid, 0, "This user is too powerful for you!")
				end
			elseif vnick then
				return false, Lang(ssid, 0, "The user you are trying to ban can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
	--		return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tSilently bans user IP and disconnects user(s) with this IP", cmd)
		end,
		["ucmd"] = {
	--		{2, 14, "Ban silent (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["ban"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+(.*)%s*$")
			local reason = reason or ""

			local vsid = tNick:GetSid(vnick)
			if vsid then
				if tUser:IsHigher(ssid,vsid) then
					local vip = tUser:GetIp(vsid)
					return tBan:BanIp(vip, reason, 0, ssid)
				else
					return false, Lang(ssid, 0, "This user is too powerful for you!")
				end
			elseif vnick then
				return false, Lang(ssid, 0, "The user you are trying to ban can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tBans user IP and disconnects user(s) with this IP", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Ban (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["kill"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+(.*)%s*$")
			local reason = reason or ""

			local vsid = tNick:GetSid(vnick)
			if vsid then
				if tUser:IsHigher(ssid,vsid) then
					return tBan:KillUser(vsid, reason, 0, ssid)
				else
					return false, Lang(ssid, 0, "This user is too powerful for you!")
				end
			elseif vnick then
				return false, Lang(ssid, 0, "The user you are trying to kill can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tBans user IP and nick and disconnects user(s) with this IP", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Kill (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["autoban"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vip = string.find(line, "^%S+%s+(%d+%.%d+%.%d+%.%d+)")
			if not vip then return false, Lang(ssid, 0, "Error: Use "..cmd.." <ip> <reason>") end

			local btime = tFuncs:Duration(tSecurity.General.DefaultAutoban or 7200)
			local ok, error = tIp:Autoban(vip)
			if ok then
				return true, Lang(ssid, 0, "IP: %s is autobanned for %s.", vip, btime)
			else
				return false, error
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <ip>\t\t\t\t\tAutoban IP, this disconnects an IP without sending a reason and resets autoban time", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Autoban IP", "!%s %[line:IP]"},
			{2, 14, "Autoban IP (selected)", "!%s %[userI4]"},
		},
	},

	["banip"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vip = string.find(line, "^%S+%s+(%d+%.%d+%.%d+%.%d+)")
			local _,_, reason = string.find(line, "^%S+%s+%d+%.%d+%.%d+%.%d+%s+(.*)%s*$")
			reason = reason or ""

			if vip then
				return tBan:BanIp(vip, reason, 0, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <ip> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <ip> <reason>\t\t\t\tBans IP and disconnects users with this IP", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Ban IP", "!%s %[line:IP] %[line:Reason (optional)]"},
			{2, 14, "Ban IP (selected)", "!%s %[userI4] %[line:Reason (optional)]"},
		},
	},

--~ 	["bannickpattern"] = {
--~ 		["allowed"] = function(ssid,line, cmd)
--~ 			return false
--~ 			return tUser:IsAllowed(ssid, "bannick")
--~ 		end,
--~ 		["cmd"] = function(ssid, line, cmd)
--~ 			local _,_, vnick = string.find(line, "^%S+ (%S+)")
--~ 			local _,_, reason = string.find(line, "^%S+ %S+ (.*)$")
--~ 			reason = reason or ""

--~ 			--// TODO: if not vsid then check if nick exists in regtable, check the profilelevel of that reg then to see if it's allowed
--~ 			local vsid = tNick:GetSid(vnick)
--~ 			if not vsid or (vsid and tUser:IsHigher(ssid,vsid)) then
--~ 				local vnick = vnick:gsub("%*", ".*")
--~ 				return tBan:BanNickPattern(vnick, reason, nil, ssid)
--~ 			elseif vsid then
--~ 				return false, Lang(ssid, 0, "This user is too powerful for you!")
--~ 			else
--~ 				return false, Lang(ssid, 0, "Error: Use "..cmd.." <ip> <reason>")
--~ 			end
--~ 		end,
--~ 		["help"] = function(ssid, line, cmd)
--~ 			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tBans nickname", cmd)
--~ 		end,
--~ 		["ucmd"] = {
--~ 			{1, 3, "Ban nickpattern", "!%s %[Nick pattern (regexp)] %[line:Reason (optional)]"},
--~ 		},
--~ 	},
--~
	["bannick"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+(.*)%s*$")
			reason = reason or ""
			if vnick and vnick ~= "" then
				return tBan:BanNick(vnick, reason, 0, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> [<reason>]")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <reason>\t\t\t\tBans nickname", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Ban nick", "!%s %[line: Nickname] %[line:Reason (optional)]"},
			{2, 14, "Ban nick (selected)", "!%s %[userNI] %[line:Reason (optional)]"},
		},
	},

	["tempban"] = {		--// works on IP like in ynhub, so you can tempban an IP of a user without having to use !getinfo for the IP
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick, stime = string.find(line, "^%S+%s+(%S+)%s+(%S+)")
			local _,_, reason = string.find(line, "^%S+%s+%S+%s+%S+%s+(.*)%s*$")
			reason = reason or ""

			--// TODO: if not vsid then check if nick exists in regtable, check the profilelevel of that reg then to see if it's allowed
			local vsid = tNick:GetSid(vnick)
			local stime = tFuncs:ConvertToSeconds(stime, "h")
			if vsid and stime then
				if tUser:IsHigher(ssid,vsid) then
					local vip  = tUser:GetIp(vsid)
					return tBan:BanIp(vip, reason, stime, ssid)
				else
					return false, Lang(ssid, 0, "This user is too powerful for you!")
				end
			elseif vnick and stime then
				return false, Lang(ssid, 0, "The user you are trying to tempban can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <time> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <time> [<reason>]\t\t\tTemporarily bans user IP and disconnects users with this IP, time is in hours if no time suffix are used", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Temp Ban (selected)", "!%s %[userNI] %[line:Time (hours)] %[line:Reason (optional)]"},
		},
	},

	["tempbannick"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick, stime = string.find(line, "^%S+ (%S+) (%S+)")
			local _,_, reason = string.find(line, "^%S+ %S+ %S+ (.*)$")
			reason = reason or ""

			--// TODO: if not vsid then check if nick exists in regtable, check the profilelevel of that reg then to see if it's allowed
			local stime = tFuncs:ConvertToSeconds(stime, "h")

			if vnick and stime then
				return tBan:BanNick(vnick, reason, stime, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <time> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <time> <reason>\t\t\tTemporarily bans nickname, time is in hours if no time suffix are used", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Temp Ban nick", "!%s %[line:Nickname] %[line:Time (hours)] %[line:Reason (optional)]"},
			{2, 14, "Temp Ban nick (selected)", "!%s %[userNI] %[line:Time (hours)] %[line:Reason (optional)]"},
		},
	},

	["tempbanip"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vip, stime = string.find(line, "^%S+ (%d+.%d+.%d+.%d+) (%S+)")
			local _,_, reason = string.find(line, "^%S+ %d+.%d+.%d+.%d+ %S+ (.*)$")
			reason = reason or ""

			--// TODO: if not vsid then check if nick exists in regtable, check the profilelevel of that reg then to see if it's allowed
			local stime = tFuncs:ConvertToSeconds(stime, "h")
			if vip and stime then
				return tBan:BanIp(vip, reason, stime, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <ip> <time> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <ip> <time> <reason>\t\t\tTemporarily bans IP (and disconnects users with this IP, time is in hours if no time suffix are used)", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Temp Ban IP", "!%s %[line:IP] %[line:Time (hours)] %[line:Reason (optional)]"},
			{2, 14, "Temp Ban IP (selected)", "!%s %[userI4] %[line:Time (hours)] %[line:Reason (optional)]"},
		},
	},

	["tempkill"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick, stime = string.find(line, "^%S+ (%S+) (%S+)")
			local _,_, reason = string.find(line, "^%S+ %S+ %S+ (.*)$")
			local reason = reason or ""

			vsid = tNick:GetSid(vnick)
			local stime = tFuncs:ConvertToSeconds(stime, "h")

			if vsid and stime then
				if tUser:IsHigher(ssid,vsid) then
					local vip = tUser:GetIp(vnick)
					return tBan:KillUser(vsid, reason, stime, ssid)
				else
					return false, Lang(ssid, 0, "This user is too powerful for you!")
				end
			elseif vnick and stime then
				return false, Lang(ssid, 0, "The user you are trying to kill can't be found.")
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <time> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <time> [<reason>]\t\t\tTemporarily bans nick and IP (and disconnects users with this IP), time is in hours if no time suffix are used", cmd)
		end,
		["ucmd"] = {
			{2, 14, "Temp Kill (selected)", "!%s %[userNI] %[line:Time (hours)] %[line:Reason (optional)]"},
		},
	},

	["banrange"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, rips, ripe = string.find(line, "^%S+ (%d+.%d+.%d+.%d+)%-(%d+.%d+.%d+.%d+)")
			local _,_, reason = string.find(line, "^%S+ %d+.%d+.%d+.%d+%-%d+.%d+.%d+.%d+ (.*)$")
			local reason = reason or ""

			--// TODO: check if a rangeban would ban the ip of a higher profile user, this should not be allowed
			if rips and ripe then
				return tBan:BanRange(rips, ripe, reason, 0, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <start IP>-<end IP> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <start IP>-<end IP> <reason>\t\t\tBans an IP range (and disconnects user(s) with IP in this range)", cmd)
		end,
		["ucmd"] = {
			{1, 3, "IP Range Ban", "!%s %[line:<start IP>-<end IP>] %[line:Reason (optional)]"},
		},
	},

	["tempbanrange"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, rips, ripe, stime, reason = string.find(line, "^%S+ (%d+%.%d+%.%d+%.%d+)[%s%-](%d+%.%d+%.%d+%.%d+) (%S+)%s-(.*)")
			local stime = tFuncs:ConvertToSeconds(stime, "h")

			--// TODO: check if a rangeban would ban the ip of a higher profile user, this should not be allowed
			if rips and ripe and stime then
				return tBan:BanRange(rips, ripe, reason, stime, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <start IP>-<end IP> <time> <reason>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <start IP>-<end IP> <time> <reason>\tTemporarily bans an IP range (and disconnects user(s) with IP in this range), time is in hours if no time suffix are used", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Temp IP Range Ban", "!%s %[line:<start IP>-<end IP>] %[line:Time] %[line:Reason (optional)]"},
		},
	},

	["unban"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,name, target = string.find(line, "^(%S+) (%S+)")
			local _,_,reason = string.find(line, "^%S+ %S+ (.*)$")
			local reason = reason or ""
			if target then
				return tBan:Unban(target, reason, ssid)
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick/ip/range> [<reason> (optional)]", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick/ip/range> <reason>\t\t\tUnbans a nick, IP or IP-range", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unban nick or IP", "!%s %[line:Nickname or IP]"},
		},
	},

	["showports"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,dport = string.find(line, "^%S+ (%d+)")
			local _,_,offset = string.find(line, "^%S+ %d+ (%d+)")
			offset = tonumber(offset or 0) or 0
			local reply = "\n\n\t"
			local portlist, setports, check = {}, {}, {}
			local usercounts = {}

			if dport then
				dport = tonumber(dport)
				if not tServer.Port[dport] then return false, "This port is not used." end

				reply = "\n\n\t"..Lang(ssid, 0, "List of users on port "..dport..":")..
				"\n\n\t"..Lang(ssid, 0, "SID\tProtocol\tNickname").."\n"..
				"\t"..string.rep("_",40).."\n"
				local replylist = {}
				for sid, data in pairs(tClient) do
					if data.iPort == dport then
						local nick = data.sNick or "<nick unknown>"
						table.insert(replylist, "$$"..nick:lower().."$$".."\t"..sid.."\t"..tostring(data.sProtocol).."\t"..nick.."\n")
					end
				end
				table.sort(replylist, SortLowercase)
				local replylist, left, total = tFuncs:TableChunk(replylist, offset)
				if #replylist > 0 then
					for _,replyline in ipairs(replylist) do
						reply = reply..replyline:gsub("^%$%$%S+%$%$", "")
					end
					if left then
						reply = reply.."\n\t"..Lang(ssid, 0, "Showing items %s-%s of %s, use %s %s %s to show more.", left[1],left[2],left[3],cmd,dport,left[4]).."\n"
					end
				else
					if left then
						reply = reply.."\n\t"..Lang(ssid, 0, "No results in the range %s-%s, total number of items: %s", left[1], left[2], left[3])
					end
				end
				return true, reply.."\n\t"..string.rep("_",40).."\n"
			else
				reply = "\n\n\t"..Lang(ssid, 0, "List of all ports (running and offline):")..
				"\n\n\t"..Lang(ssid, 0, "Port\tType\tConnected \tHandled logins \tHandled connections").."\n"..
				"\t"..string.rep("_",60).."\n"
				for port, data in pairs(tServer.Port) do
					table.insert(portlist, port)
					check[port] = true
				end
				table.sort(portlist, SortNumber)
				for num, port in ipairs(portlist) do
					local count = 0
					for sid, data in pairs(tClient) do
						if data.iPort == port then
							count = count+1
						end
					end
					reply = reply.."\t"..port.."\t"..tServer.Port[port].stype.." \t"..count.." \t\t"..tServer.Port[port].loginsok.." \t\t"..tServer.Port[port].connects.."\n"
				end

				--// if there are ports in settings that is not running show them here
				for i, j in pairs(tHubSettings.Main.tPortList) do
					for k, l in pairs(j) do
						if not check[tonumber(k)] then
							check[tonumber(k)] = i
							table.insert(setports, tonumber(k))
						end
					end
				end
				table.sort(setports, SortNumber)
				if setports[1] then
					reply = reply.."\t"..string.rep("_",60).."\n"
					for num, port in ipairs(setports) do
						reply = reply.."\t"..port.."\t"..check[port].." \tOffline\n"
					end
				end

				if portlist[1] or setports[1] then
					return true, reply.."\t"..string.rep("_",60).."\n"
				else
					return false, "There are no running ports."
				end
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<portnumber>] [<offset>]\t\t\tShow number of connections for all ports (if no port and offset specified) or connected users to specific port", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show hubports", "!%s"},
			{1, 3, "Show users on port", "!%s %[line:Portnumber]"},
			{1, 3, "Show users on port starting from <offset>", "!%s %[line:Portnumber] %[line:Offset (startnumber)]"},
		},
	},

	["showconnects"] = {    --// Todo: store disconnected users in a tClientLog
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local reply = "\n\n\t"..Lang(ssid, 0, "List of connecting users for the past 5 minutes:")..
				"\n\n\t"..Lang(ssid, 0, "State\tProtocol\tStart time\t\t\tUserdata").."\n"..
				"\t"..string.rep("_",70).."\n"
			local backtrack = os.time() - 300
			for sid, data in pairs(tClient) do
				local logintime = data.iLoginTime or 0
				if (data.iState or 0) < 7 or logintime > backtrack then
					logintime = ToDate(logintime)
					reply = reply.."\t"..data.iState.."\t"..tostring(data.sProtocol).."\t"..logintime.."\t"..tUser:GetDataLine(sid).."\n"
				end
			end
			return true, reply.."\n\t"..string.rep("_",70).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow new connections over the past five minutes", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show new connections (5 mins)", "!%s"},
		},
	},

	["shownickbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,start = string.find(line, "^%S+%s+(%d+)")
			local start = start or 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all nickbans")..
			"\n\n\t"..Lang(ssid, 0, "Type\tNick\t\t\tIP       \t\t\tExpires\t\tBy\t\tReason").."\n"..
			"\t"..string.rep("_",100).."\n"
			reply = reply..tFuncs:GetBans({"nickban"}, start, "shownickbans")
			return true, reply.."\n\t"..string.rep("_",100).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow nickbans", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show banned nicks", "!%s"},
		},
	},

	["showipbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,start = string.find(line, "^%S+%s+(%d+)")
			local start = start or 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all IP bans")..
			"\n\n\t"..Lang(ssid, 0, "Type").."\t"..Lang(ssid, 0, "IP").."\t\t\t"..Lang(ssid, 0, "Nick").."\t\t\t"..Lang(ssid, 0, "Expires")..
			"\t\t\t"..Lang(ssid, 0, "By").."\t\t"..Lang(ssid, 0, "Reason").."\n\t"..string.rep("_",100).."\n"
			return true, reply.. tFuncs:GetBans({"ipban", "autoban"}, start, "showipbans") .."\n\t"..string.rep("_",100).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow IP bans", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show banned IP's", "!%s"},
		},
	},

	["showautobans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,start = string.find(line, "^%S+%s+(%d+)")
			local start = start or 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all Auto bans").."\n\n\t"
			..Lang(ssid, 0, "Type\t\tIP\t\t\t\t\tExpires").."\n\t"..string.rep("_",100).."\n"
			..tFuncs:GetBans({"autoban"}, start, "showautobans")
			return true, reply.."\t"..string.rep("_",100).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow auto banned IP's", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show auto banned IP's", "!%s"},
		},
	},

	["showrangebans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line)
			local _,_,start = string.find(line, "^%S+%s+(%d+)")
			local start = start or 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all rangebans")..
			"\n\n\t"..Lang(ssid, 0, "Type\tRange\t\t\t\tNick    \t\t\tExpires\t\tBy\t\tReason").."\n"..
			"\t"..string.rep("_",100).."\n"
			reply = reply..tFuncs:GetBans({"rangeban"}, start, "showrangebans")
			return true, reply.."\n\t"..string.rep("_",100).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow rangebans", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show banned IP ranges", "!%s"},
		},
	},

	["showwarns"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line)
			local _,_,start = string.find(line, "^%S+ (%d+)")
			start = start or 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all warnings")..
			"\n\n\t"..Lang(ssid, 0, "Type\tValue\t\t\tNick or IP\t\t\t\t\t\tBy\t\tReason").."\n"..
			"\t"..string.rep("_",100).."\n"
			reply = reply..tFuncs:GetBans({"warn"}, start, "showwarns")
			return true, reply.."\n\t"..string.rep("_",100).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow warned users", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show warned users", "!%s"},
		},
	},

	["showallbans"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,start = string.find(line, "^%S+ (%d+)")
			start = start or 0
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all type of bans")..
			"\n\n\t"..Lang(ssid, 0, "Type\tValue\t\t\tNick or IP\t\t\tExpires\t\tBy\t\tReason").."\n"..
			"\t"..string.rep("_",100).."\n"
			reply = reply..tFuncs:GetBans({"nickban", "ipban", "rangeban", "warn", "autoban"}, start, "showallbans")
			return true, reply.."\n\t"..string.rep("_",100).."\n\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow nick/ip/rangebans/warns", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show all bans", "!%s"},
		},
	},

	["showprofiles"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local proftable = {}
			for profile,data in pairs(tProfiles) do
				local desc = data.sDescription or ""
				local level = data.iLevel
				if level < 10 and level >= 0 then
					level = "0"..tostring(level)
				elseif level < 0 and level >= -10 then
					level = "-0"..string.sub(tostring(level),2)
				end
				local prof = "\t"..tostring(level).." \t"..profile.." - "..desc
				table.insert(proftable, prof)
			end
			table.sort(proftable, SortLowercase)
			local profstring = Lang(ssid, 0, "Showing existing profiles:").."\n\n"
			for a = #proftable, 1, -1 do
				local prof = proftable[a]
				local prof = prof:gsub("^\t0", "\t  ")
				local prof = prof:gsub("^\t%-0", "\t -")
				profstring = profstring..prof.."\n"
			end
			return true, profstring
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow all profiles", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show userprofiles", "!%s"},
		},
	},

	["showregs"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,var = string.find(line, "^%S+ (%S+)")
			if not var then return false, Lang(ssid, 0, "Error: Use "..cmd.." <profilename> [<offset>] or "..cmd.." * to show all registered users.") end

			local search = var:lower()
			local profname, err = false, false
			if not search:find("%*") then
				profname, err = ProfileManager:GetName(ssid, search)
				if not profname then return false, err end
			else
				search = search:gsub('[%-%.%+%[%]%(%)%$%^%%%?]','%%%1'):gsub('%z','%%z')
				search = "^"..search:gsub("%*", ".*").."$"
--~ 				tFuncs:Debug("[showregs] Search converted to: "..search, "X")
			end

			local _,_,offset = string.find(line, "^%S+ %S+ (%d+)")
			offset = tonumber(offset or 0) or 0
			local count = 0
			local regcount = 0
			local regtable = {}

			--// Return accounts with matching profile
			if profname then
				local profmatch = profname:lower()
				for lnick, reg in pairs(tAccounts) do
					local level = tProfiles[reg.sProfile].iLevel		--// May cause error!
					if profmatch == string.lower(tAccounts[lnick].sProfile) then
						regcount = regcount + 1
					--	if level < 10 and level >= 0 then level = "  "..tostring(level) end
						level = string.format("%2d", level)
						local vsid = tNick:GetSid(reg.sNick)
						local online = "No"
						if vsid then online = "Yes" end
						local prof = "\t"..level .." \t"..online.."\t"..reg.sProfile.."\t\t"..reg.sNick
					--	tFuncs:Debug("[showregs] prof :"..prof, "X")
						table.insert(regtable, prof)
					end
					count = count + 1
				end
			--// Return accounts with matching string in nickname
			else
				for lnick,reg in pairs(tAccounts) do
					--local level = tProfiles[reg.sProfile].iLevel		--// May cause error!
					if lnick:find(search) then
						regcount = regcount + 1
					--	if level < 10 and level >= 0 then level = "  "..tostring(level) end
						local level = string.format("%2d", (tProfiles[reg.sProfile].iLevel or 0))
						local vsid = tNick:GetSid(reg.sNick)
						local online = "No"
						if vsid then online = "Yes" end
						local prof = "\t"..level .." \t"..online.."\t"..reg.sProfile.."\t\t"..reg.sNick
					--	tFuncs:Debug("[showregs] prof :"..prof, "X")
						table.insert(regtable, prof)
					end
					count = count + 1
				end
			end

--~ 		if tProfiles[prf] or prf == "*" then
			if #regtable == 0 then return false, "No accounts "..(profname and "with profile: "..profname or ("matching "..var)).." found." end
			table.sort(regtable, SortLowercase)
			local reply = ""
			local regtable, left, total = tFuncs:TableChunk(regtable, offset)

			local profstring = Lang(ssid, 0, "Showing Registered user by level:").."\n\n\t"..Lang(ssid, 0, "Level").."\t"..
				Lang(ssid, 0, "Online").."\t"..Lang(ssid, 0, "Profile").."\t\t"..Lang(ssid, 0, "Nick").."\n\n"
			local final = "\n\n\t"..Lang(ssid, 0, "There are %s matches and a total of %s users registered in this hub.\n", regcount, count)
			if search == "*" then
				final = "\n\n\t"..Lang(ssid, 0, "There are a total of %s users registered in this hub.\n", regcount)
			end

			for a = 1,#regtable do
				local prof = regtable[a]
				local prof = prof:gsub("^\t0", "\t  ")
				local prof = prof:gsub("^\t%-0", "\t -")
				profstring = profstring..prof.."\n"
			end

			if #regtable > 0 and left then
				reply = "\n\t"..Lang(ssid, 0, "Showing items %s-%s of %s, use %s %s %s to show more.", left[1],left[2],left[3],cmd,var,left[4]).."\n"
			elseif #regtable == 0 then
				reply = "\n\t"..Lang(ssid, 0, "No results in the range %s-%s, total number of items: %s", offset, offset+50, left[3])
			end

			return true, profstring..reply..final

		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profilename> [<offset>]\t\t\tShow registered users for chosen profile (or !showregs * to show all registered users)", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show accounts for profile", "!%s %[line:Profilename] %[line:Offset, start from number (optional): ]"},
		},
	},

	["reginfo"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+ (%S+)")
			if not vnick then return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> to show registration information.") end
			local ctime = socket.gettime()
			local reg, vnick = tNick:GetReg(vnick)
			if reg then
				local prof = tNick:GetProfile(vnick)
				if prof then
					if prof.iLevel < tClient[ssid].iProfile or tClient[ssid].iProfile == 10 then

						local pwd = ""
						if reg.sPassword == "" then
							pwd = "\n\t\t "..Lang(ssid, 0, "No password set for this account")
						end


						--// Total online time
						local onlinetime = false
						local targetsid = tNick:GetSid(vnick)
						if targetsid then
							onlinetime = (tClient[targetsid].iTotalOnlineTime or 0) + (ctime - tClient[targetsid].iLoginTime)
							Test("Online time from reg data, total: "..reg.iTotalOnlineTime.." + current: "..(ctime - tClient[targetsid].iLoginTime).." = "..onlinetime.." = "..tFuncs:Duration(onlinetime))
						else
							onlinetime = reg.iTotalOnlineTime
							Test("Online time from reg data, total: "..reg.iTotalOnlineTime.." = "..tFuncs:Duration(onlinetime))
						end
						local onlineduration = onlinetime and tFuncs:Duration(onlinetime) or "Not available"

						local reply = "\n\n\n\t"..string.rep("_",40)..
							"\n\t\t"..Lang(ssid, 0, "Showing info on: ")..vnick..
							"\n\t"..string.rep("_",40)..
							"\n\t"..Lang(ssid, 0, "Registration information:")..
							"\n\t"..string.rep("_",40)..
							"\n\t\t "..Lang(ssid, 0, "User: ")..vnick..
							"\n\t\t "..Lang(ssid, 0, "Profile: ")..(reg.sProfile or "unknown")..
							"\n\t\t "..Lang(ssid, 0, "Profilelevel: ")..(prof.iLevel or "unknown")..
							"\n\t\t "..Lang(ssid, 0, "Registered by: ")..(reg.sReggedBy or "")..
							"\n\t\t "..Lang(ssid, 0, "Registered date: ")..(reg.iRegDate and ToDate(reg.iRegDate) or "<not specified>")..
							"\n\t\t "..Lang(ssid, 0, "Total time online: ")..onlineduration..
							pwd..
							"\n"..
							"\n\t\t "..Lang(ssid, 0, "Last known IP: ")..(reg.sLastIP or "<not available>")..
							"\n\t\t "..Lang(ssid, 0, "Bound by IP: ")..(reg.sBindIP or "")..
							"\n"..
							"\n\t\t "..Lang(ssid, 0, "Last online: ")..(reg.iLastLogin and reg.iLastLogin > 0 and "From: "..ToDate(reg.iLastLogin) or "")..(targetsid and " (Still online)" or (reg.iLastLogout and reg.iLastLogout > 0 and " To: "..ToDate(reg.iLastLogout) or ""))..
							"\n\t\t "..Lang(ssid, 0, "Last spoke: ")..(reg.iLastSpoke and reg.iLastSpoke > 0 and ToDate(reg.iLastSpoke) or "")..
							"\n\t\t "..Lang(ssid, 0, "Last mainchat line: ")..string.gsub((reg.sLastSpoke or ""), "because:", "because")..
							"\n"..
							"\n\t\t "..Lang(ssid, 0, "Comment: ")..(reg.sComment or "")..
							"\n"..
							"\n\t"..string.rep("_",40)
						return true, reply
					else
						return false, "You are not allowed to view this information."
					end
				else
					return false, "[ERROR] Profile: "..reg.sProfile.." for user <"..vnick.."> doesn't exist."
				end
			else
				return false, "Account not found."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nickname>\t\t\t\tShow a user's registration information", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show account information of nick", "!%s %[line:Nickname]"},
			{2, 14, "Show account information (selected)", "!%s %[userNI]"},
		},
	},

	["attackstats"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local reset = string.match(line, "^%S+%s+reset")
			local portreply, ipcount, concount, tmp = {}, 0, 0, ""

			for port, data in pairs(tLogs.Counters.Ddos.Port) do
				ipcount = ipcount + data.NewCount
				concount = concount + data.Count
				tmp = Lang(ssid, 0, "\n\tPort: \t\t\t%s"..
					"\n\tAttack started:\t\t%s"..
					"\n\tAttack stopped:\t\t%s"..
					"\n\tBlocked connections:\t%s"..
					"\n\tSuccesful normal connects:\t%s (during attack)"..
					"\n\tCurrent new IP rate:\t\t%s/sec"..
					"\n\tPeak new IP rate:\t\t%s/sec"..
					"\n\tCurrent conn. rate:\t\t%s/sec"..
					"\n\tPeak conn. rate:\t\t%s/sec\n"
					,port, ((data.Started and ToDate(data.Started)) or "_"), ((data.Stopped and data.Stopped > data.Started and ToDate(data.Stopped)) or "_"),
					data.Count+data.NewCount, data.Login, math.ceil(data.Attack and data.NewAverage or 0), math.ceil(data.NewAveragePeak), math.ceil(data.Attack and data.Average or 0),
					math.ceil(data.AveragePeak) )

				if reset then
					data.Attack = false
					data.Started = false
					data.Stopped = false
					data.Count = 0
					data.NewCount = 0
					data.Login = 0
					data.NewAverage = 0
					data.NewAveragePeak = 0
					data.AveragePeak = 0
				end

				table.insert(portreply, {["port"] = port, ["result"] = tmp,})
			end

			table.sort(portreply, function(a,b) if a.port < b.port then return true else return false end end)

			local str = ""
			for indx, data in ipairs(portreply) do str = str..data.result end

			local bancount = 0
			local ctime = os.time()
			for ip, expires in pairs(tBans.autoban) do
				if ctime < expires then		--// only count unexpired IP's
					bancount = bancount+1
				end
			end

			local reply = "\n\t"..string.rep("_",40).."\n\t\t"..Lang(ssid, 0, "Showing DDOS attack statistics").."\n\t"..string.rep("_",40)..
			"\n\n\t"..Lang(ssid, 0, "Total connections blocked:").."\t\t"..tLogs.Counters.Ddos.Total..
			"\n\t"..Lang(ssid, 0, "Total IP's blocked (including expired):\t%s", ipcount)..
			"\n\t"..Lang(ssid, 0, "Currently autobanned IP's:\t\t%s", bancount)..
			"\n\n\t"..Lang(ssid, 0, "Attack sources:").."\n\t"..string.rep("_",40)

			for addy, stime in pairs (tLogs.Counters.Ddos.Source) do
				local control, addy = tCheck:ControlChars(nil, nil, addy)
				addy = addy:gsub("%x", "")
				addy = addy:gsub("%z", "")
				if not control then tFuncs:Debug("An address in attack sources contained a control char, replaced with *: "..addy) end
				reply = reply.."\t\t\t"..addy.." \t(first attack: "..ToDate(stime)..")\n\t\t\t"

				if reset then tLogs.Counters.Ddos.Source = {} end
			end

			if reset then str = str.."\n\n\tAbove the most recent attack statistics, all statistics have been reset now." end
			return true, reply.."\n"..str
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <reset>\t\t\t\t\tShows (or resets) statistics of DDOS attacks on the hub", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show DDOS statistics", "!%s"},
			{1, 3, "Reset DDOS statistics", "!%s reset"},
		},
	},

	["addreg"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick, pass = string.find(line, "^%S+%s+(%S+)%s+(%S+)")
			local _,_,prof = string.find(line, "^%S+%s+%S+%s+%S+%s+(%S+)")
			local _,_,ip = string.find(line, "(%d+%.%d+%.%d+%.%d+)%s*$")
			local ip = (tFuncs:IsIP(ip) and ip) or false
			if pass and pass == "\"\"" then pass = "" end

			--// Basic validation
			if not vnick or not pass then
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick> <password> [<profile> [<ip(optional)>]]")
			end

			return AccountManager:Add(ssid, vnick, pass, prof, ip)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <pass> <profile> [<ip>]\t\t\tRegister a user", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local cmdhelp = "Help for "..cmd..": \n\n\t "..cmd.." <nick> <password> <profile> [<ip(optional)>]\n"..
			"\tYou can add a reg without setting a password by using \"\" as password when you register the user, but ONLY if you at the same time provide a bind IP.\n"..
			"\tOnly after the user has set the password she or he will be moved from ordinary user to the set profile.\n"..
			"\tTo allow the use of blank passwords the settings 'pass_allowblank' must be set to true (current setting is '"..tostring(tSecurity.General.AllowBlankPass or "false").."')\n"..
			"\tUse !showprofiles to show avaliable user profiles"
			return true, cmdhelp
		end,
		["ucmd"] = {
			{1, 3, "Register user", "!%s %[line:Nickname] %[line:Password or \"\" for blank password] %[line:Profilename] %[line:IP or DNS (optional except when using blank pass)]"},
			{2, 14, "Register user (selected)", "!%s %[userNI] %[line:Password] %[line:Profilename] %[line:IP or DNS (optional except when using blank pass)]"},
			{2, 14, "Register user (selected) and bind IP", "!%s %[userNI] %[line:Password or \"\" for blank password] %[line:Profilename] %[userI4]"},
		},
	},

	["changereg"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _, _, vnick, nnick = line:find("^%S+%s+(%S+)%s+(%S+)")
			if not vnick or not nnick then
				return false, Lang(ssid, 0, "Error: Use %s <currentnick> <newnick>", cmd)
			elseif nnick and tNick:GetReg(nnick) then
				return false, Lang(ssid, 0, "Registered user <%s> already exists. Change cancelled.", nnick)
			elseif tNick:GetSid(nnick) then
				return false, Lang(ssid, 0, "A user with nick <%s> is already online in the hub. Change cancelled.", nnick)
			end

			local reg, vnick, lnick = tNick:GetReg(vnick)

			if reg and tProfiles[reg.sProfile] then
				if tProfiles[reg.sProfile].iLevel < tClient[ssid].iProfile or tClient[ssid].iProfile == 10 then
					local online = " The user is not online, please remember to notify the user of the change"
					local regdata = tAccounts[lnick]
					regdata.sNick = nnick
					tAccounts[lnick] = nil
					tAccounts[nnick:lower()] = regdata
					tFuncs:SaveSettings("accounts")

					--// If user is online
					local vsid = tNick:GetSid(vnick)
					if vsid then
						online = " User is online and has been told to reconnect with new nick"
						if tClient[vsid].sProtocol == "ADC" then
							tClient[vsid].sNick = nnick
						--	local ct = tInfo:SetCT(tClient[vsid], vsid)	-- isnt really needed
							tInfo:CreateADC(tClient[vsid], vsid)	-- needed on nick change
							tInfo:CreateNMDC(tClient[vsid], vsid) -- dito

							tSend:Raw("P_ADC", "BINF "..vsid.." NI"..nnick.."\n", nil, true, "Delay")

							--// Announce the change to all users
							tUser:Quit(vsid, true)
							tUser:Announce(vsid)
							tSend:OwnKey(vsid, true)
							tSend:UpdateRooms(vsid)
							tSend:ADCUserIps(vsid, true) -- only needed for adc users

						--else -- NMDC or BOT

						end

						local address = tFuncs:GetAddress()
						local details = Lang(vsid, 0, "Nickname:\t%s\n\tProfile:\t\t%s\n\tHubName:\t%s\n\tHub:\t\t%s", nnick, tClient[vsid].sProfile, SetManager:Get("hub_name"), address)
						tSend:PmToUser(Lang(vsid, 0, "Your registered nickname was changed from: %s to: %s:\n\n\t%s\n\nPlease put this in your Favorite hub options and reconnect to this hub.", vnick, nnick, details), nil, vsid)
						tSend:ChatToUser(Lang(vsid, 0, "Your registered nickname was changed from: %s to: %s, check a PM from %s for details.", vnick, nnick, SetManager:Get("bot_name")), nil, vsid)

					end
					return true, Lang(ssid, 0, "Changed registered nickname from: %s to: %s", vnick, nnick) .. online
				else
					return false, Lang(ssid, 0, "This user is too powerful for you")
				end
			else
				return false, Lang(ssid, 0, "This user is not registered")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <newnick>\t\t\tChange the nickname of a registered user", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Change nickname of account", "!%s %[line:Current nickname] %[line:New nickname]"},
			{2, 14, "Change nickname of account (selected)", "!%s %[userNI] %[line:New nickname]"},
		},
	},

	["regme"] = {
		["allowed"] = function(ssid,line, cmd)
			if tUser:GetProfileLevel(ssid) > -1 then
				if tUser:IsAllowed(ssid, "getmypass") then
					local reg = tUser:GetReg(ssid)
					return false, "You are already registered with password: "..reg.sPassword
				else
					return false, "You are already registered."
				end
			elseif tUser:IsAllowed(ssid, cmd) then
				return true
			else
				return false
			end
		end,
		["cmd"] = function(ssid, line, cmd)
			--// Check minshare_reg
			if (tClient[ssid].iShareSize or 0) < tHubSettings.tRules.iMinShareReg then
				return false, Lang(ssid, 0, "To be able to register you need to share at least %s.", tFuncs:ConvertShare(tHubSettings.tRules.iMinShareReg))
			end

			--// Check prefix
			if tHubSettings.tRules.sRegPrefix ~= "" and (string.sub(tClient[ssid].sNick, 1, #tHubSettings.tRules.sRegPrefix) ~= tHubSettings.tRules.sRegPrefix) then
				return false, Lang(ssid, 0, "You need to use the prefix %s in front of your nickname before registering, for example: %s", tHubSettings.tRules.sRegPrefix, tHubSettings.tRules.sRegPrefix..tClient[ssid].sNick)
			end

			local _,_, pass = string.find(line, "^%S+ (%S+)")
			local vnick = tUser:GetNick(ssid)
			local reg, rnick, lnick = tNick:GetReg(vnick)
			if not reg then		-- check if user is regged already
				if pass then
					local res,err = tCheck:PasswordChars(pass)
					if not res then return false, err end
					tAccounts[lnick] = {
							["sNick"] = vnick,
							["sProfile"] = "Reg",
							["iLevel"] = 0,
							["sPassword"] = pass,
							["iLastLogin"] = os.time(),		--// store os.time() here, unformatted for easy sorting
							["iLastSpoke"] = os.time(),		--// store os.time() here, unformatted for easy sorting
							["iTotalOnlineTime"] = 0,		-- total onlinetime in seconds
							["sReggedBy"] = vnick,
							["iRegDate"] = os.time(),		--// store os.time() here, unformatted for easy sorting
					}
					tClient[ssid].sProfile = "Reg"
					tClient[ssid].iProfile = 0

					tFuncs:SaveSettings("accounts")
					tSend:Feedroom(Lang(ssid, 0, "User <%s> successfully selfregistered.", vnick),"AAAA","CCCC")

					local address = tFuncs:GetAddress()
					return true, Lang(ssid, 0, "You are registered with the following details:\n\tNickname:\t%s\n\tPassword:\t%s\n\tProfile:\tRegistered user\n\tHub:\t%s\n\nPlease put this in your Favorite hub options and reconnect to this hub", vnick, pass, address)
				else
					return false, Lang(ssid, 0, "To register yourself use %s <password>\t\tExample: !regme 123ABC", cmd)
				end
			else
				return false, Lang(ssid, 0, "You are already registered with password: %s", reg.sPassword)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <password>\t\t\t\t\tRegister yourself", cmd)
		end,
		["xhelp"] = function(ssid, line, cmd)
			local cmdhelp = "Help for "..cmd..": \n\n\tHow To Register in this hub\n\n"

			if tHubSettings.tRules.sRegPrefix ~= "" and (string.sub(tClient[ssid].sNick, 1, #tHubSettings.tRules.sRegPrefix) ~= tHubSettings.tRules.sRegPrefix) then
				cmdhelp = cmdhelp.."\t* Remove all other prefixes and just add "..tHubSettings.tRules.sRegPrefix.." infront of your nick. For example "..tHubSettings.tRules.sRegPrefix..tClient[ssid].sNick.."\n"
			end

			cmdhelp = cmdhelp.."\t* Type !regme password (Choose a reasonable secure password, you need to remember it for now)\n"..
			"\t* Type /fav in mainchat\n"..
			"\t* Click the Yellow Star in the toolbar at the top\n"..
			"\t* Find this hub in the list ("..tHubSettings.Main.sName.."), right click and choose properties\n"..
			"\t* If the password box is empty you must enter the same password you entered above and click ok\n"..
			"\t* Close the main chat window and reconnect from your Favourites (where you just set your pass)\n"..
			"\n"..
			"\tYou should now be registered. Need help? Ask in main.\n"..
			"\n\n"

			if tHubSettings.Main.sRegmeMessage and tHubSettings.Main.sRegmeMessage ~= "" then cmdhelp = "Help for "..cmd..": \n\n"..tHubSettings.Main.sRegmeMessage end
			return true, cmdhelp
		end,
		["ucmd"] = {
			{1, 3, "Register yourself", "!%s %[line:Password]"},
		},
	},

	["setpass"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
		--	local vnick, pass = false, false
		--	if string.find(line, "^%S+ (%S+)%s(%S+)")  then
		--		_,_,vnick, pass = string.find(line, "^%S+%s+(%S+)%s+(%S+)")
			--else
			--	_,_,pass = string.find(line, "^%S+ (%S+)")
		--	end
			local vnick, pass = line:match("^%S+ (%S+)%s(%S+)")

			if pass and pass == "\"\"" then pass = "" end

			--// Change someone elses password
			if vnick and pass then
				if pass == "" and not tSecurity.General.AllowBlankPass then
					return false, Lang(ssid, 0, "You must specify a password.")
				elseif pass ~= "" then
					local res, err = tCheck:PasswordChars(pass)
					if not res then
						return false, err
					end
				end

				local reg, rnick, lnick = tNick:GetReg(vnick)
				if reg then
					if tClient[ssid].iProfile > reg.iLevel or tClient[ssid].iProfile == 10 then
						tAccounts[lnick].sPassword = pass
						local vsid = tNick:GetSid(rnick)
						if vsid then	-- user is online and needs to be updated
							if tClient[vsid].bNotifySetOwnPass or tClient[vsid].sProfile ~= reg.sProfile then
								tClient[vsid].sProfile = reg.sProfile
								tClient[vsid].iProfile = reg.iLevel
								tClient[vsid].bOperator = reg.bOperator or (reg.iLevel == 10 and true)
								tClient[vsid].bHasKey = reg.bHasKey
								tClient[vsid].bHasFakeKey = reg.bHasFakeKey
								tClient[vsid].bNotifySetOwnPass = nil

								if tSecurity.General.CleanBindIP and tAccounts[lnick].sBindIP then
									tAccounts[lnick].sBindIP = nil
								end
								local ct = tInfo:SetCT(tClient[vsid], vsid)
								tInfo:CreateADC(tClient[vsid], vsid)
								tInfo:CreateNMDC(tClient[vsid], vsid)
								--// Announce the change to all users
								tUser:Quit(vsid, true)
								tUser:Announce(vsid)
								tSend:OwnKey(vsid, true)
								tSend:UpdateRooms(vsid)
								tSend:ADCUserIps(vsid, true)
							elseif pass == "" then
								-- online user got his password set blank
								-- for now we let him keep his own profile
								tClient[vsid].bNotifySetOwnPass = true
							end
--~ 						tSend:PmToUser(Lang(ssid, 0, "Your password is changed to %s, Please change this in your Favorite hub options.", pass), nil, vsid)
						end
						tFuncs:SaveSettings("accounts")
						return true, Lang(ssid, 0, "Password of <%s> is set to: ",rnick)..pass
					else
						return false, Lang(ssid, 0, "This user is too powerful for you")
					end
				else
					return false, Lang(ssid, 0, "User: %s isn't registered yet!",vnick)
				end
			else
				return false, Lang(ssid, 0, "Error: Use %s <nickname> <password>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick>] <password>\t\t\t\tChange a user's password", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Change a user's password", "!%s %[line:Nickname] %[line:Password]"},
			{2,14, "Change a user's password (selected)", "!%s %[userNI] %[line:Password]"},
		},
	},

	["setmypass"] = {
		["allowed"] = function(ssid, line, cmd)
			return tClient[ssid].bNotifySetOwnPass or tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,pass = string.find(line, "^%S+%s+(%S+)")
			if pass then
				local res, err = tCheck:PasswordChars(pass)
				if not res then return false, err end

				local vnick = tClient[ssid].sNick
				local reg, rnick, lnick = tNick:GetReg(vnick)
				if reg then -- check if user is regged allready
					tAccounts[lnick].sPassword = pass

					if tClient[ssid].bNotifySetOwnPass or tClient[ssid].sProfile ~= reg.sProfile then
						-- update and do same for setpass cmd
						tClient[ssid].sProfile = reg.sProfile
						tClient[ssid].iProfile = reg.iLevel
						tClient[ssid].bOperator = reg.bOperator or (reg.iLevel == 10 and true)
						tClient[ssid].bHasKey = reg.bHasKey
						tClient[ssid].bHasFakeKey = reg.bHasFakeKey
						tClient[ssid].bNotifySetOwnPass = nil

						if tSecurity.General.CleanBindIP and tAccounts[lnick].sBindIP then
							tAccounts[lnick].sBindIP = nil
						end

						local ct = tInfo:SetCT(tClient[ssid], ssid)
						tInfo:CreateADC(tClient[ssid], ssid)
						tInfo:CreateNMDC(tClient[ssid], ssid)

						--// Announce the change to all users
						tUser:Quit(ssid, true)
						tUser:Announce(ssid)
						tSend:OwnKey(ssid, true)
						tSend:UpdateRooms(ssid)
						tSend:ADCUserIps(ssid, true)
					end
					tFuncs:SaveSettings("accounts")
					return true, Lang(ssid, 0, "You successfully entered the new password: ")..pass
				else
					return false, Lang(ssid, 0, "You are not registered yet!")
				end
			else
				return false, Lang(ssid, 0, "Error: Use %s <password>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <password>\t\t\t\tChange your own password", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Change your own password", "!%s %[line:Password]"},
		},
	},

	["unhideshare"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local vsid = ssid
			local _,_,nick = string.find(line, "^%S+ (%S+)")
			if nick then vsid = tNick:GetSid(nick) end

			if not vsid or not tClient[vsid] then
				return false, "User <"..nick.."> not found."
			end

			if vsid ~= ssid and not tUser:IsHigher(ssid, vsid) then
				return false, "You are not allowed to hide <"..nick..">."
			end

			if tClient[vsid].bShareHidden then
				tClient[vsid].bShareHidden = nil
				local reg, vnick, lnick = tUser:GetReg(vsid)
				if reg and vnick then
					tAccounts[lnick].bShareHidden = nil
					tFuncs:SaveSettings("accounts")
				end

				local ss = tClient[vsid].tInfo.tOrig.SS or 0
				tClient[vsid].tInfo.tNew = { ["SS"] = ss}
				tClient[vsid].tInfo.tFiltered.SS = ss
				tInfo:CreateADC(tClient[vsid], vsid)
				tInfo:CreateNMDC(tClient[vsid], vsid)
				tClient[vsid].tInfo.tNew = {}
				tSend:MyInfUpdate(vsid)

				if vsid == ssid then
					return true, "Your share is now visible again, incoming connection requests will be allowed."
				else
					return true, "Share of <"..nick.."> is now unhidden."
				end
			else
				if vsid == ssid then
					return false, "Your share is already visible."
				else
					return false, "Share of <"..nick.."> is already visible."
				end
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick>]\t\t\t\tUnhide share, this will allow incoming connection attempts", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unhide your share", "!%s"},
			{1, 3, "Unhide user's share", "!%s %[line:Nickname]"},
			{2, 14, "Unhide user's share (selected)", "!%s %[userNI]"},
		},
	},

	["hideshare"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local vsid = ssid
			local _,_,nick = string.find(line, "^%S+%s+(%S+)")
			if nick then
				vsid = tNick:GetSid(nick)
			else
				nick = tClient[ssid].sNick
			end

			if not vsid or not tClient[vsid] then
				return false, "User <"..nick.."> not found."
			end

			if vsid ~= ssid and not tUser:IsHigher(ssid, vsid) then
				return false, "You are not allowed to hide <"..nick..">."
			end

			if not tClient[vsid].bShareHidden then
				tClient[vsid].bShareHidden = true
				local reg, rnick, lnick = tUser:GetReg(vsid)
				if reg and rnick then
					tAccounts[lnick].bShareHidden = true
					tFuncs:SaveSettings("accounts")
				end

				tClient[vsid].tInfo.tNew = { ["SS"] = 0, ["FS"] = 0,}
				tClient[vsid].tInfo.tFiltered.SS = 0
				tClient[vsid].tInfo.tFiltered.FS = 0
				tInfo:CreateADC(tClient[vsid], vsid)
				tInfo:CreateNMDC(tClient[vsid], vsid)
				tClient[vsid].tInfo.tNew = {}
				tSend:MyInfUpdate(vsid)

				if vsid == ssid then
					return true, "Your share is now hidden, incoming connection requests will be blocked by the hub."
				else
					return true, "You've hidden the share of <"..nick..">."
				end
			else
				return false, "Share is already hidden."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick>]\t\t\t\t\tHide someone's share, without <nick> it will hide your own share.", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Hide your share", "!%s"},
			{1, 3, "Hide user's share", "!%s %[line:Nickname]"},
			{2, 14, "Hide user's share (selected)", "!%s %[userNI]"},
		},
	},

	["getpass"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local msg = ""
			local _,_,vnick = string.find(line, "^%S+ (%S+)")
--~ 			if vnick and vnick:lower() == string.lower(tClient[ssid].sNick) then vnick = false end 	--// Check nick matches own nick

			--// Getting pass of another user
			if vnick then
				local vreg, rnick = tNick:GetReg(vnick)
				if vreg and rnick then
					local vlevel = tProfiles[vreg.sProfile].iLevel
					if tClient[ssid].iProfile > vlevel or tClient[ssid].iProfile == 10 then
						if vreg.sPassword == "" then
							return true, Lang(ssid, 0, "No password set, please ask the user to set a new password with !setmypass <new password>, or set it for him/her with !setpass nick <new password>",rnick)
						else
							return true, Lang(ssid, 0, "The password of user %s is: ",rnick)..vreg.sPassword
						end
					else
						return false, Lang(ssid, 0, "You are not allowed to view the password of this user.")
					end
				else
					return false, Lang(ssid, 0, "User not registered.")
				end
			else
				return false, Lang(ssid, 0, "Error: Use %s <nickname>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tShow a user's password", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show a user's password", "!%s %[line:Nickname]"},
			{2, 14, "Show a user's password (selected)", "!%s %[userNI]"},
		},
	},

	["getmypass"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local vreg = tUser:GetReg(ssid)
			if vreg then
				if vreg.sPassword == "" then
					return true, Lang(ssid, 0, "No password set, please set a new password with !setmypass <new password>")
				else
					return true, Lang(ssid, 0, "Your password is: ")..vreg.sPassword
				end
			else
				--// This shouldn not be possible to occur, unregged users can't use getpass
				return false, Lang(ssid, 0, "You aren't registered.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s \t\t\t\t\tShow your own password", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show your own password", "!%s"},
		},
	},

	["setprofile"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick,prof = string.find(line, "^%S+%s+(%S+)%s+(%S+)")
			if vnick and prof then
				return AccountManager:SetProfile(ssid, vnick, prof)
			else
				return false, "Error: Use "..cmd.." <nick> <profilename>"
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <newprofile>\t\t\t\tChange profile of registered user", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Set profile of user", "!%s %[line:Nickname] %[line:Profilename]"},
			{2,14, "Set profile (selected)", "!%s %[userNI] %[line:Profilename]"},
		},
	},

	["comment"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+ (%S+)")
			local _,_,comment = string.find(line, "^%S+ %S+ (.*)")
			if not vnick then
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick>")
			end
			return AccountManager:AddComment(ssid, vnick, comment)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <text>\t\t\t\tAdd a comment to reginfo, or %s <nick> to clear all comments", cmd, cmd)
		end,
		["ucmd"] = {
			{1, 3, "Add comment to reginfo", "!%s %[line:Nickname] %[line:Comment]"},
			{2,14, "Add comment to reginfo (selected)", "!%s %[userNI] %[line:Comment]"},
		},
	},

	["delreg"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)

			local _,_,vnick = string.find(line, "^%S+ (%S+)")
			if not vnick then
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick>")
			end

			return AccountManager:Delete(ssid, vnick)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tRemove a registered account", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Delete account", "!%s %[line:Nickname]"},
			{2,14, "Delete account (selected)", "!%s %[userNI]"},
		},
	},

	["dns"] = {		--// TODO:  Make this threaded with a callback when ready
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,dns = string.find(line, "^%S+ (.+)$")
			local msg, ip = "", false

--~ 			local ok, err = geoip.open("./lib/GeoIP.dat")
			if dns then
				local ip,err = socket.dns.toip(dns)
				if not ip then return false, "Couldn't resolve adress: "..dns.." to an IP number: "..err end
				msg = msg.."\n\n\t\t"..Lang(ssid, 0, "Dns information").."\n"
				msg = msg.."\n\t\t"..string.rep("_",25)
				msg = msg.."\n\t\t"..Lang(ssid, 0, "Requested").."      \t: "..dns
				msg = msg.."\n\t\t"..Lang(ssid, 0, "Resolved to IP").." \t: "..ip

				--// Get countrycode
				if tHubSettings.Main.bUseGeoIP and tGeoIP.loaded then
					local cc, cn = tGeoIP.loaded:lookup(ip)
					msg = msg.."\n\t\t"..Lang(ssid, 0, "Countrycode").."  \t: "..(cc or "Not found")
					msg = msg.."\n\t\t"..Lang(ssid, 0, "Countryname").." \t: "..(cn or "Not found")
				end

				msg = msg.."\n\t\t"..string.rep("_",25).."\n\n"
				return true, Lang(ssid, 0, msg)
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <dns>")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <dns-address>\t\t\t\t\tResolve a dns-address to an IP", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Resolve DNS-address", "!%s %[line:DNS Address]"},
		},
	},

	["revdns"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			--// This needs to run in a separate (fire-and-forget) thread which shows the result on completion
			--// Right now it's freezing the hub until it's found a revdns

			local _,_,ip = string.find(line, "^%S+%s+(%d+%.%d+%.%d+%.%d+)$")
			local test = tFuncs:IsIP(ip or "")

			if not test then
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <ip>")
			elseif tThreads.RevDNS then
				return false, "RevDNS process already running, please wait for it to complete before starting another."
			else
				tThreads.RevDNS = ThreadCreate(tFuncs.RevDNSStart, {1, ssid, ip})
				return true, "RevDNS lookup started, output will appear in mainchat when it's finished."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <ip>\t\t\t\t\tReverses an IP to a dns-address", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Get reverse DNS Adress from IP", "!%s %[line:IP]"},
			{2,14, "Get reverse DNS Adress (selected)", "!%s %[userI4]"},
		},
	},

	["hubinfo"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local opchat = ""
			if tUser:IsOp(ssid) then opchat = "\n\t"..Lang(ssid, 0, "OpChat").." \t\t: "..SetManager:Get("opchat_name") end

			local port = tonumber(string.match(SetManager:Get("hub_address"), ":(%d+)") or 411)
			local stype = tServer.Port[port] and tServer.Port[port].stype or "Port not running!"
--~ 			local res, msg = tFuncs:GetPorts(ssid, {["Regular"] = "[R]",["Secure"] = "[S]", ["Gui"] = "[G]",})

			return true, string.gsub("\n\n\t"..string.rep("*",80)..
				"\n\t\t"..Lang(ssid, 0, "General Hub information").."\n\t"..string.rep("*",80)..
				"\n"..
				"\n\t"..Lang(ssid, 0, "HubName").." \t: "..(SetManager:Get("hub_name") or "")..
				"\n\t"..Lang(ssid, 0, "HubAddress").." \t: "..(SetManager:Get("hub_address") or "unknown").." ("..stype..")"..
--~ 				"\n\t"..Lang(ssid, 0, "Running Ports").." \t: "..msg..
				"\n\t"..Lang(ssid, 0, "Description").." \t: "..(SetManager:Get("hub_description") or "")..
				"\n\t"..Lang(ssid, 0, "Topic").." \t\t: "..(SetManager:Get("hub_topic")or "")..
				"\n\t"..Lang(ssid, 0, "Hubowner").." \t: "..(SetManager:Get("hub_owner") or "")..
				"\n"..
				"\n\t"..Lang(ssid, 0, "Hubsoft").." \t\t: FlexHub"..
				"\n\t"..Lang(ssid, 0, "Version").." \t\t: ".._APPVER..
				"\n\t"..Lang(ssid, 0, "Revision").." \t\t: "..(tSettings.General.svnversion or "Not available")..(tSettings.General.svnpublic and " (public)" or " (private)")..
				"\n\t"..Lang(ssid, 0, "Supports").." \t\t: NMDC/ADC"..
				"\n\t"..Lang(ssid, 0, "Planned features").."\t: SSL for both protocols, Ptokax Lua API"..
				"\n"..
				"\n\t"..Lang(ssid, 0, "HubBot name").." \t: "..SetManager:Get("bot_name")..
				opchat..
				"\n\t"..Lang(ssid, 0, "Hub Email").." \t: "..SetManager:Get("hub_email")..
				"\n"..
				"\n\t"..Lang(ssid, 0, "Hub Uptime").." \t: "..tFuncs:HubUpTime()..
				"\n\t"..Lang(ssid, 0, "Local Time").." \t: "..ToDate(os.time())..
--				"\n\t"..Lang(ssid, 0, "Local Time").." \t: "..ToDate(os.time()).." (UTC"..tFuncs:GetUTCOffset().." hours)"..
				"\n\n\t"..string.rep("*",80)..
				"\n"..(tFuncs:FillVars(tHubSettings.Main.sHubInfo, ssid) or "") .."\n" ,"because:", "because")
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tShow general hub information", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show Hub Information", "!%s"},
		},
	},

	["stats"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
		--	local port = tonumber(string.match(SetManager:Get("hub_address"), ":(%d+)") or 411)
		--	local stype = tServer.Port[port] and tServer.Port[port].stype or "Port not running!"

--[[
		["SentMergedRaw"] = 0, 		--// Total data before compression
		["SentMerged"] = 0,			--// Total data after optional compression
		["SentMergedADC"] = 0,		--// Total data after optional compression
		["SentMergedNMDC"] = 0,		--// Total data after optional compression
		["SentRawNMDC"] = 0,			--// Data before compression evaluation
		["SentRawADC"] = 0,			--// Data before compression evaluation
		["SentNoZlibSupportNMDC"] = 0,	--// Data can't be compressed, client doesn't support it
		["SentNoZlibSupportADC"] = 0,	--// Data can't be compressed, client doesn't support it
		["SentZlibInputNMDC"] = 0,		--// Data before compressing
		["SentZlibInputADC"] = 0,		--// Data before compressing
		["SentZlibOutputNMDC"] = 0,		--// Data after compressing
		["SentZlibOutputADC"] = 0,		--// Data after compressing
		["SentUncompressedNMDC"] = 0,	--// Data not compressed, too short
		["SentUncompressedADC"] = 0,	--// Data not compressed, too short
]]--

			local div = tStats.Tcp.SentRawADC
			local crtotalADC = div > 0 and ((tStats.Tcp.SentMergedADC) / div) or 1
			local div = tStats.Tcp.SentRawNMDC
			local crtotalNMDC = div > 0 and ((tStats.Tcp.SentMergedNMDC) / div) or 1
			local div = tStats.Tcp.SentMergedRaw
			local crtotal = div > 0 and (tStats.Tcp.Sent / div) or 1

			local div = (tStats.Tcp.SentUncompressedADC + tStats.Tcp.SentZlibInputADC)
			local crsupportADC = div > 0 and ((tStats.Tcp.SentUncompressedADC + tStats.Tcp.SentZlibOutputADC) / div) or 1
			local div = (tStats.Tcp.SentUncompressedNMDC + tStats.Tcp.SentZlibInputNMDC)
			local crsupportNMDC = div > 0 and ((tStats.Tcp.SentUncompressedNMDC + tStats.Tcp.SentZlibOutputNMDC) / div) or 1
			local div =  tStats.Tcp.SentZlibInputADC
			local crpureADC = div > 0 and (tStats.Tcp.SentZlibOutputADC / div) or 1
			local div =  tStats.Tcp.SentZlibInputNMDC
			local crpureNMDC = div > 0 and (tStats.Tcp.SentZlibOutputNMDC / div) or 1
			local saved = tStats.Tcp.SentMergedRaw - tStats.Tcp.Sent
			saved = saved > 0 and saved or 0	--// workaround to ignore the sent traffic during the no-protocol stage

			return true, Lang(ssid, 0, tHub:GetBotName().."\n\n\n\t"..string.rep("*",80)..
				"\n\t\t\t"..Lang(ssid, 0, "Current Hub Status:").."\n\t"..string.rep("*",80)..
				"\n"..
				"\n\t"..Lang(ssid, 0, "Hub Uptime").."    \t: "..tFuncs:HubUpTime()..
				"\n\t"..Lang(ssid, 0, "Hub Soft").."       \t: ".._APPNAME..
				"\n\t"..Lang(ssid, 0, "Hub Version").."   \t: ".._APPVER..
				"\n\t"..Lang(ssid, 0, "Hub Revision").."  \t: "..(tSettings.General.svnversion or "Not available")..
				"\n\t"..Lang(ssid, 0, "Mem Usage").."    \t: (Current) "..tFuncs:ConvertShare(math.ceil(collectgarbage("count")*1024)).." (Peak) "..tFuncs:ConvertShare(math.ceil((tStats.Hub.MaxMem or 0)*1024))..
				(_NOGUI and "" or "\n\t"..Lang(ssid, 0, "Mem Free").."      \t: "..((wx.wxGetFreeMemory():ToDouble() ~= -1 and tFuncs:ConvertShare(math.abs(wx.wxGetFreeMemory():ToDouble()))) or "unknown"))..
				"\n"..
				"\n\t"..Lang(ssid, 0, "Online Users").." \t\t: Total: "..(tStats.Hub.Users.Total or 0).." \tNMDC: "..(tStats.Hub.Users.NMDC or 0).." \tADC: "..(tStats.Hub.Users.ADC or 0)..
				"\n\t"..Lang(ssid, 0, "Max Users").."\t\t: Total: "..(tStats.Hub.Users.TotalMax or 0).." \tNMDC: "..(tStats.Hub.Users.NMDCMax or 0).." \tADC: "..(tStats.Hub.Users.ADCMax or 0)..

				"\n\t"..Lang(ssid, 0, "Total bytes sent").." \t\t: "..Padding((tFuncs:ConvertShare(tStats.Tcp.Sent) or ""), 12).." \tRaw: \t"..Padding(tFuncs:ConvertShare(tStats.Tcp.SentMergedRaw), 12).." \tCompression ratio: "..string.format('%.02f', crtotal).." \t\tSaved: "..tFuncs:ConvertShare(saved)..
				"\n\t"..Lang(ssid, 0, "Total bytes sent ADC").." \t: "..Padding(tFuncs:ConvertShare(tStats.Tcp.SentMergedADC), 12).." \tRaw: \t"..Padding(tFuncs:ConvertShare(tStats.Tcp.SentRawADC), 12).."\tCompression (in/out/too small): \t"..tFuncs:ConvertShare(tStats.Tcp.SentZlibInputADC).." / "..tFuncs:ConvertShare(tStats.Tcp.SentZlibOutputADC).." / "..tFuncs:ConvertShare(tStats.Tcp.SentUncompressedADC)..
				"\n\t"..Lang(ssid, 0, "Total bytes sent NMDC").." \t: "..Padding(tFuncs:ConvertShare(tStats.Tcp.SentMergedNMDC), 12).." \tRaw: \t"..Padding(tFuncs:ConvertShare(tStats.Tcp.SentRawNMDC), 12).."\tCompression (in/out/too small): \t"..tFuncs:ConvertShare(tStats.Tcp.SentZlibInputNMDC).." / "..tFuncs:ConvertShare(tStats.Tcp.SentZlibOutputNMDC).." / "..tFuncs:ConvertShare(tStats.Tcp.SentUncompressedNMDC)..
				"\n\t"..Lang(ssid, 0, "Compression ratio ADC").."\t: Total: \t"..string.format('%.02f', crtotalADC).." \tTotal ratio of supporting clients: "..string.format('%.02f', crsupportADC).." \tPure compression: "..string.format('%.02f', crpureADC)..
				"\n\t"..Lang(ssid, 0, "Compression ratio NMDC").."\t: Total: \t"..string.format('%.02f', crtotalNMDC).." \tTotal ratio of supporting clients: "..string.format('%.02f', crsupportNMDC).." \tPure compression: "..string.format('%.02f', crpureNMDC)..

				"\n\t"..Lang(ssid, 0, "Total bytes recieved").."\t: "..(tFuncs:ConvertShare(tStats.Tcp.Received) or "")..
				"\n\t"..Lang(ssid, 0, "Sending\t\t").."\t: (Current) "..(tFuncs:ConvertShare(tStats.Tcp.SentSpeed or 0)).."/s".." (Peak) "..(tFuncs:ConvertShare(tStats.Tcp.SentPeak or 0)).."/s"..
				"\n\t"..Lang(ssid, 0, "Receiving\t").."\t: (Current) "..(tFuncs:ConvertShare(tStats.Tcp.ReceivedSpeed or 0)).."/s".." (Peak) "..(tFuncs:ConvertShare(tStats.Tcp.ReceivedPeak or 0)).."/s"..
				"\n"..
				"\n\t"..Lang(ssid, 0, "Local Time").."\t\t: "..ToDate(os.time())..
				"\n\n\t"..string.rep("*",80)
			)
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tShow hub statistics", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show Hub Statistics", "!%s"},
		},
	},

	["corestats"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,option = string.find(line, "^%S+ %-(%S+)")

			local port = tonumber(string.match(SetManager:Get("hub_address"), ":(%d+)") or 411)
			local stype = tServer.Port[port] and tServer.Port[port].stype or "Port not running!"
 			local _, ports = tFuncs:GetPorts(ssid, {["Regular"] = "[R]",["Secure"] = "[S]", ["Gui"] = "[G]",})

			if not option then
				local res = Lang(ssid, 0, tHub:GetBotName()).."\n\n\n\t"..string.rep("*",80)..
					"\n\t\t\t"..Lang(ssid, 0, "Current hub status:").."\n\t"..string.rep("*",80)..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Hub name").."\t\t: "..SetManager:Get("hub_name")..""..
					"\n\t"..Lang(ssid, 0, "Hub address").."\t\t: "..(SetManager:Get("hub_address") or "unknown").." ("..stype..")"..
					"\n\t"..Lang(ssid, 0, "Hub IP").."\t\t\t: "..(tHubSettings.Main.sHubIp or "unknown")..
					"\n\t"..Lang(ssid, 0, "Running Ports").."\t\t: "..(ports or "")..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Hub uptime").."\t\t: "..tFuncs:HubUpTime()..
					"\n\t"..Lang(ssid, 0, "Application uptime").."\t\t: "..(tStats.Hub.AppStarted and tFuncs:ConvertOsTime(tStats.Hub.AppStarted) or "Not available")..
					"\n\t"..Lang(ssid, 0, "Hub soft").."\t\t\t: ".._APPNAME..
					"\n\t"..Lang(ssid, 0, "Hub version").."\t\t: ".._APPVER..
					"\n\t"..Lang(ssid, 0, "Hub revision").."\t\t: ".. (tSettings.General.svnversion or "Not available")..
					"\n\t"..Lang(ssid, 0, "Memory usage").."\t\t: (Current) "..tFuncs:ConvertShare(math.ceil(collectgarbage("count")*1024)).." (Peak) "..tFuncs:ConvertShare(math.ceil((tStats.Hub.MaxMem or 0)*1024))..
					"\n"..
--					"\n\t"..Lang(ssid, 0, "Server Operating System").."\t: "..osdesc..
--					"\n\t"..Lang(ssid, 0, "Server Os Release         ").."\t: "..uname("release")..
--					"\n"..
--					"\n\t"..Lang(ssid, 0, "OS clock time: ")..socket.gettime()..
					"\n\t"..Lang(ssid, 0, "Full cycle times:")..
					"\n\t"..Lang(ssid, 0, "Complete thread").."\t\t: "..tStats.Timers.Thread.." (Max: "..tStats.Timers.ThreadMax..")"..
					"\n\t"..Lang(ssid, 0, "Receiving data").."\t\t: "..tStats.Timers.ReadSockets.." (Max: "..tStats.Timers.ReadSocketsMax..")"..
					"\n\t"..Lang(ssid, 0, "Parsing commands").."\t\t: "..tStats.Timers.ParseCmds.." (Max: "..tStats.Timers.ParseCmdsMax..")"..
					"\n\t"..Lang(ssid, 0, "Sending function").."\t\t: "..tStats.Timers.SendFunction.." (Max: "..tStats.Timers.SendFunctionMax..")"..
					"\n\t"..Lang(ssid, 0, "Sending sockets").."\t\t: "..tStats.Timers.SendSockets.." (Max: "..tStats.Timers.SendSocketsMax..")"
				res = res..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Single client max times:")..
					"\n\t"..Lang(ssid, 0, "Receiving data").."\t\t: "..tStats.Timers.ReadSingleSocketMax.." "..(tStats.Timers.ReadSingleSocketMaxId or "")..
					"\n\t"..Lang(ssid, 0, "Parsing commands").."\t\t: "..tStats.Timers.ParseSingleCmdMax.." "..(tStats.Timers.ParseSingleCmdMaxId or "")..
					"\n\t"..Lang(ssid, 0, "Sending function").."\t\t: "..tStats.Timers.SendSingleMax.." "..(tStats.Timers.SendSingleMaxId or "")..
					"\n\t"..Lang(ssid, 0, "Sending zlib").."\t\t: "..tStats.Timers.SendZlibMax.." "..(tStats.Timers.SendZlibMaxId or "")..
					"\n\t"..Lang(ssid, 0, "Sending socket").."\t\t: "..tStats.Timers.SendSingleSocketMax.." "..(tStats.Timers.SendSingleSocketMaxId or "")..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Online users").."\t\t: (Current) "..(tStats.Hub.Users.Total or 0).." (Peak) "..(tStats.Hub.Users.TotalMax or 0)..
					"\n\t"..Lang(ssid, 0, "Total bytes send").."\t\t: "..(tFuncs:ConvertShare(tStats.Tcp.Sent) or "")..
					"\n\t"..Lang(ssid, 0, "Total bytes recieved").."\t: "..(tFuncs:ConvertShare(tStats.Tcp.Received) or "")..
					"\n\t"..Lang(ssid, 0, "Sending\t\t").."\t: (Current) "..(tFuncs:ConvertShare(tStats.Tcp.SentSpeed or 0)).."/s".." (Peak) "..(tFuncs:ConvertShare(tStats.Tcp.SentPeak or 0)).."/s"..
					"\n\t"..Lang(ssid, 0, "Receiving\t").."\t: (Current) "..(tFuncs:ConvertShare(tStats.Tcp.ReceivedSpeed or 0)).."/s".." (Peak) "..(tFuncs:ConvertShare(tStats.Tcp.ReceivedPeak or 0)).."/s"..
					"\n\t"..Lang(ssid, 0, "Number of socketsends").."\t: "..tStats.Tcp.SendNow.." (Delayed: "..(tStats.Tcp.SendDelayed or "")..")"..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Errors:")..
					"\n\t"..Lang(ssid, 0, "Login Timeouts").."\t\t: "..(tLogs.Errors.LoginTimeouts or 0)..
					"\n\t"..Lang(ssid, 0, "Socket Create").."\t\t: "..(tLogs.Errors.CreateSocket or 0)..
					"\n\t"..Lang(ssid, 0, "Socket Timeouts").."\t\t: "..(tLogs.Errors.Socket or 0)..
					"\n\t"..Lang(ssid, 0, "Secure Socket Initiation").."\t: "..(tLogs.Errors.SSLWrap or 0)..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Local time").."\t\t: "..ToDate(os.time())..
					"\n\n\t"..string.rep("*",80)
				return true, res
			elseif option == "r" or option == "reset" then
				tStats.Timers.ThreadMax = 0
				tStats.Timers.ReadSocketsMax = 0
				tStats.Timers.ParseCmdsMax = 0
				tStats.Timers.SendFunctionMax = 0
				tStats.Timers.SendSocketsMax = 0
				tStats.Timers.ReadSingleSocketMax = 0
				tStats.Timers.ParseSingleCmdMax = 0
				tStats.Timers.SendSingleMax = 0
				tStats.Timers.SendZlibMax = 0
				tStats.Timers.SendSingleSocketMax = 0
				tStats.Timers.ReadSingleSocketMaxId = false
				tStats.Timers.ParseSingleCmdMaxId = false
				tStats.Timers.SendSingleMaxId = false
				tStats.Timers.SendZlibMaxId = false
				tStats.Timers.SendSingleSocketMaxId = false
				return true, Lang(ssid, 0, "Core statistics have been reset.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tShow hubcore statistics", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show Hub Core Statistics", "!%s"},
		},
	},

	["convstats"] = {	--// temporary addon to help track down conversion errors
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,option = string.find(line, "^%S+ %-(%S+)")
 			local _, ports = tFuncs:GetPorts(ssid, {["Regular"] = "[R]",["Secure"] = "[S]", ["Gui"] = "[G]",})

			if not option then
				local str = "\n\n\t"..string.rep("*",80)..
					"\n\t\t\t"..Lang(ssid, 0, "Current hub status:").."\n\t"..string.rep("*",80).."\n"..
					"\n\t"..Lang(ssid, 0, "Hub name").."\t\t: "..SetManager:Get("hub_name")..""..
					"\n\t"..Lang(ssid, 0, "Hub IP").."\t\t\t: "..(tHubSettings.Main.sHubIp or "unknown")..
					"\n\t"..Lang(ssid, 0, "Running Ports").."\t\t: "..(ports or "")..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Hub soft").."\t\t\t: ".._APPNAME.." ".._APPVER.." ".. (tSettings.General.svnversion or "Not available")..
					"\n\t"..Lang(ssid, 0, "Memory usage").."\t\t: (Current) "..tFuncs:ConvertShare(math.ceil(collectgarbage("count")*1024)).." (Peak) "..tFuncs:ConvertShare(math.ceil((tStats.Hub.MaxMem or 0)*1024))..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Hub NMDC charset").."\t\t: "..tHubSettings.Main.sCharset..
					"\n\t"..Lang(ssid, 0, "Hub Locale").."\t\t: "..tHubSettings.Main.sLocale..
					"\n\t"..Lang(ssid, 0, "Lua SetLocale").."\t\t: all\t\t"..os.setlocale(nil, "all") ..
					"\n\t\t\t\t: collate\t\t"..(os.setlocale(nil, "collate") or "unknown")..
					"\n\t\t\t\t: ctype\t\t"..(os.setlocale(nil, "ctype") or "unknown")..
					"\n\t\t\t\t: monetary\t"..(os.setlocale(nil, "monetary") or "unknown")..
					"\n\t\t\t\t: numeric\t\t"..(os.setlocale(nil, "numeric") or "unknown")..
					"\n\t\t\t\t: time\t\t"..(os.setlocale(nil, "time") or "unknown")..
					"\n"

					if wx and wx.wxLocale then
						local plat = wx.wxPlatformInfo.Get()
						str = str.."\n\t"..Lang(ssid, 0, "WX Information:")..
						"\n\tGetSystemEncodingName\t: "..wx.wxLocale.GetSystemEncodingName()..
						"\n\tGetLanguageName\t\t: "..wx.wxLocale.GetLanguageName(wx.wxLocale.GetSystemLanguage())..
						"\n\tGetOSMajorVersion\t\t: "..plat:GetOSMajorVersion()..
						"\n\tGetOSMinorVersion\t\t: "..plat:GetOSMinorVersion()..
						"\n\tGetOperatingSystemId\t: "..plat:GetOperatingSystemId()..
						"\n\tGetOSFamilyName\t\t: "..plat:GetOperatingSystemFamilyName()..
						"\n\tGetArchName\t\t: "..plat:GetArchName()..
						"\n\tGetEndiannessName\t: "..plat:GetEndiannessName()..
						"\n\tGetOsDescription\t\t: "..wx.wxGetOsDescription()..
						"\n\tGetOSVersion\t\t: "..table.concat({wx.wxGetOsVersion()}, ", ")..
						"\n"
					end

					str = str.."\n\t"..Lang(ssid, 0, "Conversion Error Count:")..
					"\n\t"..Lang(ssid, 0, "From UTF-8 to local charset").."\t: "..(tLogs.Errors.ToLocale or 0)..
					"\n\t"..Lang(ssid, 0, "From local charset to UTF-8").."\t: "..(tLogs.Errors.ToUTF8 or 0)..
					"\n"..
					"\n\t"..Lang(ssid, 0, "Local time").."\t\t: "..ToDate(os.time())..
					"\n\n\t"..string.rep("*",80)

				return true, str
			elseif option == "r" or option == "reset" then
				tLogs.Errors.ToLocale = 0
				tLogs.Errors.ToUTF8 = 0
				return true, Lang(ssid, 0, "Protocol Conversion statistics have been reset.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tShow Protocol Conversion Statistics", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show Protocol Conversion Statistics", "!%s"},
		},
	},

	["showsid"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vsid = string.find(line, "^%S+ (%w%w%w%w)$")
			if vsid then
				if tClient[vsid] then
					return true, "User with SID: "..vsid.." found: Nick: <"..(tUser:GetNick(vsid) or "<nick unknown>").."> Ip: "..tUser:GetIp(vsid)
				else
					return false, "User with SID: "..vsid.." not found."
				end
			else
				return false, Lang(ssid, 0, "Error: Use %s <sid>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <sid>\t\t\t\t\tShow nick and IP of a user-ID (SID)", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show SID info", "!%s %[line:SID]"},
		},
	},

	["trafficstats"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			--// Build stats
			local nmdcstats = {}
			for name, data in pairs(NMDC_Commands) do
				table.insert(nmdcstats, Padding(name, 30).."\t"..Padding((tStats.Counters.NMDC[name] and tStats.Counters.NMDC[name].Received or 0).." ("..(tStats.Counters.NMDC[name] and tStats.Counters.NMDC[name].Blocked or 0)..") ", 14).." \t"..Padding(tFuncs:ConvertShare(tStats.Counters.NMDC[name] and tStats.Counters.NMDC[name].DataIn or 0),14))
			end
			table.sort(nmdcstats, SortLowercase)
			local adcstats = {}
			for name, data in pairs(ADC_Commands) do
				table.insert(adcstats, Padding(name, 30).."\t"..Padding((tStats.Counters.ADC[name] and tStats.Counters.ADC[name].Received or 0).." ("..(tStats.Counters.ADC[name] and tStats.Counters.ADC[name].Blocked or 0)..") ", 14).." \t"..Padding(tFuncs:ConvertShare(tStats.Counters.ADC[name] and tStats.Counters.ADC[name].DataIn or 0),14))
			end
			table.sort(adcstats, SortLowercase)

--~ 			tStats.Counters[protocol][stype].Received = (tStats.Counters[protocol][stype].Received or 0) + number
--~			tStats.Counters[protocol][stype].DataIn = (tStats.Counters[protocol][stype].DataIn or 0)+(msg and #msg or 0)

			local looptable = { {"MC", "MC"}, {"PM", "PM"}, {"HC", "HC"}, {"$MyINFO", "BINF"}, {"SA", "FSCH"},{"SP", "BSCH"}, {"$SR", "DRES"}, {"$ConnectToMe", "DCTM"}, {"$RevConnectToMe", "DRCM"} }
			local restable = {}
			for _,vars in ipairs(looptable) do
				table.insert(restable,
					Padding((tStats.Counters.NMDC[vars[1]] and tStats.Counters.NMDC[vars[1]].Received or 0).." ("..(tStats.Counters.NMDC[vars[1]] and tStats.Counters.NMDC[vars[1]].Blocked or 0)..")", 14).." \t"..Padding(tFuncs:ConvertShare(tStats.Counters.NMDC[vars[1]] and tStats.Counters.NMDC[vars[1]].DataIn or 0), 14).." \t"..
					Padding((tStats.Counters.ADC[vars[2]] and tStats.Counters.ADC[vars[2]].Received or 0).." ("..(tStats.Counters.ADC[vars[2]] and tStats.Counters.ADC[vars[2]].Blocked or 0)..")", 14).." \t"..Padding(tFuncs:ConvertShare(tStats.Counters.ADC[vars[2]] and tStats.Counters.ADC[vars[2]].DataIn or 0), 14)
				)
			end

			local reply = "\n\n\t"..string.rep("_",80)..
				"\n\t"..Lang(ssid, 0, "Incoming traffic statistics: ")..
				"\n\n\t"..string.rep("_",80)..
				"\n\t\t\t\tNMDC\t\t\t\tADC"..
				"\n\tOnline users:             \t"..Padding(tStats.Hub.Users.NMDC, 14).." \t\t\t"..Padding(tStats.Hub.Users.ADC, 14)..
				"\n\tMainchat messages: \t"..restable[1]..
				"\n\tPersonal messages:  \t"..restable[2]..
				"\n\tHub commands:       \t"..restable[3]..
				"\n\tMyINFO / BINF:        \t"..restable[4]..
				"\n\tActive searches:      \t"..restable[5]..
				"\n\tPassive searches:      \t"..restable[6]..
				"\n\tPassive searchresults \t"..restable[7]..
				"\n\tActive connections: \t"..restable[8]..
				"\n\tPassive connections \t"..restable[9]..
				"\n\n\t"..string.rep("_",80)..
				"\n\tNMDC Details:\n\t\t"..table.concat(nmdcstats, "\n\t\t")..
				"\n\n\t"..string.rep("_",80)..
				"\n\tADC Details: \n\t\t"..table.concat(adcstats, "\n\t\t")..
				"\n\n\t"..string.rep("_",80)
			return true, reply
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow traffic statistics of the hub", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Traffic statistics", "!%s"},
		},
	},

	["userstats"] = {
		["allowed"] = function(ssid,line, cmd)
--~ 			return false
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+ (%S+)")
			if vnick then
				local vsid, nick = tNick:GetSid(vnick)
				if vsid then
					--// Build stats
					local userstats = {}
					local base = tClient[vsid].tStats.Counters

					if tClient[vsid].sProtocol == "NMDC" and not tClient[vsid].bBot then
						for name, data in pairs(NMDC_Commands) do
							table.insert(userstats, Padding(name, 30).."\t"..Padding((base[name] and base[name].Received or 0).." ("..(base[name] and base[name].Blocked or 0)..") ", 14).." \t"..Padding(tFuncs:ConvertShare(base[name] and base[name].DataIn or 0),14))
						end
					elseif tClient[vsid].sProtocol == "ADC" and not tClient[vsid].bBot then
						for name, data in pairs(ADC_Commands) do
							table.insert(userstats, Padding(name, 30).."\t"..Padding((base[name] and base[name].Received or 0).." ("..(base[name] and base[name].Blocked or 0)..") ", 14).." \t"..Padding(tFuncs:ConvertShare(base[name] and base[name].DataIn or 0),14))
						end
					else
						return false, "No statistics available."
					end
					table.sort(userstats, SortLowercase)

		--~ 			tStats.Counters[protocol][stype].Received = (tStats.Counters[protocol][stype].Received or 0) + number
		--~			tStats.Counters[protocol][stype].DataIn = (tStats.Counters[protocol][stype].DataIn or 0)+(msg and #msg or 0)

					local looptable = { {"MC", "MC"}, {"PM", "PM"}, {"HC", "HC"}, {"$MyINFO", "BINF"}, {"SA", "SA"},{"SP", "SP"}, {"$SR", "DRES"}, {"$ConnectToMe", "DCTM"}, {"$RevConnectToMe", "DRCM"} }
					local restable = {}
					local num = 1
					if tClient[vsid].sProtocol == "ADC" then num = 2 end

					for _,vars in ipairs(looptable) do
						table.insert(restable, Padding((base[vars[num]] and base[vars[num]].Received or 0).." ("..(base[vars[num]] and base[vars[num]].Blocked or 0)..")", 14).." \t"..Padding(tFuncs:ConvertShare(base[vars[num]] and base[vars[num]].DataIn or 0), 14))
					end

					local reply = "\n\n\t"..string.rep("_",80)..
						"\n\t"..Lang(ssid, 0, "Incoming traffic statistics for: ")..nick..
						"\n\n\t"..string.rep("_",80)..
						"\n\tMainchat messages: \t"..restable[1]..
						"\n\tPersonal messages:  \t"..restable[2]..
						"\n\tHub commands:       \t"..restable[3]..
						"\n\tMyINFO / BINF:        \t"..restable[4]..
						"\n\tActive searches:      \t"..restable[5]..
						"\n\tPassive searches:      \t"..restable[6]..
						"\n\tPassive searchresults \t"..restable[7]..
						"\n\tActive connections: \t"..restable[8]..
						"\n\tPassive connections \t"..restable[9]..
						"\n\n\t"..string.rep("_",80)..
						"\n\tDetails:\n\t\t"..table.concat(userstats, "\n\t\t")..
						"\n\n\t"..string.rep("_",80)
					return true, reply
				else
					return false, "User not found."
				end
			else
				--// Show hub traffic stats
				return false, Lang(ssid, 0, "Error: Use %s <nickname>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nickname>\t\t\t\tShow traffic statistics of user", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Traffic statistics of user", "!%s %[line:Nickname]"},
			{2, 14, "Traffic statistics of user (selected)", "!%s %[userNI]"},
		},
	},

	["userinfo"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,targetnick = string.find(line, "^%S+%s+(%S+)")
			local targetsid, targetlevel, reply, data, history = ssid, -1, "", false, {}
			local ctime = socket.gettime()

			if not targetnick or targetnick == "" then
				targetnick = tClient[ssid].sNick
			else
				targetsid = tNick:GetSid(targetnick)
				if targetsid then
					targetnick = tClient[targetsid].sNick or "<nick unknown>"
				end
			end

			--// Check permissions
			local reg, rnick = tNick:GetReg(targetnick)	-- tNick:GetReg returns table with reg info {reg.sProfile, reg.iLevel} + real nick
			if reg then
				targetlevel = reg.iLevel
				targetnick = rnick
			end

			if targetlevel >= tClient[ssid].iProfile and targetsid ~= ssid and tClient[ssid].iProfile ~= 10 then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			local regged, reginfo = HubCmd["reginfo"]["cmd"](ssid, "!reginfo "..targetnick, "!reginfo")
			if not regged then reginfo = "" end
			local banned, baninfo = HubCmd["baninfo"]["cmd"](ssid, "!baninfo "..targetnick, "!baninfo", true)
			if not banned then baninfo = "" end

			if targetsid then
				data = tClient[targetsid]

				-- Chatrooms
				if data.sProtocol == "ChatRoom" then
					local levelroom = false

					-- private rooms
					if data.bPrivate then
						local room = tPrivateRooms[tClient[targetsid].sNick]
						for lnick, nick in pairs(room.tMembers) do
							if tNick:GetSid(nick) then
								reply = reply.."\n\t\t\t"..Padding(nick, 20).."\t(Online)"
							else
								reply = reply.."\n\t\t\t"..Padding(nick, 20).."\t(Offline)"
							end
						end
						for lnick, nick in pairs(data.tMembersTemp) do
							if tNick:GetSid(nick) then
								reply = reply.."\n\t\t\t"..Padding(nick, 20).."\t(Unregged Online)"
							else
								reply = reply.."\n\t\t\t"..Padding(nick, 20).." \t(Unregged Offline)"
							end
						end
					else

						-- opchat or feed room
						if targetsid == "BBBB" or targetsid == "CCCC" then --// opchat and feed
							local name = (targetsid == "BBBB" and "opchat" or "feed")
							local users = {}
							for vnick, vdata in pairs(tAccounts) do
								if tProfiles[vdata.sProfile] and (tProfiles[vdata.sProfile].iLevel == 10 or (tProfiles[vdata.sProfile].tCmds and (tProfiles[vdata.sProfile].tCmds.all or tProfiles[vdata.sProfile].tCmds[name]))) then
									table.insert(users, Padding(vdata.sNick, 20).."\t("..(tNick:GetSid(vdata.sNick) and "Online" or "Offline, last online: "..(ToDate(vdata.iLastLogin) or "never"))..")")	--// Needs fixing: iLastOnline needs to be added to accountdata
								end
							end
							table.sort(users, SortLowercase)
							reply = "\n\t\t\t"..table.concat(users, "\n\t\t\t").."\n\n"

						--// Normal chatroom based on profilelevel
						else
							levelroom = true
							local profiles = tUser:GetHighProfiles(data.iProfile)
							table.sort(profiles, SortLowercase)
							reply = Lang(ssid, 0,"All users with profilelevel %s or higher", data.iProfile).."\n\t "..Lang(ssid, 0, "Profiles:\t\t")..table.concat(profiles, ", ").."\n"
						end
					end

					reply = "\n\t"..string.rep("_",80)..
					"\n\t"..Lang(ssid, 0, "Showing info of: ")..targetnick ..
					(targetsid == "BBBB" and "\n\t "..Lang(ssid, 0, "Operator chatroom") or "")..
					(targetsid == "CCCC" and "\n\t "..Lang(ssid, 0, "Feedroom") or "")..
					(data.bPrivate and "\n\t Private room created by: "..data.sOwner or "")..
--~ 					"\n\t\t "..Lang(ssid, 0, "Chatroom:\t")..data.sNick..
					(levelroom and "\n\t "..Lang(ssid, 0, "Level:\t\t")..data.iProfile or "")..
					"\n\t "..Lang(ssid, 0, "Members:\t")..reply
					return true, reply
				elseif data.bBot then
					return true, "This is a bot, no userinfo available."
				end

			-- offline users
			elseif targetnick and tLogs.Nick[string.lower(targetnick)] then
				data = tLogs.Nick[string.lower(targetnick)]
				local regg = tNick:GetReg(targetnick)
--~ 				if regg then
--~ 					data.sProfile = regg.sProfile
--~ 					data.iProfile = regg.iLevel
--~ 				else
--~ 					data.sProfile = "User"
--~ 					data.iProfile = -1

--~ 				end
			else
				if regged then
					reply = reply..reginfo
				end
				if banned then
					reply = reply..baninfo
				end
				if reply == "" then
					return false, Lang(ssid, 0, "User <%s> can't be found.", targetnick)
				else
					return true, reply
				end
			end

			if tLogs.Nick[string.lower(targetnick)] then
				history = tLogs.Nick[string.lower(targetnick)].tIPHistory
				--// If history = nil then the userdata has been purged from tLogs.Nick[nick]
			end

			if data then
--~ 					local ok, err = geoip.open("./lib/GeoIP.dat")

				--// Gather information
				local rev = data.sReverseDNS or "N/A"
				--// tag = <++ V:0.674,M:A,H:0/9/19,S:1>
				--local _,_,client,ap,us,regged,op,slots = string.find((data.sTag or ""),"<(.+)%,M:(%S+)%,H:(%d+)%/(%d+)%/(%d+)%,S:(%d+).*>")

				--// Get kicks
				local kicks = 0
				local times = 0
				if tBans.ipban[data.sIP] then
					times = tBans.ipban[data.sIP].iCount or 0
					kicks = times..(times==0 and "" or (" (Last reason: "..(tBans.ipban[data.sIP].sReason or "<not specified>") ..")"))
				end

				--// Get warnings
				local warns = 0
				if tBans.warn[data.sNick:lower()] then warns = (tBans.warn[data.sNick:lower()].iCount or 0) .." (Last reason: "..(tBans.warn[data.sNick:lower()].sReason or "<not specified>") ..")" end

				--// Get TLS state
				local secure = ""
				if data.bSecure then secure = " (Secure: TLS)" end

				--// Get IP Binding
				local ipbind = ""
				local lnick = targetnick:lower()
				if tAccounts[lnick] and tAccounts[lnick].sBindIP and tAccounts[lnick].sBindIP ~= "" then
					ipbind = Lang(ssid, 0, " (Account bound to IP: %s)", tAccounts[lnick].sBindIP)
				end

				--// Protected
				local protected = ""
				if tUser:IsProtected(nil, data.sNick, ip) then protected = " (Protected)" end

				--//Users with same IP
				local ipmore = ""
				if tIp:GetSids(data.sIP) then
					for num, targetsid in ipairs(tIp:GetSids(data.sIP)) do
						local nick = tClient[targetsid].sNick or "<nick unknown/login>"
						if nick:lower() ~= data.sNick:lower() then
							ipmore = ipmore .." "..nick
						end
					end
					if ipmore ~= "" then ipmore = " (Online with same IP: "..ipmore .." )" end
				end

				--~ iphistory[client.iLoginTime]  = { ["sIP"] = client.sIP, ["iLogoutTime"] = os.time() }
				local histtxt = ""
				local histsort = {}
				if history then
					for num, hdata in ipairs(history) do
						table.insert(histsort, hdata.iLoginTime .." \t "..hdata.sIP.."  \t"..(hdata.sProtocol or "    ").."  \tFrom: "..ToDate(hdata.iLoginTime) .."\t\t\To: "..ToDate(hdata.iLogoutTime) .."\n")
					end
					table.sort(histsort)
					for i, line in ipairs(histsort) do
						histtxt = string.gsub(line, "^%d+%s", "") ..histtxt
					end
				elseif not targetsid then
					histtxt = "\t "..data.sIP.."  \t"..(data.sProtocol or "    ").."  \tFrom: "..(data.iLoginTime and ToDate(data.iLoginTime) or "\t").."\t\t\To: "..(data.iLogoutTime and ToDate(data.iLogoutTime) or "")
				end
				if targetsid then histtxt = "\t "..data.sIP.."  \t"..(data.sProtocol or "    ").."  \tFrom: "..ToDate(data.iLoginTime) .."\t\t\To: Still connected\n"..histtxt end

				--// Blocks
				local blocktext = "<none>"
				local userblocks = tFuncs:GetBlocks(data.sIP)
				if #userblocks > 0 then
					blocktext = "\n\t\t\t "..table.concat(userblocks, "\n\t\t\t ")
				end

				local pwd = ""
				if reg and reg.sPassword == "" then
					pwd = "\n\t\t "..Lang(ssid, 0, " (The user has not yet set a password for this account)")
				end

				local locale = ""
				if data.sLocale then
					locale = "\n\t\t "..Lang(ssid, 0, "Locale: ")..(data.sLocale or "")
				end

				--// SU field in INF (client-client supports)
				local suinf,supportsinf = {}, ""
				if data.tSupportsInf then
					for name,value in pairs(data.tSupportsInf) do
						if value then table.insert(suinf, name) end
					end
					if #suinf > 0 then
						supportsinf = " / INF: "..table.concat(suinf, ",")
					end
				end

				--// Total online time
				local onlinetime = false
				if targetsid then
					onlinetime = data.iTotalOnlineTime + (ctime - data.iLoginTime)
				else
					if reg then
						onlinetime = reg.iTotalOnlineTime
					else
						onlinetime = data and data.iTotalOnlineTime
					end
				end
				local onlineduration = onlinetime and tFuncs:Duration(onlinetime) or "Not available"
				if targetsid then onlineduration = onlineduration.."   (This session: "..(tFuncs:Duration(ctime-data.iLoginTime) or "not available")..")" end

				--// Output information
				local reply = "\n\n\n\t"..Lang(ssid, 0, "Showing info of: ")..targetnick..
					"\n\t"..string.rep("_",80)..
					"\n\t"..Lang(ssid, 0, "General information:")..
					"\n\t"..string.rep("_",80)..
					"\n\t\t "..Lang(ssid, 0, "User: ")..data.sNick..protected..(not history and " \t(User data purged: offline for more than "..tHubSettings.Main.iKeepDays.." days)" or "")..
					pwd..
					((history or targetsid) and
					"\n\t\t "..Lang(ssid, 0, "SID: ")..(targetsid or "<USER OFFLINE>") ..
					"\n\t\t "..Lang(ssid, 0, "CID: ")..(data.sCID or "") ..
					"\n\t\t "..Lang(ssid, 0, "Protocol: ")..tostring(data.sProtocol or "")..secure
					or "")..
					"\n\t\t "..Lang(ssid, 0, "Profile: ")..tostring(reg and reg.sProfile or "User").." ("..(reg and reg.iLevel or -1) ..")"..
					"\n\t\t "..Lang(ssid, 0, "IP: ")..data.sIP ..ipbind ..ipmore..
					((history or targetsid) and
					"\n\t\t "..Lang(ssid, 0, "HubPort: ")..tostring(data.iPort) ..
					"\n\t\t "..Lang(ssid, 0, "Active Port: ")..tostring(data.sActivePort or "not available").." "..tostring(data.sActiveProtocol or "")..
					"\n\t\t "..Lang(ssid, 0, "Search Port UDP: ")..tostring(data.sSearchPort or "not available") ..
					"\n\t\t "..Lang(ssid, 0, "Supports: ")..tostring(data.sSupports)..tostring(supportsinf)..
					"\n\t\t "..Lang(ssid, 0, "Countrycode: ")..tostring(data.sCountryCode or "")..
					locale..

--~ 					"\n\t\t "..Lang(ssid, 0, "DNS: ")..(rev or "N/A")..
					"\n"..
					"\n\t\t "..Lang(ssid, 0, "Tag: ")..(data.sTag or "")..
					"\n\t\t "..Lang(ssid, 0, "Description: ")..(data.sDescription or "")
					or "")..
					"\n\t\t "..Lang(ssid, 0, "Email: ")..(data.sEmail or "")..
					((history or targetsid) and
					"\n\t\t "..Lang(ssid, 0, "Active / Passive: ")..(data.bActive and "Active" or "Passive")..
					"\n\t\t "..Lang(ssid, 0, "In hubs as user: ")..(data.iNormalHubs or "N/A") ..
					"\n\t\t "..Lang(ssid, 0, "In hubs as regged: ")..(data.iRegHubs or "N/A") ..
					"\n\t\t "..Lang(ssid, 0, "In hubs with key: ")..(data.iOpHubs or "N/A") ..
					"\n\t\t "..Lang(ssid, 0, "Slots open: ")..(data.iSlots or "N/A") ..
					"\n\t\t "..Lang(ssid, 0, "Sharesize: ")..tFuncs:ConvertShare(tonumber(data.bShareHidden and 0 or (data.iShareSize or 0))) .." ("..(data.bShareHidden and "Hidden" or (data.iShareSize or 0))..")"
					or "")..
					"\n\t\t "..Lang(ssid, 0, "Total time online: ")..onlineduration..
					"\n"..
					"\n\t\t "..Lang(ssid, 0, "Times warned: ")..warns ..
					"\n\t\t "..Lang(ssid, 0, "Times kicked: ")..kicks ..
					"\n"..
					"\n\t\t "..Lang(ssid, 0, "Blocks/Filters: ")..blocktext ..
					"\n"..
					"\n\t"..Lang(ssid, 0, "IP History: ")..
					"\n\t"..string.rep("_",80)..
					"\n"..histtxt..
					"\n\t"..string.rep("_",80).."\n\n"
--				return true, reply ..(reginfo or "") .. (baninfo or "")
				return true, reply ..(baninfo or "")
			elseif not targetnick then
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <nick>")
			else
				if banned then
					return true, "User not found, but related ban(s) exist:\n"..baninfo
				else
					return false, Lang(ssid, 0, "User <"..targetnick.."> can't be found.")
				end
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> \t\t\t\t\tShow online and offline user information", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Show user information", "!%s %[line:Nickname]"},
			{2, 14, "Show user information (selected)", "!%s %[userNI]"},
		},
	},

	["ipinfo"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local tIPs = {}
			local partial, vip, escvip = false, false, false

			if string.find(line, "%*$") then
				partial = true
				if string.find(line, "^%S+ (%d+%.%d+%.%d+%.%d+%.?)%*$") then
					vip = string.match(line, "^%S+ (%d+%.%d+%.%d+%.%d+%.?)%*$")
				elseif string.find(line, "^%S+ (%d+%.%d+%.%d+%.?)%*$") then
					vip = string.match(line, "^%S+ (%d+%.%d+%.%d+%.?)%*$")
				elseif string.find(line, "^%S+ (%d+%.%d+%.?)%*$") then
					vip = string.match(line, "^%S+ (%d+%.%d+%.?)%*$")
				elseif string.find(line, "^%S+ (%d+%.?)%*$") then
					vip = string.match(line, "^%S+ (%d+%.?)%*$")
				else
					partial = false
				end
				if partial then
					escvip = string.gsub(vip, "%.", "%%.")
				end
			else
				vip = string.match(line, "^%S+ (%d+%.%d+%.%d+%.%d+)")
			end

			if vip then
				local match = false
				local output = ""
				if partial then
					output = "\n\n\t"..Lang(ssid, 0, "Showing info on ip's matching pattern: ")..vip.."*\n\t"..string.rep("_",80).."\n"
				else
					output = "\n\n\t"..Lang(ssid, 0, "Showing info on: ")..vip.."\n\t"..string.rep("_",80).."\n"
				end

				local tOnline = {}
				for i,v in pairs(tClient) do
					if (not partial and v.sIP == vip) or (partial and string.find(v.sIP, "^"..escvip)) then
						match = true
						table.insert(tOnline, tFuncs:IP2DEC(v.sIP).." \t "..v.sIP .."\t<"..(v.sNick or "pre-login / nick unknown")..">\t\tOnline since: "..ToDate(v.iLoginTime).."\n")
					end
				end

				local histsort = {}
				ThreadLockMutex(mutexiplog)
				if not partial and tLogs.IP[vip] then
					for nick,v in pairs(tLogs.IP[vip]) do
						if nick ~= "$" then
							table.insert(histsort, tFuncs:IP2DEC(vip)..(v.iLoginTime or 0).." \t <"..nick..">\t\tUsed "..v.iCounter.." times, last From: "..ToDate(v.iLoginTime or 0).."\t\tTo: "..ToDate(v.iLogoutTime).."\n")
							match = true
						end
					end
				elseif partial then
					for ip, data in pairs(tLogs.IP) do
						if string.find(ip, "^"..escvip) then
							table.insert(tIPs, ip)
							for nick, v in pairs(tLogs.IP[ip]) do
								if nick ~= "$" then
									table.insert(histsort, tFuncs:IP2DEC(ip)..(v.iLoginTime or 0).." \t "..ip .."\t<"..nick..">\t\tUsed "..v.iCounter.." times, last From: "..ToDate(v.iLoginTime or 0).."\tTo: "..ToDate(v.iLogoutTime).."\n")
									match = true
								end
							end
						end
					end
				end
				ThreadUnlockMutex(mutexiplog)

				if tOnline[1] then
					table.sort(tOnline)
					for _,line in ipairs(tOnline) do
						output = output..(string.gsub(line, "^%S+%s", ""))
					end
				end

				if histsort[1] then
					output = output.."\n\t\t"..Lang(ssid, 0, "History: ").."\n\t"..string.rep("_",80).."\n"
					table.sort(histsort)
					for _,line in ipairs(histsort) do
						output = output..(string.gsub(line, "^%S+%s", ""))
					end
				end

				if tIPs[2] then table.sort(tIPs) end

				local banned = ""

				if not partial then
					local res, ban = HubCmd["baninfo"]["cmd"](ssid, "!baninfo "..vip, "!baninfo")
					if res then banned = ban end
				else
					for indx, ip in ipairs(tIPs) do
						local res, ban = HubCmd["baninfo"]["cmd"](ssid, "!baninfo "..ip, "!baninfo", true)
						if res then banned = banned .. ban end
					end
				end

				if match then
					return true, output..banned
				else
					if banned ~= "" then
						return true, Lang(ssid, 0, " IP [ "..vip.." ] not found, but related ban(s) exist:\n"..banned)
					elseif partial then
						return true, Lang(ssid, 0, " No IP matching pattern [ "..vip.."* ] found.")
					else
						return true, Lang(ssid, 0, " IP [ "..vip.." ] not found.")
					end
				end
			else
				return false, Lang(ssid, 0, "Error: Use "..cmd.." <ip>")
			end
		end,
		["outputtype"] = "info",
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <ip>\t\t\t\t\tShow online and offline IP information", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show IP information", "!%s %[line:IP]"},
			{2, 14, "Show IP information (selected)", "!%s %[userI4]"},
		},
	},

	["topic"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local topic, silent = false, false
			local ln = line:match("^%S+ (.*)")

			if ln and ln:find("^%-[sS]") then
				silent = true
				ln = ln:match("^%-[sS] (.*)") or false
			end

			if ln then
				topic = ln:gsub("\n", "")
			else
				topic = ""
			end

			if topic:find("%%") then
				topic = tFuncs:FillVars(topic, nil, "topic")
			end

			if #topic > tSecurity.General.TopicMaxLength then
				return false, Lang(ssid, 0, "Topic is too long, max. %s characters.", tSecurity.General.TopicMaxLength)
			end

			local old = tHubSettings.Main.sTopic or false

			if old and old == topic then
				return false, Lang(ssid, 0, "New topic same as old, nothing changed.")
			end

			local res, err = SetManager:Set("hub_topic", topic, nil, true)
			if not res then return false, err end

			local nick = tUser:GetNick(ssid) or "<unknown nick>"
			local msg = ""
			if topic == "" then
				tSend:Raw("P_ADC", "IINF DE"..ADC_E(topic).."\n")
				tSend:Raw("P_NMDC", "$HubName "..NMDC_E(SetManager:Get("hub_name")).."|")
				msg = Lang(ssid, 0, "The topic has been reset.")
			else
				tSend:Raw("P_ADC", "IINF DE"..ADC_E(topic).."\n")
				tSend:Raw("P_NMDC", "$HubName "..NMDC_E(SetManager:Get("hub_name")).." - "..NMDC_E(topic).."|")
				msg = Lang(ssid, 0, "The topic has been changed by %s to: %s", nick, topic)
			end

			if old and old ~= "" then
				msg = msg.."\n"..Lang(ssid, 0, "Old topic was : %s", old)
			end

			if tHubSettings.Main.bTopicBroadcast and not silent then
				tSend:ChatToAll(msg)
				if ssid ~= "AAAA" then
					return true, Lang(ssid, 0, "Done")
				else
					return true, err
				end
			else
				if silent then
				msg = msg:gsub("has been", "has been silently", 1)
				end
				return true, msg
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [-s] <message>\t\t\t\t\tSet topic, or use !topic without any message to clear it, -s means dont show the change in main (use !showtopic to view the current topic)", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Change topic", "!%s %[line:New topic]"},
		},
	},

	["say"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _, _, name, msg = string.find(line, "^%S+%s+(%S+) (.+)")
			if name and msg then
				local tsid = tNick:GetSid(name)
				local reg, sNick, lnick = tNick:GetReg(name)	--return reg, reg.sNick, lnick
				if reg then name = sNick end

				if tsid and tUser:IsHigher(ssid, tsid) then	-- user is online - easy check
					tSend:ChatToAll(msg, tsid)
					tFuncs:Debug("<"..tUser:GetNick(ssid).."> said as <"..name.."> : "..msg)
					return true
				elseif not reg or (reg and tClient[ssid].iProfile > reg.iLevel) then -- user not online and not regged or is regged with lower level
					--// NMDC
					tSend:Raw("P_NMDC", "<"..name.."> "..msg.."|")
					--tSend:Raw("P_NMDC", "$Quit "..name.."|")	-- dcpp bug :P
					--// ADC
					local fakesid = tFuncs:CreateSid()
					tSend:Raw("P_ADC", "BINF "..fakesid.." ID"..dc.hashpid(fakesid).." NI"..ADC_E(name).."\n".."BMSG "..fakesid.." "..ADC_E(msg).."\nIQUI "..fakesid.."\n")

					tFuncs:Debug("<"..tUser:GetNick(ssid).."> said as <"..name.."> : "..msg)
					return true
				else
					return false, Lang(ssid, 0, "This user is too powerful for you.")
				end
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick> <msg>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick> <message>\t\t\t\tImpersonate someone in mainchat", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Say as nick", "!%s %[line:Nickname to use say] %[line:Message]"},
			{2,14, "Say as nick (selected)", "!%s %[userNI] %[line:Message]"},
		},
	},

	["mc"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)

			local _,_,msg = string.find(line, "^%S+%s+(.*)")
			if msg then
				tSend:ChatToAll(msg, "AAAA")
				tFuncs:Debug("<"..tUser:GetNick(ssid).."> said as <"..tClient["AAAA"].sNick.."> : "..msg)
				return true
			else
				return false, Lang(ssid, 0, "Error: Use %s <msg>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <message>\t\t\t\t\tSend a mainchat message from hubbot", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Mainchat message from bot", "!%s %[line:Message]" },
		},
	},

	["mass"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)

			local _,_,msg = string.find(line, "^%S+ (.*)")
			if msg then
				tSend:PmToAll(msg, "AAAA")
				tFuncs:Debug(Lang(ssid, 0 , "Mass message sent from <%s> to all users: %s", tUser:GetNick(ssid), msg))
				return true, Lang(ssid, 0, "Mass message sent to all: %s", msg)
			else
				-- return error in cmd
				return false, Lang(ssid, 0, "Error: Use %s <message>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <message>\t\t\t\t\tSends a mass PM to all users", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Mass PM to everyone", "!%s %[line:Message]"},
		},
	},

	["massprofile"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,prof, msg = string.find(line, "^%S+ (%S+) (.*)")
			if prof and msg then
				local name, err = ProfileManager:GetName(ssid, prof)
				if not name then return false, err end

				tSend:PmToProf(msg, "AAAA", name)
				tFuncs:Debug(Lang(ssid, 0 , "Mass message sent from <%s> to profile %s: %s", tUser:GetNick(ssid), prof, msg))
				return true, Lang(ssid, 0, "Mass message sent to all users with profile %s: %s", prof, msg)
			else
				return false, Lang(ssid, 0, "Error: Use %s <profile> <msg>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <profile> <message>\t\t\tSends a mass PM to all users with a certain profile", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Mass PM to a certain profile", "!%s %[line:Profilename] %[line:Message]"},
		},
	},

	["report"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
--~ 			local _,_,nick, msg = string.find(line, "^%S+%s+(%S+)%s+(.+)")
			local _,_,msg = string.find(line, "^%S+%s+(.+)")
			if msg then
				--// Check if report contains nickname
				local _,_,nick = string.find(msg, "^(%S+)%s+.+")
				local vsid = false
				if nick then vsid = tNick:GetSid(nick) end
				if vsid then	msg = string.gsub(msg, "^(%S+%s)", "") 	end

				--// Pattern check the report
				local res, retmsg = tCheck:Pattern(ssid, "RE", msg)
				if not res then return false, retmsg end

				if vsid then
					tSend:ChatRoom("User report from <"..tClient[ssid].sNick.."> with IP: "..tClient[ssid].sIP.." about user <"..tClient[vsid].sNick.."> with IP: "..tClient[vsid].sIP.." because: "..retmsg, "AAAA", "BBBB")
				else
					tSend:ChatRoom("User report from <"..tClient[ssid].sNick.."> with IP: "..tClient[ssid].sIP..": "..retmsg, "AAAA", "BBBB")
				end
				return true, Lang(ssid, 0, "Report delivered.")
			else
				return false, Lang(ssid, 0, "Error: Use %s <nick> <reason>", cmd)
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick>] <reason>\t\t\t\tReport someone or something to the operator chatroom", cmd)
		end,
		["ucmd"] = {
			{2,14, "Report user to operators (selected)", "!%s %[userNI] %[line:Reason]"},
			{1,3, "Send a message to operators", "!%s %[line:Message]"},
		},
	},

	["myip"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local ip = tUser:GetIp(ssid)
			local msg = ""

			msg = msg.."\n\n\t\tYour information:\n"
			msg = msg.."\n\t\tIp: "..ip
			msg = msg.."\n\t\t"..string.rep("_",25)

--~ 			local rev, err = socket.dns.tohostname(ip)
--~ 			if not rev then rev = "Error: "..err end

--~ 			msg = msg.."\n\n\t\t"..Lang(ssid, 0, "Dns information:").."\n"
--~ 			msg = msg.."\n\t\t"..string.rep("_",25)
--~ 			msg = msg.."\n\t\t"..Lang(ssid, 0, "Requested").."    \t: "..ip
--~ 			msg = msg.."\n\t\t"..Lang(ssid, 0, "Reversed to").."  \t: "..rev

			--// Get countrycode
			if tHubSettings.Main.bUseGeoIP and tGeoIP.loaded then
				local cc, cn = tGeoIP.loaded:lookup(ip)
				msg = msg.."\n\t\t"..Lang(ssid, 0, "Countrycode").."  \t: "..(cc or "Not found")
				msg = msg.."\n\t\t"..Lang(ssid, 0, "Countryname").." \t: "..(cn or "Not found")
			end

			msg = msg.."\n\t\t"..string.rep("_",25).."\n\n"

			return true, Lang(ssid, 0, msg)
		end,
		["outputtype"] = "info",
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tShow your own IP address", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show my IP", "!%s"},
		},
	},

	["lockpm"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			if SetManager:Get("lock_pm") then return false, "Personal messages are already locked." end
			SetManager:Set("lock_pm", true)
			tSend:ChatToAll("Personal messages are now locked.")
			return true, "You have locked personal messages, only operators can PM now."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tLock personal messages, only operators can PM", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Lock personal messages", "!%s"},
		},
	},

	["unlockpm"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			if not SetManager:Get("lock_pm") then return false, "Personal messages aren't locked." end
			SetManager:Set("lock_pm", false)
			tSend:ChatToAll("Personal messages are unlocked, you can PM again.")
			return true, "You have unlocked personal messages, everyone can PM again."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tUnlock personal messages", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unlock personal messages", "!%s"},
		},
	},

	["lockmain"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			if SetManager:Get("lock_main") then return false, "Mainchat is already locked." end
			SetManager:Set("lock_main", true)
			tSend:ChatToAll("Mainchat is now locked.")
			return true, "You have locked mainchat, only operators can speak now."
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tLock mainchat, only operators can speak", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Lock mainchat", "!%s"},
		},
	},

	["unlockmain"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			if not SetManager:Get("lock_main") then return false, "Mainchat isn't locked." end

			local ret, err = SetManager:Set("lock_main", false)
			if ret then
				tSend:ChatToAll("Mainchat is unlocked, you can speak again.")
				return true, "You have unlocked mainchat, everyone can speak again."
			else
				return false, err
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tUnlock mainchat", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unlock mainchat", "!%s"},
		},
	},

	["lockhub"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)

			local res, msg = tFuncs:LockHub()
			return res, msg
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\t\tLock hub, no new connections will be allowed", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Lock hub (block new connections)", "!%s"},
		},
	},

	["unlockhub"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)

			local res, msg = tFuncs:UnlockHub()
			return res, msg
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tUnlock hub", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unlock hub (allow new connections)", "!%s"},
		},
	},

	["blockpm"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_,option = string.find(line, "^%S+%s+%S+%s+(%S+)")
			local silent = false
			if option and option == "-s" then silent = true end

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick or IP>", cmd)
			else
				local ret, msg = tFuncs:AddBlocks(ssid, vnick, silent, "pm")
				return ret, msg
			end
	--[[
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid then
				local lNick = vnick:lower()
				if tLogs.Nick[lNick] and tLogs.Nick[lNick].sIP and tLogs.Nick[lNick].sNick then
					if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then
						local res, retmsg = tBlock:AddBlock(ssid, "pm", tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP, option)
						tFuncs:Debug(retmsg)
						return res, retmsg
					end
				end
				return false, Lang(ssid, 0, "User can't be found.")
			elseif not tUser:IsHigher(ssid,vsid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			else
				local res, retmsg = tBlock:AddBlock(ssid, "pm", vnick, vip, option)
				tFuncs:Debug(retmsg)
				return res, retmsg
			end
	--]]
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tBlock user's private messages, use %s <nick> -s for stealth block, if user is offline, use ip instead of nick", cmd, cmd)
		end,
		["ucmd"] = {
			{1, 3, "Block PM of user or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Block PM of user (selected)", "!%s %[userNI]"},
		},
	},

	["showblocks"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local btables = {
				["mute"] = {"tMute", "Muted"},
				["kenny"] = {"tKenny",  "Kennylized"},
				["muppet"] = {"tMuppet", "Muppetized"},
				["connects"] = {"tConnects", "Downloads/searches blocked"},
				["pm"] = {"tPm", "PM's blocked"},
			}
			local reply = "\n\n\t"..Lang(ssid, 0, "List of all blocks:")
			for _,name in ipairs({"connects", "pm", "mute", "kenny", "muppet"}) do
				reply = reply.."\n\n\t"..btables[name][2].."\n\t"..string.rep("_",50)
				for vip, vdata in pairs(tBlocks[btables[name][1]]) do
					local silent = vdata.bSilent and "\tSILENT" or "\t"
					local vnick = vdata.sNick or ""
					local hinick, tnicks, snicks = tIp:GetNick(vip)
					reply = reply.."\n\t"..silent.."\t"..vip.."\t"..(hinick and snicks or "<user offline>").." ("..vnick..")"
				end
			end
			return true, reply
		end,
		["outputtype"] = "info",
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tShow all (partial) blocks (download, PM, mute, kenny, muppetize)", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Show all blocks", "!%s"},
		},
	},

	["unblock"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)
			local value = false

			local btables = {
				["mute"] = {"tMute", "Muted\t\t"},
				["kenny"] = {"tKenny",  "Kennylized\t"},
				["muppet"] = {"tMuppet", "Muppetized\t"},
				["connects"] = {"tConnects", "Downloads/searches"},
				["pm"] = {"tPm", "PM's\t\t"},
			}

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "Only operators or higher can unblock offline users.")
			elseif vsid and not tUser:IsHigher(ssid, vsid) and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			if not vip then value = vnick else value = vip end	--// If user isn't online, use the nick to find the proper IP in blocktable
			local reply = "\n\n\t"..Lang(ssid, 0, "Result of clearing all blocks using search string '"..value.." ':\n\t"..string.rep("_",50))
			local found = false

			if vnick == "*" then
				if not tUser:IsAllowed(ssid, "setcmd") then
					return false, Lang(ssid, 0, "You're not allowed to remove multiple blocks at once. Your profile is not allowed to use 'setcmd'.")
				end
				local _,_,btype = string.find(line, "^%S+%s+%S+%s+(%S+)")
				if not btype or not btables[btype] then return false, Lang(ssid, 0, "Missing or invalid block type, use %s * <blocktype>  Blocktypes: mute,kenny,muppet,connects,pm", cmd) end
				tBlocks[btables[btype][1]] = {}
				tFuncs:SaveSettings("blocks")
				return true, Lang(ssid, 0, "Cleared all blocks for type: %s.", btype)
			else

				for _,name in ipairs({"connects", "pm", "mute", "kenny", "muppet"}) do
					reply = reply.."\n\t"..btables[name][2].."\t"
					local res, retmsg, silent = tBlock:DelBlock(ssid, name, value)
					if res then found = true end
					reply = reply..retmsg
				end
			end

			if found then
				return true, reply.."\n"
			else
				return false,  "User or IP wasn't found."
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tRemove all (partial) blocks for a user or Ip (download, PM, mute, kenny, muppetize\n\t%s * <blocktype>\t\t\tRemove all blocks of a certain type. Blocktypes: mute,kenny,muppet,connects,pm", cmd,cmd)
		end,
		["ucmd"] = {
			{1, 3, "Remove all blocks for a user or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Remove all blocks for a user (selected)", "!%s %[userNI]"},
		},
	},

	["unblockpm"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)
			local value = false

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "Only operators or higher can unblock offline users.")
			elseif vsid and not tUser:IsHigher(ssid,vsid) and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			if not vip then value = vnick else value = vip end	--// If user isn't online, use the nick to find the proper IP in blocktable
			local res,retmsg,silent = tBlock:DelBlock(ssid, "pm", value)
			tFuncs:Debug(retmsg)
			return res, retmsg
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tAllow a user to send private messages again", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unblock PM of user or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Unblock PM of user (selected)", "!%s %[userNI]"},
		},
	},

	["blockdownload"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_,option = string.find(line, "^%S+%s+%S+%s+(%S+)")
			local silent = false
			if option and option == "-s" then silent = true end

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick or IP>", cmd)
			else
				local ret, msg = tFuncs:AddBlocks(ssid, vnick, silent, "connects")
				return ret, msg
			end

	--[[
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid then
				local lNick = vnick:lower()
				if tLogs.Nick[lNick] and tLogs.Nick[lNick].sIP and tLogs.Nick[lNick].sNick then
					if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then
						local res, retmsg = tBlock:AddBlock(ssid, "connects", tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP)
						if res then tSend:ADCUserIps(vsid, 1) end
						tFuncs:Debug(retmsg)
						return res, retmsg
					else
						return false, Lang(ssid, 0, "This user is too powerful for you.")
					end
				end
				return false, Lang(ssid, 0, "User can't be found.")
			elseif not tUser:IsHigher(ssid,vsid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			else
				local res, retmsg = tBlock:AddBlock(ssid, "connects", vnick, vip)
				if res then tSend:ADCUserIps(vsid, 1) end
				tFuncs:Debug(retmsg)
				return res, retmsg
			end
		--]]
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\tBlock a user from making new connections to users, use %s <nick> -s for stealth block, if user is offline, use ip instead of nick", cmd, cmd)
		end,
		["ucmd"] = {
			{1, 3, "Block downloads//searches of user or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Block downloads//searches of user (selected)", "!%s %[userNI]"},
		},
	},

	["unblockdownload"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)
			local value = false

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "Only operators or higher can unblock offline users.")
			elseif vsid and not tUser:IsHigher(ssid,vsid) and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			if not vip then value = vnick else value = vip end	--// If user isn't online, use the nick to find the proper IP in blocktable
			local res,retmsg = tBlock:DelBlock(ssid, "connects", value)
			tFuncs:Debug(retmsg)
			return res, retmsg
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\tUnblock a user from making new connections", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unblock downloads//searches of user or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Unblock downloads//searches of user (selected)", "!%s %[userNI]"},
		},
	},

	["mute"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_,option = string.find(line, "^%S+%s+%S+%s+(%S+)")
			local silent = false
			if option and option == "-s" then silent = true end

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick or IP>", cmd)
			else
				local ret, msg = tFuncs:AddBlocks(ssid, vnick, silent, "mute")
				return ret, msg
			end

--[[
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick or IP>", cmd)
			else
				local check, res = tUser:IsProtected(nil, vnick, vnick)
				if check then return false, res end
			end

			if not vsid then
				local lNick = vnick:lower()

				if tAccounts[lNick] then
					if (tClient[ssid].iProfile or -2) > (tAccounts[lNick].iLevel or -2) or tClient[ssid].iProfile == 10 then

						local check, res = tUser:IsProtected(nil, lNick, tAccounts[lNick].sLastIP)
						if check then return false, res end

						local res, retmsg = tBlock:AddBlock(ssid, "mute", lNick, tAccounts[lNick].sLastIP, silent)
						tFuncs:Debug(retmsg)
						if res and not silent then
							tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now muted.", tAccounts[lNick].sNick), "AAAA")
						end

						return res, retmsg
					end
				elseif tLogs.Nick[lNick] and tLogs.Nick[lNick].sIP and tLogs.Nick[lNick].sNick then
					if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then

						local check, res = tUser:IsProtected(nil, tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP)
						if check then return false, res end

						local res, retmsg = tBlock:AddBlock(ssid, "mute", tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP, silent)
						tFuncs:Debug(retmsg)
						if res and not silent then
							tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now muted.", vnick), "AAAA")
							return res
						end

						return res, retmsg
					end
				elseif tFuncs:IsIP(vnick) then
					--if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then
						--local check, res = tUser:IsProtected(nil, nil, vnick)	- added the check at the start instead
						--if check then return true, res end

						local res, retmsg = tBlock:AddBlock(ssid, "mute", "<unknown>", vnick, silent)
						tFuncs:Debug(retmsg)
--~ 						if res and not silent then		--// No broadcast on mute IP
--~ 							tSend:ChatToAll(Lang(ssid, 0, "User with ip %s is now muted.", vnick), "AAAA")
--~ 							return res
--~ 						end

						return res, retmsg
					--end
--~ 					return res, nil
				end
				return false, Lang(ssid, 0, "User can't be found.")
			elseif not tUser:IsHigherUser(ssid, vsid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			elseif tNick:GetSid(vnick) == ssid then
				return false, Lang(ssid, 0, "You can't mute yourself.")
			else
				local res, retmsg = tBlock:AddBlock(ssid, "mute", vnick, vip, silent)
				tFuncs:Debug(retmsg)
				if res and not silent then
					tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now muted.", vnick), "AAAA")
					return res
				end
				return res, retmsg
			end
	--]]
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tMute a user in mainchat, use %s <nick> -s for stealth mute, if user is offline, use ip instead of nick", cmd, cmd)
		end,
		["ucmd"] = {
			{1, 3, "Mute nick or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Mute user (selected)", "!%s %[userNI]"},
		},
	},

	["unmute"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local vsid,fnick = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)
			local value = false

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "Only operators or higher can unmute offline users.")
			elseif vsid and not tUser:IsHigher(ssid,vsid) and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			--// If user isn't online, use the vnick to find the proper IP in blocktable
			if not vip then
				value = vnick
			else
				vnick = fnick
				value = vip
			end

			local res,retmsg, silent = tBlock:DelBlock(ssid, "mute", value)
			tFuncs:Debug(retmsg)
			if res and vip and not silent then
				tSend:ChatToAll("User <"..vnick.."> is now unmuted.", "AAAA")
				return true
			end
			return res, retmsg
--~ 			if found and vip then
--~ 				return true, retmsg
--~ 			elseif found then
--~ 				return true, "User <"..vnick.."> is now unmuted."
--~ 			else
--~ 				return false, "User or IP wasn't muted."
--~ 			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tUnmute a user or Ip in mainchat", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unmute nick or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Unmute user (selected)", "!%s %[userNI]"},
		},
	},

	["kenny"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_,option = string.find(line, "^%S+%s+%S+%s+(%S+)")
			local silent = false
			if option and option == "-s" then silent = true end

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick or IP>", cmd)
			else
				local ret, msg = tFuncs:AddBlocks(ssid, vnick, silent, "kenny")
				return ret, msg
			end
--[[
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			else
				local check, res = tUser:IsProtected(nil, vnick, vnick)
				if check then return true, res end
			end

			if not vsid then
				local lNick = vnick:lower()
				if tAccounts[lNick] then
					if (tClient[ssid].iProfile or -2) > (tAccounts[lNick].iLevel or -2) or tClient[ssid].iProfile == 10 then

						local check, res = tUser:IsProtected(nil, lNick, tAccounts[lNick].sLastIP)
						if check then return false, res end

						local res, retmsg = tBlock:AddBlock(ssid, "kenny", lNick, tAccounts[lNick].sLastIP, silent)
						tFuncs:Debug(retmsg)
						if res and not silent then
							tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now kennylized.", vnick), "AAAA")
							return res
						end
						return res, retmsg
					end
				elseif tLogs.Nick[lNick] and tLogs.Nick[lNick].sIP and tLogs.Nick[lNick].sNick then
					if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then

						local check, res = tUser:IsProtected(nil, tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP)
						if check then return false, res end

						local res, retmsg = tBlock:AddBlock(ssid, "kenny", tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP, silent)
						tFuncs:Debug(retmsg)
						if res and not silent then
							tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now kennylized.", vnick), "AAAA")
							return res
						end
						return res, retmsg
					end
				elseif tFuncs:IsIP(vnick) then
					local check, res = tUser:IsProtected(nil, nil, vnick)
					if check then return false, res end

					local res, retmsg = tBlock:AddBlock(ssid, "kenny", "<unknown>", vnick, silent)
					tFuncs:Debug(retmsg)
					return res, retmsg
--~ 					if res then tSend:ChatToAll(Lang(ssid, 0, "User with ip %s is now kennylized.", vnick), "AAAA") end
--~ 					return res, nil
				end
				return false, Lang(ssid, 0, "User can't be found.")
			elseif not tUser:IsHigher(ssid,vsid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			else
				local res, retmsg = tBlock:AddBlock(ssid, "kenny", vnick, vip)
				tFuncs:Debug(retmsg)
				if res then tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now kennylized.", vnick), "AAAA") end
				return res
			end
		--]]
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tKennylize a user in mainchat, if user is offline, use ip instead of nick", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Kenny nick or Ip", "!%s %[line:Nickname or Ip]"},
			{2, 14, "Kenny (selected)", "!%s %[userNI]"},
		},
	},

	["unkenny"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)
			local value = false

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "Only operators or higher can unmute offline users.")
			elseif vsid and not tUser:IsHigher(ssid,vsid) and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			--// If user isn't online, use the nick to find the proper IP in blocktable
			if not vip then
				value = vnick
			else
				value = vip
			end

			local res,retmsg, silent = tBlock:DelBlock(ssid, "kenny", value)
			tFuncs:Debug(retmsg)
			if res and vip and not silent then tSend:ChatToAll("User <"..vnick.."> is now unkennylized.", "AAAA") return true end
			return res, retmsg
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tUnkennylize a user or Ip in mainchat", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unkenny nick or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Unkenny user (selected)", "!%s %[userNI]"},
		},
	},

	["muppetize"] = {
		["allowed"] = function(ssid, line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_, vnick = string.find(line, "^%S+%s+(%S+)")
			local _,_, option = string.find(line, "^%S+%s+%S+%s+(%S+)")
			local silent = false
			if option and option == "-s" then silent = true end

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick or IP>", cmd)
			else
				local ret, msg = tFuncs:AddBlocks(ssid, vnick, silent, "muppet")
				return ret, msg
			end

--[[			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			else
				local check, res = tUser:IsProtected(nil, vnick, vnick)
				if check then return true, res end
			end

			if not vsid then
				local lNick = vnick:lower()
				if tAccounts[lNick] then
					if (tClient[ssid].iProfile or -2) > (tAccounts[lNick].iLevel or -2) or tClient[ssid].iProfile == 10 then

						local check, res = tUser:IsProtected(nil, lNick, tAccounts[lNick].sLastIP)
						if check then return true, res end

						local res, retmsg = tBlock:AddBlock(ssid, "muppet", lNick, tAccounts[lNick].sLastIP)
						tFuncs:Debug(retmsg)
						if res then tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now muppetized.", vnick), "AAAA") end
						return res, nil
					end
				elseif tLogs.Nick[lNick] and tLogs.Nick[lNick].sIP and tLogs.Nick[lNick].sNick then
					if (tClient[ssid].iProfile or -2) > (tLogs.Nick[lNick].iProfile or -2) or tClient[ssid].iProfile == 10 then

						local check, res = tUser:IsProtected(nil, tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP)
						if check then return true, res end

						local res, retmsg = tBlock:AddBlock(ssid, "muppet", tLogs.Nick[lNick].sNick, tLogs.Nick[lNick].sIP)
						tFuncs:Debug(retmsg)
						if res then tSend:ChatToAll(Lang(ssid, 0, "User <%s> is now muppetized.", vnick), "AAAA") end
						return res, nil
					end
				elseif tFuncs:IsIP(vnick) then
					local res, retmsg = tBlock:AddBlock(ssid, "muppet", "<unknown>", vnick)
					tFuncs:Debug(retmsg)
--~ 					if res then tSend:ChatToAll(Lang(ssid, 0, "User with ip %s is now muppetized.", vnick), "AAAA") end
					return res, retmsg
				end
				return false, Lang(ssid, 0, "User can't be found.")
			elseif not tUser:IsHigher(ssid,vsid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			else
				local res, retmsg = tBlock:AddBlock(ssid, "muppet", vnick, vip)
				tFuncs:Debug(retmsg)
				if res then
					tSend:ChatToAll(Lang(ssid, 0, (tFilters.Muppets.BroadcastOn or "User <%s> is now muppetized."), vnick), "AAAA")   --// When translation is done, that can be used for customizing messages as well
					return res
				end
				return res, retmsg
			end
		--]]


		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\t\tMuppetize a user in mainchat, if user is offline, use ip instead of nick", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Muppetize nick or Ip", "!%s %[line:Nickname or Ip]"},
			{2, 14, "Muppetize (selected)", "!%s %[userNI]"},
		},
	},

	["unmuppetize"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,vnick = string.find(line, "^%S+%s+(%S+)")
			local vsid = tNick:GetSid(vnick)
			local vip = tUser:GetIp(vsid)
			local value = false

			if not vnick then
				return false, Lang(ssid, 0, "Error: Use %s <nick>", cmd)
			elseif not vsid and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "Only operators or higher can unmute offline users.")
			elseif vsid and not tUser:IsHigher(ssid,vsid) and not tUser:IsOp(ssid) then
				return false, Lang(ssid, 0, "This user is too powerful for you.")
			end

			if not vip then value = vnick else value = vip end	--// If user isn't online, use the nick to find the proper IP in blocktable

			local res, retmsg = tBlock:DelBlock(ssid, "muppet", value)
			tFuncs:Debug(retmsg)

			if res then
				tSend:ChatToAll(Lang(ssid, 0, (tFilters.Muppets.BroadcastOff or "User <%s> is now unmuppetized."), vnick), "AAAA")
				return res
			else
				return res, retmsg
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s <nick>\t\t\t\tUnmuppetize a user or Ip in mainchat", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unmuppetize nick or Ip", "!%s %[line:Nickname or Ip]"},
			{2,14, "Unmuppetize user (selected)", "!%s %[userNI]"},
		},
	},

	["checksums"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local sBigBuffer = ""
			local ctime = os.time()
			local tbl = {}
			local folders = {"core", "api", "protocols",}	-- only look in these folders for more files
			local res = "\n\n\tFile Scanner Output generated on "..os.date(tHubSettings.Main.sDateFormat, ctime) .." "..os.date(tHubSettings.Main.sTimeFormat, ctime) .."\n\tFor hub "..tHubSettings.Main.sName .." at "..tHubSettings.Main.sAddress .."\n"
			local tResult = {}

			local counter = 0
			for i, path in ipairs(folders) do
				local folder = _APPPATH..path
				for file in lfs.dir(folder) do
					if string.sub(file, #file - 3) == ".lua" and string.sub(file, 1, 1) ~= "." then
						local attr = lfs.attributes(folder.."/"..file)
						if attr and attr.mode == "file" then
							counter = counter + 1
							local tth, buff = tFuncs:GetHashFromFile(folder.."/"..file, true)
							if tth then
								sBigBuffer = sBigBuffer .. buff
								tResult[file] = {["tth"] = tth, ["path"] = path.."/"..file, ["modification"] = ToDate(attr.modification), ["size"] = tFuncs:ConvertShare(attr.size)}
							else
								tResult[file] = {["err"] = "No TTH for file", ["path"] = folder.."/"..file}
								tFuncs:Debug("Couldnt get TTH for "..file)
							end
						end
					end
				end
			end

			local sBigTTH = tth.buffer(sBigBuffer)
			res = res .."\tCombined TTH for all "..counter .." files is "..sBigTTH .."\n\n"

			tbl = {}
			for i, data in pairs(tResult) do
				if not data.err then
					table.insert(tbl, "\t"..Padding(data.path, 32).."\t"..data.tth.."\t"..data.modification.."\t"..data.size)
				else
					table.insert(tbl, "\t"..Padding(data.path, 32).."\tERROR : "..data.err)
				end
			end
			table.sort(tbl)
			return true, res..table.concat(tbl,"\n") .."\n"
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s\t\t\t\t\tGenerates a list of check sums of the hub files, useful for support issues ", cmd)
		end,
		["outputtype"] = "info",
		["ucmd"] = {
			{1, 3, "Generate checksums ", "!%s"},
		},
	},

	["unhide"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,nick = string.find(line, "^%S+ (%S+)")
			local result1, result2, profname = false, false, ""
			local onick = tClient[ssid].sNick
			local nick = nick or onick
			local vsid = tNick:GetSid(nick)

			local regname, reg = AccountManager:GetName(nick)

			--// Check if user is online
			if vsid and tClient[vsid].bBot then
				return false, "Bots aren't hidden, chatrooms are forced hidden to anyone who can't use them."
			elseif vsid then
				profname = tClient[vsid].sProfile
				if tUser:IsHigher(ssid,vsid) or onick:lower() == nick:lower() then
					if not tClient[vsid].bHidden then
						return false, Lang(ssid, 0, "<%s> isn't hidden.", nick)
					end
					tClient[vsid].bHidden = nil
					tUser:Announce(vsid)
					result1 = reg.sNick
				else
					return false, Lang(ssid, 0, "This user is too powerful for you.")
				end
			end

			--// Check if user is registered
			if regname and reg then
				if reg.iLevel < tClient[ssid].iProfile or nick:lower() == onick:lower() or tClient[ssid].iProfile == 10 then
					if not reg.bHidden then
						return false, Lang(ssid, 0, "<%s> isn't hidden.", nick)
					end
					reg.bHidden = nil
					result2 = reg.sNick
					tFuncs:SaveSettings("accounts")
				else
					return false, Lang(ssid, 0, "<%s> is too powerful for you.", nick)
				end
			end

			if result1 and result2 then
				tSend:PmToUser("You are now unhidden by "..tClient[ssid].sNick, nil, vsid)
				return true, Lang(ssid, 0, "<%s> is now unhidden", result1)
			elseif not result1 and result2 then
				return true, Lang(ssid, 0, "Offline registered user %s is unhidden ", result2)
			elseif not result1 and not result2 then
				return false, Lang(ssid, 0, "<%s> not found.", nick)
			elseif result1 and not result2 then
				tSend:PmToUser("You are now unhidden by "..tClient[ssid].sNick, nil, vsid)
				return true, Lang(ssid, 0, "<%s> is now visible to everyone.", result1)
			else
				return false, Lang(ssid, 0, "Crap, an unexpected problem.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick>]\t\t\t\t\tUnhide yourself (use without nick), or unhide a user/chatroom", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Unhide yourself", "!%s"},
			{1, 3, "Unhide nick", "!%s %[line:Nickname]"},
			{2,14, "Unhide user (selected)", "!%s %[userNI]"},
		},
	},

	["hide"] = {
		["allowed"] = function(ssid,line, cmd)
			return tUser:IsAllowed(ssid, cmd)
		end,
		["cmd"] = function(ssid, line, cmd)
			local _,_,nick = string.find(line, "^%S+ (%S+)")
			local result1, result2, profname = false, false, ""
			local onick = tClient[ssid].sNick
			local nick = nick or onick
			local vsid = tNick:GetSid(nick)

			local regname, reg = AccountManager:GetName(nick)

			--// Check if user is online
			if vsid and tClient[vsid].bBot then
				return false, "Bots can't be hidden, chatrooms are forced hidden to anyone who can't use them."
			elseif vsid then
				if not regname then
					return false, Lang(ssid, 0, "Only registered users can be hidden.")
				elseif tUser:IsHigher(ssid,vsid) or nick:lower() == onick:lower() then
					if tClient[vsid].bHidden then
						return false, Lang(ssid, 0, "<%s> is already hidden.", nick)
					end
					tClient[vsid].bHidden = true
					tUser:Hide(vsid)
					result1 = reg.sNick
				else
					return false, Lang(ssid, 0, "<%s> is too powerful for you.", nick)
				end
			end

			--// Check if user is registered
			if regname and reg then
				if reg.iLevel < tClient[ssid].iProfile or nick:lower() == onick:lower() or tClient[ssid].iProfile == 10 then
					if reg.bHidden then
						return false, Lang(ssid, 0, "<%s> is already hidden.", reg.sNick)
					end
					reg.bHidden = true
					result2 = reg.sNick
					tFuncs:SaveSettings("accounts")
				else
					return false, Lang(ssid, 0, "This user is too powerful for you.")
				end
			else
				return false, Lang(ssid, 0, "Only registered users can be hidden.")
			end

			if result1 and result2 then
				tSend:PmToUser("You are now hidden by "..tClient[ssid].sNick, nil, vsid)
				return true, Lang(ssid, 0, "<%s> is now hidden to everyone below level: %s", result2, reg.iLevel.." ("..reg.sProfile..")")
			elseif not result1 and result2 then
				return true, Lang(ssid, 0, "Offline registered user <%s> is hidden to everyone below level: %s", result2, reg.iLevel.." ("..reg.sProfile..")")
			elseif not result1 and not result2 then
				return false, Lang(ssid, 0, "User not found.")
			elseif result1 and not result2 then
				tSend:PmToUser("You are now hidden by "..tClient[ssid].sNick, nil, vsid)
				return true, Lang(ssid, 0, "<%s> is now hidden to everyone below level: %s.", result1, tClient[vsid].iProfile)
			else
				return false, Lang(ssid, 0, "Crap, an unexpected problem.")
			end
		end,
		["help"] = function(ssid, line, cmd)
			return Lang(ssid, 0, "%s [<nick>]\t\t\t\t\tHide yourself (use without nick), or hide a user/chatroom, from users lower than it's own", cmd)
		end,
		["ucmd"] = {
			{1, 3, "Hide yourself", "!%s"},
			{1, 3, "Hide nick", "!%s %[line:Nickname]"},
			{2,14, "Hide user (selected)", "!%s %[userNI]"},
		},
	},

}
--[[
--~ 	["worldmap"] = {
--~ 		["allowed"] = function(ssid, line, cmd)
--~ 			return tUser:IsAllowed(ssid, cmd)
--~ 		end,
--~ 		["cmd"] = function (ssid, line, cmd)
--~ 			--// Build country-table
--~ 			local tCountryTable = {}
--~ 			for vsid,vclient in pairs(tClient) do
--~ 				if vclient.iState == 7 and vclient.sCountryCode and vclient.sCountryCode ~= "" and vclient.sCountryCode ~= "ZZ" and vclient.sCountryCode ~=  "--" then
--~ 					tCountryTable[vclient.sCountryCode] = tCountryTable[vclient.sCountryCode] or {["Users"] = {}}
--~ 					table.insert(tCountryTable[vclient.sCountryCode].Users, vclient.sNick)
--~ 					tCountryTable[vclient.sCountryCode].UsersTxt = (tCountryTable[vclient.sCountryCode].UsersTxt or "")..vclient.sNick..", "
--~ 					tCountryTable[vclient.sCountryCode].Count = (tCountryTable[vclient.sCountryCode].Count or 0)+1
--~ 				end
--~ 			end
--~
--~ 			--// Load country-coordinates
--~ 			if not tCountryCoords then
--~ 				local f = io.open("./scripts/worldcoords.txt","rt")
--~ 				if f == nil then return false , "Worldmap Error: Can't load ./scripts/worldcoords.txt" end
--~ 				tCountryCoords = {}
--~ 				local cc,lat,lng
--~ 				for str in f:lines() do
--~ 					cc,lat,lng = str:match("^(%w%w),([+-]?%d+%.%d+),([+-]?%d+%.%d+)$")
--~ 					if cc and lat and lng then
--~ 						tCountryCoords[cc] = { ["Lat"] = lat, ["Lng"] = lng}
--~ 					end
--~ 				end
--~ 				f:close()
--~ 			end
--~
--~ 			--// Create markers
--~ 			local markers = ""
--~ 			local counter = 0
--~ 			for name,data in pairs(tCountryTable) do
--~ 				if tCountryCoords[name] then
--~ 					local res, nicks = tFuncs:ToUTF8(string.sub(tCountryTable[name].UsersTxt, 1,-3))
--~ 					local coord1 = tCountryCoords[name].Lat
--~ 					local coord2 = tCountryCoords[name].Lng
--~ 					markers = markers.."markers["..counter.."] = ["..coord1..","..coord2..",\"Country: "..name.."<br>Users: "..tCountryTable[name].Count.."<br>Nicknames: "..nicks.."\"];\n"
--~ 					counter = counter+1
--~ 				end
--~ 			end

--~ 			if markers ~= "" then
--~ 				local parta, partb = "",""

--~ 				--// Construct the html file
--~ 				local afile, err = io.open("./scripts/worldmap-A.txt", "r")
--~ 				if afile then
--~ 					parta = afile:read("*all")
--~ 					afile:close()
--~ 				else
--~ 					return false, "Worldmap Error: Can't load ./scripts/worldmap-A.txt"
--~ 				end

--~ 				local bfile, err = io.open("./scripts/worldmap-B.txt", "r")
--~ 				if bfile then
--~ 					partb = bfile:read("*all")
--~ 					bfile:close()
--~ 				else
--~ 					return false, "Worldmap Error: Can't load ./scripts/worldmap-B.txt"
--~ 				end

--~ 				tWebServer.Files["worldmap.html"].Content = parta..markers..partb
--~
--~ 				--// Save the file for debugging:
--~ 				local f = io.open("./scripts/worldmap.html","w")
--~ 				if f then
--~ 					f:write(tWebServer.Files["worldmap.html"].Content)
--~ 					f:close()
--~ 				end

--~ 				--// Show the url the user can use:
--~ 				return true, "http://"..SetManager:Get("hub_address").."/worldmap.html"
--~ 			else
--~ 				return false, "No country information available."
--~ 			end
--~ 		end,
--~ 		["help"] = function(ssid, line, cmd)
--~ 			return Lang(ssid, 0, "%s\t\t\t\t\tBuild a Google Map showing all online users", cmd)
--~ 		end,
--~ 		["ucmd"] = {
--~ 			{1,3,"Build Google Map showing users","!%s"},
--~ 		},
--~ 	},	--]]
